/*
 * Copyright JJ Furman
 * User: user
 * Date: May 10, 2002
 * Time: 1:09:16 PM
 */
package freebase.api;

import freebase.sql.ColumnDefinition;
import freebase.util.SqlTypes;

import java.io.Serializable;
import java.sql.SQLException;
import java.sql.Types;

/**
 * An Immutable package of information about columns for returning with the results of
 * a query.<BR><BR>
 *
 * To save a few it might be possible to keep a flag that determines if this is mutable, and have
 * a few operations that either create a new one or just change it based on the flag.
 * TODO:  add correct defaults other than null
 */
public class ColumnInfo implements Serializable {
    String schemaName;
    String tableName;
    String columnName;

    int type;
    int size = 0;  //  same as precision
    int scale = 0;

    boolean nullable = true;

    //  TODO:  support other constraints here

    /**
     * Construct a ColumnInfo with all data
     */
    public ColumnInfo(String columnName, String tableName, String schemaName,
                      int type, boolean nullable, int size, int scale) {
        this.columnName = columnName;
        this.nullable = nullable;
        this.size = size;
        this.scale = scale;
        this.schemaName = schemaName;
        this.tableName = tableName;
        this.type = type;
    }

    /**
     * Used from the create table statement.  Seems silly to have both this and ColumnDefinition but oh well
     */
    public ColumnInfo(String schemaName, String tableName, ColumnDefinition def) {
        this(def.getFieldName(), tableName, schemaName,
                def.getSqlType(), def.isNullable(), def.getSize(), def.getScale());
    }

    /**
     * Construct a ColumnInfo with name, type, and precision
     */
    public ColumnInfo(String columnName, int type, int size, int scale, boolean nullable) {
        this.columnName = columnName;
        this.type = type;
        this.size = size;
        this.scale = scale;
        this.nullable = nullable;
    }

    /**
     * Construct a ColumnInfo with name, type, and precision
     */
    public ColumnInfo(String columnName, int type, int size) {
        this.columnName = columnName;
        this.size = size;
        this.type = type;
    }

    /**
     * Creates a
     */
    public ColumnInfo(String columnName, int type) {
        this.columnName = columnName;
        this.type = type;
    }

    /**
     * Construct a new ColumnInfo with just the name changed
     */
    public ColumnInfo(String newColumnName, ColumnInfo otherInfo) {
        this(newColumnName, otherInfo.getTableName(), otherInfo.getSchemaName(),
                otherInfo.getType(), otherInfo.isNullable(), otherInfo.getSize(), otherInfo.getScale());
    }

    /*************************************************************
     * static efficiency methods
     */

    /**
     * Current implementation creates a new ColumnInfo, but later could (if appropriate)
     * just change the name locally and return the same object passed in
     */
    public static ColumnInfo rename(ColumnInfo colInfo, String newName) {
        return new ColumnInfo(newName, colInfo);
    }

    /*************************************************************
     * getters
     */
    public String getName() {
        return columnName;
    }

    public String getColumnName() {
        return columnName;
    }

    public boolean isNullable() {
        return nullable;
    }

    public int getSize() {
        return size;
    }

    public int getScale() {
        return scale;
    }

    public String getSchemaName() {
        return schemaName;
    }

    public String getTableName() {
        return tableName;
    }

    public int getType() {
        return type;
    }

    /**
     * TODO:  return the default!
     * @return
     */
    public Comparable getDefault() {
        return null;
    }

    /**
     * Gets the max possible number of visible characters ... ie 3.45 is 4
     */
    public int getDisplaySize() {
        switch(type) {
            case Types.BIT:
                return 5;  //  max len is 5, ie "false"

            case Types.NUMERIC:
                if(size == 0) {
                    return 40;
                } else if(scale == 0) {
                    return size;  //  integer
                } else {
                    return size + 1;  //  include the decimal point
                }

            case Types.CHAR:
            case Types.VARCHAR:
            case Types.LONGVARCHAR:
            default:
                return size;

        }
    }

    public String getFullName() {
        if(tableName != null) {
            return tableName + "." + columnName;
        }
        return columnName;
    }

    public String toString() {
        try{
            return "ColumnInfo: " + columnName + " " + SqlTypes.getNameForType(type);
        }catch(SQLException e) {
            return e.getMessage();
        }
    }
}
