/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: Sep 9, 2001
 * Time: 5:20:51 PM
 * To change template for new interface use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.api;

import java.sql.SQLException;

/**
 * Makes it faster to find your data.  Indexes have varying degrees of support for multiversion concurrency
 * but no matter how well integrated they are they can never make the final call for recently modified data.
 * This is because:
 *
 * 1.  They will not have enough information to keep track of whether the current transaction has
 * locked the particular row (nor can they find out easily without sometimes hitting disk)
 * 2.  They do not know, even if the current transaction is the writer of the newest data value, if
 *    the current transaction view can see the data.
 *
 * Appears that the only thing the index can decide is that if the transaction that appears to have last
 * modified the data cannot be found in the transaction manager, then it can age out the transaction and
 * mark the entry as visible to all.  Otherwise it must warn via the isPossiblyNotVisible flag
 *
 * Should the index know about what data it is indexing?
 * TODO:  should the transaction be removed from all of the methods
 *         since it is already supplied??  Probably not, explicit is always more clear
 */
public interface IndexCursor {
      //  NONE_FLAG used when
    final public static byte NONE_FLAG = 0;
    final public static byte NORMAL_FLAG = 1;
    final public static byte INSERTED_FLAG = 2; //  a flag saying that this is pending insert
    final public static byte DELETED_FLAG = 3;  //  a flag saying that this is pending delete
    final public static byte REMOVED_FLAG = 4;  //  a flag saying that this has been removed already
    final public static byte UPDATED_FLAG = 5;  //  probably don't need this value anymore!!
    final public static byte MOVED_FLAG = 6;    //  may not be this value yet!

    //  all passing in transaction:

      //  most are not legal anyway
    final public static String[] statusNames = new String[] {"NONE", "NORMAL", "INSERTED", "DELETED",
                                                             "REMOVED", "UPDATED", "MOVED"};

    /**
     * Value is either an Object[] or a straight value.  Pass null if the transaction
     * that created the data is guarenteed to be commited (ie, during creation)
     *
     * TODO:  looks like this needs to return an array of potential collisions (for a unique index only)
     * Does the collision list need to have the statuses too?  How will the resolution be determined by
     * caller?
     */
    void insertEntry(Object value, long rowId) throws SQLException;

    /**
     * Operates with the current row, changing the current rowid to the value
     * given
     *
     * TODO:  looks like this needs to return an array of potential collisions (for a unique index only)
     */
    void updateEntry(Object newValue) throws SQLException;

    /**
     * Operates with the current row
     */
    void deleteEntry() throws SQLException;

    /**
     * Set the current position to the first entry with the passed value, or
     * the first entry with the next high value if it doesn't exist.
     *
     * Passed value is either a Comparable value or an Comparable[].  If passed a Comparable[1], it should auto
     *   convert it internally (for now)
     * @value null means beginning.  Any component null means go to first of that group
     */
    boolean setPos(Object value) throws SQLException;

    // NOTE:  need: setPos(value, rowId) ?

    /**
     * set the position of the cursor before the beginning
     */
    void reset() throws SQLException;

    /**
     * @return true if there is a next in this index.
     *
     * I assume that this still works in a Hash style index, just that
     * the next only means fetch next with the same value.
     */
    boolean next() throws SQLException;

    /**
     *  Operates with the current row.
     * Either this, or a no argument version that returns an Object[]
     */
    Comparable getCurValue(int pos) throws SQLException;

    /**
     * get the row id of the current cursor position
     */
    long getCurRowId() throws SQLException;

    /**
     * It is the responsibility of the caller to check the status
     * of the entry.  If it is not normal, then the caller must verify the
     * status of the transaction that last modified it.
     *
     * TODO:  probably deprecate this
     */
    byte getCurFlags() throws SQLException;

    /**
     * When the current position of this index cursor is possibly not visible to the transaction
     * that validated this
     */
    boolean isPossiblyNotVisible() throws SQLException;

    ////////////////  following methods should go into the index interface


    //  find...  is rather dependent.  Later the index should be able to make suggestions
    //  to a smart planner about how it expects it can help with a query.  It should also
    //  quantify how many units of work it expects to take and how many results it expects.

    /**
     * TO be called when the user is done.
     */
    void close() throws SQLException;
}
