/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Nov 11, 2001
 * Time: 7:56:52 AM
 * To change template for new interface use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.api;

import freebase.engine.Transaction;
import freebase.lock.LockWait;

import java.sql.SQLException;

/**
 * Holds locks for a table.  table locks are gotten before row locks.
 * Since all row locks require a ROW_EXCLUSIVE_LOCK to be obtained first,
 * only the table forces a registration for a callback on transaction commit.
 */
public interface TableLockManager extends TransactionDoneCallback {
    /**
     *  For select for update.  Local only lock.
     */
    final public static byte ROW_SHARE_LOCK = 0;

    /**
     * For update, insert, delete.  Local only lock.
     */
    final public static byte ROW_EXCLUSIVE_LOCK = 1;

    //  get these locks on all nodes/partitions that contain table data.
    final public static byte SHARE_LOCK = 2;
    final public static byte SHARE_ROW_EXCLUSIVE_LOCK = 3;
    final public static byte EXCLUSIVE_LOCK = 4;

    final public static byte NUM_TABLE_LOCKS = 5;

    /**
     * For each type of lock i this contains an array: incompatibleLocks[i] of
     * locks that are not allowed with i
     */
    final public static byte[][] incompatibleLocks = new byte[][]
        {{4},          //  RS  incompatible with X
         {2,3,4},      //  RX  incompatible with S,SRX,X
         {1,3,4},      //  S   incompatible with RX,SRX,X
         {1,2,3,4},    //  SRX incompatible with RX,S,SRX,X
         {0,1,2,3,4}}; //  X   incompatible with all others

    final public static String[] tableLockNames = new String[] {"ROW_SHARE_LOCK",
        "ROW_EXCLUSIVE_LOCK", "SHARE_LOCK", "SHARE_ROW_EXCLUSIVE_LOCK", "EXCLUSIVE_LOCK"};

    /**
     * only undone by an unlock
     */
    boolean lockTable(Transaction t, byte mode, boolean nowait) throws SQLException;

    void unlockTable(Transaction t, byte mode) throws SQLException;

    /**
     * Should only be called when a small rollback is happening.  On Commit,
     * call unlockAll(Transaction t), or likewise on a complete rollback.
     *
     * When releasing locks, the simplest design is to notify on the transactions
     * finish lock.  The alternative is to look and see what transactions are exactly
     * waiting for and see if this matches.  On a complete commit or rollback, just
     * loop through all.  This is more concurrent.
     */
    public void unlockRow(long rowId) throws SQLException;

    /**
     * actually, this probably will be the final implementation, though it is
     * called just in case it is useful later.
     */
    public void unlockAll(Transaction t) throws SQLException;

    /**
     *   unlockTo(SavePoint ...) Appears we probably don't need this.
     * After all, we have to walk over all of the rows anyway to actually
     * undo the data, why not call unlockRow then...
     *
     * When are waiters notified?  as they go, or at the end?
     */


    /**
     * Obtain write lock on the row.
     *
     * Note the simplest design is to explicitly lock the row with a log entry.  That
     * way when the log entry is un-done then unlockRow is called.  Alternatively, the
     * row lock could hold a bit as to whether this is the first time lockRow has been
     * called.  The data updater can then either query or this method can be modified
     * to return boolean as to if this is the first time.  This flag is marked in the
     * row modification log entry, so that when undoing locks can be properly released.
     *
     * And then, consider the implications of this pattern for the table locks.
     *
     * @param expectedHolder a hint to the system for who to wait for if there is no queue found
     * for the current rowl
     * @return the previous holder of the lock.  (this will go away)
     */
    public LockWait lockRow(Transaction t, Transaction expectedHolder, long rowId) throws SQLException;

}
