/*
 * Copyright (c) 1999,2006 JJ Furman
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name "Freebase" may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 4. No portion of this software may be re-released under the GPL license.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package freebase.catalog;

import freebase.api.*;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.log.LogEntry;
import freebase.log.data.TableLogEntry;
import freebase.parser.ParseException;
import freebase.parser.SqlParser;
import freebase.sql.*;
import freebase.sql.plan.QueryProcessor;
import freebase.util.*;

import java.io.FileInputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;

/**
 * The Catalog is where you discover database objects, like tables and indexes. It is essentially
 * the same as the Data Dictionary<BR><BR>
 *
 * Should this class be an interface, as above?
 *
 * TODO:  needs a lot of work.  Move processInsert to another place!
 * TODO:  get rid of constants interface pattern that people whine about.
 */
public class Catalog implements LogEntryProcessor, CatalogDD {

    /**
     * Where all of our data is actually stored.
     */
    StorageManager sm;

    TransactionManager tm;

    /**
     * map from schemaname to HashMap of tablenames to the actual tables.
     * TODO:  although this information may be cached eventually, this is not the
     * main store of this information.  The actual schema tables will be
     */
    HashMap schemas = new HashMap();

    /**
     * Helper class manages storage into the actual data dictionary tables
     */
    DataDictionary dataDictionary = new DataDictionary();

    boolean initialized = false;

    /**
     * Query engine currently used to execute high performance plans on updates.
     * (eventually higher performance than always table scan)
     */
    QueryProcessor queryEngine;

    /**
     * The catalog should only need the StorageManager at startup
     */
    public Catalog(TransactionManager tm, StorageManager sm, QueryProcessor queryEngine) throws SQLException {
        this.tm = tm;
        this.sm = sm;
        this.queryEngine = queryEngine;

        if(sm.getMaxTableId() >= 0) {
            loadCatalog();
        }
    }

    /**
     * Internal helper method to make code easier to read
     * @param t
     * @param table
     * @return
     * @throws SQLException
     */
    private TypedTupleIterator getTypedCursor(Transaction t, Table table) throws SQLException {
        return new TypedTupleIterator(table.getTableTupleIterator(t, false));
    }

    /**
     * Loads the catalog from storage.  Does this with the following steps (so far):<BR>
     * 1.  load TABLES.  create Tables and hash all by id<BR>
     * 2.  load COLUMNS.  add to Table objects
     * 3.  load USERS.  make all Schemas, and hook the tables in by name.
     *
     * TODO:  move this to DataDictionary or elsewhere, use compiled constants
     */
    protected void loadCatalog() throws SQLException {
        Transaction t = tm.beginTransaction(null, "SYS", Connection.TRANSACTION_READ_COMMITTED);
        LongHashMap tablesById = new LongHashMap();

          //  load TABLES
        TableStorage tablesStorage = sm.getTableStorage(0);
        Table tablesTable = new Table(this, tablesStorage, dataDictionary.getCreateTABLES());

        TypedTupleIterator cursor = getTypedCursor(t, tablesTable);
        int tableIdOffset   = tablesTable.getColumnOffset("TABLE_ID");
        int tableNameOffset = tablesTable.getColumnOffset("TABLE_NAME");
        int userNameOffset  = tablesTable.getColumnOffset("USER_NAME");
        while(cursor.next()) {
              //  create the table
            int tableId = ((Number)cursor.getValue(tableIdOffset)).intValue();
            TableStorage storage = sm.getTableStorage(tableId);
            String tableName = (String)cursor.getValue(tableNameOffset);
            String schema = (String)cursor.getValue(userNameOffset);

            Table table = null;
            if(tableId == 0) {
                table = tablesTable;
            } else {
                table = new Table(this, storage, schema, tableName);
            }

            tablesById.put(tableId, table);
            if(DataDictionary.SYS_USER.equals(schema)) {
                dataDictionary.registerSysTable(table);  //  seems redundant, cleanup eventually
            }
        }
        cursor.close();

        Table columnsTable = dataDictionary.getSysTable(COLUMNS_TABLE_OFFSET);
        Table usersTable = dataDictionary.getSysTable(USERS_TABLE_OFFSET);
        Table sequencesTable = dataDictionary.getSysTable(SEQUENCES_TABLE_OFFSET);
        Table indexesTable = dataDictionary.getSysTable(INDEXES_TABLE_OFFSET);
        Table indexColumnsTable = dataDictionary.getSysTable(INDEX_COLUMNS_TABLE_OFFSET);

          //  load COLUMNS
        dataDictionary.addCOLUMNSColumns(columnsTable);
        cursor = getTypedCursor(t, columnsTable);
        int columnIdOffset  = columnsTable.getColumnOffset("COLUMN_ID");  //  TODO:  may not be needed ever
        int positionOffset  = columnsTable.getColumnOffset("POSITION");
        int nameOffset      = columnsTable.getColumnOffset("NAME");
        int typeOffset      = columnsTable.getColumnOffset("TYPE");
        int sizeOffset      = columnsTable.getColumnOffset("SIZE");
        int precisionOffset = columnsTable.getColumnOffset("PRECISION");
        int nullableOffset  = columnsTable.getColumnOffset("NULLABLE");
        tableIdOffset       = columnsTable.getColumnOffset("TABLE_ID");

        while(cursor.next()) {
              //  TODO:  columnId not obtained at all.  Perhaps would be
              //  useful later if needing to change any value
            ColumnInfo colInfo = new ColumnInfo((String)cursor.getValue(nameOffset),
                    cursor.getInt(typeOffset), cursor.getInt(sizeOffset), cursor.getInt(precisionOffset),
                    ((Bool)cursor.getValue(nullableOffset)).getValue());

            Table table = (Table)tablesById.get(cursor.getInt(tableIdOffset));
            table.addColumn(colInfo, cursor.getInt(positionOffset));
        }
        cursor.close();

          //  load USERS
        cursor = getTypedCursor(t, usersTable);
        nameOffset = usersTable.getColumnOffset("NAME");
        synchronized(schemas) {
            while(cursor.next()) {
                //  TODO:  support password and other good stuff in the schema
                Schema schema = new Schema(this, (String)cursor.getValue(nameOffset));
                schemas.put(schema.getUser(), schema);
            }
        }
        cursor.close();

          //  now place the tables into the appropriate schemas
        Enumeration en = tablesById.keys();
        while(en.hasMoreElements()) {
            Long value = (Long)en.nextElement();
            Table table = (Table)tablesById.get(value.longValue());
            Schema schema = getSchema(table.getSchema());
            schema.addTable(table);
            //System.out.println(table);
        }

          //  load INDEX_COLUMNS into ArrayList (by index_id) -> ArrayList (column list)
        ListOfLists indexColumns = new ListOfLists();
        cursor = getTypedCursor(t, indexColumnsTable);
        while(cursor.next()) {
            int indexId = cursor.getInt(IndexColumnsColumns.INDEX_ID_OFFSET);
            int pos = cursor.getInt(IndexColumnsColumns.POSITION_OFFSET);
            String col = cursor.getString(IndexColumnsColumns.COLUMN_NAME_OFFSET);

              //  put the column in
            indexColumns.set(indexId, pos, col);
        }
        cursor.close();

          //  load INDEXES
        cursor = getTypedCursor(t, indexesTable);
        while(cursor.next()) {
            int indexId = cursor.getInt(IndexesColumns.INDEX_ID_OFFSET);
            String indexName = cursor.getString(IndexesColumns.INDEX_NAME_OFFSET);
            Table table = (Table)tablesById.get(cursor.getInt(IndexesColumns.TABLE_ID_OFFSET));
            boolean isUnqiue = cursor.getBoolean(IndexesColumns.UNIQUE_FLAG_OFFSET);

            List colList = indexColumns.get(indexId);
            String[] fields = new String[colList.size()];
            colList.toArray(fields);

            Schema schema = getSchema(table.getSchema());
            Index index = new Index(this, schema, indexName, table, fields, isUnqiue);
            index.setStorage(sm.getIndexStorage(indexId));
            table.addIndex(index);
            schema.addIndex(index);
        }
        cursor.close();

          //  load SEQUENCES
        cursor = getTypedCursor(t, sequencesTable);
        while(cursor.next()) {
            int seqId = cursor.getInt(SequencesColumns.SEQUENCE_ID_OFFSET);
            SequenceStorage ss = sm.getSequenceStorage(seqId);
            Sequence s = new Sequence(this, ss, cursor);

              //  create the table
            Schema schema = getSchema(s.getSchema());
            schema.addSequence(s);
        }
        cursor.close();

        t.commit();
        initialized = true;
    }

    public boolean isInitialized() {
        return initialized;
    }


    /**
     * Runs initialization scripts that prepares this catalog for general use.
     * Currently called from DatabaseNode.
     */
    public void initializeNew(Transaction t) throws SQLException {
        //  TODO:  double check to see that this is a new catalog

          //  parse the create_tables.  For now, this is hard coded as such, later use F3_HOME or whatever
        SqlParser parser = null;
        try{
            FileInputStream in = new FileInputStream("catalog/create_system.sql");
            parser = new SqlParser(in);
        }catch(IOException ioe) {
            throw new SQLException("Failed to open create_system file: " + ioe.getMessage());
        }

        ArrayList allStatements = new ArrayList();
        ArrayList allTables = new ArrayList();
        CreateTableStatement tablesCreate = null;
        boolean done = false;
        while(! done) {
            SqlStatement s = null;
            try{
            s = parser.TerminatedStatement();
            }catch(ParseException pe) {
                throw new SQLException("Failed to parse statement: " + pe.getMessage());
            }

            if(s != null) {
                allStatements.add(s);
                if(s instanceof CreateTableStatement) {
                    CreateTableStatement c = (CreateTableStatement)s;
                    if(c.getTableName().equals(DataDictionary.TABLES_TABLE)) {
                        tablesCreate = c;
                    }

                    allTables.add(s);
                }
            } else {
                done = true;
            }
        }

          //  first create the TABLES table, must be ID 0
        TableStorage ts = sm.createTableStorage(t);
        Table tablesTable = new Table(this, ts, tablesCreate);
        dataDictionary.addSysTable(t, tablesTable);

        //  TODO:  possibly move the row insertion here and out of the DataDictionary

          //  create all of the rest of the tables
        for(int i = 0; i < allTables.size(); i++) {
            CreateTableStatement createTable = (CreateTableStatement)allTables.get(i);
            if(createTable.getTableName().equals(DataDictionary.TABLES_TABLE)) {
                continue;
            }

            ts = sm.createTableStorage(t);
            Table table = new Table(this, ts, createTable);
            dataDictionary.addSysTable(t, table);
        }

        dataDictionary.putAllTables(createSchema(t, DataDictionary.SYS_USER));

          //  insert all rows into the "COLUMNS" table
        for(int i = 0; i < allTables.size(); i++) {
            CreateTableStatement createTable = (CreateTableStatement)allTables.get(i);
            insertColumnsIntoDD(t, getTable(DataDictionary.SYS_USER, createTable.getTableName()).getTableId(), createTable);
        }
    }


    /********************************************************************
     * basic Catalog methods
     */

    /**
     * Currently called from QueryProcessor, will probably change when
     * the select uses the standard form of table name dd.dd@link.
     * Errors are handled inside each method.
     */
    public Table getTable(String schema, String tableName) throws SQLException {
        Schema s = getSchema(schema);
        return s.getTable(tableName);
    }

    public Table getTable(TableRef ref, String curUser) throws SQLException {
        return getTable(ref.getSchema(curUser), ref.getTable());
    }

    /**
     * Return the tables for this schema name
     */
    public Schema getSchema(String schema) throws SQLException {
        synchronized(schemas) {
            Schema s = (Schema)schemas.get(schema);
            if(s == null) {
                throw new SQLException("Schema '" + schema + "' does not exist", "none", ErrorCode.SCHEMA_DOES_NOT_EXIST);
            }
            return s;
        }
    }

    public boolean schemaExists(String schema) throws SQLException {
        synchronized(schemas) {
            Schema s = (Schema)schemas.get(schema);
            return s != null;
        }
    }

    public ArrayList getTableNames(String schema) throws SQLException {
        Schema s = null;

        synchronized(schemas) {
            s = getSchema(schema);
        }

        return s.getAllTableNames();
    }

    /*****************************************  DDL   ***************************************************/

    /**
     * Creates a standard index on the table's columns.
     * TODO:  try to create two indexes concurrently with same name on table, to test failure
     * (can try over and over with intermittent sleeps after every region)
     */
    public void createIndex(Transaction t, CreateIndexStatement createIndex) throws SQLException {
        String[] fields = createIndex.getColumnsArray();

        Schema s = getSchema(createIndex.getTable().getSchema(t.getUser()));

        //  check the name by putting an entry into the constraints or indexes table
        //      unique constraint will be (schema,indexname)

        //  TODO:  check to see if there is already an index on the given fields

        Table table = s.getTable(createIndex.getTable().getTable());

          //  used to protect the structure so we can verify the columns etc
        table.getUseLock().lockShared();
        try{
            table.lockTable(t, TableLockManager.SHARE_LOCK, false);

            Index index = new Index(this, s, createIndex.getIndexName(), table, fields, createIndex.isUnique());
            t.registerDoneCallback(new CreateIndexCallback(), index);

            s.addIndex(index);

              //  will be save because index creation writes it's own log message
            IndexStorage indexStore = sm.createIndexStorage(t, createIndex.isUnique());
            index.setStorage(indexStore);
            //  TODO:  write log entry  what???

              //  insert into catalog data
            Integer indexIdObj = new Integer(indexStore.getIndexId());
            dataDictionary.insertIntoINDEXES(t, indexIdObj, createIndex.getIndexName(),
                    new Integer(table.getTableId()), Bool.getBool(createIndex.isUnique()));
            String[] columns = createIndex.getColumnsArray();
            for(int i = 0; i < columns.length; i++) {
                dataDictionary.insertIntoINDEX_COLUMNS(t, new Integer(-1), indexIdObj, new Short((short)i), columns[i]);
            }

              //  actually create the index structure
            table.createIndex(t, fields, index, createIndex.isUnique());

            //  put index into
        } finally {
            //  NOW called in the commit callback... to ensure safety until commit
            //table.getUseLock().unlockShared();
        }

        if(createIndex.getIndexName().equals("DIE")) {
            ErrorCode.throwException(ErrorCode.DEBUG_EXCEPTION);
        }
          //  TODO:  consider now grabbing the X lock or at least notifying prepared statements and other holders
          //  so that they can rewrite their plans
    }

    /**
     * inserts a row into the "USERS" table as well as creating the actual schema object etc...
     * TODO:  add and use the UseLock!  add drop schema!
     */
    public Schema createSchema(Transaction t, String schema) throws SQLException {
          //  technically this should throw an exception on a unique key violation if
          //  the name is already in use
        dataDictionary.insertIntoUSERS(t, schema, null, DateHelper.now());

        synchronized(schemas) {
            Schema s = (Schema)schemas.get(schema);

            if(s != null) {
                throw new SQLException("Schema '" + schema + "' already exists", "none", ErrorCode.SCHEMA_ALREADY_EXISTS);
            }

            s = new Schema(this, schema);
            schemas.put(schema, s);
            return s;
        }
    }

    /**
     * TODO:  eventually just use a sequence here
     */
    Object columnSequenceLatch = new Object();
    int seqNum = 0;

    /**
     * Inserts columns into the data dictionary
     */
    private void insertColumnsIntoDD(Transaction t, int tableId, CreateTableStatement createTable)
                throws SQLException {
        int firstSeq = -1;
          //  TODO:  eventually use a sequence for this
        synchronized(columnSequenceLatch) {
            firstSeq = seqNum;
            seqNum += createTable.getNumColumns();
        }

        for(int i = 0; i < createTable.getNumColumns(); i++) {
            ColumnDefinition colDef = createTable.getColumn(i);
            /*
            dataDictionary.insertRowIntoCOLUMNS(t, firstSeq + i, i, colDef.getFieldName(), colDef.getSqlType(),
                    colDef.getSize(), colDef.getScale(), colDef.isNullable(), tableId);
                    */
              //  TODO:  restore global unique ID, perhaps by fetching highest columnID on load?
            dataDictionary.insertIntoCOLUMNS(t, tableId, i, colDef);
        }
    }

    /**
     * Create the table in the underlying storage and write the data to the catalog.
     */
    public void createTable(Transaction t, CreateTableStatement createTable) throws SQLException {
          //  TODO:  lock the schema from being removed, maybe a light lock
        Schema s = getSchema(createTable.getTable().getSchema(t.getUser()));

        TableStorage ts = sm.createTableStorage(t);
        ts.setName(createTable.getTableName());  //  for debugging purposes

        dataDictionary.insertIntoTABLES(t, new Integer(ts.getTableId()),
                createTable.getTableName(), createTable.getSchema(), DateHelper.now());

          //  TODO:  should this command automatically cause the rest, thus making replication easier?
          //  if all Data Dictionary operations also did something to the other structures, that would
          //  be pretty neat
        insertColumnsIntoDD(t, ts.getTableId(), createTable);
        Table newTable = new Table(this, ts, createTable);

          //  prevents others that find the table from using it until it is certainly created (post commit)
        newTable.getUseLock().lockExclusive(t);

        //  TODO:  create all primary and other keys, indexes etc

        t.registerDoneCallback(new CreateTableCallback(), newTable);
        s.addTable(newTable);

        /*if(createTable.getTableName().equals("DIE")) {
            ErrorCode.throwException(ErrorCode.DEBUG_EXCEPTION);
        }
        */
    }

    /**
     *  Drop the table.  Gets the use lock to prevent any queries from planning themselves until
     * this commits.  Then they will fail because the table isn't around anymore.
     */
    public void dropTable(Transaction t, TableRef ref) throws SQLException {
        Schema s = getSchema(ref.getSchema(t.getUser()));
        Table table = s.getTable(ref.getTable());

          //  grab exclusive use lock, to prevent query plans from being made!
        table.getUseLock().lockExclusive(t);
          //  grab EXCLUSIVE LOCK on the table, in NOWAIT mode
        table.lockTable(t, TableLockManager.EXCLUSIVE_LOCK, true);

        s.removeTable(table);
        t.registerDoneCallback(new DropTableCallback(), table);

          //  TODO:  delete constraints, indexes etc TOO
        //  delete columns first, then table entry.
        deleteById(t, DataDictionary.COLUMNS_TABLE, "TABLE_ID", table.getTableId());
        deleteById(t, DataDictionary.TABLES_TABLE, "TABLE_ID", table.getTableId());

          //  drop all indexes as well
        ArrayList allIndexes = table.getAllIndexes();
        for(int i = 0; i < allIndexes.size(); i++) {
            Index index = (Index)allIndexes.get(i);
            _dropIndex(t, s, index);
        }

          //  technically we should wait for the pending query count to drop to 0 before nuking the table storage
        sm.dropTableStorage(t, table.getTableId());

        //ErrorCode.throwException(ErrorCode.DEBUG_EXCEPTION);
    }

    public void alterTable(Transaction t, AlterTableStatement alter) throws SQLException {
        Schema s = getSchema(alter.getTable().getSchema(t.getUser()));
        Table table = s.getTable(alter.getTable().getTable());
        table.getUseLock().lockExclusive(t);
        table.lockTable(t, TableLockManager.EXCLUSIVE_LOCK, true);

        ColumnDefinition colDef = alter.getNewColumn();
        if(colDef == null) {
            ErrorCode.throwException(ErrorCode.FUNCTION_NOT_COMPLETED, "ALTER TABLE w/ Foreign key constraints");
        }

        t.registerDoneCallback(new AlterAddColumnCallback(), new ColumnInfo(s.getName(), table.getName(), colDef));
        dataDictionary.insertIntoCOLUMNS(t, table.getTableId(), table.getNumColumns(), colDef);

        //ErrorCode.throwException(ErrorCode.FUNCTION_NOT_COMPLETED, "ALTER TABLE");
    }

    /**
     *  grabs the exclusive table use lock, so probably don't need the SHARE table DML lock
     *   since readers and writers alike will be excluded temporarily
     * @param t
     * @param schema will not be null (like TableRef can be)
     * @param indexName
     */
    public void dropIndex(Transaction t, String schema, String indexName) throws SQLException {
        Schema s = getSchema(schema);
        Index index = s.getIndex(indexName);
        index.getTable().getUseLock().lockExclusive(t);

        //Don't think we need this anymore:  index.getTable().lockTable(t, TableLockManager.SHARE_LOCK, false);

        _dropIndex(t, s, index);

        deleteById(t, DataDictionary.INDEX_COLUMNS_TABLE, "INDEX_ID", index.getIndexId());
        deleteById(t, DataDictionary.INDEXES_TABLE, "INDEX_ID", index.getIndexId());
    }

    /**
     * Do the work of actually dropping the index.
     * @param t
     * @param index
     */
    private void _dropIndex(Transaction t, Schema s, Index index) throws SQLException {
          //  first always must add the callback so that in any case the thing will get replaced
        t.registerDoneCallback(new DropIndexCallback(), index);

          //  remove it from the schema and table
        s.removeIndex(index);
        index.getTable().removeIndex(index);

        //  TODO: delete all appropriate catalog data

          //  drop the storage associated
        sm.dropIndexStorage(t, index.getIndexId());
    }

    /**
     * TODO:  journally, etc!
     * @param t
     * @param user
     * @param seq
     * @throws SQLException
     */
    public void createSequence(Transaction t, String user, CreateOrAlterSequenceStatement seq) throws SQLException {
        Schema s = getSchema(seq.getSchemaName(user));
        SequenceStorage seqStore = sm.createSequenceStorage(t, seq.getStart());

        Sequence sequence = new Sequence(this, seqStore, s.getName(), seq.getName(),
                seq.getIncrement(), seq.getCache().intValue(), seq.isFlush());

        sequence.lockExclusive(t);  //  equivalent to UseLock but slightly different implementation for performance
        t.registerDoneCallback(new CreateSequenceCallback(), sequence);

        //  TODO:  handle other features like min/max

        dataDictionary.insertIntoSEQUENCES(t, new Integer(seqStore.getSequenceId()), seq.getName(), s.getName(), seq.getIncrement(),
                seq.getMinValue(), seq.getMaxValue(), new Integer(seq.getCache().intValue()),
                Bool.getBool(seq.isCycle()), Bool.getBool(seq.isOrder()), Bool.getBool(seq.isFlush()));
        s.addSequence(sequence);
    }

    /**
     *
     * @param t
     * @param user
     * @param seq
     * @throws SQLException
     */
    public void alterSequence(Transaction t, String user, CreateOrAlterSequenceStatement seq) throws SQLException {
        ErrorCode.throwException(ErrorCode.FUNCTION_NOT_COMPLETED, "ALTER SEQUENCE");

        Schema s = getSchema(seq.getSchemaName(user));

        Sequence sequence = s.getSequence(s.getName(), true);
        sequence.lockExclusive(t);
        //TODO:  register alterSequnce callback

        //  TODO:  update the sequence structure, internally it will update
//        Sequence sequence = new Sequence(this, seqStore, s.getName(), seq.getName(),
  //              seq.getIncrement(), seq.getCache().intValue());


        //  TODO:  handle other features like min/max

         //  TODO:  must be an update statement!
        /*
        dataDictionary.insertIntoSEQUENCES(t, new Integer(seqStore.getSequenceId()), seq.getName(), s.getName(), seq.getIncrement(),
                seq.getMinValue(), seq.getMaxValue(), new Integer(seq.getCache().intValue()),
                Bool.getBool(seq.isCycle()), Bool.getBool(seq.isOrder()), Bool.getBool(seq.isFlush()));
        /*
        dataDictionary.insertRowIntoSEQUENCES(t, seqStore.getSequenceId(), seq.getName(), s.getName(), seq.getIncrement(),
                seq.getMinValue(), seq.getMaxValue(), seq.getCache(), seq.isCycle(), seq.isOrder(), seq.isFlush());
        */
    }

    public void dropSequence(Transaction t, String user, DropSequenceStatement drop) throws SQLException {
        Schema s = getSchema(drop.getSchema(user));
        Sequence seq = s.getSequence(drop.getName(), true);

        seq.lockExclusive(t);
        t.registerDoneCallback(new DropSequenceCallback(), seq);

        sm.dropSequenceStorage(t, seq.getSequenceId());
        deleteById(t, SEQUENCES_TABLE, "SEQUENCE_ID", seq.getSequenceId());
    }

    /**
     * Helper method, deletes data from the catalog by id.
     * TODO:  move this to dataDictionary
     * @param table
     * @param field
     * @param id
     */
    protected void deleteById(Transaction t, String table, String field, long id) throws SQLException {
        queryEngine.executeDelete(t, DataDictionary.SYS_USER, new DeleteStatement(table,
                SqlCondition.fieldEqualsLiteral(field, new BigDecimal(id))));
    }

    /**
     * Truncate the table completely.  Grabs the Exclusive use lock to ensure there is no corruption.
     * @param t
     * @param ref
     * @throws SQLException
     */
    public void truncateTable(Transaction t, TableRef ref) throws SQLException {
        Table table = getTable(ref, t.getUser());

          //  prevent query plans from being made
        table.getUseLock().lockExclusive(t);

          //  if the above is in place, do we need this???
        table.lockTable(t, TableLockManager.EXCLUSIVE_LOCK, true);

        table.getStorage().truncateTable(t);

          //  truncate indexes  (NOTE:  no specific reason these are after the table truncate)
        ArrayList list = table.getAllIndexes();
        for(int i = 0; i < list.size(); i++) {
            Index index = (Index)list.get(i);
            index.getStorage().truncateIndex(t);
        }
    }

    /*********************************************************
     * Log processing methods here
     */

    /**
     * This method will get rewritten when we go to a table numbering scheme.
     */
    public void processLogEntry(int applyMode, long lsn, Transaction t, LogEntry logEntry) throws SQLException {
        if(logEntry instanceof TableLogEntry) {
            TableLogEntry logE = (TableLogEntry)logEntry;

            System.out.println("Catalog.processLogEntry: ignored log entry: " + logE);

            //  TODO:  whatever needed, make this work
            //Table t = getTable(logE.getSchemaName(), logE.getTableName());
            //t.processLogEntry(applyMode, transactionId, logEntry);
        } else {
            System.out.println("Catalog.processLogEntry: ignored log entry: " + logEntry);
        }
    }

    /****************************************************************************
     * Transaction completion callbacks for DDL operations
     * TODO:  Many of these could be replaced with a mechanism that would get callbacks
     * from specific tables so that the library cache (data dictionary whatever)
     * would be kept in sync.  This would also allow replication to be easier.
     * In a way, this is avoiding a somewhat simpler solution of just adding another log message
     * though I must admit it seems silly to have 2 log messages for the same thing (in addition
     * to the actual row in the catalog).  Plus, this way we can actually keep our internal data
     * structure without having to rebuild it at all.
     *
     * NOTE:  these could all be in Schema!
     */

    //  TABLE callbacks

    /**
     * used only on rollback, ensures that the table reference is dele
     * TODO:  finish drop etc && sequences etc too
     */
    class CreateTableCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            if(finalStatus == Transaction.ROLLED_BACK) {
                try{
                    Table table = (Table)callbackData;
                    getSchema(table.getSchema()).removeTable(table);
                }catch(SQLException s) {
                    throw new OhShitException("failed to undo create table", s);
                }
            }
        }
    }

    /**
     * put the table back in if the drop fails.  Rare, needs to be explicitly tested.
     * TODO:  consider the possibility of leaving in explicit testing exceptions
     */
    class DropTableCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            if(finalStatus == Transaction.ROLLED_BACK) {
                try{
                    Table table = (Table)callbackData;
                    getSchema(table.getSchema()).addTable(table);
                }catch(SQLException s) {
                    throw new OhShitException("failed to undo drop table", s);
                }
            }
        }
    }

    /**
     * adds the column definition in!
     */
    class AlterAddColumnCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            if(finalStatus == Transaction.COMMITED) {
                try{
                    ColumnInfo col = (ColumnInfo)callbackData;
                    getSchema(col.getSchemaName()).getTable(col.getTableName()).addColumn(col);
                }catch(SQLException s) {
                    throw new OhShitException("failed to add a column to a table", s);
                }
            }
        }
    }

    //  INDEX callbacks

    class CreateIndexCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            Index index = (Index)callbackData;

            if(finalStatus == Transaction.ROLLED_BACK) {
                try{
                    index.getSchema().removeIndex(index);
                    index.getTable().removeIndex(index);
                }catch(SQLException s) {
                    throw new OhShitException("failed to undo create index", s);
                }
            }

              //  commit or rollback, unlock!
            index.getTable().getUseLock().unlockShared();
        }
    }

    class DropIndexCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            Index index = (Index)callbackData;

            if(finalStatus == Transaction.ROLLED_BACK) {
                try{
                    index.getSchema().addIndex(index);
                    index.getTable().addIndex(index);
                }catch(SQLException s) {
                    throw new OhShitException("failed to undo drop index", s);
                }
            }
        }
    }

    //  SEQUENCE callbacks

    class CreateSequenceCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            Sequence seq = (Sequence)callbackData;

            if(finalStatus == Transaction.ROLLED_BACK) {
                try{
                    getSchema(seq.getSchema()).removeSequence(seq);
                }catch(SQLException s) {
                    throw new OhShitException("failed to undo create sequence", s);
                }
            }
        }
    }

    /**
     * does the actual dropping.  This pattern was used rather than undoing the drop
     * for whatever reason, but it should work at least as well.
     */
    class DropSequenceCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            if(finalStatus == Transaction.COMMITED) {
                try{
                    Sequence seq = (Sequence)callbackData;
                    getSchema(seq.getSchema()).removeSequence(seq);
                    seq.dropSequence();
                }catch(SQLException s) {
                    throw new OhShitException("failed to undo drop sequence", s);
                }
            }
        }
    }
}
