/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Dec 28, 2001
 * Time: 2:16:13 PM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.catalog;

import freebase.api.IndexCursor;
import freebase.api.IndexStorage;
import freebase.api.LogEntryProcessor;
import freebase.engine.IndexScan;
import freebase.engine.ScanRegionList;
import freebase.engine.Transaction;
import freebase.log.LogEntry;
import freebase.sql.SqlCondition;
import freebase.sql.SqlExpression;
import freebase.sql.expression.BasicCondition;
import freebase.sql.expression.BetweenCondition;
import freebase.sql.expression.ColumnExpression;
import freebase.sql.expression.LiteralExpression;
import freebase.util.SqlTypes;

import java.sql.SQLException;

/**
 * Holds the information about an index.  Eventually this should definitely be an
 * interface so that multiple different indexing implementations can sit side by
 * side.  This is called directly to modify the index values because eventually the index
 * will actually be distributed across multiple nodes, and will be called through a wrapper.<BR><RB>
 *
 * TODO:  eventually think about dealing with dry indexes (no support for multiple versions)
 *         and indexes that do not deal with internal housekeeping themselves but support it
 *         externally
 *
 * TODO:  currently indexes are hardcoded to not index NULL.  Technically there is no reason for this
 *   in the future.  Start to fix it by changing extractValue
 */
public class Index implements LogEntryProcessor {
    Catalog catalog;
    Schema schema;
    String name;
    Table table;
    IndexStorage storage;
    String[] fields;
    boolean unique;

    /**
     * Construct the index structure for runtime
     */
    public Index(Catalog catalog, Schema schema, String name, Table table, String[] fields, boolean unique) {
        this.catalog = catalog;
        this.schema = schema;
        this.name = name;
        this.table = table;
        this.fields = fields;
        this.unique = unique;
    }

    public Schema getSchema() {
        return schema;
    }

    public void setStorage(IndexStorage storage) {
        this.storage = storage;
    }

    public IndexStorage getStorage() {
        return storage;
    }

    public String getName() {
        return name;
    }

    /**
     * get the id code for this one
     */
    public int getIndexId() {
        return storage.getIndexId();
    }

    /**
     * Returns the table that this index is associated with
     */
    public Table getTable() {
        return table;
    }

    public int getNumFields() {
        return fields.length;
    }

    public String getField(int pos) {
        return fields[pos];
    }

    /**
     * Similar to Table's getTableCursor, this will eventually only be used internally
     */
    protected IndexCursor getIndexCursor(Transaction t, boolean readWrite) throws SQLException {
        return storage.getIndexCursor(t, readWrite);
    }

    /**
     * Return an object that will do a range scan within the range specified.
     * TODO:  eventually this will need to support more than a simple min/max range.  It will
     *       need to support some kind of arbitrary expression
     */
    public IndexScan getIndexScan(Transaction t, boolean readWrite, ScanRegionList scanRegions) throws SQLException {
        return new IndexScan(t, this, getIndexCursor(t, readWrite),
                table.getTableTupleIterator(t, readWrite),  scanRegions);
    }

    /**
     * Helper method, pass in a row of values and this method returns an object
     * appropriate for use as the index value.  Could be called extractIndexValue.
     *
     * TODO:  When doing this many times is there a way to cache the current positions of the fields
     *     so that it is faster.  Technically this might be doable globally, but how can it update?
     * TODO:  this index could be a listener that invalidates its copy of the positions of the columns.  steps:
     *      register as a listener, but set the valid flag to false.  When the shared lock is next grabbed, test
     *     the valid flag.  If not, lookup the values (right down here!) and cache em.
     *      Note:  when being notified of invalid column positions, re-register for next time.
     *      Or, just have the table explicitly call this when its
     * TODO:  eventually input a tuple, and return a tuple
     * @return null if all the values are null
     */
    public Object extractValue(Comparable[] rowData) throws SQLException {
        if(fields.length == 1) {
            // System.out.println("FIELD IS : " + fields[0]);
            return rowData[table.getColumnOffset(fields[0])];
        } else {
            int numNulls = 0;
            Comparable[] extraction = new Comparable[fields.length];
            for(int i = 0; i < fields.length; i++) {
                extraction[i] = rowData[table.getColumnOffset(fields[i])];
                if(extraction[i] == null) {
                    numNulls++;
                }
            }

            if(numNulls == fields.length) {
                return null;
            }
            return extraction;
        }
    }


    /**
     * Returns true if this index can be used to quicky find entries that satisfy the
     * condition passed in.
     * TODO:  support AND conditions, with ranges etc, support mirror of current 1=id
     *    perhaps have all anded expressions flattened, and thus on the same level
     * TODO:  eventually this needs to be delegated out to another object
     * TODO:  Eventually return some kind of cost.
     *
     * Eventually this is the heart of high performance... make it ROCK!!!
     *
     * @return null if there is no good fit for a scan region
     */
    public ScanRegionList findScanRegions(SqlCondition condition) {
          //  TODO:  expect a flattened condition list or array
        if(condition == null) {
            return null;
        }

          //  TODO:  support   5 = a!  not so hard :)
        if(condition instanceof BasicCondition) {
            BasicCondition bc = (BasicCondition)condition;
            if(canUseIndex(bc.getLeftExpression()) &&
                        bc.getRightExpression() instanceof LiteralExpression ) {
                LiteralExpression val = (LiteralExpression)bc.getRightExpression();

                ScanRegionList scanRegion = new ScanRegionList();
                if(bc.getOperator() == BasicCondition.EQUALS) {
                    scanRegion.addScanRegion(val.getValue(), val.getValue());
                } else if(bc.getOperator() == BasicCondition.GREATER_THAN ||  //  TODO:  set bounds as exclusive
                          bc.getOperator() == BasicCondition.GREATER_THAN_OR_EQUALS) {
                    scanRegion.addScanRegion(val.getValue(), null);
                } else if(bc.getOperator() == BasicCondition.LESS_THAN ||    //  TODO:  set bounds as exclusive
                          bc.getOperator() == BasicCondition.LESS_THAN_OR_EQUALS) {
                    scanRegion.addScanRegion(null, val.getValue());
                } else {
                      //  not too good an implementation!!
                    return null;
                }

                return scanRegion;
            }
        } else if(condition instanceof BetweenCondition) {
            BetweenCondition bc = (BetweenCondition)condition;

            if(canUseIndex(bc.getLeft()) && bc.getMin() instanceof LiteralExpression &&
                                            bc.getMax() instanceof LiteralExpression) {
                ScanRegionList scanRegion = new ScanRegionList();
                scanRegion.addScanRegion(((LiteralExpression)bc.getMin()).getValue(),
                                         ((LiteralExpression)bc.getMax()).getValue());
                return scanRegion;
            }
        }

        return null;
    }

    /**
     * True if the expression is a column expression and is the first part of this index
     * @param x
     * @return
     */
    private boolean canUseIndex(SqlExpression x) {
        if(x instanceof ColumnExpression) {
            ColumnExpression ce = (ColumnExpression)x;

            if(ce.getColumn().getColumn().equals(fields[0])) {
                return true;
            }
        }

        return false;
    }

    /**
     * Deletes the entry with the given specification.  Seems like these cursors could be recycled a bit
     * better, perhaps pooled or something.
     */
    public void deleteEntry(Transaction t, Comparable[] rowData, long rowId) throws SQLException {
        //System.out.println("ROW DATA IS: " + rowData[0] + " " + rowData[1]);

          //  TODO:  could extract value posibly not create a new object every time?
        Object value = extractValue(rowData);
          //  value is null means not in the index
        if(value != null) {
            IndexCursor indexCursor = getIndexCursor(t, true);
            try{
                _deleteEntry(indexCursor, value, rowId);
            } finally {
                indexCursor.close();
            }
        }
    }

    /**
     * Does the actual work of deleting the entries.  Is called by deleteEntry and updateEntry
     */
    private void _deleteEntry(IndexCursor indexCursor, Object value, long rowId)
            throws SQLException {
        indexCursor.setPos(value);

          //  scan for our exact row
        while(indexCursor.getCurRowId() != rowId) {
            if(! indexCursor.next()) {
                //  TODO:  technically this is a critical error
                throw new SQLException("Failed to find ourselves in the index!!");
            }
        }

        indexCursor.deleteEntry();
    }

    /**
     * Inserts an entry with the given specification.
     */
    public void insertEntry(Transaction t, Comparable[] rowData, long rowId) throws SQLException {
        Object value = extractValue(rowData);
        if(value != null) {
            IndexCursor indexCursor = getIndexCursor(t, true);
            try{
                _insertEntry(indexCursor, value, rowId);
            } finally {
                indexCursor.close();
            }
        }
    }

    /**
     * internal insert entry, is called from insertEntry() and updateEntry()
     */
    private void _insertEntry(IndexCursor indexCursor, Object value, long rowId)
            throws SQLException {
        indexCursor.insertEntry(value, rowId);
    }

    /**
     * Currently just deletes and inserts.  Eventually may mark the old one as UPDATED.
     * This checks that the old value and the new value are not the same for this index...
     * it might be faster TODO this check externally but...
     */
    public void updateEntry(Transaction t, Comparable[] oldRowData, Comparable[] newRowData, long rowId)
            throws SQLException {
        Object oldValue = extractValue(oldRowData);
        Object newValue = extractValue(newRowData);

          //  compare the values.  If not equal, return
        if(oldValue instanceof Comparable[]) {
            if(IndexScan.compareIndexValues((Comparable[])oldValue, (Comparable[])newValue) == 0) {
                return;
            }
        } else {
            if(SqlTypes.compare((Comparable)oldValue, (Comparable)newValue) == 0) {
                return;
            }
        }

        if(oldValue != null || newValue != null) {
            IndexCursor indexCursor = getIndexCursor(t, true);
            try{
                if(oldValue != null) _deleteEntry(indexCursor, oldValue, rowId);
                if(newValue != null) _insertEntry(indexCursor, newValue, rowId);
            } finally {
                indexCursor.close();
            }
        }
    }

    public void processLogEntry(int applyMode, long lsn, Transaction t, LogEntry logEntry)
            throws SQLException {
        //storage.pro
        System.out.println("Index.processLogEntry called but not implemented");
    }
}
