/*
 * Copyright JJ Furman
 * User: user
 * Date: Mar 11, 2002
 * Time: 9:48:11 PM
 */
package freebase.catalog;

import freebase.api.SequenceStorage;
import freebase.api.TransactionDoneCallback;
import freebase.api.TypedTupleIterator;
import freebase.engine.Transaction;
import freebase.util.ErrorCode;
import freebase.util.OhShitException;

import java.math.BigInteger;
import java.sql.SQLException;

/**
 * Handles sequence information.  Uses storage to store the underlying value.  Handles its own locking similar
 * but subtly different that UseLock
 */
public class Sequence implements CatalogDD.SequencesColumns {
    Catalog catalog;

    SequenceStorage storage;

    String schema;
    String name;

    BigInteger nextVal;
    BigInteger increment;

    /**
     * When this value goes to zero, a log record will be written out to the actual table defining what
     * the next next is so that upon reboot it will reflect the correct value.  The exclusivity of this
     * will be enforced by a local latch.
     */
    int numRemainingInCache;

    int cacheSize;

    /**
     * equal to increment * cacheSize, this is the interval that the log will be written in
     */
    BigInteger jumpAmount;

    boolean isDropped = false;

    boolean flush;

    Transaction isLockedExclusivelyBy = null;

    FinishCallback callbackHelper = new FinishCallback();

    /**
     * TODO:  support max/min value, cycle/nocycle, nocache, and order  (add to other constructor toO!!)
     */
    public Sequence(Catalog catalog, SequenceStorage storage, String schema, String name,
                    BigInteger increment, int cacheSize, boolean flush) throws SQLException {
        this.catalog = catalog;
        this.storage = storage;
        this.schema = schema;
        this.name = name;
        this.nextVal = storage.getValue();
        this.increment = increment;
        this.cacheSize = cacheSize;
        this.flush = flush;
        numRemainingInCache = 0;

        jumpAmount = increment.multiply(BigInteger.valueOf(cacheSize));
    }

    /**
     * Construct a sequence object from a catalog sequence tuple
     * @param seqTuple
     */
    public Sequence(Catalog catalog, SequenceStorage storage, TypedTupleIterator seqTuple) throws SQLException {
        this(catalog, storage, seqTuple.getString(SCHEMA_OFFSET), seqTuple.getString(NAME_OFFSET),
                seqTuple.getBigInteger(INC_OFFSET), seqTuple.getInt(CACHE_OFFSET),
                seqTuple.getBoolean(FLUSH_FLAG_OFFSET));
    }

    public SequenceStorage getStorage() {
        return storage;
    }

    public int getSequenceId() {
        return storage.getSequenceId();
    }

    public String getName() {
        return name;
    }

    public String getSchema() {
        return schema;
    }

    /**
     * This simply sets the sequence as dropped so that readers about to get values will get a rude awakening!
     * Called after commit and before releasing the lock
     */
    public void dropSequence() {
        isDropped = true;
    }

    /**
     * Lock this sequence exclusively for dropping or altering.
     */
    public synchronized void lockExclusive(Transaction t) throws SQLException {
        waitUntilNotLocked();
        isLockedExclusivelyBy = t;
        t.registerDoneCallback(callbackHelper, null);
    }

    private void waitUntilNotLocked() throws SQLException {
        while(true) {
            if(isLockedExclusivelyBy == null) {
                if(isDropped) {
                    ErrorCode.throwException(ErrorCode.SEQUENCE_DROPPED);
                }
                return;
            }

            try{
                wait();
            }catch(InterruptedException i) {
                throw new SQLException("thread interrupted while waiting for lock (rare indeed!)");
            }
        }
    }

    /**
     * Gives the passed transaction a value for the sequence.  It will be stored in the transaction itself for now,
     * though later it could be stored in the sequence manager if a transaction commit/rollback event gets fired
     * for cleanup.
     *
     * TODO:  maybe make this a use lock instead (or optimize by checking for the shared lock and keeping the sync
     *    all right here!).  Ie, subclass from UseLock here and use that subclass, but let the rest of the world
     *    still see UseLock objects for consistency
     */
    public synchronized BigInteger getNextValue(Transaction t) throws SQLException {
        waitUntilNotLocked();

        if(numRemainingInCache == 0) {
            nextVal = storage.getValue();
              //  if we fail, the recovery should pick up a value outside of this cached group
            BigInteger journalNextValue = jumpAmount.add(nextVal);
            storage.setValue(journalNextValue, flush);
            numRemainingInCache = cacheSize;
        }

        BigInteger returnValue = nextVal;
        t.setSequenceValue(schema, name, nextVal);
        nextVal = nextVal.add(increment);
        numRemainingInCache--;

        return returnValue;
    }

    public BigInteger getCurValue(Transaction t) {
        return t.getSequenceValue(schema, name);
    }

    /***************************************************************************
     * Transaction callbacks
     */

    /**
     * Used to release the exclusive lock
     */
    class FinishCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
              //  unlock
            synchronized(Sequence.this) {
                if(isLockedExclusivelyBy != t) {
                    throw new OhShitException("WE didn't have the lock!!!!!");
                }

                isLockedExclusivelyBy = null;
                Sequence.this.notifyAll();
            }
        }
    }
}
