/*
 * Copyright JJ Furman
 * User: user
 * Date: Mar 22, 2002
 * Time: 9:58:36 AM
 */
package freebase.engine;

import java.sql.Connection;

/**
 * A Consistent View consists of a Transaction, a scn (view time), and a local log position.<BR><BR>
 *
 * A particular row is considered in our view if:
 *     it has been recently modified by our transaction and it's log position was before or equal to our current
 *  -or-
 *     the transaction that wrote the last change is committed and it's scn (view time) is less or equal to our scn<BR><BR>
 *
 * The view is shared by the transaction and a number of cursors.  Each increments the use count to prevent it from
 * being given away, and decrements the use count when done.  When the use count goes to zero the view is dissolved
 * by removing it from the system.  Any transaction information being held for it can then also be released.<BR><BR>
 *
 * TODO: consider the ability for the system to POP views.  If it is determined that a view is just too
 * old then the system should have the ability to BREAK a view, thus destroying success for the operation,
 * though not necessarily in all cases the transaction.
 */
public class ConsistentView {
    /**
     * Transaction of the cursor using this View
     */
    Transaction t;

    /**
     * Equivalent to SCN for oracle folks
     */
    long viewTime;

    /**
     * logPos for local transaction modifications
     */
    long logPos;

    /**
     * When use count goes to zero this view is released
     */
    volatile int useCount = 0;

    /**
     * Construct a consistent view or snapshot of data
     */
    public ConsistentView(Transaction t, long viewTime, long logPos) {
        this.t = t;
        this.viewTime = viewTime;
        this.logPos = logPos;

        //System.out.println("View created: " + this);
        //new Exception("CONSISTENT VIEW CREATED").printStackTrace();
    }

    /*  SOME NOTES:

    // checks to see if this view is still legitimate.
    public boolean isViewLegal() {
        //  could check to see if we are higher than the lowest legal view
        //    in the ViewManager.  Otherwise, the view manager could explicitly
        //    set this as invalid
    }
    */

    /**
     * Called in the case of repeatable reads and serializable at the beginning of a new statement
     */
    public void updateLogPosition(long lsn) {
        this.logPos = lsn;
    }

    public long getViewTime() {
        return viewTime;
    }

    /**
     * Logic for determining if this view can this see the changes from the
     * passed transaction and state write?
     */
    public boolean canSeeChangesFrom(Transaction other, long otherLogPos) {
          /**
           * TODO:  consider the case that the trans is in READ UNCOMMITTED mode
           */

        /**
         * If the change was ours, we don't see it even if the other guy is already
         * committed.
         */
        if(t == other) {
            return otherLogPos <= logPos;
        }

          //  are they so old, or are we very nosey
        if(other == null ||
                t.getIsolationLevel() == Connection.TRANSACTION_READ_UNCOMMITTED) {
            return true;
        }

          //  if the other is not finished yet, we can't see it's changes
        if(other.isActive()) {
            return false;
        }

        /**
         *  TODO:  make a switch statement.  Deal with PREPARED read blocking here
         * if the finish time of the other transaction is such that we won't see the
         * results anyway, then we should return false
         */

        if(other.getStatus() == Transaction.ROLLED_BACK) {
            System.out.println("checking against a rolled back transaction");
        }

        long otherFinishTime = other.getFinishTime();
        if(otherFinishTime == -1)  //  TODO:  get rid of this at some point
            return false;

          //  if we grabbed a view time equal or after it's finish time, yes
        if(otherFinishTime <= viewTime) {
            return true;
        }

          //  we must have a view time less that it's finish time, so no
        return false;
    }

    /**
     * Makes sure that this view is no longer active.
     */
    private void dissolveView() {
        t.getTM().getViewManager().removeView(this);
    }

    public void incrementUseCount() {
        useCount++;
    }

    public void decrementUseCount() {
        useCount--;

        if(useCount == 0) {
            dissolveView();
        }

        if(useCount < 0) {
            System.out.println("ALERT! use count going negative!!!");
        }
    }

    public String toString() {
        return "(View on " + t + ", view time:" + viewTime + ", log pos:" + logPos + ")";
    }
}
