/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 3, 2002
 * Time: 10:10:05 PM
 */
package freebase.engine;

import freebase.api.ColumnInfo;
import freebase.api.IndexCursor;
import freebase.api.LockableTupleIterator;
import freebase.catalog.Index;
import freebase.server.ExecutionPlan;
import freebase.util.SqlTypes;

import java.sql.SQLException;

/**
 * The index scan is the location for code that applies to all index scans in general.
 * The index cursor itself should be able to keep state about the index specific location
 * but the scan will do the job of validating that a row should indeed be part of the
 * return set.<BR><BR>
 *
 * For some index types, like hash, this will still be used but technically "scanning" will
 * not occur
 */
public class IndexScan implements LockableTupleIterator {
    Transaction t;
    Index index;
    IndexCursor indexCursor;
    TableTupleIterator tableCursor;

      //  todo:  support a series of disjoin scan regions
    Comparable[] min;
    Comparable[] max;

      //  currently for use in the explain plan
    ScanRegionList scanRegions;

    /**
     * Derived map from table columns to index columns.  -1 in an entry means
     * that the column is not supported in the index
     */
    int[] columnOffsets;

    /**
     * map from index pos to actual table column
     */
    int[] tableOffsets;

    /**
     * Set to true when the first entry has been hit
     */
    boolean initialized = false;

    /**
     * Current index value.  TODO:  get rid of this, only used in one method!!
     */
    Comparable[] recentValue = null;

    /**
     * True when the table cursor has been consulted and the data has been loaded.
     * This will happen when a column not in the index is needed, or when an entry
     * is in question
     */
    boolean tableCursorCurrent = false;

    //  store:  offsets of the index columns, in case no other columns are requested.
    //  this will make the lookups much faster

    /**
     * Null's are not indexed for now.  Min of null means the beginning, and max null means the end.
     * Null values within the actual array may be a different story.
     *
     * @param scanRegions currently only supports ONE scan region, to change later!!  null means full index scan
     */
    public IndexScan(Transaction t, Index index, IndexCursor cursor, TableTupleIterator tableCursor,
                     ScanRegionList scanRegions) {
        this.t = t;
        this.index = index;
        this.indexCursor = cursor;
        this.tableCursor = tableCursor;
        this.scanRegions = scanRegions;

          //  TODO:  support multiple scan regions
        if(scanRegions != null) {
            this.min = scanRegions.getMin(0);
            this.max = scanRegions.getMax(0);
        }
        //this.columnOffsets = columnOffsets;

        //  TODO:  grab a consistent view or something, still need to work this out

        recentValue = new Comparable[index.getNumFields()];

        //  Determine what the offsets are of the index fields are and create a reverse map
        //  from table offset to index offset.  Use this map to determine if it is necessary
        //  to go into the tableCursor or not for requests
        columnOffsets = new int[index.getTable().getNumColumns()];
        tableOffsets = new int[index.getNumFields()];
        for(int i = 0; i < columnOffsets.length; i++) {
            columnOffsets[i] = -1;
        }

        for(int i = 0; i < index.getNumFields(); i++) {
            tableOffsets[i] = index.getTable().getColumnOffset(index.getField(i));
            columnOffsets[tableOffsets[i]] = i;
        }

        //  Now I gotta stand with the Glock in my hand
    }

    /**
     * Pass on to the table Cursor
     * @throws SQLException
     */
    public void lockRow() throws SQLException {
        tableCursor.lockRow();
    }

    /**
     * Pass on to the table Cursor
     */
    public void unlockRow() throws SQLException {
        tableCursor.unlockRow();
    }

    /**
     * Reset this to the uninitialized state
     */
    public void reset() throws SQLException {
        tableCursor.reset();
        initialized = false;
    }

    /**
     * Used by nested loops to dynamically change the usage of this index scan.
     * Also resets the cursor to this range.
     */
    public void setSingleKeyRange(Comparable[] key) throws SQLException {
        min = key;
        max = key;
        reset();
    }

    /**
     * Get the next value from this index.  Will traverse the entire index until the end.
     * They come off in sorted order.
     */
    public boolean next() throws SQLException {
        while(true) {
            tableCursorCurrent = false;

              //  fetch the next (or first) entry from an index cursor
            boolean foundAny;
            if(! initialized) {
                  //  if there is no min, just start scanning the cursor

                if(min == null) {
                    foundAny = indexCursor.next();  //  get first
                } else {
                    if(recentValue.length == 1) {
                        foundAny = indexCursor.setPos(min[0]);
                    } else {
                        foundAny = indexCursor.setPos(min);
                    }
                }
                //System.out.println("Found any: " + foundAny);
                initialized = true;
            } else {
                foundAny = indexCursor.next();
            }

            if(! foundAny)
                return false;

            for(int i = 0; i < recentValue.length; i++) {
                recentValue[i] = indexCursor.getCurValue(i);
            }

              //  check to see if we are still in range (if max==null there is no upper limit)
            if(!(max == null) && compareIndexValues(recentValue, max) > 0) {
                //  NOTE:  the tableCursor may have a pending write.  This is a bit problematic.
                // One solution to ensure it gets written is to send a message to the lower level.
                //  Current solution is to ensure that .close is called.  It really matters
                //  for checking unique key constraints.
                return false;
            }


              //  if the entry is possibly deleted or possibly not commited yet
            //if(indexCursor.isPossiblyNotVisible()) {

              //  is the row even there at all?  If not, just move on.
              //  note, this can be false even if the current status flags is NORMAL in the case of a rollback
              //  of a delete and a specific race condition
            if(makeTableCursorCurrent()) {
                //  TODO:  If we had more information from the index than possibly
                //  not visible, such as was updated, was deleted, or was inserted
                //  then we can save this last lookup

                  //  check the current value.  This is strictly to ensure ordering.  When being modified,
                  //  a row will appear in two places
                boolean valueIsCorrect = true;
                for(int i = 0; i < index.getNumFields(); i++) {
                    //System.out.println("comparing index value: " + recentValue[i] + "  with  " + tableCursor.getValue(tableOffsets[i]));

                    if(SqlTypes.compareNullsEqual(recentValue[i], tableCursor.getValue(tableOffsets[i])) != 0) {
                        valueIsCorrect = false;
                        break;
                    }
                }

                if(valueIsCorrect) {
                    return true;
                } else {
                    //System.out.println("Table Value not [" + StringUtil.dump(recentValue) + "]");
                    //  but not error?
                    //  this needs more thinking through
                }
            }
        }
    }

    /**
     * Move the table cursor to the position defined by the index cursor.
     * @return true if the row even exists (at this location), false if not
     */
    private boolean makeTableCursorCurrent() throws SQLException {
        if(! tableCursorCurrent) {
            tableCursorCurrent = true;
            return tableCursor.gotoRow(indexCursor.getCurRowId());
        }

        return true;
    }

    /*************************************************************************
     *  Tuple Iterator methods here
     */
    public void close() throws SQLException {
        //  TODO:  close this TOO!

        indexCursor.close();
        tableCursor.close();
    }

    public Comparable getValue(int i) throws SQLException {
        //if(columnOffsets[i] != -1) {  //  THIS IS WRONG, but left here in case there is some value in it
        //    return recentValue[i];
        //} else {
            //makeTableCursorCurrent();  //  NOTE:  pretty sure this is never needed anymore, since cursor must be on spot always now!

            return tableCursor.getValue(i);

        //}
    }

    public void setValue(int i, Comparable value) throws SQLException {
        tableCursor.setValue(i, value);
    }

    public int getNumColumns() throws SQLException {
        return tableCursor.getNumColumns();
    }

    public String getColumnName(int i) throws SQLException {
        return tableCursor.getColumnName(i);
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return tableCursor.getColumnInfo(i);
    }

    public boolean isUpdateable() throws SQLException {
        return tableCursor.isUpdateable();
    }

    public void deleteTuple() throws SQLException {
        tableCursor.deleteTuple();
    }

    /**
     * Returns a plan for this query.
     */
    public ExecutionPlan getExecutionPlan() throws SQLException {
        String fieldString = null;
        if(index.getNumFields() == 1) {
            fieldString = index.getField(0);
        } else {
            StringBuffer sb = new StringBuffer();
            for(int i = 0; i < index.getNumFields(); i++) {
                if(i != 0)
                    sb.append(',');
                sb.append(index.getField(i));
            }

            fieldString = sb.toString();
        }

        String scanRegionsString = "";
        if(scanRegions != null) {
            scanRegionsString = " " + scanRegions.toString();
        }
        return new ExecutionPlan("Index Scan [" + index.getTable().getName() + "(" + fieldString + ")]" + scanRegionsString);
    }
    /*************************************************************************
     *
     */


    /**
     *  Same as java compare, but compares these entire values, even if they are arrays of values.
     * This could just go on into SqlTypes
     * NOTE:  uses the length of the second value (b), because in one calling case b is the max and the
     * max can have fewer entries than the index has.  In those cases, the value a is found to be
     * "sufficiently equal" to b for the purposes of scanning.  This is a messy artifact, so I wanted to
     * document it.
     */
    public static int compareIndexValues(Comparable[] a, Comparable[] b) {
        for(int i = 0; i < b.length; i++) {
            int comp = SqlTypes.compareNullsEqual(a[i], b[i]);
            if(comp != 0) {
                return comp;
            }
        }

        return 0;
    }
}
