/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Nov 10, 2001
 * Time: 6:01:56 PM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.engine;

import freebase.api.*;
import freebase.catalog.Index;
import freebase.catalog.Table;
import freebase.server.ExecutionPlan;
import freebase.util.ErrorCode;
import freebase.util.StringUtil;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Classic Table Scan / data access interface.  The role of the table tuple iterator (for now) is
 * to perform all of the index updates and constraint checking.  This relieves the actual
 * cursor itself to deal with the actual storage details.  When we get index scans going some
 * of the functionality here might get mixed into the IndexTupleIterator
 *
 * Possibly rename this to TableTupleDelegate, Adapter or Interface, or TableScan
 *
 * This is the class that is going to convert data into byte[],
 *  perform the direct TableScan... and perform:
 *
 *  update() {
 *      lockRow();
 *      checkBasicConstraints();
 *      checkExternalConstraints();
 *      updateData in TableCursor
 *      update all indexes.
 *  }
 */
public class TableTupleIterator implements LockableTupleIterator {
    /**
     * passed fieldscurRowData
     */
    protected Transaction t;
    protected Table table;

    /**
     * extracted fields
     */
    TableCursor cursor;
    TableLockManager locks;

    //  TODO:  eventually do not want to materialize all data
    //  just to get any one value.  Delegate this down.
    boolean isDataLoaded = false;
    boolean isDirty = false;

    /**
     * keeps the true previous value in case of an update, thus allowing the indexes to be updated<BR>
     * Only used in the case of an overwrite
     */
    Comparable[] prevRowData;

    /**
     * Current row data, the primary data used unless there is an overwrite
     */
    Comparable[] curRowData;

    ColumnInfo[] columnInfo;

    /**
     * TODO:  pass in the cursor for which this Iterator is supposed to walk.
     * TODO:  think about implementing some group operations, especially for reading quickly,
     * and holding locks for shorter time
     */
    public TableTupleIterator(Transaction t, Table table, boolean readWrite) throws SQLException {
        this.t = t;
        this.table = table;

        cursor = table.getStorage().getTableCursor(t, readWrite);
        locks = table.getLockManager();

          //  TODO:  eventually grab a row object from the table
        curRowData = new Comparable[table.getNumColumns()];
        prevRowData = new Comparable[table.getNumColumns()];

          //  get the column info, since we have the Use lock in Shared mode (required!!)
        columnInfo = table.getColumnInfos();
    }

    /**
     * moves this to the underlying cursor to this rowid.
     * Called from IndexScan.  TODO:  make sure that any locks are released
     * @return true if there is a row there
     */
    public boolean gotoRow(long rowId) throws SQLException {
        if(isDirty) {
            writeDirtyRow();
        }

        isDataLoaded = false;
        return cursor.setCurrentRowNum(t, rowId);
    }

    public void reset() throws SQLException {
        if(isDirty) {
            writeDirtyRow();
        }

        cursor.reset();
        isDataLoaded = false;
    }

    /***************************************
     * TableIterator methods
     */

    /**
     * Writes out dirty pending data, and trys to go to the next row.
     * @return
     * @throws SQLException
     */
    public boolean next() throws SQLException {
        if(isDirty) {
            writeDirtyRow();
        }

        boolean hasMore = cursor.next(t);
        if(! hasMore)
            return false;

        isDataLoaded = false;
        return true;
    }

    /**
     * In turn, close the cursor.
     */
    public void close() throws SQLException {
        if(isDirty) {
            writeDirtyRow();
        }

        cursor.close();
    }

    /**
     * since this is single threaded, go ahead and get the data
     * TODO:  delegate this to the lower depths... perhaps fetch
     * row data up to pos i... since that is how the implementation is
     * physically working.
     */
    public Comparable getValue(int pos) throws SQLException {
        assureCurDataLoaded();
        return curRowData[pos];
    }

    /**
     * Assumed that the lock is already obtained.
     * NOTE:  could mark individual dirty flags in boolean array
     *   for faster determination of index update requirements
     */
    public void setValue(int i, Comparable value) throws SQLException {
        assureCurDataLoaded();

        if(! isDirty) {
            for(int pos = 0; pos < prevRowData.length; pos++) {
                prevRowData[pos] = curRowData[pos];
            }

            isDirty = true;
        }
        curRowData[i] = value;
    }

    /**
     * Loads the current data if it hasn't been done yet
     */
    private void assureCurDataLoaded() throws SQLException {
        if(! isDataLoaded) {
            cursor.getRowData(t, curRowData);
            isDataLoaded = true;
        }
    }

    /**
     * Same as assureCurDataLoaded but forces the load
     */
    private boolean reloadCurData() throws SQLException {
        return cursor.getRowData(t, curRowData);
    }

    public int getNumColumns() throws SQLException {
        return columnInfo.length;
    }

    //  TODO:  deprecate this, ColumnInfo should have all the data
    public String getColumnName(int i) throws SQLException {
        return columnInfo[i].getName();
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return columnInfo[i];
    }

    public boolean isUpdateable() throws SQLException {
          //  TODO:  depends upon the mode it was opened in
        return true;
    }

    /**
     * This is called by RowFilterIterator.  Eventually all this
     * can be rewritten since Filter Iterator should work on any
     * underlying TupleIterator.  Use Value style locks and enqueues<BR><BR>
     *
     * Returns false if the row has been deleted
     */
    public void lockRow() throws SQLException {
          //  TODO: eventually support tableRowShare
        cursor.reserveRowLock(t, true);

        //  TODO:  what if we already have the lock?

        if(! reloadCurData()) {
              //  row has been deleted.  I think this unlock is meaningless but its safer to be safe
            unlockRow();

            if(t.getIsolationLevel() == Connection.TRANSACTION_SERIALIZABLE) {
                ErrorCode.throwException(ErrorCode.CANT_SERIALIZE);
            } else {
              // so what does that mean?  I assume throw a bad view exception
                throw new ViewInconsistentException();
            }
        }

        //  TODO:  this might be a rather excellent time check the row condition again.
        //  by putting the condition down here, it might clean everything up

          //  TODO:  -or- select for update!!!
        if(t.getIsolationLevel() == Connection.TRANSACTION_SERIALIZABLE) {
            if(cursor.getLastSCN() > t.getCurrentConsistentView().getViewTime()) {
                ErrorCode.throwException(ErrorCode.CANT_SERIALIZE);
            }
        }
    }

    /**
     * Used to release a pending (temporary) lock when it
     * is decided that the row no longer fulfills the query/update
     * criteria
     */
    public void unlockRow() throws SQLException {
        cursor.releaseRowLock();
    }

    /**
     * Lock is obtained before
     */
    public void deleteTuple() throws SQLException {
          //  if no reading has happened (because there is no condition)
        assureCurDataLoaded();

        //  lockRow();
        //  check our child tables for any dependencies

        cursor.deleteRow(t);

          //  now update the indexes

          //  TODO:  eventually do this in parallel for remote indexes
        ArrayList allIndexes = table.getAllIndexes();
        for(int i = 0; i < allIndexes.size(); i++) {
            Index index = (Index)allIndexes.get(i);
              //  because technically curRowData has not been nulled out, this is still ok
            index.deleteEntry(t, curRowData, cursor.getCurrentRowNum());
        }
    }

    /**
     * Granted, questionable that this is here... but
     * it makes sense since all other type checking stuff is here.<BR><BR>
     *
     * Note: For rollback to work correctly indexes must be updated after the
     * main table.  This is because in reverse, the indexes information must be undone before
     * the lock is released.
     */
    public void insertTuple(Comparable[] data) throws SQLException {
        //  TODO:  need to lock row for uniqueness constraint collision
        //  lock table in appropriate mode (part of the lock row?)
        //  check all our parents exist

          //  first, insert the row into the table
        long rowNum = cursor.insertRow(t, data);

          //  lock the row... hmm, probably not necessary at all
        //table.getLockManager().lockRow(t, rowNum);

        //  check constraints, ie that all parents exist etc

          //  TODO:  eventually do this in parallel for remote indexes
          //  then insert into indexes
        ArrayList allIndexes = table.getAllIndexes();
        for(int i = 0; i < allIndexes.size(); i++) {
            Index index = (Index)allIndexes.get(i);
            index.insertEntry(t, data, rowNum);

            //  TODO:  check the unique constraint here so that we know which constraint failed!!
        }

        //  if a possible unique constraint is detected, maybe the Transaction should be notified
        //  delayed constraint validation is reasonable
    }

    /**
     * calls the method above, but automatically converts the Object data to the correct internal
     * type, in case they are not java types.  For now they are, but later it might change
     */
    public void insertTupleWithJavaObjects(Comparable[] data) throws SQLException {
        //SqlValue[] v = new SqlValue[data.length];
        //for(int i = 0; i < data.length; i++) {
        //    v[i] = new SqlValue(table.getColumnDef(i).getType(), (Comparable)data[i]);
        //}
        insertTuple(data);
    }

    /*********************************
     * writeDirtyRow.  flushes changes out.
     * The row should already be locked in the caller.  This is true because the check condition
     * needs to make sure that the condition is still true after the lock.  No other way really works:
     * Even oracle will block if there is another party that has the lock on the object.
     */
    protected void writeDirtyRow() throws SQLException {
        //  checkBasicConstraints();
        //  checkExternalConstraints();
        try{
            cursor.setRowData(t, curRowData);
        }catch(RuntimeException re) {
            re.printStackTrace();
            System.out.println("writeDirtyRow has:" + StringUtil.dump(curRowData) + "   prev:" + StringUtil.dump(prevRowData));
            throw re;
        }

        long curRowId = cursor.getCurrentRowNum();

          //  update all indexes.
        ArrayList allIndexes = table.getAllIndexes();
        for(int i = 0; i < allIndexes.size(); i++) {
            Index index = (Index)allIndexes.get(i);
            index.updateEntry(t, prevRowData, curRowData, curRowId);
        }

        isDirty = false;
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        return new ExecutionPlan("Full Table Scan [" + table.getName() + "]");
    }
}
