package freebase.engine;

import freebase.api.TableLockManager;
import freebase.api.TransactionDoneCallback;
import freebase.log.LogEntry;
import freebase.log.LogEntryHeader;
import freebase.log.tm.SavepointLogEntry;
import freebase.server.DBConnectionImpl;
import freebase.util.ErrorCode;
import org.apache.log4j.Category;

import java.lang.ref.WeakReference;
import java.math.BigInteger;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;


/**
 * Encapsulates all of the information of a transaction.  Many methods might
 * delegate actual implementation to the transaction manager.
 *
 * Note:  some methods are implemented here, some in TransactionManager. At
 *    some point it might be nice to have everything in one place.
 *
 * TODO:  for the ability to keep ResultSets open, should we implement a
 *   lowWater Mark for transaction view times?
 *
 * TODO:  add current value of sequence numbers.  Parallel update seems to be
 *    a problem:  what happens to curval (lastval)?
 */
public class Transaction {
    static Category cat = Category.getInstance(Transaction.class.getName());

    /**
     *  TODO: PREPARED will be here eventually
     * Should we have a being rolledback state?  It seems that other
     * transactions should not see the thing in rolled_back state until it
     * is completely done.  For example, delete holes should not be filled
     * until
     */
    final public static int ACTIVE      = 0;
    final public static int COMMITED    = 1;
    final public static int ROLLED_BACK = 2;

    /**
     * For printing the name
     */
    final public static String[] statusName = new String[] {"ACTIVE","COMMITED", "ROLLED_BACK"};

    TransactionManager tm;

    int isolationLevel;
    boolean anyWrites = false;

    long firstLSN;
    long beginningOfStatementLSN = 0;
    long begOfSubStatementLSN = 0;
    long lastLSN;

    /**
     * Used only during recovery, represents the number of undo records found after a rollback entry.
     * In theory helps recovery pick up a rollback where it left off (in case of failure during rollback)
     * but I'm not really convinced that it matters much... maybe on really large rollbacks, but scanning back
     * through is half the slow part and that still needs to be done.
     */
    long undoCount;

    /**
     * not just using "this" to reduce exposure risk
     */
    Object statusChangeLock = new Object();

    /**
     * our transaction id.  If zero, then not initialized yet.
     */
    long transactionId = 0;

    /**
     * map of name to Long LSN.
     * TODO:  allocate this lazy
     */
    HashMap savepoints = new HashMap();


    /**  gives the current viewTime of queries  FSN is finish sequence number (or time)
     *  could eventually be an object for distributed stuff
     *  could use an int, and just roll over.  Would need to use a separate
     *  flag for finished.  For doing > and <, would be best to automatically
     *  translate the view time into 0, max  (offset by a low water mark)
     * @deprecated
     */
    //long viewTime = -1;

    /**
     * FSN
     */
    long finishTime = -1;

    /**
     * Transaction current state, protected by statusChangeLock,
     * to be gotten when the state is PREPARED
     */
    volatile int status = ACTIVE;



    /**
     * Callbacks to be notified when this transaction commits or rollback. They will
     * be called in the reverse order that they were added.
     * Lazy Allocated in first register.
     */
    ArrayList commitCallbackList = new ArrayList();

    /**
     * The same as above but for eliminating dupes.
     * Is its own sync lock.
     */
    HashSet commitCallbacks;


    /**
     * Per callback objects that the callbacks have requested get included in the callback.
     * Lazy Allocated in first register.
     */
    ArrayList callbackObjects;

      //  same as touched tables
    //HashSet touchedIndexes = new HashSet();

    /**
     * HashMap table -> boolean[] for flags as to if we have gotten
     * particular table locks.  Warning, currently we use the lockmanager
     * pointer as the key, which might be dangerous
     * TODO:  allocate this lazy
     */
    HashMap tableLocks = new HashMap();

    /**
     * Contains a map of all current sequence values.  TODO:  cleanup this when done
     */
    HashMap sequenceValues;

    /**
     * The connection that spawned us.  TODO:  make this non-generic (java 1.5)
     */
    WeakReference connection;

    /**
     * The user that this connection is spawning.
     * Perhaps we can just get this from the connection?
     */
    String user;

    /**
     * This is the current view of this transaction.  It is established at the
     * beginning of the statement.  The only problem with this method is that trigger
     * statements might need their own view.
     */
    ConsistentView currentView;

    /**
     * Not sure yet is this is correct.  For dealing with updates of the form:<BR>
     * update test set id=id+1<BR>
     * Maintains the number of potential collisions while
     * doing updates to unique keys.  For each duplicate key added, this is incremented by one.
     * For each removed, it is decremented by one.  At the end of the statement, it should
     * be balanced.
     */
    long numUniqueContentions;

    /**
     *  creates an empty transaction
     */
    protected Transaction(TransactionManager tm, DBConnectionImpl connection, String user, int isolationLevel)
            throws SQLException {
        this.tm = tm;
        //this.connection = connection;
        this.connection = new WeakReference(connection);
        this.user = user.intern();
        this.isolationLevel = isolationLevel;

        //beginStatement();
    }

    /**
     * return the first log entry LSN of this transaction.
     * @return
     */
    public long getFirstLSN() {
        return firstLSN;
    }

    public TransactionManager getTM() {
        return tm;
    }

    public int getIsolationLevel() {
        return isolationLevel;
    }

    /**
     * sets the isolation level for this transaction.  Note, for serializable it will lock in the
     * view time right now due to the consistent view created automatically for us for each statement.
     * To fix this use a flag demarking this as a new isolation level so that it can be discarded
     * until next time through (to get the final view)
     * TODO:  verify that it is only setable in the correct circumstances
     * @param i
     */
    public void setIsolationLevel(int i) throws SQLException {
        if(transactionId != 0) {
            throw new SQLException("Not allowed to change isolation level after an update");
        }
        isolationLevel = i;
    }

    protected void setStatus(int s) {
        //System.out.println("Transaction " + transactionId + " is being set to " + s);
        synchronized(commitCallbackList) {
            this.status = s;
        }
    }

    /*******************************************************************************
     * numUniqueContentions methods, for detecting unique key violation
     * Sinced these are all accessed by the same thread,
     */

    /**
     *
     */
    public void incNumUniqueContentions() {
        numUniqueContentions++;
    }

    public void decNumUniqueContentions() {
        numUniqueContentions--;
    }

    public void clearNumUniqueContentions() {
        numUniqueContentions = 0;
    }

    public long getNumUniqueContentions() {
        return numUniqueContentions;
    }

    /*******************************************************************************
     *  UNDO count methods
     */

    /**
     * called by recovery every time an undo message appears.
     */
    protected void incUndoCount() {
        undoCount++;
    }

    /**
     * Called when recover finds rollback complete message
     */
    protected void clearUndoCount() {
        undoCount = 0;
    }

    /**
     * called when recovery decides to rollback incomplete transactions.  It can use this to ensure
     * @return
     */
    protected long getUndoCount() {
        return undoCount;
    }

    /*******************************************************************************
     *  END UNDO count methods
     */

    /**
     * sets the sequence value in this transaction.  This is for seq_name.curval or whatever.
     */
    public void setSequenceValue(String schema, String sequence, BigInteger value) {
          //  technically could lock on anything
        synchronized(this) {
            if(sequenceValues == null) {
                sequenceValues = new HashMap();
            }
            sequenceValues.put(schema + "." + sequence, value);
        }
    }

    /**
     * Get the current sequence value.  Must have been set by a nextval operation on a sequence
     */
    public BigInteger getSequenceValue(String schema, String sequence) {
        synchronized(this) {
            if(sequenceValues == null) {
                return null;
            }
            return (BigInteger)sequenceValues.get(schema + "." + sequence);
        }
    }

    /**
     * Return the user that is doing this transaction.  Doesn't get the user from the connection because
     * it may be null
     */
    public String getUser() {
        return user;
    }

    /**
     * Get the connection implemenation.  This may eventually be a security hole, if we
     * ever pass around the transaction object directly.  May be null if this is an internal/bootstrap transaction
     */
    public DBConnectionImpl getConnection() {
        if(connection == null) {
            return null;
        }
        //return connection;
        return (DBConnectionImpl)connection.get();
    }

    /**
     *  return our transaction id.
     * TODO:  rewrite this in such a way as to get one
     * lazy if the need arrises, so that we are never caught without
     * one
     * could rename getID() (shorter, easier to read actually)
     */
    public long getTransactionId() {
        return transactionId;
    }

    /**
     * Returns the last lsn that this transaction wrote.  Called
     * by the log manager
     */
    public long getLastLSN() {
        return lastLSN;
    }


    protected void setLastLSN(long LSN) {
        lastLSN = LSN;
    }

    /**
     * Returns one of (ACTIVE, COMMITED, ROLLED_BACK)
     * TODO: When adding PREPARED, check callers to see if their logic is still good
     * TODO:  may need to be protected with a sync
     */
    public int getStatus() {
        return status;
    }

    /**
     * Gets the status in a readable form
     */
    public String getStatusReadable() {
        return statusName[status];
    }

    /**
     * Will return true in active or prepared state, ie is this still alive.  Called by
     */
    public boolean isActive() {
        if(status == ACTIVE) {
            return true;
        }
        return false;
    }

    /**
     * Returns the statement's consistent view.  The view is calculated once at the beginning
     * of the statement and is used by multiple cursors during the course.  Each cursor
     * increments the useCount by one and decrements it when done.
     */
    public ConsistentView getCurrentConsistentView() {
        return currentView;
    }

    /**
     * This waits for a commit or a rollback; even a partial rollback is ok.
     * Currently this is the only method for blocking on another transaction.
     * Eventually a more complex scheme could be put in place.  Current thought
     * is that rollbacks are rare enough that this should be fine.
     *
     * Good though:  a method blockIfLockIsValid(...)  or blockIfLSNIsValid(...)
     *  NOTE:  1/2/2004 (see features.html doc)  While this is currently deprecated, it may be useful for implementing
     *  unique key on indexes.  If a transaction detects that there is already an entry before it (but the
     * transaction that wrote that is not completed) then it doesn't know yet if it can proceed with the insert, and
     * thus must block.  It must wait for the firstmost transaction to finish (commit or rollback).  When that
     * transaction is completed it must check the current firstmost transaction.  If it is not us, then it
     *
    public void waitForTransaction(Transaction t) {
        synchronized(t.statusChangeLock) {
            //  TODO:  check to see if the transaction is already finished  (ie, a while loop)
            //         something wrong with this.  need to work on the
            //         lock grabbing aspect of this!!!
            //  this is done by putting this in a while loop

            //  TODO:  add information to deadlock detector.
            try{
            t.statusChangeLock.wait();
            }catch(InterruptedException i) {
                i.printStackTrace();
            }
        }
    }*/

    /**
     * return the finish time of this transaction, or -1 if
     * it has not finished yet.
     *
     * There is not a race condition with calling this because the caller
     * will be comparing the time with it's own viewTime.  If this is called
     * almost at the same time as the transaction is commiting, then the result
     * will be the same reguardless of order.  The lock is there because
     * view time is a long
     */
    public long getFinishTime() {
        synchronized(statusChangeLock) {
            return finishTime;
        }
    }

    /**
     * Set the SCN for this transaction.
     */
    protected void setFinishTime(long time) {
        synchronized(statusChangeLock) {
            finishTime = time;
        }
    }

    /**
     * Gets the table lock flags for this transaction, or creates
     * it if it does not already exist.  The lock flags are expected to
     * be modified by the caller, lockTable and unlockTable
     */
    public boolean[] getTableLocks(Object tableKey) {
        synchronized(tableLocks) {
            boolean[] locks = (boolean[])tableLocks.get(tableKey);
            if(locks == null) {
                locks = new boolean[TableLockManager.NUM_TABLE_LOCKS];
                tableLocks.put(tableKey, locks);
            }

            return locks;
        }
    }


    /**
     *  Called at the start of the transaction and after each statement,
     *  tells this transaction object to demark the beginning
     *  of a statement, for the purposes of putting
     *  info into the log and for updating view times and such.<BR>
     *
     * TODO:  handle serializable statement views.  A separate one must exist
     *    if for no other reason than to allow early cursors to not see later data.
     *    could be as easy as decrementing the previous count after establishing the new
     */
    public void beginStatement() throws SQLException {
        //System.out.println("BEGIN STATEMENT " + getTransactionId() + "  status: " +statusName[status]);
        if(getStatus() != ACTIVE) {
            throw new SQLException("Transaction not in ACTIVE status:  Cannot begin statement");
        }

          //  Serializable transactions need new views every statement too,
          //  for the log position.  That way past reads don't see future inserts
        switch(isolationLevel) {
            case Connection.TRANSACTION_READ_COMMITTED:
                currentView = tm.getViewManager().createCurrentView(this);
                //currentView.incrementUseCount();    no longer needed
                //viewTime = tm.getCurrentViewTime();  //  <-  don't know if this still has a use
                break;

            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                if(currentView == null) {
                    currentView = tm.getViewManager().createCurrentView(this);
                    //currentView.incrementUseCount();  no longer needed
                } else {
                    currentView.updateLogPosition(getLastLSN());
                }

                //  old ideas:
                //  create a view using the existing one's time
                //  increment the count
                //  decrement the currentView count
                //  replace the currentView with the new one
                break;
            case Connection.TRANSACTION_READ_UNCOMMITTED:
                //  TODO:  currently uses the same currentView as was created before the first statement!!!
                //  TODO:  currenly a given dirty view prevents log cleaning as badly as a serializable!!!
                //  TODO:  in setIsolationLevel() or something fix this, perhaps put in as a view time forever in the
                     //  future or something (or just decrement right away but keep the actual object for posterity)
                //  exists because it was created during the set transaction statement.  Hmm, a little subtle but works
                currentView.updateLogPosition(getLastLSN());
                break;
            default:
                throw new SQLException("isolation level " + isolationLevel + " unrecognized");
        }
          //  Set the location of where rollback occurs
        beginningOfStatementLSN = lastLSN;
    }

    /**
     * Called for each component of a batch update.  A sub statement is a single statement that will
     * be rolled back alone only in the case of a ViewInconsistentException.  If this patterns seems
     * to be expanding a general push/pop model can be employed.
     * @throws SQLException
     */
    public void beginSubStatement() throws SQLException {
        begOfSubStatementLSN = lastLSN;
    }

    /**
     * Rolls back all changes since the substatement demark, and
     * pushes the view all the way forward for another attempt, and resets the beginning of
     * the sub statement (phew!)<BR>
     * This is used when a view inconsistency is found and another attempt to update or select for update is
     * made.  Common.
     */
    public void resetSubStatement() throws SQLException {

        if(isolationLevel == Connection.TRANSACTION_SERIALIZABLE) {
            ErrorCode.throwException(ErrorCode.CANT_SERIALIZE, "resetting sub statement but isolation serializable.");
        }

        tm.rollback(this, beginningOfStatementLSN);

        currentView.decrementUseCount();
        currentView = tm.getViewManager().createCurrentView(this);
        begOfSubStatementLSN = lastLSN;
    }

    /**
     * Called at the end of the statement, is most useful non-SERIALIZABLE
     * mode to release the statement level view.
     */
    public void endStatement() throws SQLException {
        //  if not serializable, remove our use of the current view

        if(isolationLevel == Connection.TRANSACTION_READ_COMMITTED) {
            currentView.decrementUseCount();
            currentView = null;
        } else {
            //  we can't decrement the view until next pass, otherwise the transactions committed after our time
            // will be released and our view time will be corrupted.
            //throw new SQLException("SERIALIZABLE NOT YET SUPPORTED");
        }
    }

    /**
     * Register a commit/rollback callback
     * @param callback
     * @param callbackData data to be attached to the callback.  Technically this isn't needed since
     *    things that need a special piece of data can record it in the callback object they are passing in.
     * @return false if the transaction is already commited or rolled back.
     */
    public boolean registerDoneCallback(TransactionDoneCallback callback, Object callbackData) {
        synchronized(commitCallbackList) {
              //  lazy allocate these extra objects
            if(commitCallbacks == null) {
                commitCallbacks = new HashSet();
                callbackObjects = new ArrayList();
            }

            int status = getStatus();
            if(status == COMMITED || status == ROLLED_BACK) {
                return false;
            }

              //  if wasn't already in the set, add to list as well
            if(commitCallbacks.add(callback)) {
                commitCallbackList.add(callback);
                callbackObjects.add(callbackData);
            }
            return true;
        }
    }

    //public long getTransaction

    /**
     * Adds whatever this logEntry refers to (Table) to the
     * touched list.
     *
    protected void addToTouchedTableIndexSet(LogEntry logEntry) {
        /*
          //  TODO: eventually generalize this for other than MemoryTable
          //  it is sloppy.  Also, don't forget to optimize imports!
        if(logEntry instanceof TableLogEntry) {
            TableLogEntry tableEntry = (TableLogEntry)logEntry;
            Integer tableId = new Integer(tableEntry.getTableId());
            touchedTables.add(tableId);
        }

        if(logEntry instanceof IndexLogEntry) {
            IndexLogEntry indexEntry = (IndexLogEntry)logEntry;
            Integer tableId = new Integer(indexEntry.getIndexId());
            touchedIndexes.add(tableId);
        }*

        //  TODO:  remove this completely
    }*/

    /**
     * the log record is currently written in the TransactionManager
     */
    public void commit() throws SQLException {
        //System.out.println("TRANSACTION COMMIT: " + this + "  Start LSN:" + getFirstLSN());
          //  protect the finish time, it is a long thus not atomic
        /**
         * This synchronized body was removed.  The finishTime is only read
         * if the transaction is in the COMMITTED state
         */
        //synchronized(statusChangeLock) {

          //  flush the log if we wrote anything at all
        if(getTransactionId() != 0) {
            tm.commit(this);
        }

        cleanup();
    }


    /**
     *  rollback to the savepoint, or null if all the way.
     */
    public void rollback(String savepoint) throws SQLException {
        if(savepoint == null) {  //  ie, rollback all
            //System.out.println("TRANSACTION ROLLBACK: " + this + "  Start LSN:" + getFirstLSN());
              //  rollback all changes, including table locks, row locks etc.  Very stable, no need
              //  to see status ROLLED_BACK, but only do it if something was actually done.

            if(getTransactionId() != 0) {
                tm.rollback(this, 0);
                //tm.removeHardLogStopper(firstLSN);  now called inside rollback
            }

            //  must be set before the share lock on checkpoint is released, thus inside already
            //setStatus(ROLLED_BACK);

              //  remove the transaction from the transactions map
            //tm.removeReference(this);

            cleanup();
        } else {
            Long point = (Long)savepoints.get(savepoint);
            if(point == null) {
                throw new SQLException("savepoint '" + savepoint + "' never defined", null,
                        ErrorCode.SAVEPOINT_NEVER_DEFINED);
            }

            tm.rollback(this, point.longValue());

              //  remove all savepoints after the one, but leave the one we just went back to
            Iterator it = savepoints.keySet().iterator();
            while(it.hasNext()) {
                String saveName = (String)it.next();
                Long pos = (Long)savepoints.get(saveName);
                if(pos.longValue() > point.longValue()) {
                    it.remove();
                }
            }
        }
    }

    /**
     * if any work as been done since this statement beginning was declared, ensure that it is all
     * rolled back.  If this was the first statement, rollback whole transaction and return false.
     * DDL's are single statement transactions that need to be rolled back right away.<BR>
     * NOTE:  see also the useful resetSubStatement
     * @throws SQLException
     * @return true if the transaction is still active.
     */
    public boolean rollbackStatement() throws SQLException {
        //  ExternalStats.incNumStatementRollbacks();

        if(getLastLSN() != beginningOfStatementLSN) {
              //  if this was the only statement, rollback whole transaction.
              //  this is most common in DDL
            if(beginningOfStatementLSN == firstLSN) {
                rollback(null);
                return false;
            } else {
                tm.rollback(this, beginningOfStatementLSN);
                return true;
            }
        }

          //  no change, so still active.
        return true;
    }

    /**
     * Called at the end of rollback and commit, Nulls out all unneeded
     * references for the post active life of this moribund transaction.
     */
    void cleanup() {
          //  NOTE:  could save one volitale read by saving this value!
        int finalStatus = getStatus();

          //  now unlock all waiters
          //  since after transaction is set to COMMITTED/ROLLBACK, this can;t be added to so it is safe
        for(int i = commitCallbackList.size() - 1; i >= 0; i--) {
            TransactionDoneCallback cc = (TransactionDoneCallback)commitCallbackList.get(i);
            cc.transactionDone(this, finalStatus, callbackObjects.get(i));
        }

          //  decrementing the use count here allows the view to be released immediately.
          //  currentView will be != null when in SERIALIZABLE mode
        if(currentView != null) {
            currentView.decrementUseCount();
            currentView = null;
        }

        commitCallbacks = null;
        savepoints = null;
        connection = null;
        tableLocks = null;

        sequenceValues = null;
    }

    /**
     * must mark a new generation (epoc) for this transaction, so that new locks obtained
     * are put into that epoc, and when partial rollback occurs they are also rolled back.
     * Overwrites any old savepoint entries by the same name.
     */
    public void savepoint(String name) throws SQLException {
          //  could validate name format (ie bad chars)  (may be defunct)
        if(savepoints.containsKey(name)) {
            throw new SQLException("Savepoint '" + name + "' already exists");
        }
        writeLogEntry(new SavepointLogEntry(name), true);
        addSavepoint(name, lastLSN);
    }

    /**
     * Adds a savepoint locally.  Called from savepoint (above) and from recovery mode in
     * the case of failover.
     */
    protected void addSavepoint(String name, long lsn) {
        savepoints.put(name, new Long(lsn));
    }

    /**
     * Write a non flushing log entry on behalf of this transaction.
     */
    public long writeLogEntry(LogEntry logEntry) throws SQLException {
        return writeLogEntry(logEntry, false);
    }

    /**
     * Write a log entry on behalf of this transaction.  NOTE:  when rolling back, currently this is
     * not called so if this changes you must also change the rollback call.
     * NOTE:  if there every become parallel writers, this needs to do some locking (to protect
     * touched tables at least)
     * @param flush true if this must flush/sync the log before returning
     */
    public long writeLogEntry(LogEntry logEntry, boolean flush) throws SQLException {
          //  is this the first write?
        if(getTransactionId() == 0) {
            setTransactionId(tm.assignTransactionId(this));

            firstLSN = tm.writeBeginLogEntry(this);
            beginningOfStatementLSN = firstLSN;

            //  record the time and add it as an entry in the low water (or, per query?)
            //  mark table in TM.
        }

        //addToTouchedTableIndexSet(logEntry);  @deprecated, though still not a
          //  horrid idea... if it was possible to ensure all could be added here for all Enqueues

          //  TODO:  should this line be in the log header itself?  (cuz now it's in two places)
        lastLSN = tm.getLog().writeEntry(this, logEntry, flush);
        return lastLSN;
    }

    /**
     * Called from TransactionManager.  TODO:  what about synchronization?
     */
    protected void setTransactionId(long tid) {
        transactionId = tid;
    }

    /**
     * Gets the log entry associated with this LSN, in this transactions
     * log.
     */
    public LogEntryHeader getLogEntry(long LSN) throws SQLException {
        return tm.getLog().getLogEntry(LSN);
    }

    /**
     *  Later add more debugging info if log mode is set
     */
    public String toString() {
        if(transactionId == 0) {
            return "Trans. id: (none)";
        }

        return "Trans. id:" + transactionId;
    }
}