package freebase.engine;

import freebase.api.DeadlockDetector;
import freebase.api.LogEntryProcessor;
import freebase.api.LogManager;
import freebase.api.StorageManager;
import freebase.lock.LocalDeadlockDetector;
import freebase.log.DefaultLogEntryCodes;
import freebase.log.LogEntry;
import freebase.log.LogEntryHeader;
import freebase.log.UndoLogEntry;
import freebase.log.tm.*;
import freebase.server.DBConnectionImpl;
import freebase.util.FauxShareSemaphore;
import freebase.util.LogStop;
import freebase.util.LongHashMap;
import freebase.util.ShareLock;
import org.apache.log4j.Category;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Enumeration;

/**
 *  TransactionManager is the central location for the
 *  entire database.<BR><BR>
 *
 * There should be events for when the low active transaction has risen.<BR><BR>
 *
 * This class needs to be generalized and put into the api section.
 * It will handle general checkpoint capability.<BR><BR>
 *
 * Consider making low water mark raising occur in a separate thread
 */
public class TransactionManager {
    static Category cat = Category.getInstance(TransactionManager.class.getName());

    LogManager log;

    /**
     * Where we send log entries...
     */
    //Catalog catalog;
    StorageManager sm;

    //  to keep track of
    //    the highest FSN known
    //    all transactions by local id
    //    map of all remote to local transation ids

    /**
     * Lock that snapshots the active transactions while writing the start checkpoint log entry.
     * It prevents any transaction from starting, committing, or rolling back (complete) for the short
     * time it takes the checkpoint process to write out a checkpoint start log entry
     * TODO:  make sure interrupts behave correct.  Since it isn't possible that some other lock is obtained
     *     after this, perhaps ignoring is ok.
     * TODO:  implement this correctly, its easy easy.  Add statistics as well
     */
    ShareLock startCheckpointLock = new FauxShareSemaphore();
    
    boolean shuttingDown = false;
    boolean forceAtClose = false;

    /**
     * used for fetching the next transaction id.
     */
    long lastTransactionId = 0;

    Object transIdLock = new Object();

    /**
     * Variables for assigning view times (SCNs) to all finished transactions
     */
    Object scnLock = new Object();
    long highestViewTime = 0;

    /**
     * Is used
     */
    ViewManager viewManager;

    /**
     * Holds all transactions that have not completed and whose finish
     * time is still greater than some other transactions view time.
     */
    LongHashMap transactions = new LongHashMap();
    Object transactionMapLatch = new Object();

    /**
     * The deadlock detector that we are using.
     */
    DeadlockDetector deadlockDetector;

    //  TODO:  make hardStops and softStops optional, since they may consume some time
    //  Perhaps the log implementation can be asked if it needs this feature at all.

    /**
     * Log stoppers are used to
     * prevent a section of the log from being taken offline, as they are still
     * needed by active transacions.  The hardStops are the lowest is the low water mark for
     * online required log
     * NOTE:  both hard and soft are protected by the hardStops monitor for simplicity
     */
    LogStop hardStops = new LogStop();

    /**
     * These stops are not absolutely required for correctness, but are used to ensure that all active
     * views can be fulfilled by data available from the log.  As views are no longer needed
     */
    LogStop softStops = new LogStop();

    /**
     * Construct the transaction manager with the log.
     * TODO:  pass in the log position controller
     */
    public TransactionManager(LogManager log, StorageManager sm) throws SQLException {
        this.log = log;
        this.sm = sm;
        this.viewManager = new ViewManager(this);

        sm.setTransactionManager(this);

          //  TODO:  put the class of this entity in a config file
        deadlockDetector = new LocalDeadlockDetector();

          //  TODO:  eventually must get the position from the log anchor or whatever it is called
        recover(0);

        //  TODO:  after recovery, initialize viewManager state to current
    }

    /**
     * Set the Log Processor so we can do things like undo and redo.
     */
    public void setStorageManager(StorageManager sm) {
        this.sm = sm;
    }

    public StorageManager getStorageManager() {
        return sm;
    }

    public ViewManager getViewManager() {
        return viewManager;
    }

    /**
     * Permanently remove all reference to this transaction.
     */
    protected void removeReference(Transaction t) {

        synchronized(transactionMapLatch) {
            if(t == null) {
                System.out.println("WARNING!!! t is null");
                new Exception("DEBUG STACK TRACE").printStackTrace();
            }

            Object whatWasRemoved = transactions.remove(t.getTransactionId());
            if(whatWasRemoved == null) {
                System.out.println("Alert, removing transaction " + t + " but not in list!!");
            }
        }

        synchronized(hardStops) {
            //System.out.println("removing " + t + "  @" + t.getFirstLSN());
            if(softStops.removeStop(t.getFirstLSN())) {
                log.raiseLowWaterMark(hardStops.getMinStop(), softStops.getMinStop());
            }
        }
    }

    //public void setCatalog(Catalog cat) {
    //    this.catalog = cat;
    //}

    //public Catalog getCatalog() {
    //    return catalog;
    //}

    /**
     * Create a new Transaction.  The transaction manager needs to know
     * the username for posterity.
     */
    public Transaction beginTransaction(DBConnectionImpl con, String user, int isolationLevel) throws SQLException {
        startCheckpointLock.getLockShared();
        try {
            if(shuttingDown)
                throw new SQLException("DB SHUTTING DOWN!!");
            
            //  should probably keep track of the view time or something here
            //  should pass in the transaction id, the current view time etc
            Transaction t = new Transaction(this, con, user, isolationLevel);
            t.beginStatement();
            return t;
        } finally {
            startCheckpointLock.releaseLockShared();
        }
    }

    /**
     * Called internally after a recovery, used for forget about all transactions!
     */
    private void removeAllTransactions() {
        synchronized(transactionMapLatch) {
            transactions = new LongHashMap();
        }
    }

    /**
     * Return the transaction object given it's id.  Null if the id
     * is old and the transaction is historical.
     */
    public Transaction getTransaction(long transactionId) {
        synchronized(transactionMapLatch) {
            return (Transaction)transactions.get(transactionId);
        }
    }

    /**
     * Is called when a query is about to happen.  This fetches
     * the current view time associated with all completed transactions.
     */
    public long getCurrentViewTime() {
        synchronized(scnLock) {
            return highestViewTime;
        }
    }

    /**
     * Is called at the end of a transaction place it's finishing in
     * historical order.
     */
    public long assignFinishViewTime() {
        synchronized(scnLock) {
            return ++highestViewTime;
        }
    }

    /**
     * Returns the log for the given transation
     */
    public LogManager getLog() {
        return log;
    }

    /**
     * Returns the deadlock detector used by this transaction manager.
     */
    public DeadlockDetector getDeadlockDetector() {
        return deadlockDetector;
    }

    /**
     * Should be called by Transaction when it is about to do it's first
     * write.
     */
    public long assignTransactionId(Transaction t) {
        long tId;
        synchronized(transIdLock) {
            tId = ++lastTransactionId;
        }

        synchronized(transactionMapLatch) {
            transactions.put(tId, t);
        }

        return tId;
    }

    /************************************************************************
     * Methods to manage low water mark for online log requirements
     */

    /**
     * This method was created to guarenttee that log stoppers get created in increasing order.
     * It does this correctly, but at the expense of preventing other transactions from
     * beginning until this one is done.  The extra time is pretty small, but eventually
     * we may just relax the requirement and instead say... low water mark is only mostly
     * correct.  In this case, sometimes a low water mark is set and then a lower one is set...
     * FUTURE_TUNE<BR><BR>
     *
     * Alternatively use a linked list to keep them all in order, and a hashmap to record
     * removed stoppers.  Then traverse the list until.. no, too much memory for one long transaction.
     */
    protected long writeBeginLogEntry(Transaction t) throws SQLException {
          //  atomically write the entry and prevent the log cleaning
        synchronized(hardStops) {
            long pos = getLog().writeEntry(t, new BeginTransactionLogEntry(), false);
            //System.out.println("adding " + t + " @" + pos);
            hardStops.addStop(pos);
            softStops.addStop(pos);
            return pos;
        }
    }

    /**
     * Called when the transaction is committed or rolled back.
     * TODO:  pass in the LSN of the lowest transaction for the soft
     */
    protected void removeHardLogStopper(long logPos) {
        //  System.out.println("LOG MOVING UP to " + logPos);

        synchronized(hardStops) {
            if(hardStops.removeStop(logPos)) {
                getLog().raiseLowWaterMark(hardStops.getMinStop(), softStops.getMinStop());
            }
        }
    }

    /**
     * Any transaction that is not ROLLEDBACK or COMMITTED.  Since this is protected by
     * the startCheckpointLock it can't be modified, though a more paranoid would also lock it
     * by transactionMapLatch as well.
     * @return ArrayList of Transaction Objects
     */
    protected ArrayList getActiveTransactions() {
        ArrayList activeTransactions = new ArrayList();
        Enumeration en = transactions.elements();
        while(en.hasMoreElements()) {
            Transaction t = (Transaction)en.nextElement();
            int status = t.getStatus();
            if(status != Transaction.ROLLED_BACK && status != Transaction.COMMITED) {
                if(t.getTransactionId() != 0) {  //  ie, has this written any data yet!?
                    activeTransactions.add(t);
                }
            }
        }
        return activeTransactions;
    }
    /************************************************************************
     * Recovery Methods
     */

    /**
     *  shutdown the database.  
     */
    public void shutdown(boolean force) {
        startCheckpointLock.getLockExclusive();
        try {
            shuttingDown = true;
            forceAtClose = force;

              //  TODO:  when checkpointing, don't regrab lock!!!
            //  allow no new transactions to begin.
            //  if force, 
            //   begin a checkpoint, allow no new transactions end... arg!
        } finally {
            startCheckpointLock.releaseLockExclusive();
        }
    }

    /**
     * Checkpoint the system as per the REDO/UNDO logging rules.  Should be called internally only.
     * TODO:  checkpointing should NOT be part of the transaction manager, or
     * if it is it should only be at the highest level of abstraction.
     */
    public void checkpoint() throws SQLException {
        //  check to see if we are currently the primary node
        //  if not, then we need to send a note to the primary node
        
        startCheckpointLock.getLockExclusive();
        try {
            ArrayList activeTrans = getActiveTransactions();

              //  pull up all transactions that are still alive.  Write them
              //  in the start checkpoint entry.
            log.writeEntry(null, new CheckpointStartLogEntry(activeTrans), false);

            //  write start checkpoint log entry
        } finally {
            startCheckpointLock.releaseLockExclusive();
        }
        
          //  do the checkpoint.
        sm.checkpoint();

        //  TODO:  does this happen in a separate thread?  answer: probably not, since this is already called in separate thread

        //  write end checkpoint end log entry (and flush)
        log.writeEntry(null, new CheckpointEndLogEntry(), true);

        // TODO:  update the log anchor
    }

    /**
     * Called only from Transaction.commit(...)
     */
    void commit(Transaction t) throws SQLException {
          //  grab the lock to assure that the transaction does 
          //  not commit before the start checkpoint entry goes in
          //  the log.  This could be done by having the log be aware
          //  of the transactions directly.
        startCheckpointLock.getLockShared();
        try {
              //  put a stake in the ground for "when" we committed.
              //  this will play a bigger part in the 2PC setting
            t.setFinishTime(assignFinishViewTime());  //  now the actual finish time is established (if successful)

            getLog().writeEntry(t, new CommitLogEntry(t.getFinishTime()), true);

              //  raise low water mark if we are the lowest.
            removeHardLogStopper(t.getFirstLSN());

              //  map the view time to the transaction
              //  or remove the transaction immediately if need be
            getViewManager().transactionFinished(t);

            t.setStatus(Transaction.COMMITED);
        } finally {
            startCheckpointLock.releaseLockShared();
        }
    }

    /**
     * NOT IMPLEMENTED BECAUSE not needed
     * Undos a bunch of things that have happened to the database.
     * If savepoint is null, then undo everything.
     *
    protected void rollback(Transaction t, String savepoint) throws SQLException {
        //  scan through and find appropriate savepoint's LSN
        //  call rollback(savepointLSN);
    }
    */

    /**
     * Rollback to the given LSN
     */
    protected void rollback(Transaction t, long toLSN) throws SQLException {
        if(cat.isInfoEnabled()) {
            cat.info("ROLLING BACK " + t + " to LSN: " + toLSN);
        }

        if(t.getTransactionId() != 0) {
            LogManager log = getLog();

              //  get previous last, rollback from there
            long curLSN = t.getLastLSN();

              //  first write a rollback message, then do the rollbacks, then write a rollback complete message
            t.writeLogEntry(new RollbackLogEntry());
            rollback(log, t, curLSN, toLSN);

              //  guarentee that no transactions get lost...  We will either be in the list of transactions
              //  only grab the lock if the transaction is finishing
            if(toLSN != 0) {
                t.writeLogEntry(new RollbackCompleteLogEntry(toLSN));
            } else {
                try{
                    startCheckpointLock.getLockShared();
                    t.writeLogEntry(new RollbackCompleteLogEntry(toLSN));
                    removeHardLogStopper(t.getFirstLSN());

                    t.setStatus(Transaction.ROLLED_BACK);

                    //  remove the transaction from the transactions map TODO:  WRONG! (unless reconstructs all old data so this isn't needed)
                    //  was:
                    //removeReference(t);
                    getViewManager().transactionFinished(t);
                } finally {
                    startCheckpointLock.releaseLockShared();
                }
            }
        } else {
            t.setStatus(Transaction.ROLLED_BACK);
            //  TODO:  is this ever called???????  If so, does it have a view?
        }
    }


    /**
     *  Undoes the operations in the log from the fromLSN back to the toLSN.<BR><BR>
     *
     * TODO:  eventually we will need to not ignore rollback messages, but instead
     * use RollbackComplete messages to skip over rollbacks that have already been done.
     * In normal operation, all Rollbacks that have been completed should be skipped.  During
     * recovery only rollback completed before the last checkpoint start should be skipped.
     */
    protected void rollback(LogManager log, Transaction t, long fromLSN, long toLSN) throws SQLException {
        long curLSN = fromLSN;

        while(curLSN != toLSN && curLSN != 0) {
            if(cat.isDebugEnabled()) {
                cat.debug("  Found entry @" + curLSN);
            }

            LogEntryHeader logHeader = log.getLogEntry(curLSN);
            LogEntry entry = logHeader.getLogEntry();
            if(! (entry instanceof TMLogEntry)) {
                if(cat.isDebugEnabled()) {
                    cat.debug("  undoing entry @" + curLSN + ":  " + logHeader.getLogEntry());
                }

                  //  invert the log entry here.
                LogEntry undo = null;
                if(entry.getCode() == DefaultLogEntryCodes.UNDO_LOG_CODE) {
                    undo = ((UndoLogEntry)entry).getOperationUndone();
                } else {
                    undo = new UndoLogEntry(entry);
                }

                  //  explicitly write the undo message out
                long undoLSN = log.writeEntry(t, undo, false);
                t.setLastLSN(undoLSN);

                  //  undo the original event (or redo the undone thing, as appropriate)
                if(entry.getCode() == DefaultLogEntryCodes.UNDO_LOG_CODE) {
                    sm.processLogEntry(LogEntryProcessor.REDO_ENTRY, undoLSN, t, undo);
                } else {
                    sm.processLogEntry(LogEntryProcessor.UNDO_ENTRY, undoLSN, t, entry);
                }
            } else if(entry instanceof RollbackCompleteLogEntry) {
                  //  skip over the regions we have already rolled back!
                RollbackCompleteLogEntry rol = (RollbackCompleteLogEntry)entry;
                curLSN = rol.getToLSN();

                  //  don't go for the previous log entry!!!
                continue;
            }

            curLSN = logHeader.getPrevTransactionLSN();
        }
    }

    /**
     *  Rolls the database forward from the lastCheckpoint position.
     *  the checkpoint lsn should point to the end checkpoint entry.
     *
     * TODO:
     * deal with checkpointing, especially the rollback log entries that complete after the checkpoint
     *
     * @param lastCheckpoint the last start checkpoint entry in the log
     */
    public void recover(long lastCheckpoint) throws SQLException {
        LogEntryHeader logHeader = log.getLogEntry(lastCheckpoint);

        //  If the log entry is start checkpoint, initialize all of the transactions
        //  enclosed with the correct information, especially the lastLSN!

          //  apply all log entries to the end of the log
        while(logHeader != null) {
            applyLogEntry(logHeader);

            logHeader = log.getLogEntry(logHeader.getNextLSN());
        }

          //  cleanup the remaining transaction objects.  (no need to lock on transactionMapLatch cuz we did in this thread already)
        Enumeration en = transactions.keys();
        while(en.hasMoreElements()) {
            Long tid = (Long)en.nextElement();
            Transaction t = getTransaction(tid.longValue());

              //  check all prepared transactions, commit them if need be
            if(t.getStatus() == Transaction.ACTIVE) {
                //  rollback all changes
                t.rollback(null);
            }

            //System.out.println("TRANSACTION leftover: " + t + "  " + t.getStatus());
        }


        //    for all transactions that are not committed, roll them back.
        //    if we are in the correct mode, attempt to only rollback to a specific (or the most recent)
        //       savepoint

        //  explicitly remove all transaction objects
        removeAllTransactions();

          //  currently doesn't checkpoint immediately upon startup, but can easily
        checkpoint();

        //  TODO:  execute checkpoint.  If the checkpoint position points to the end checkpoint spot, then
        //  how does the next recovery process determine the active transactions?  By storing the lsn
        //  of the begin checkpoint?  Technically, the TM will need to replay everything from the
        //  checkpoint start on since there is no guarentee that any changes made afterward are actually
        //  stored anywhere.  SO... it seems that a checkpoint end log message has less importance.


          //  we are done
        System.gc();
    }
    
    
    /**
     *  FILL IN:  to be used by recover, and from the remote node live.  Thus single threaded
     */
    public void applyLogEntry(LogEntryHeader logHeader) throws SQLException {
        LogEntry entry = logHeader.getLogEntry();

        long tid = logHeader.getTransactionId();
          //  update the last LSN so that it is correct.  This could happen at the beginning
        Transaction t = getTransaction(tid);
        if(t != null) {
            t.setLastLSN(logHeader.getLSN());

            //  now done via the operations themselves.
            //t.addToTouchedTableIndexSet(entry);
        }

            //  if log entry is a transaction entry, should we keep the transaction here?
        if(entry instanceof TMLogEntry) {
            switch(entry.getCode()) {
                case DefaultLogEntryCodes.BEGIN_TRANSACTION_LOG_CODE:
                    //BeginTransactionLogEntry begin = (BeginTransactionLogEntry)entry;

                    //  create transaction and add it to the list

                    //  TODO:  add isolation mode and user_id to BeginLogEntry
                    t = new Transaction(this, null, "TODO", Connection.TRANSACTION_READ_COMMITTED);
                    t.setTransactionId(tid);

                    if(tid > lastTransactionId) {
                        lastTransactionId = tid;
                    }

                    transactions.put(tid, t);
                    t.setLastLSN(logHeader.getLSN());

                    break;

                case DefaultLogEntryCodes.COMMIT_TRANSACTION_LOG_CODE:
                    CommitLogEntry commit = (CommitLogEntry)entry;
//System.out.println(logHeader.getTransactionId() + " " + t);
                    t.setStatus(Transaction.COMMITED);
                    t.setFinishTime(commit.getScn());
                    t.cleanup();

                      //  move view time if necessary
                      //  in reality the numbers might come slightly out of
                      //  order but never causally out of order ;)
                    if(highestViewTime < commit.getScn()) {
                        highestViewTime = commit.getScn();
                    }

                    //  TODO:  why not call this:
                    //removeReference(t);

                    /** for now, leave the transaction in place, post recovery will clean it up
                     * could easily see it being removed right away.  Will queries failover?
                     * If so, then we will need to keep it around... otherwise we can dump it here
                     */
                    break;


                /**
                 * Currently do nothing.  These may eventually be used as hints for how much to eventually
                 * rollback when rolling back
                 */
                case DefaultLogEntryCodes.ROLLBACK_LOG_CODE:
                    break;
                case DefaultLogEntryCodes.ROLLBACK_COMPLETE_LOG_CODE:
                    RollbackCompleteLogEntry roll = (RollbackCompleteLogEntry)entry;
                    if(roll.getToLSN() == 0) {  //  if complete rollback and abort
                        t.setStatus(Transaction.ROLLED_BACK);
                        t.cleanup();
                        removeReference(t);
                    }
                    break;

                case DefaultLogEntryCodes.SAVEPOINT_LOG_CODE:
                    SavepointLogEntry savepoint = (SavepointLogEntry)entry;

                    t.addSavepoint(savepoint.getName(), logHeader.getLSN());
                    break;


                //  case PrepareLogEntry:
                //    mark this guy in the prepared state
            }
        } else {  //
          //  System.out.println(entry.getCode() + "    " + DefaultLogEntryCodes.UNDO_LOG_CODE);
            if(entry.getCode() == DefaultLogEntryCodes.UNDO_LOG_CODE) {
                UndoLogEntry undoEntry = (UndoLogEntry)entry;
                LogEntry undoneOperation = undoEntry.getOperationUndone();

                  //  apply the reverse of this.  Only Storage Manager messages are undoable, thus send it on.
                sm.processLogEntry(LogEntryProcessor.UNDO_ENTRY, logHeader.getLSN(), t, undoneOperation);
            } else {
                sm.processLogEntry(LogEntryProcessor.REDO_ENTRY, logHeader.getLSN(), t, entry);
            }
        }

        /**
         * Old notes:
         */
        //  not sure exactly how this is going to work yet
        //  ideally it should be able to work in conjunction with live transactions,
        //  itself doing the locking and updating just as the OldTableCursor and IndexCursor
        //  would.  Maybe it uses a special method on those guys?  This would be better
        //  for alternate future indexes being added to the system, sort of allow them to each
        //  manage the changes that they define.
    }
}
