/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 7, 2002
 * Time: 10:27:05 AM
 */
package freebase.engine;

import freebase.util.LongHashMap;
import freebase.util.MutableInteger;

import java.util.TreeMap;

/**
 * The View Manager works with the transaction manager to keep track of active views of
 * the database and to age old transaction that have no effect anymore.  A transaction
 * object T stays in memory as long as it has some effect on active views; when the view
 * that is furthest back can see a particular transactions changes, then the transaction
 * object T is no longer needed.  Forever in the future all views created will also be able
 * to see the T's changes.<BR><BR>
 *
 * The view manager deals with aging the old transactions.  It eventually will keep track of
 * lowest view times (SCNs) on other servers as well to manage a global aging process.<BR><BR>
 *
 * We learn from history that we learn nothing from history. --George Bernard Shaw<BR><BR>
 *
 * NOTE:  this implemenation is very memory heavy and going to only be acceptable for the first round.
 * What will happen if there is a whole day query and the system runs all day online?   Some consolidation
 * of views will need to happen then to reduce memory footprint.  In fact the ConsistentView algorithm will
 * probably need to change slightly.
 */
public class ViewManager {
    TransactionManager tm;

    //  variables:
    /**
     * map of all view times (SCN's) to the count of views.
     * TODO:  FUTURE_TUNE This data structure is bound to be slow, especially because
     * we need objects to pass in.  Eventually at least rewrite it with a number (long) as
     * the key directly
     */
    TreeMap activeViewCounts = new TreeMap();

    /**
     * explicity store the min of the tree externally.  Note that nothing can ever get
     * put into the tree lower than the min because the min was obtained at one point from
     * the highest scn, which must by definition be higher now.
     */
    Long minSCN;

    /**
     * This is the highest SCN we have seen so far
     */
    long highestSCN = 0;

    /**
     * The highest SCN that we have actually cleaned up.
     */
    long lastCleanupSCN = 0;

    //  map from SCN to actual transactions?  Could be stored in the TransactionManager
    //    needs to be determined if the completion of the transaction should happen first?

    /**
     * A map of all finished transactions by finish time.  Could just as easily be
     * some kind of linked list or whatever since they are always removed in linear
     * order and added likewise, honestly I prefer a growing ring structure.  If they
     * are added to a linked list then the call to add it will need to be under the lock
     * so that everything gets added in exactly the right order.
     * TODO:  reimplement this as a circular list
     */
    LongHashMap scnToTransaction = new LongHashMap();

    /**
     * only used for debugging when there is one instance of this.  It will
     * contain the most recently allocated ViewManager
     */
    public static ViewManager debugSingleton;

    /**
     * The view manager should be created after the TM has recovered
     */
    public ViewManager(TransactionManager tm) {
        this.tm = tm;

        debugSingleton = this;
    }

    //  methods:

    /**
     * In the safty of a monitor, creates a view for this transaction and assures
     * that changes made after this point in time are preserved until all of the users
     * of the view are done.<BR><BR>
     *
     * I'm a little concerned that since assigning a view time and putting the transaction
     * are not atomic that the there might be a race condition.  But since it is only cleanup
     * if something is missed for a short while there may be no problem.  Once certain solution
     * is to move the scn to Transaction map stuff all to the TM, and just call clean over there
     * periodically.  The only concern with that is that the view assignment happens before the
     * transaction actually commits, so we wouldn't want anything assuming that a transaction is
     * commited (because there is no reference to it) before .<BR><BR>
     *
     * TODO:  support serializable transactions by fetching the current view time from
     *         the transaction instead of the transaction manager
     *         ... just have a different method, or easier:  pass in the time!
     */
    public synchronized ConsistentView createCurrentView(Transaction t) {
//        System.out.println("ViewManager.createCurrentView: " + t);
        //DebugLog.log(this, "createCurrentView: " + t);

        long currentViewTime = tm.getCurrentViewTime();
        ConsistentView view = new ConsistentView(t, currentViewTime, t.getLastLSN());

        //System.out.println("add view: " + view);

          //  TODO:  one minor thing possible is to keep the Long around so that two
          //  don't get created every time.
        Long viewTime = new Long(currentViewTime);
        MutableInteger mi = (MutableInteger)activeViewCounts.get(viewTime);
        if(mi == null) {
            activeViewCounts.put(viewTime, new MutableInteger(1));

              //  keep track of the lowest scn view around
            if(minSCN == null) {
                minSCN = viewTime;
            }
        } else {
            mi.value++;
        }

          //  TODO:  debug test, remove it eventually
        if(viewTime.longValue() < minSCN.longValue()) {
            System.out.println("ViewManager:  new view time less than legal minimum:  " + viewTime.longValue() +
                    "    min is : " + minSCN.longValue());
        }

        view.incrementUseCount();
        return view;
    }

    /**
     * Is the process of cleaning up views asynchronous?
     */
    public synchronized void removeView(ConsistentView view) {
        //System.out.println("ViewManager.removeView: " + view);
        //DebugLog.log(this, "removeView: " + view);

          //  TOO bad this is getting created again... could just be sitting in the view.
        Long viewTime = new Long(view.viewTime);

        MutableInteger mi = (MutableInteger)activeViewCounts.get(viewTime);
        if(mi == null) {
            System.out.println("ERROR:  view removed but not currently taken account for");
        }

        mi.value--;

        //System.out.println("Removing view @ time " + viewTime + " count is now " + mi.value);
        if(mi.value == 0) {
            activeViewCounts.remove(viewTime);

              //  if we are getting rid of the minimum SCN, then we should clean up a bit
            if(viewTime.equals(minSCN)) {
                  //  find the next min
                Long nextMin = null;
                if(activeViewCounts.size() != 0) {
                    nextMin = (Long)activeViewCounts.firstKey();
                }

                  //  fro viewTime to but not
                performCleanup(viewTime, nextMin);

                minSCN = nextMin;
            }
        }
    }

    /**
     * removes references to old transactions that have no views behind them.  Can be called either
     * from a periodic thread or directly when a min is detected.  Null "to" means to the highest seen so far
     * @param from the low point we should clean from
     * @param to (current thinking) the highest point to clean up to but not including... ie, leave
     *    the transaction with scn "to" .  if To is null, clear out all.
     */
    private void performCleanup(Long from, Long to) {
        //System.out.println("ViewManager:performCleanup : from " + from + " to " + to + " max is: " + highestSCN + "  last cleanup:" + lastCleanupSCN);
        //DebugLog.log(this, "performCleanup : from " + from + " to " + to + " max is: " + highestSCN + "  last cleanup:" + lastCleanupSCN);

        long fromPos = from.longValue();
        long toPos = (to == null) ? highestSCN + 1  : to.longValue();

        for(long i = fromPos; i < toPos; i++) {
            if(i > lastCleanupSCN) {
                Transaction t = (Transaction)scnToTransaction.remove(i);
                if(t == null) {  //  DEBUG
                    System.out.println("Found transaction t == null for pos " + i);
                    //DebugLog.printLog(this);
                }
                tm.removeReference(t);
            }
        }

          //  since lastCleanupSCN is the highest scn we have actually cleaned out, its -1
        lastCleanupSCN = toPos - 1;
    }

    /**
     * Called by the transaction manager when a transaction finishes.  If the transaction does
     * not need to go into the map, then it needs to be removed from the transaction manager
     * immediately
     */
    public synchronized void transactionFinished(Transaction t) {
       // System.out.println("ViewManager.transactionFinished: " + t + " w/ finish time of " + t.getFinishTime());

          //  increment the highest scn
        if(t.getFinishTime() > highestSCN) {
            highestSCN = t.getFinishTime();
        }

          //  if there are no views at all, ignore transaction.  This is just a boundardy case optimization
          //  and should not be critical for overall correctness
        if(minSCN == null) {
              //  since we are not adding out transaction assume it is already done (possibly plus some)
            lastCleanupSCN = highestSCN;
            tm.removeReference(t);
            return;
        }

          //  if cleanup has already gone past us then we need to expunge ourselves immediately
        if(lastCleanupSCN >= t.getFinishTime()) {
            tm.removeReference(t);
            return;
        }

          //  add the transaction reference for future removal
        scnToTransaction.put(t.getFinishTime(), t);
        //System.out.println("# views: " + scnToTransaction.size());
    }

    /***************************************************************************************
     * debug methods
     */

    /**
     * return the earliest scn that is still active
     */
    public synchronized Long getLowSCNView() {
        return minSCN;
    }
}


