/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 24, 2002
 * Time: 6:23:53 PM
 */
package freebase.jdbc;

import freebase.server.DBConnection;

import java.rmi.RemoteException;
import java.sql.*;
import java.util.Map;
import java.util.Properties;

/**
 * A database connection.  Currently supports local and remote connections,
 * but will also eventually need to support internal connections for stored
 * procedures
 */
public class F3Connection implements Connection {
    /**
     * Initial properties
     */
    DBConnection physicalCon;
    String user;
    String password;
    Properties props;

      //  when false, each statement does not act in it's own transaction
    boolean autoCommit = true;
    boolean isClosed = false;  //  either explicitly or via failure


      //  hint as to whether a transaction is currently active.  should err toward true, just in case
    boolean transactionActive = false;

    //  a hack to get savepoint id to be mostly unique across the life of a connection.
    int savepointCounter;

    int isolationLevel = TRANSACTION_READ_COMMITTED;

    /**
     * Creates a new connection via the physical connection
     */
    protected F3Connection(DBConnection physicalCon, String user, String password, Properties props)
            throws SQLException {
        this.physicalCon = physicalCon;
    }

    /**
     * Returns the physical connection.  May need to make this public eventually, to make it possible
     * to expose underlying capabilities (like close hard)
     */
    public DBConnection getPhysicalConnection() {
        return physicalCon;
    }

    protected void setTransactionActive(boolean t) {
        transactionActive = t;
    }

    protected boolean isTransactionActive() {
        return transactionActive;
    }

    /************************************************************************
     * Create statement & call methods
     */

    public Statement createStatement() throws SQLException {
        return new F3Statement(this);
    }

    public PreparedStatement prepareStatement(String sql)
            throws SQLException {
        return new F3PreparedStatement(this, sql);
    }

    public CallableStatement prepareCall(String sql) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public Statement createStatement(int resultSetType, int resultSetConcurrency,
                                     int resultSetHoldability) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public PreparedStatement prepareStatement(String sql, int resultSetType,
                                              int resultSetConcurrency, int resultSetHoldability)
            throws SQLException {
        throw new UnsupportedOperationException();
    }

    public CallableStatement prepareCall(String sql, int resultSetType,
                                         int resultSetConcurrency,
                                         int resultSetHoldability) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys)
            throws SQLException {
        throw new UnsupportedOperationException();
    }

    public PreparedStatement prepareStatement(String sql, int columnIndexes[])
            throws SQLException {
        throw new UnsupportedOperationException();
    }

    public PreparedStatement prepareStatement(String sql, String columnNames[])
            throws SQLException {
        throw new UnsupportedOperationException();
    }

    public Statement createStatement(int resultSetType, int resultSetConcurrency)
            throws SQLException {
        throw new UnsupportedOperationException();
    }

    public PreparedStatement prepareStatement(String sql, int resultSetType,
                                              int resultSetConcurrency)
            throws SQLException {
        throw new UnsupportedOperationException();
    }

    public CallableStatement prepareCall(String sql, int resultSetType,
                                         int resultSetConcurrency) throws SQLException {
        throw new UnsupportedOperationException();
    }

    /****************************************************************************
     * Other important methods
     */

    public void setAutoCommit(boolean autoCommit) throws SQLException {
        this.autoCommit = autoCommit;
    }

    public boolean getAutoCommit() throws SQLException {
        return autoCommit;
    }

    public void commit() throws SQLException {
        if(autoCommit) {
            throw new SQLException("Commit not legal in AutoCommit mode");
        }

        if(! isTransactionActive()) {
            return;
        }

        try{
            physicalCon.commit();
            setTransactionActive(false);
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);
        }
    }

    public void rollback() throws SQLException {
        if(! isTransactionActive()) {
            //  TODO:  something more drastic if a specific savepoint is mentioned
            return;
        }

        if(autoCommit) {
            throw new SQLException("Rollback not legal in AutoCommit mode");
        }

        try{
        physicalCon.rollback(null);
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);
        }
    }

    /*****************************************************
     * JDK1.4 methods
     */

    public Savepoint setSavepoint() throws SQLException {
        return markSavepoint(new F3Savepoint(savepointCounter++));
    }

    public Savepoint setSavepoint(String name) throws SQLException {
        return markSavepoint(new F3Savepoint(name, savepointCounter++));
    }

    private Savepoint markSavepoint(Savepoint s) throws SQLException {
        if(autoCommit) {
            throw new SQLException("Savepoint not legal in AutoCommit mode");
        }
        try{
            physicalCon.executeStatement("SAVEPOINT " + s.getSavepointName(), false);
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);
        }
        return s;
    }

    public void rollback(Savepoint savepoint) throws SQLException {
        if(! isTransactionActive()) {
            //  TODO:  something more drastic if a specific savepoint is mentioned
            return;
        }
        throw new UnsupportedOperationException("SAVEPOINTS EXIST IN THE DATABASE BUT NOT YET IN JDBC!!");
    }


    public void releaseSavepoint(Savepoint savepoint) throws SQLException {
        //  TODO:  implement this in the actual DB?.  Currently ignore it.  No reason to do this
    }


    int holdability = ResultSet.HOLD_CURSORS_OVER_COMMIT;
    public void setHoldability(int h) throws SQLException {
        if(h != ResultSet.HOLD_CURSORS_OVER_COMMIT &&
           h != ResultSet.CLOSE_CURSORS_AT_COMMIT) {
            throw new SQLException("value " + h + " not legal");
        }
    }

    public int getHoldability() throws SQLException {
        return holdability;
    }

     /*
     END JDK1.3
     */

    /**
     * Closes the connection and rolls back any pending transaction.  This may not be the ideal situation, but
     * that is what it does!  It specifically does this so that system test writing can be easier.  Change this
     * and many fail.
     */
    public void close() throws SQLException {
        if(isClosed()) {
            return;
        }
        try{
            physicalCon.close(false);
            isClosed = true;
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);
        }
    }

    public void finalize() {
        try{
            close();
        }catch(SQLException e) { }
    }

    public boolean isClosed() throws SQLException {
        return isClosed;
    }

    public DatabaseMetaData getMetaData() throws SQLException {
        return new F3DatabaseMetaData(this);
    }

    public void setTransactionIsolation(int level) throws SQLException {
        String mode = null;

        switch(level) {
            case TRANSACTION_READ_UNCOMMITTED:  mode = "READ UNCOMMITTED";  break;
            case TRANSACTION_READ_COMMITTED:    mode = "READ COMMITTED";    break;
            case TRANSACTION_REPEATABLE_READ:   mode = "REPEATABLE READ";   break;
            case TRANSACTION_SERIALIZABLE:      mode = "SERIALIZABLE";      break;
            default:  throw new SQLException("Don't recognize isolation level " + level);
        }

        try{
            physicalCon.executeStatement("set transaction isolation level " + mode, false);
            isolationLevel = level;
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);
        }
    }

    /**
     * Currently only reflects the isolation level that was set by setTransactionIsolationLevel, not
     * by direct statements
     * @return
     * @throws SQLException
     */
    public int getTransactionIsolation() throws SQLException {
        //  TODO:  the first time, this should go through to the underlying DB?
        return isolationLevel;
    }

    public SQLWarning getWarnings() throws SQLException {
        return null;
    }

    public void clearWarnings() throws SQLException {
    }

    public Map getTypeMap() throws SQLException {
        return null;
    }

    public void setTypeMap(Map map) throws SQLException {
    }







    /************************************************************************
     * ignored features
     */
    public void setReadOnly(boolean readOnly) throws SQLException {
        //  ignore, no special read only capabilities
    }

    public boolean isReadOnly() throws SQLException {
        return false;
    }

    public void setCatalog(String catalog) throws SQLException {
    }

    public String getCatalog() throws SQLException {
        return null;
    }

    /**
     * Not sure what this does exactly
     */
    public String nativeSQL(String sql) throws SQLException {
        return sql;
    }
}
