/*
 * Copyright JJ Furman
 * User: user
 * Date: May 14, 2002
 * Time: 3:24:29 PM
 */
package freebase.jdbc;

import freebase.server.BatchResult;
import freebase.util.Bool;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.net.URL;
import java.rmi.RemoteException;
import java.sql.*;
import java.util.ArrayList;
import java.util.Calendar;

public class F3PreparedStatement extends F3Statement implements PreparedStatement {
    /**
     * Represents a true null value, where null means not set.
     */
    final static private Object UNASSIGNED_PARAMETER = new Object();

    /**
     * The preparedSql
     */
    String sql;

    /**
     * Current parameters, where the first is @ 0 (unlike JDBC). A null means null value,
     * UNASSIGNED_PARAMETER means that the value is unassigned as of yet.
     */
    ArrayList parameters = new ArrayList();

    /**
     * lazy allocated, create when addBatch is used
     */
    ArrayList batchParams;

    /**
     * Construct as new PreparedStatement.
     * This does not prepare the statement on the server until the last
     * possible moment.
     */
    protected F3PreparedStatement(F3Connection con, String sql) {
        super(con);

        this.sql = sql;
    }

    /*********************************************************************
     *
     */

    /**
     * Performs only the most basic validation
     */
    private void validateParameters() throws SQLException {
        StringBuffer errorBuffer = null;  //  lazy allocation this when needed

        for(int i = 0; i < parameters.size(); i++) {
            if(parameters.get(i) == UNASSIGNED_PARAMETER) {
                if(errorBuffer == null) {
                    errorBuffer = new StringBuffer("The following parameters not set: ");
                } else {
                    errorBuffer.append(',');
                }

                errorBuffer.append(i);
            }
        }

        if(errorBuffer != null) {
            throw new SQLException(errorBuffer.toString());
        }
    }

    /**
     * Will be supported in the future.
     */
    public void addBatch() throws SQLException {
        validateParameters();
        if(batchParams == null) {
            batchParams = new ArrayList();
        }

        Comparable[] row = new Comparable[parameters.size()];
        parameters.toArray(row);
        batchParams.add(row);
        parameters.clear();

        //  TODO:  consider sending the data now if the set is large enough!
    }

    public void clearParameters() throws SQLException {
        parameters.clear();
    }

    public boolean execute() throws SQLException {
        validateParameters();

        return executeCommand(sql, parameters);
    }

    public ResultSet executeQuery() throws SQLException {
        validateParameters();

        if(! execute()) {
            throw new SQLException("'" + sql + "' is not a query.  Warning, it may have made changes!");
        }

        return getResultSet();
    }

    public int executeUpdate() throws SQLException {
        validateParameters();

        if(execute()) {
            throw new SQLException("'" + sql + "' is not a query");
        }

        return getUpdateCount();
    }


    /**
     * Not currently supported.<BR>
     * I don't see how it is always possible to know what data this prepared statement
     * is always going to return.  What if the stet
     */
    public ResultSetMetaData getMetaData() throws SQLException {
        return null;
    }

    /**
     * Not currently supported.
     */
    public ParameterMetaData getParameterMetaData() throws SQLException {
        return null;
    }

    /**
     * All statements in the batch are executed atomically.  Thus if one fails,
     * they all fail.  This may not be the prefered implementation, but that is how it
     * is done currently.
     */
    public int[] executeBatch() throws SQLException {
        beginCommand();

        try{
            if(batchParams == null) {
                 return new int[0];
            }

            //String[] statements = new String[statementBatch.size()];
            //statementBatch.toArray(statements);

            BatchResult result = con.getPhysicalConnection().executeBatch(sql, batchParams, con.getAutoCommit());
            batchParams.clear();
            
            return result.getUpdateCounts();
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);

              //  note: this will never get run...
            return new int[0];
        }
    }

    /**********************************************************************
     * All the various setters
     */

    /**
     * The chief setter.
     * TODO:  probably deal with Boolean object here, since currently it is dealt with in setBoolean
     */
    public void setObject(int parameterIndex, Object x) throws SQLException {
        if(parameters.size() < parameterIndex) {
            for(int i = parameters.size(); i < parameterIndex; i++) {
                parameters.add(UNASSIGNED_PARAMETER);
            }

            parameters.set(parameterIndex - 1, x);
        }
    }

    ///////  the rest...

    public void setArray(int i, Array x) throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setAsciiStream(int parameterIndex, InputStream x, int length)
            throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setBinaryStream(int parameterIndex, InputStream x,
                                int length) throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setBlob(int i, Blob x) throws SQLException {
        //  TODO:  wrap the Blob object with some kind of reference or something.
        throw new SQLException("This type not yet supported");
    }

    public void setBoolean(int parameterIndex, boolean x) throws SQLException {
        setObject(parameterIndex, Bool.getBool(x));
    }

    public void setBytes(int parameterIndex, byte x[]) throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setCharacterStream(int parameterIndex,
                                   Reader reader,
                                   int length) throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setClob(int i, Clob x) throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setDate(int parameterIndex, Date x)
            throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setDate(int parameterIndex, Date x, Calendar cal)
            throws SQLException {
        throw new SQLException("This method not yet supported");
    }

    public void setDouble(int parameterIndex, double x) throws SQLException {
        setObject(parameterIndex, new Double(x));
    }

    public void setFloat(int parameterIndex, float x) throws SQLException {
        setObject(parameterIndex, new Float(x));
    }

    public void setByte(int parameterIndex, byte x) throws SQLException {
        setObject(parameterIndex, new Byte(x));
    }

    public void setShort(int parameterIndex, short x) throws SQLException {
        setObject(parameterIndex, new Short(x));
    }

    public void setInt(int parameterIndex, int x) throws SQLException {
        setObject(parameterIndex, new Integer(x));
    }

    public void setLong(int parameterIndex, long x) throws SQLException {
        setObject(parameterIndex, new Long(x));
    }

    public void setNull(int parameterIndex, int sqlType) throws SQLException {
        setObject(parameterIndex, null);
    }

    public void setNull(int paramIndex, int sqlType, String typeName)
            throws SQLException {
        setObject(paramIndex, null);
    }

    public void setObject(int parameterIndex, Object x, int targetSqlType) throws SQLException {
          //  TODO:  validate the type I guess!
        setObject(parameterIndex, x);
    }

    public void setObject(int parameterIndex, Object x, int targetSqlType, int scale)
            throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setRef(int i, Ref x) throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setString(int parameterIndex, String x) throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setTime(int parameterIndex, Time x)
            throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setTime(int parameterIndex, Time x, Calendar cal)
            throws SQLException {
        throw new SQLException("This method not yet supported");
    }

    public void setTimestamp(int parameterIndex, Timestamp x)
            throws SQLException {
        setObject(parameterIndex, x);
    }

    public void setTimestamp(int parameterIndex, Timestamp x, Calendar cal)
            throws SQLException {
        throw new SQLException("This method not yet supported");
    }

    public void setUnicodeStream(int parameterIndex, InputStream x,
                                 int length) throws SQLException {
        throw new SQLException("This type not yet supported");
    }

    public void setURL(int parameterIndex, URL x) throws SQLException {
        throw new SQLException("This type not yet supported");
    }
}
