/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 25, 2002
 * Time: 12:01:13 PM
 */
package freebase.jdbc;

import freebase.api.TupleIterator;
import freebase.server.QueryResult;
import freebase.util.NumberUtil;

import java.io.InputStream;
import java.io.Reader;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.URL;
import java.sql.*;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Map;

public class F3ResultSet implements ResultSet {
    F3Statement st;
    TupleIterator source;

      //  if it has been created, then this will be set
    F3ResultSetMetaData metaData;

    boolean isClosed = false;
    int recentReadColumn = 0;

    HashMap columnMap = null;

    /**
     *
     */
    protected F3ResultSet(F3Statement st, TupleIterator source) throws SQLException {
        this.st = st;
        this.source = source;
    }

    /*********************************************************************
     * standard methods
     */


    public boolean next() throws SQLException {
        checkClosed();

        boolean hasMore = source.next();
        if(! hasMore)
            return false;
        recentReadColumn = 0;

        return true;
    }

    /**
     * Close the result set.
     */
    public void close() throws SQLException {
        source.close();

        isClosed = true;

        st.clearCurrentResults();
    }

    public Statement getStatement() throws SQLException {
        return st;
    }

    /**
     * Finds the columnId for the named column.
     *
     * TODO:  do this first, then the rest should be easy!
     */
    public int findColumn(String columnName) throws SQLException {
        if(columnMap == null) {
            HashMap map = new HashMap();

            for(int i = 0; i < source.getNumColumns(); i++) {
                String column = source.getColumnName(i);
                map.put(column, new Integer(i + 1));
            }

            columnMap = map;
        }

        return ((Integer) columnMap.get(columnName)).intValue();
    }

    public void setFetchSize(int rows) throws SQLException {
        if(source instanceof QueryResult) {
            ((QueryResult)source).setFetchSize(rows);
        }
    }

    public int getFetchSize() throws SQLException {
        if(source instanceof QueryResult) {
            return ((QueryResult)source).getFetchSize();
        } else {
            return Integer.MAX_VALUE;
        }
    }

    /****************************
     * Internal helper methods
     */

    /**
     * Records the fact that the named (numbered) column is about to be read.
     * If the result set is closed, then an error occurs.  Also checks the value
     * range of this column, and throws an error if it is bad
     */
    private void aboutToRead(int column) throws SQLException {
        checkClosed();

        //if(column < 1 || column > source.getNumColumns()) {
       //     throw new SQLException()
        //}

        recentReadColumn = column;
    }

    private void checkClosed() throws SQLException {
        if(isClosed)
            throw new SQLException("Result Set has already been closed");
    }

    /**
     * Possibly called from the F3Statement
     */
    public boolean isClosed() {
        return isClosed;
    }

    /******************************************************************
     * reading methods
     */

    Number getNumber(int columnIndex) throws SQLException {
        Number num = (Number)getObject(columnIndex);
        if(num == null) {
            return BigInteger.ZERO;
        }
        return num;
    }

    Number getNumber(String columnName) throws SQLException {
        Number num = (Number)getObject(columnName);
        if(num == null) {
            return BigInteger.ZERO;
        }
        return num;
    }

    /**
     * For the current implementation, this is the central method
     */
    public Object getObject(int columnIndex) throws SQLException {
        aboutToRead(columnIndex);
        return source.getValue(columnIndex - 1);
    }

    public boolean wasNull() throws SQLException {
        return getObject(recentReadColumn) == null;
    }

    public String getString(int columnIndex) throws SQLException {
        return (String)getObject(columnIndex);
    }

    public boolean getBoolean(int columnIndex) throws SQLException {
        return false;
    }

    public byte getByte(int columnIndex) throws SQLException {
        return getNumber(columnIndex).byteValue();
    }

    public short getShort(int columnIndex) throws SQLException {
        return getNumber(columnIndex).shortValue();
    }

    //  TODO:  rewrite all these with care WRT overflow values
    public int getInt(int columnIndex) throws SQLException {
        return getNumber(columnIndex).intValue();
    }

    public long getLong(int columnIndex) throws SQLException {
        return getNumber(columnIndex).longValue();
    }

    public float getFloat(int columnIndex) throws SQLException {
        return getNumber(columnIndex).floatValue();
    }

    public double getDouble(int columnIndex) throws SQLException {
        return getNumber(columnIndex).doubleValue();
    }

    public BigDecimal getBigDecimal(int columnIndex, int scale) throws SQLException {
        return NumberUtil.convertToBigDecimal(getNumber(columnIndex));
    }

    public byte[] getBytes(int columnIndex) throws SQLException {
        throw new UnsupportedOperationException("no blobs just just yet, though not too hard to add in a simple way");
        //return new byte[0];
    }

    public Date getDate(int columnIndex) throws SQLException {
        return (Date)getObject(columnIndex);
    }

    public Time getTime(int columnIndex) throws SQLException {
        return (Time)getObject(columnIndex);
    }

    public Timestamp getTimestamp(int columnIndex) throws SQLException {
        return (Timestamp)getObject(columnIndex);
    }

    public InputStream getAsciiStream(int columnIndex) throws SQLException {
        throw new UnsupportedOperationException("no blobs just just yet, though not too hard to add in a simple way");
        //return null;
    }

    public InputStream getUnicodeStream(int columnIndex) throws SQLException {
        throw new UnsupportedOperationException("no blobs just just yet, though not too hard to add in a simple way");
        //return null;
    }

    public InputStream getBinaryStream(int columnIndex)
            throws SQLException {
        //return null;
        throw new UnsupportedOperationException("no blobs just just yet, though not too hard to add in a simple way");
    }

    /**
     * TODO:  incorporate the map
     * @param i
     * @param map
     * @return
     * @throws SQLException
     */
    public Object getObject(int i, Map map) throws SQLException {
        return getObject(i);
    }

    public Ref getRef(int i) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public Blob getBlob(int i) throws SQLException {
        throw new UnsupportedOperationException();
        //turn null;
    }

    public Clob getClob(int i) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public Array getArray(int i) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public Date getDate(int columnIndex, Calendar cal) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public Time getTime(int columnIndex, Calendar cal) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public Timestamp getTimestamp(int columnIndex, Calendar cal)
            throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public URL getURL(int columnIndex) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public Reader getCharacterStream(int columnIndex) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    public BigDecimal getBigDecimal(int columnIndex) throws SQLException {
        throw new UnsupportedOperationException();
        //return null;
    }

    /**********************************************************
     * Named columns, probably don't need to change ever again!
     */

    public String getString(String columnName) throws SQLException {
        return getString(findColumn(columnName));
    }

    public boolean getBoolean(String columnName) throws SQLException {
        return getBoolean(findColumn(columnName));
    }

    public byte getByte(String columnName) throws SQLException {
        return getByte(findColumn(columnName));
    }

    public short getShort(String columnName) throws SQLException {
        return getShort(findColumn(columnName));
    }

    public int getInt(String columnName) throws SQLException {
        return getInt(findColumn(columnName));
    }

    public long getLong(String columnName) throws SQLException {
        return getLong(findColumn(columnName));
    }

    public float getFloat(String columnName) throws SQLException {
        return getFloat(findColumn(columnName));
    }

    public double getDouble(String columnName) throws SQLException {
        return getDouble(findColumn(columnName));
    }

    public BigDecimal getBigDecimal(String columnName, int scale) throws SQLException {
        return getBigDecimal(findColumn(columnName));
    }

    public byte[] getBytes(String columnName) throws SQLException {
        return getBytes(findColumn(columnName));
    }

    public Date getDate(String columnName) throws SQLException {
        return getDate(findColumn(columnName));
    }

    public Time getTime(String columnName) throws SQLException {
        return getTime(findColumn(columnName));
    }

    public Timestamp getTimestamp(String columnName) throws SQLException {
        return getTimestamp(findColumn(columnName));
    }

    public InputStream getAsciiStream(String columnName) throws SQLException {
        return getAsciiStream(findColumn(columnName));
    }

    public InputStream getUnicodeStream(String columnName) throws SQLException {
        return getUnicodeStream(findColumn(columnName));
    }

    public InputStream getBinaryStream(String columnName)
            throws SQLException {
        return getBinaryStream(findColumn(columnName));
    }

    public Object getObject(String colName, Map map) throws SQLException {
        return getObject(findColumn(colName), map);
    }

    public Ref getRef(String colName) throws SQLException {
        return getRef(findColumn(colName));
    }

    public Blob getBlob(String colName) throws SQLException {
        return getBlob(findColumn(colName));
    }

    public Clob getClob(String colName) throws SQLException {
        return getClob(findColumn(colName));
    }

    public Array getArray(String colName) throws SQLException {
        return getArray(findColumn(colName));
    }

    public Timestamp getTimestamp(String columnName, Calendar cal)
            throws SQLException {
        return getTimestamp(findColumn(columnName), cal);
    }

    public URL getURL(String columnName) throws SQLException {
        return getURL(findColumn(columnName));
    }

    public Object getObject(String columnName) throws SQLException {
        return getObject(findColumn(columnName));
    }

    public Reader getCharacterStream(String columnName) throws SQLException {
        return getCharacterStream(findColumn(columnName));
    }

    public BigDecimal getBigDecimal(String columnName) throws SQLException {
            return getBigDecimal(findColumn(columnName));
    }

    public Date getDate(String columnName, Calendar cal) throws SQLException {
        return getDate(findColumn(columnName), cal);
    }

    public Time getTime(String columnName, Calendar cal) throws SQLException {
        return getTime(findColumn(columnName), cal);
    }

    /**************************************************************************
     * Non read/write methods
     */

    public SQLWarning getWarnings() throws SQLException {
        return null;
    }

    public void clearWarnings() throws SQLException {
    }

    public String getCursorName() throws SQLException {
        return null;
    }

    /**
     * Note that this is not locked correctly for concurrent use
     */
    public ResultSetMetaData getMetaData() throws SQLException {
        if(metaData == null) {
            metaData = new F3ResultSetMetaData(this, source);
        }
        return metaData;
    }

    public boolean isBeforeFirst() throws SQLException {
        return false;
    }

    public boolean isAfterLast() throws SQLException {
        return false;
    }

    public boolean isFirst() throws SQLException {
        return false;
    }

    public boolean isLast() throws SQLException {
        return false;
    }

    public void beforeFirst() throws SQLException {
    }

    public void afterLast() throws SQLException {
    }

    public boolean first() throws SQLException {
        return false;
    }

    public boolean last() throws SQLException {
        return false;
    }

    public int getRow() throws SQLException {
        return 0;
    }

    public boolean absolute(int row) throws SQLException {
        return false;
    }

    public boolean relative(int rows) throws SQLException {
        return false;
    }

    public boolean previous() throws SQLException {
        return false;
    }

    public void setFetchDirection(int direction) throws SQLException {
    }

    public int getFetchDirection() throws SQLException {
        return 0;
    }

    public int getType() throws SQLException {
        return 0;
    }

    public int getConcurrency() throws SQLException {
        return 0;
    }

    public boolean rowUpdated() throws SQLException {
        return false;
    }

    public boolean rowInserted() throws SQLException {
        return false;
    }

    public boolean rowDeleted() throws SQLException {
        return false;
    }

    /*****************************************************************************************
     * Update methods, for now ignored  TODO:  don't ignore these...
     */

    public void updateNull(int columnIndex) throws SQLException {
    }

    public void updateBoolean(int columnIndex, boolean x) throws SQLException {
    }

    public void updateByte(int columnIndex, byte x) throws SQLException {
    }

    public void updateShort(int columnIndex, short x) throws SQLException {
    }

    public void updateInt(int columnIndex, int x) throws SQLException {
    }

    public void updateLong(int columnIndex, long x) throws SQLException {
    }

    public void updateFloat(int columnIndex, float x) throws SQLException {
    }

    public void updateDouble(int columnIndex, double x) throws SQLException {
    }

    public void updateBigDecimal(int columnIndex, BigDecimal x) throws SQLException {
    }

    public void updateString(int columnIndex, String x) throws SQLException {
    }

    public void updateBytes(int columnIndex, byte x[]) throws SQLException {
    }

    public void updateDate(int columnIndex, Date x) throws SQLException {
    }

    public void updateTime(int columnIndex, Time x) throws SQLException {
    }

    public void updateTimestamp(int columnIndex, Timestamp x)
            throws SQLException {
    }

    public void updateAsciiStream(int columnIndex,
                                  InputStream x,
                                  int length) throws SQLException {
    }

    public void updateBinaryStream(int columnIndex,
                                   InputStream x,
                                   int length) throws SQLException {
    }

    public void updateCharacterStream(int columnIndex,
                                      Reader x,
                                      int length) throws SQLException {
    }

    public void updateObject(int columnIndex, Object x, int scale)
            throws SQLException {
    }

    public void updateObject(int columnIndex, Object x) throws SQLException {
    }

    public void updateNull(String columnName) throws SQLException {
    }

    public void updateBoolean(String columnName, boolean x) throws SQLException {
    }

    public void updateByte(String columnName, byte x) throws SQLException {
    }

    public void updateShort(String columnName, short x) throws SQLException {
    }

    public void updateInt(String columnName, int x) throws SQLException {
    }

    public void updateLong(String columnName, long x) throws SQLException {
    }

    public void updateFloat(String columnName, float x) throws SQLException {
    }

    public void updateDouble(String columnName, double x) throws SQLException {
    }

    public void updateBigDecimal(String columnName, BigDecimal x) throws SQLException {
    }

    public void updateString(String columnName, String x) throws SQLException {
    }

    public void updateBytes(String columnName, byte x[]) throws SQLException {
    }

    public void updateDate(String columnName, Date x) throws SQLException {
    }

    public void updateTime(String columnName, Time x) throws SQLException {
    }

    public void updateTimestamp(String columnName, Timestamp x)
            throws SQLException {
    }

    public void updateAsciiStream(String columnName,
                                  InputStream x,
                                  int length) throws SQLException {
    }

    public void updateBinaryStream(String columnName,
                                   InputStream x,
                                   int length) throws SQLException {
    }

    public void updateCharacterStream(String columnName,
                                      Reader reader,
                                      int length) throws SQLException {
    }

    public void updateObject(String columnName, Object x, int scale)
            throws SQLException {
    }

    public void updateObject(String columnName, Object x) throws SQLException {
    }

    public void insertRow() throws SQLException {
    }

    public void updateRow() throws SQLException {
    }

    public void deleteRow() throws SQLException {
    }

    public void refreshRow() throws SQLException {
    }

    public void cancelRowUpdates() throws SQLException {
    }

    public void moveToInsertRow() throws SQLException {
    }

    public void moveToCurrentRow() throws SQLException {
    }

    public void updateRef(int columnIndex, Ref x) throws SQLException {
    }

    public void updateRef(String columnName, Ref x) throws SQLException {
    }

    public void updateBlob(int columnIndex, Blob x) throws SQLException {
    }

    public void updateBlob(String columnName, Blob x) throws SQLException {
    }

    public void updateClob(int columnIndex, Clob x) throws SQLException {
    }

    public void updateClob(String columnName, Clob x) throws SQLException {
    }

    public void updateArray(int columnIndex, Array x) throws SQLException {
    }

    public void updateArray(String columnName, Array x) throws SQLException {
    }
}
