/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 25, 2002
 * Time: 11:51:42 AM
 */
package freebase.jdbc;

import freebase.api.TupleIterator;
import freebase.server.ActionResult;
import freebase.server.BatchResult;

import java.rmi.RemoteException;
import java.sql.*;
import java.util.ArrayList;

public class F3Statement implements Statement {
    F3Connection con;

    Object recentStatementResults;

    ResultSet currentResults = null;
    long numericalResult = 0;

    ArrayList statementBatch;


    protected F3Statement(F3Connection con) {
        this.con = con;
    }

    /*************************************************************************
     * Execute methods
     */

    public ResultSet executeQuery(String sql) throws SQLException {
        if(! execute(sql)) {
            throw new SQLException("'" + sql + "' is not a query.  Warning, it may have made changes!");
        }

        return getResultSet();
    }

    public int executeUpdate(String sql) throws SQLException {
        if(execute(sql)) {
            throw new SQLException("'" + sql + "' is not a query");
        }

        return getUpdateCount();
    }

    /**
     * Currently the core execution method
     */
    public boolean execute(String sql) throws SQLException {
        return executeCommand(sql, null);
    }

    /**
     * Executes the command on the server.  If the parameters are null, it
     * will assume this is a normal command
     */
    protected boolean executeCommand(String sql, ArrayList params) throws SQLException {
        beginCommand();

        try{
            Object result = con.getPhysicalConnection().executeStatement(sql, params, con.getAutoCommit());
            if(result == null) {
                return false;
            } else if(result instanceof ActionResult) {
                ActionResult action = (ActionResult)result;
                numericalResult = action.getNumRowsEffected();

                //  TODO:  update local log position from action result!!!!
                //    and, make lock table return an action result!

                return false;
            } else if(result instanceof TupleIterator) {
                currentResults = new F3ResultSet(this, (TupleIterator)result);

                return true;
            } else {
                throw new SQLException("Critical Error:  Returned object " + result +
                        "  of type: " + result.getClass() + " not recognized by JDBC driver");
            }

              //  this is needed even in the case that only queries were executed,
              //  since those resources need to be released too

        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);
              //  note: this will never happen
            return false;
        }
    }

    /***************************************
     * Currently unsupported variants
     */

    public int executeUpdate(String sql, int autoGeneratedKeys) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public int executeUpdate(String sql, int columnIndexes[]) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public int executeUpdate(String sql, String columnNames[]) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public boolean execute(String sql, int autoGeneratedKeys) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public boolean execute(String sql, int columnIndexes[]) throws SQLException {
        throw new UnsupportedOperationException();
    }

    public boolean execute(String sql, String columnNames[]) throws SQLException {
        throw new UnsupportedOperationException();
    }


    /**************************************************
     * Batch methods
     */

    /**
     * To implement batch, just add the command to an ArrayList.
     * Then execute the whole thing at once... much faster.
     */
    public void addBatch(String sql) throws SQLException {
        if(statementBatch == null) {
            statementBatch = new ArrayList();
        }

        statementBatch.add(sql);
    }

    public void clearBatch() throws SQLException {
        statementBatch = null;
    }

    /**
     * All statements in the batch are executed atomically.  Thus if one fails,
     * they all fail.  This may not be the prefered implementation, but that is how it
     * is done currently.
     */
    public int[] executeBatch() throws SQLException {
        beginCommand();

        try{
            if(statementBatch == null) {
                 return new int[0];
            }

            String[] statements = new String[statementBatch.size()];
            statementBatch.toArray(statements);

            BatchResult result = con.getPhysicalConnection().executeBatch(statements, con.getAutoCommit());
            statementBatch.clear();
            return result.getUpdateCounts();
        }catch(RemoteException re) {
            F3Driver.convertRemoteException(re);

              //  note: this will never get run...
            return new int[0];
        }
    }

    /*************************************************************************
     * Other methods
     */

    public ResultSet getResultSet() throws SQLException {
        return currentResults;
    }

    public int getUpdateCount() throws SQLException {
        if(numericalResult > Integer.MAX_VALUE) {
            //  TODO:  add SQLWarning in this case
        }
        return (int)numericalResult;
    }

    public boolean getMoreResults() throws SQLException {
        return false;
    }

    public void close() throws SQLException {
        if(currentResults != null) {
            currentResults.close();
        }
    }

    /**
     * What is this nonsense?
     */
    public int getMaxFieldSize() throws SQLException {
        return 0;
    }

    /**
     * What is this nonsense?
     */
    public void setMaxFieldSize(int max) throws SQLException {
    }

    public int getMaxRows() throws SQLException {
        return 0;
    }

    public void setMaxRows(int max) throws SQLException {
    }

    public void setEscapeProcessing(boolean enable) throws SQLException {
    }

    public int getQueryTimeout() throws SQLException {
        return 0;
    }

    public void setQueryTimeout(int seconds) throws SQLException {
    }

    public void cancel() throws SQLException {
    }

    public SQLWarning getWarnings() throws SQLException {
        return null;
    }

    public void clearWarnings() throws SQLException {
    }

    public void setCursorName(String name) throws SQLException {
    }

    public void setFetchDirection(int direction) throws SQLException {
    }

    public int getFetchDirection() throws SQLException {
        return 0;
    }

    /**
     * TODO:  make this value the starting value...
     */
    public void setFetchSize(int rows) throws SQLException {
    }

    public int getFetchSize() throws SQLException {
        return 0;
    }

    public int getResultSetConcurrency() throws SQLException {
        return 0;
    }

    public int getResultSetType() throws SQLException {
        return 0;
    }

    public Connection getConnection() throws SQLException {
        return con;
    }

    public boolean getMoreResults(int current) throws SQLException {
        return false;
    }

    public ResultSet getGeneratedKeys() throws SQLException {
        return null;
    }

    public int getResultSetHoldability() throws SQLException {
        return 0;
    }

    /**********************************************************************
     * local helper methods
     */
    protected void clearCurrentResults() {
        currentResults = null;
    }

    /**
     * Will fail if another result set is currently active (in this statement presumably)
     */
    protected void beginCommand() throws SQLException {
        if(currentResults != null) {
            throw new SQLException("Another Result Set is still active");
        }
        numericalResult = 0;

        con.setTransactionActive(true);
    }
}
