package freebase.lock;

import freebase.api.DeadlockableResource;
import freebase.engine.Transaction;

import java.sql.SQLException;

/**
 * Represents a wait for an exclusive locked resource.  When the resource is ours,
 * this will be called.
 *
 * TODO:  add a   boolean haveLock() method that can allow a system to quickly detect whether it has
 * a lock immediately.  This will be used when applying this pattern to table locks to prevent always needing
 * to pause the scan thread for every row.  Alternative is to return "null" meaning no wait is required.
 *
 * Copyright 2004 JJ Furman
 * Time: 8:08:23 PM
 */
public class EnqueuedWait implements DeadlockableResource, LockWait {
    /**
     * the transaction that initated this request
     */
    Transaction initiator;

    /**
     * set to true when this queued waiter has been granted the lock.  Prevents spurrious
     * wait wakeups from assuming they got the lock.
     */
    boolean gotLock = false;

    /**
     * set to true when the transaction waiting for this resource is in deadlock with another transaction,
     */
    boolean isInDeadlock = false;

    public EnqueuedWait next;

    /**
     * the wait queue we live in.  Kept track of in case a deadlock is detected, in
     * which case the
     */
    WaitQueue waitQueue;

    /**
     * When this is set, the thread waiting on the even has already been interrupted.  Thus they are no longer
     * waiting on this monitor anymore.  The next candidate must be selected to wake up.
     */
    boolean wasKilled = false;

    public EnqueuedWait(WaitQueue q, Transaction initiator) {
        this.waitQueue = q;
        this.initiator = initiator;
        if(initiator == null) {
            System.out.println("EnqueuedWait.constructor initiator is null");
        }
    }

    /**
     * return the owner transaction, the one that initiated this lock
     * @return
     */
    public Transaction getInitiator() {
        return initiator;
    }

    /**
     * Gives the lock in question to this waiter
     * @return false if the waiter no longer actually wants the resource, ie has been explicitly instructed
     * not to have it.
     */
    synchronized public boolean giveLock() {
        if(wasKilled) {
            return false;
        }

        gotLock = true;
        notify();
        return true;
    }

    public void notifyOfDeadlock() {
        //System.out.println("NOTIFYING OF DEADLOCK: " + assigned);

        waitQueue.queuedWaitInDeadlock(this);

        synchronized(this) {
            isInDeadlock = true;
            this.notify();
        }
    }

    /**
     * waits for the lock to be available to us.  Only one thread should ever be awaiting this resource
     * per Enqueue.
     * @exception SQLException throws SQLException if there is a deadlock with the current holder
     */
    synchronized public void await() throws SQLException {
        int numWakeups = 0;
        while(! gotLock) {
              //  now safe in case the deadlock already occurred!!!  Can happen fast after edge added
            if(isInDeadlock) {
                throw new SQLException("DEADLOCK DETECTED with : " + waitQueue.getCurrentLockOwner() +
                        " (during lock)");
            }

            try{
                  //  TODO:  get rid of all this crud and leave a single nice wait()
                Transaction prev = this.waitQueue.getCurrentLockOwner();
                wait(1000);

                //  DEBUG
                numWakeups++;
                if(numWakeups > 1000 && numWakeups < 1005) {
                    if(prev == null) {
                        System.out.println("waiting on: null Transaction!!!");
                    } else {
                        System.out.println("waiting on: " + Transaction.statusName[prev.getStatus()]);
                    }
                }

            }catch(InterruptedException e) {
                wasKilled = true;
                throw new SQLException("interrupted (abandoned sql thread)");  //  called if the connection was abandoned
            }
        }
    }
}
