/**
 * Copyright 2004 JJ Furman
 * Time: 8:45:58 AM
 */
package freebase.lock;

import freebase.api.TransactionDoneCallback;
import freebase.engine.Transaction;
import freebase.util.OhShitException;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * A use lock is very similar to Oracles Library Cache Pins.  It prevents structure from changing while
 * a query plan or update plan is being drawn up.  The lock will be released by casual users as soon as
 * they are done, but by structure change events only on transaction completion.  It is a relatively
 * straightforward share/exclusive lock.<BR>
 *
 * This lock also prevents some kinds of transitory weirdness.  For example, if one was to truncate a table
 * while another user is doing a single table self outer join (for example, give me all employees and their
 * managers, even if they don't have one (CEO etc)) then the query planner would need to get two cursors to
 * perform this query.  If the truncate table happened after the first query, but before the second the user
 * might temporarily see all employess with no managers at all.  This should not happened, no matter how rare.<BR>
 *
 * Currently a pretty simple lock implementation.  Use Doug Lea's later for queued operation of shared lock.<p>
 *
 * Probably need to store a deleted flag in here so that grabbing the share lock fails 
 */
public class UseLock {
    /**
     * To ensure that the methods of this are not reachable, despite being public.
     */
    private TransFinishCallback callbackHelper = new TransFinishCallback();

    int numSharedOwners = 0;
    Transaction exclusiveOwner = null;

    /**
     * Lock this resource in shared mode.  This is done before:
     * 1.  creating a query plan
     * 2.  creating an update plan
     * 3.  using a
     *
     * NOTE: a given Use lock can be obtained multiple times by the same caller (and will)
     * for simplicty.
     * @throws SQLException if a wait is interrupted, it will throw an SQLException
     *      An alternative is to assume that locks will eventually be gotten soon since
     *      exclusive locks shouldn't be held for too long.
     */
    public synchronized void lockShared() throws SQLException {
        while(true) {
            if(exclusiveOwner == null) {
                numSharedOwners++;
                return;
            }

            try{
                wait();
            }catch(InterruptedException i) {
                throw new SQLException("thread interrupted while waiting for share lock (rare indeed!)");
            }
        }
    }

    public synchronized void unlockShared() {
        numSharedOwners--;

          //  only need to notify one because there can only be exclusive waiters on this simple lock
        notify();
    }

      //  TODO: add unlockRegisterChangeCallback to demote the share lock to a touch lock

    /**
     * Lock when changing the structure (or structure of sufficiently related component).
     * TODO:  consider reimplementing this only in NO_WAIT mode (?)
     * @param t This will register for notification of transaction commit
     * @throws SQLException if a wait is interrupted, it will throw an SQLException
     */
    public synchronized void lockExclusive(Transaction t) throws SQLException {
          //  do we already have the lock?
        if(exclusiveOwner == t) {
            return;
        }

        while(true) {
            if(numSharedOwners == 0 && exclusiveOwner == null) {
                exclusiveOwner = t;
                t.registerDoneCallback(callbackHelper, null);
                return;
            }

            try{
                wait();
            }catch(InterruptedException i) {
                throw new SQLException("thread interrupted while waiting for share lock (rare indeed!)");
            }
        }
    }

    private synchronized void unlockExclusive(Transaction t) {
          //  basically an assert
        if(exclusiveOwner != t) {
            throw new OhShitException("Trying to unlock this but we don't even own it!!!");
        }

        exclusiveOwner = null;
          //  multiple shared etc
        notifyAll();
    }

    /***********************************************************************/

    /**
     * Atomically locks all the UseLocks passed in.  It will try to grab them all, and if
     * there is a problem, it will back out.
     * @param locks
     * @throws SQLException If there is an interrupt on any of the locks
     */
    public static void lockAllShared(ArrayList locks) throws SQLException {
        int i = 0;
        try{
            for(i = 0; i < locks.size(); i++) {
                ((UseLock)locks.get(i)).lockShared();
            }
        }catch(SQLException s) {
              //  back out, starting with the last one we successfully got, ie i-1
            for(int j = 0; j < i; j++) {
                ((UseLock)locks.get(j)).unlockShared();
            }

            throw s;
        }
    }

    /**
     * Unlocks all the resources previous obtained in shared mode.
     * @param locks
     */
    public static void unlockAllShared(ArrayList locks) {
        for(int i = 0; i < locks.size(); i++) {
            ((UseLock)locks.get(i)).unlockShared();
        }
    }

    /***********************************************************************/

    private class TransFinishCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            unlockExclusive(t);
        }
    }
}
