package freebase.lock;

import freebase.api.TransactionDoneCallback;
import freebase.engine.Transaction;
import freebase.util.LongHashMap;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

/**
 * Lock manager for locking particular values.  Currently in use for implementing unique indexes, but may
 * be generally applicable to foreign keys, as well as a reimplement of row locks.  (simplifying the disk
 * streaming aspect)<BR><BR>
 *
 * For unique keys it works like this:  The "first" element in a list for a particular value are considered
 * to own the lock.  Other entries get added afterwards, in order (the order they appear in the log as well).
 * If a later entry is added while the first entry is still in an active transaction, it waits on that.
 *
 * <BR><BR>  when implementing this for table locks, make sure that the enqueue mechanism is used even when
 * it is suspected that a transaction is committed.  Immediately after a transaction is committed, there may
 * be
 * Copyright Mar 4, 2004 JJ Furman
 * Time: 5:59:27 PM
 */
public class ValueLock implements TransactionDoneCallback {
    /**
     * locks per value.
     * map from value -> WaitLockHolder
     *
     */
    HashMap valueLocks = new HashMap();

    /**
     * Map from transactionId -> LongHashMap acting as a set of rowNums
     * that have waiters for this transaction.  This is a required
     * structure for making commits fast... we don't want to rummage
     * through the log!
     */
    LongHashMap waitsPerTransaction = new LongHashMap();

    /**
     * Unlock the value.  Called either when a rollback or a commit happens
     */
    public void unlock(Object value) {
        //System.out.println("unlock " + value);
        //  wake up the waiters on this value

        boolean alreadyAwokeWaiterIfAvailable = false;
        WaitQueue waitQueue = null;
        EnqueuedWait wait = null;
        Transaction previousOwner = null;

          //  keep looping
        while(! alreadyAwokeWaiterIfAvailable) {
              //  big candidate for ConcurrentHashMap
            synchronized(valueLocks) {
                  //  first pass only
                if(waitQueue == null) {
                    waitQueue = (WaitQueue)valueLocks.get(value);
                      //  if there is nothing possibly waiting
                    if(waitQueue == null) {
                        return;
                    }
                }

                previousOwner = waitQueue.getCurrentLockOwner();
                wait = waitQueue.popNextInLine();

                  //  safe.  Concurrent grabbers can just go ahead and take the lock.
                  //  If the arrived before us (got valueLocks) then this would not have been empty!
                if(wait == null) {
                    valueLocks.remove(value);
                }
            }

              //  go on, cowboy!  Wakeup!
            if(wait != null) {
                  //  TODO:  if the wait has just been interrupted, it will be "killed"
                  //  or something.  In that case, wake up the next guy
                  //  could restructure the code to explicitly syn(wait) and check isKilled before
                boolean stillWantedLock = wait.giveLock();

                if(stillWantedLock) {  //  ie if wasn't killed by an abandoned connection
                    alreadyAwokeWaiterIfAvailable = true;

                      //  after we give the lock to the current owner, it may get sufficient time to commit.
                      //  the method inside ensures that if it commits, we'll unlock the next value ourselves.
                      //  of course, the point is moot if the releaser is rolling back because the resource
                      //  semaphore is already owned, preventing the next transaction thread from doing anything
                      //  for a while.
                    addToTransactionWaits(wait.getInitiator(), value);
                      //  the above is strange because it calls unlock in a nested fashion every so often.
                      //  I find it strange but I think it works ok

                    //  TODO:  THINK! is it possible that above transaction isn't the one we just woke up?
                      //  I guess not since this can't be unlocked twice until the transaction knows
                      //  it is the holder

                      //  must be called if the lock is legal in the loop because the edges are about to be removed above if it loops
                    //waitQueue.addDDWaitingEdgesToLockOwner();
                }
            } else {
                  //  waiter not available, so no problem.
                alreadyAwokeWaiterIfAvailable = true;
            }
        }

          //  TODO:  could have a transferTransactionWait(old, new, value), and call this above
                //  (before giveLock(), and insteadof addToTransactionWaits)
          //  TODO:  could at least have removeFromTransactionWaits
          //  NOTE:  since this will eventually use a concurrent hashmap, it is not critical
          //    to only grab the waitsPerTransaction lock once... it will only do so anyway.
          //  since there is a waitHolder, then remove it from the transaction too
          //  probably can go last for performance, but for now ok
        synchronized(waitsPerTransaction) {
            // System.out.println("Lock released: " + waitHolder);
            long transId = previousOwner.getTransactionId();
            HashSet allWaiters = (HashSet)waitsPerTransaction.get(transId);

              //  only in a rollback situation will this matter
            if(allWaiters != null) {
                allWaiters.remove(value);
            }
        }

        //  NOTE:  could be
        /*
        synchronized(waitsPerTransaction) {
              //  get mutex on waits for this transaction to ensure that these don't get
              //  notified before we get to hook our event listener (our lock wakeup listener)
              //  in
            boolean isOkCandidate = wait.giveLock();

            if(isOkCandidate) {
                _addToTransactionWaits(wait.getInitiator(), value);
            }

              //  NOTE:  this will eventually need its own synchronized body to ensure that
            _removeFromTransactionWaits(previousOwner.getTransactionId(), value);
        }

        */
    }

    /**
     * release all resources.
     * @param t
     */
    public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
        if(finalStatus == Transaction.COMMITED) {
            unlockAll(t);
        }
    }

    /**
     * Unlocks all values associated with a given owner.
     * @param owner
     */
    public void unlockAll(Transaction owner) {
          //  eventually the following code could be in transaction.
        HashSet allWaiters = null;
        synchronized(waitsPerTransaction) {
            long transId = owner.getTransactionId();
            allWaiters = (HashSet)waitsPerTransaction.remove(transId);
        }

        if(allWaiters != null) {
              //  now, for each one unlock...
            Iterator it = allWaiters.iterator();
            while(it.hasNext()) {
                unlock(it.next());
            }
        }
    }

    /**
     * Essentially a lock operation, but it is up to the decision of the caller to decide if it
     * has the lock or not.  If it doesn't, then it re-calls this function.  In the case of indexes,
     * the "next" value in the list will automatically
     * get the lock.  In the case of a row lock, whatever returning waiter that actually gets the block
     * semaphore and does the next update will obtain the lock.  This may be slightly inefficient if there
     * are large numbers of waiters on a piece of data.  <BR><BR>
     *
     * Perhaps a later implemention of this can automatically wake up waiters in the order that
     * they are called, thus being a true "lock" function.  For now, waitOnUnlock() will do.
     * NOTE:  a row lock probably doesn't need to guarentee the ordering, so just passing in the owner is
     * sufficient and makes it easy to implement scalable locks without needing to store lock objects for
     * everything.  NICE!  (but reimplementing that will have to wait another day.)
     *
     * When going to reimplement Table Lock Manager, (now movable to this package cuz generic) the caller
     * must promise to own the block/row semaphor.  This is to ensure that only one thread enters this at a
     * time so that if the lockOwner is indeed still alive that it is the correct next owner.
     *  To do this, an enqueue must occur, and then release the
     * block lock and then wait on the queued object<BR><BR>
     *
     * IMPORTANT NOTE:Assumes that a higher level lock exists to guarentee multiple enqueues
     * for the same thing do not occur.  This is critical to understanding the correctness of this function.
     * The local monitors are for ensuring the HashTables etc don't get corrupted.<BR><BR>
     *
     * This method always returns immediately, ie no waiting.
     * @param value the value to lock on
     * @param lockOwner a hint to the system for who to wait for if there is no queue found
     * for the current row.  This can even be null, in which case if there is no queue just obtain the lock.
     * This is allowed *only* because a higher level lock is assumed when getting the wait object here.
     * @return I've gone back and forth about returning null.  Its very "C" like, which is both quaint
     *  and horrid.   Null return means that the lock has already been gotten, and to go ahead (please)!
     * The alternative is to create a dummy object, which is probably not so bad.  But this is faster!
     */
    public LockWait getLockWait(Object value, Transaction lockOwner, Transaction myTransaction) throws SQLException {
        //System.out.println("lock " + value + "  by " + myTransaction);
        WaitQueue waitQueue = null;
        boolean waitQueueIsNew = false;
        EnqueuedWait wait = null;

          //  grab master lock (currently only one)
        synchronized(valueLocks) {
            waitQueue = (WaitQueue)valueLocks.get(value);

            if(waitQueue == null) {
                  //  this may happen VERY infrequelty, so much so to remove it
                if(lockOwner == null || ! lockOwner.isActive()) {
                    return null;
                }

                  //  if we are the first one to wait for this resource
                waitQueue = new WaitQueue(lockOwner);

                  //  mark us as the current transaction holding this lock
                valueLocks.put(value, waitQueue);
                waitQueueIsNew = true;
            }

              //  This enqueue must be in the valueLocks monitor to ensure that the queue isn't removed
              //  by a committing transaction getting too aggressive.
            wait = new EnqueuedWait(waitQueue, myTransaction);
            waitQueue.enqueue(wait, lockOwner);
        }

          //  deals with it's own concurrency.  it does so by assuring that a transaction is set to COMMITTED
          //  state strictly before any locks are released.  There is no way for the committing transaction
          //  to release the locks before we get in there causing us to hang, because to release the locks requires
          //  first grabbing the transactionWaits monitor, thus the state is COMMITTED.
        if(waitQueueIsNew) {
            addToTransactionWaits(lockOwner, value);
        }

        return wait;
    }

    /**
     * Adds the given value to the list of locks currently held (by virtue of being waited for) to the
     * list of things held by the transaction t.  Since the transaction that we are registering is on another
     * thread, it may concurrently rollback or commit.  If it does so, it will do one of two things:
     * 1.  Not have any other waits scheduled for this ValueLock, thus set will be null
     * 2.  Will have but have already cleared them out in which case set will be null
     * 3.  Will be in the process, but not have cleared them out.  In the case of a ROLL_BACK
     *     this may be problematic!!!!!!!!!!!!!!!
     * @param t
     * @param value
     */
    protected void addToTransactionWaits(Transaction t, Object value) {
          //  this monitor protects the lockOwner from running in and releasing its
          //  owners without including our value if appropriate
        synchronized(waitsPerTransaction) {
              //  if the transaction is rolling back, then this value will get the works by the definition
              //  of the fact that if it did anything, it will be undone.
            if(t.getStatus() == Transaction.ROLLED_BACK) {
                return;
            }

            long transId = t.getTransactionId();

            HashSet set = (HashSet)waitsPerTransaction.get(transId);
            if(set == null) {
                  //  this way it only needs to be registered once... I hope
                  //  if it returns false, then this transaction has already committed
                if(! t.registerDoneCallback(this, null)) {
                    unlock(value);
                    return;
                }
                
                set = new HashSet();
                waitsPerTransaction.put(transId, set);
            }

            set.add(value);
        }
    }
}
