package freebase.lock;

import freebase.api.DeadlockDetector;
import freebase.engine.Transaction;

import java.sql.SQLException;

/**
 * A queue for a specific resource.  Enqueued waiters will be processed in the order they arrive.
 * If the resource is something that multiple transactions can hold at a time (like a block for inittrans)
 * then there is no problem.  As the transactions are complete just call unlock and the next waiter will
 * get the resource.<BR><BR>
 *
 * Note:  a block lock manager will need to keep multiple transactions as holders of a block queue, as opposed
 *   to the simple single holder.
 * Copyright Mar 4, 2004 JJ Furman
 * Time: 8:21:45 PM
 */
public class WaitQueue {
    /**
     * single linked list
     */
    EnqueuedWait head;
    EnqueuedWait tail;

    /**
     * The current lock owner
     */
    Transaction currentLockOwner;

    /**
     * just here for convenience, technically could fetch it every time
     */
    DeadlockDetector detector;

    /**
     * True when the current edges are registered with the DD.  This prevents
     * unnecessary removals when they aren't even currently queued up.  If
     * another waiter is added when these are not queued, it can be assumed that
     * the addDDEdges is going to be called later so don't bother adding the edge
     * at that time.
     * NOTE:  now that removing and adding is done in the same method this variable is not needed
     */
    boolean queuedElementsHaveDeadlockEdges = true;

    /**
     * construct the queue with an owner
     * @param owner
     */
    public WaitQueue(Transaction owner) {
        currentLockOwner = owner;
        detector  = currentLockOwner.getTM().getDeadlockDetector();
    }

    public Transaction getCurrentLockOwner() {
        return currentLockOwner;
    }

    /**
     * Tells the wait queue that one of its elements was found to be in deadlock, and therefore
     * with the current owner.  (since how else could it get stuck?)  Thus the current owner is
     * blocked and won't continue until this transaction rolls back or commits anyway.<BR><BR>
     *
     * Now, remove the deadlocked wait from the queue.
     */
    synchronized public void queuedWaitInDeadlock(EnqueuedWait wait) {
          //  loop over all element until we file
        if(wait == head) {
            head = head.next;
            if(head == null) {
                tail = null;
            }

            return;
        }

          //  find "wait" node
        EnqueuedWait cur = head;
        while(cur.next != wait) {
            cur = cur.next;
        }

          //  skip it
        cur.next = wait.next;
        if(cur.next == null) {
            tail = cur;
        }

        return;
    }

    /**
     * add this wait to the queue.  Synchronized to ensure that adding and leaving the queue is safe.
     * @param wait
     * @param expectedOwner
     * @throws SQLException
     */
    synchronized public void enqueue(EnqueuedWait wait, Transaction expectedOwner) throws SQLException {
        if(expectedOwner != currentLockOwner) {
            //  TODO:  reinstate this
            //throw new SQLException("internal error.  expected owner to be " + expectedOwner + " but was " + currentLockOwner);
        }

        if(head == null) {
            head = wait;
            tail = wait;
        } else {
            tail.next = wait;
            tail = wait;
        }

          //  add waiting edge to deadlock detector
        if(queuedElementsHaveDeadlockEdges) {
            detector.addWaitingEdge(wait.getInitiator(), currentLockOwner, wait);
        }
    }

    /**
     * Called from the lock manager when the top owner is done with the lock, and is thus releasing it.
     * The head will get returned, to be notified by the caller.  It is removed from the list now.  The
     * reason it isn't notified now is that it hasn't been added to the transactions list of value locks.
     * @return false if the list is now empty
     */
    synchronized public EnqueuedWait popNextInLine() {
        if(head == null) {
              //  after the last waiter for a resource, head will be null.  It is used to bridge
              //  the time gap between, since any doubt by a requestor requires an enqueue.
              //  this can happen if the other guy in the list was notified for deadlock, and so this wait
              //  queue is left an empty shell.  Basically, there is nobody else in line!
            currentLockOwner = null;
            return null;
        }

          //  remove all deadlock edges to previous lock holder
        if(queuedElementsHaveDeadlockEdges) {
            EnqueuedWait wait = head;
            while(wait != null) {
                detector.removeWaitingEdge(wait.getInitiator(), currentLockOwner, wait);
                wait = wait.next;
            }

            queuedElementsHaveDeadlockEdges = false;
        }

          //  transfer ownership of lock
        EnqueuedWait newOwner = head;
        currentLockOwner = newOwner.getInitiator();
        if(currentLockOwner == null) {  //  assert that should never happen
            System.out.println("ERROR:  ALERT!!!  currentLockOwner is null somehow!");
        }

       //   head.giveLock();  now called later, from the caller

        head = head.next;
        if(head == null) {
            tail = null;
        }

          //  it might be possible to move this to after the giveLock as it was before to
          //  make only one thread wait for the short work to be done, but probably not
          //  a big optimization
        addDDWaitingEdgesToLockOwner();

        return newOwner;
    }

    /**
     * add all deadlock wait edges.  This can be called anytime after the existing owners are removed,
     * and we are certain that the current owner is indeed legal
     */
    synchronized public void addDDWaitingEdgesToLockOwner() {
        EnqueuedWait wait = head;
        while(wait != null) {
            detector.addWaitingEdge(wait.getInitiator(), currentLockOwner, wait);
            wait = wait.next;
        }

        queuedElementsHaveDeadlockEdges = true;
    }
}
