/**
 * Copyright 2004 JJ Furman
 * Time: 7:08:39 PM
 */
package freebase.log.impl;

import freebase.api.LogManager;
import freebase.engine.Transaction;
import freebase.log.DefaultLogEntryCodes;
import freebase.log.LogEntry;
import freebase.log.LogEntryHeader;
import freebase.log.LogEntryManager;
import freebase.log.tm.CommitLogEntry;
import freebase.util.FileHelper;
import freebase.util.LongHashMap;
import freebase.util.Props;

import java.io.*;
import java.lang.ref.SoftReference;
import java.sql.SQLException;

/**
 * as the name implies, not ideal but better than the simple one.
 *
 * Ultimately a block buffered implementation of this is the way to go.  But for now, in simple cases
 * these kinds of cheesy implementations will work fine for basic testing.  For slightly higher concurrency,
 * while flushing one could record all new entries in a List, and when done flushing could write them all out
 * again. To do this, maybe create instance of a subclass of LogEntryHeader that also have flush flags
 *
 * TODO:  support minor log corruption at the tail.  IE, chop it down to last complete entry
 *
 * Qualities in a fast log:
 *   1.  Flush should wait a while before doing its work to improve boxcarring
 *   2.  Flush should not prevent basic log writes from happening.  Don't know what to do yet about flushing writes.
 *   3.  Reads should try to hit a memory buffer
 *   4.  sequential reads should perform prefetch
 */
public class BetterFileLogManager implements LogManager {
    final public static String LOG_FILENAME_PROPERTY = "LOG_FILENAME";

    final public static int OUTBUFFER_SIZE = 256 * 1024;  //  256K for now

    DataOutputStream out;
    RandomAccessFile in;

    String logFilename;

    /**
     * map from codes to log entries
     */
    LogEntryManager logEntryManager;

    long lastLSN = 0;

    /**
     * curPos of the file (ie, the next LSN)
     */
    long nextLSN = 0;

    Object logTailLatch = new Object();

    long lastFlushLSN = 0;
    //  TODO:  add highest LSN flushed, for later the question isFlushTo(LSN)

    /**
     * Cached buffer -> experimental only, but damn it speeds thing up!
     * SoftReference of LongHashMap (lsn) -> LogEntryHeader
     * <P>
     * this allows the whole thing to be thrown away any time
     */
    SoftReference rowCacheHolder = null;

    /**
     * Prepare the log for reading.
     * TODO:  pass in the cur pos from the log anchor or something
     * @throws IOException
     */
    public BetterFileLogManager() throws IOException, SQLException {
        logFilename = Props.get(LOG_FILENAME_PROPERTY);

        synchronized(logTailLatch) {
            File f = new File(logFilename);
            FileHelper.ensureParentDirExists(f);
            if(! f.exists()) {
                f.createNewFile();
            }

            nextLSN = f.length();  //  set this because if it is zero nothing reads!
            lastFlushLSN = f.length();

            if(nextLSN  != 0) {
                //  TODO:  read the file in?  or just wait for the calls to come?
                //  must find out what the last LSN number is (unless the end of the log always
                //     has that number written)
            }

            //  wait until the logEntryHandler is set
        }
    }

    public long writeEntry(Transaction t, LogEntry entry, boolean flush) throws SQLException {
        try{
        synchronized(logTailLatch) {
            if(entry instanceof CommitLogEntry) {
                if(t.getTransactionId() == 0)
                    new Exception("STACK TRACE").printStackTrace();
            }

            if(entry.getCode() == DefaultLogEntryCodes.ABSTRACT_LOG_CODE) {
                System.out.println("ERROR writing log entry:  type " + entry.getClass() + " does not have it's own log code.");
            }

              //  write out the basic logheader stuff
            writeLogEntryHeader(out, nextLSN, lastLSN, System.currentTimeMillis(), t,
                    entry.getCode(), entry.getSize());

            getEntryCache().put(nextLSN, new LogEntryHeader(nextLSN, lastLSN, nextLSN + entry.getSize() + getHeaderSize(), t, entry));

            entry.writeOut(out);

            lastLSN = nextLSN;
            nextLSN += entry.getSize() + getHeaderSize();

              //  a high performance version of this would not hold the tail lock while writing out!
              //  but I'm not sure how RandomAccessFile was implemented, most likely not multithreaded
              //  TODO:  in fact, while flushing, buffer the entries in memory!!!
            if(flush) {
                //System.out.println("Flush write");
                out.flush();
                lastFlushLSN = lastLSN;
            }

            return lastLSN;
        }
        }catch(IOException ioe) {
            throw new SQLException(ioe.getMessage());
        }
    }

    /**
     * Eventually this will need to internally deal with EOF errors itself
     * @param LSN
     * @return
     * @throws SQLException
     */
    public LogEntryHeader getLogEntry(long LSN) throws SQLException {
        try{
        synchronized(logTailLatch) {
            Object value = getEntryCache().get(LSN);
            if(value != null) {
                return (LogEntryHeader)value;
            }

            //  oh well, go to disk

              //  ensure the underlying system has the data
            if(LSN > lastFlushLSN) {
                out.flush();
                lastFlushLSN = lastLSN;
            }

              //  when reached the end, return null.
              //  TODO:  I don't like the meaning of this...
            if(LSN == nextLSN)
                return null;

            in.seek(LSN);

            //try{
                return readLogEntryHeader(in);
            //}catch(EOFException uhoh) {
                  //  should only happen on first read :)
                //nextLSN = LSN;
             //   return null;
            //}
        }
        }catch(IOException ioe) {
            ioe.printStackTrace();
            throw new SQLException(ioe.getMessage());
        }
    }

    /**
     * Currently, nothing stored in memory so no problem...
     */
    public void raiseLowWaterMark(long hardLSN, long softLSN) {
        //System.out.println("raise " + softLSN);
    }


    public void setLogEntryManager(LogEntryManager logEntryManager) throws SQLException, IOException {
        this.logEntryManager = logEntryManager;

          // a hack for now, since we don't have fast forward scroll prefetch... but ok
        readEntireLog();
    }

    /**
     * reads the log
     */
    private void readEntireLog() throws SQLException, IOException {
        DataInputStream inStream = new DataInputStream(new BufferedInputStream(new FileInputStream(logFilename), 1024*1024));;

          //  TODO:  this technique needs to know the log anchor lsn, not something it
          //  technically should know at this point... ie, this is a trashy implementation
        long lsn = 0;
        LogEntryHeader prevHeader = null;
        LogEntryHeader header = null;

        try{
            while((header = readLogEntryHeader(inStream)) != null) {
                getEntryCache().put(lsn, header);
                //System.out.println(header);
                prevHeader = header;
                lsn = header.getNextLSN();
            }
        }catch(EOFException e) { /*  do nothing */ }

        if(prevHeader != null) {
            nextLSN = prevHeader.getNextLSN();
        }

        inStream.close();

          // set the length
        in = new RandomAccessFile(logFilename, "rw");
        in.setLength(nextLSN);
        in.close();

        in = new RandomAccessFile(logFilename, "r");
          //  this is here because the size of the file may change first, ie chop off the bad partial log entry
        out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(logFilename, true), OUTBUFFER_SIZE));

       // System.out.println("next: " + nextLSN);
    }

    /**
     *
     * @return
     */
    private LongHashMap getEntryCache() {
        if(rowCacheHolder == null) {
            LongHashMap entryCache = new LongHashMap();
            rowCacheHolder = new SoftReference(entryCache);
            return entryCache;
        }

        LongHashMap entryCache = (LongHashMap)rowCacheHolder.get();
        if(entryCache == null) {
            entryCache = new LongHashMap();
            rowCacheHolder = new SoftReference(entryCache);
            return entryCache;
        }

        return entryCache;
    }

    /********************
     * Header stuff
     */

    /**
     * In the case of writing, the data written does not even
     * need to come from a
     * TODO:  consider in this implementation not writing out the lsn and prevLSN, since it is a bit
     * redundant.
     */
    public void writeLogEntryHeader(DataOutput out, long lsn, long prevLsn, long timestamp, Transaction t,
                                    short logEntryType, long logEntryLength) throws IOException {
        out.writeLong(lsn);
        out.writeLong(prevLsn);
        out.writeLong(timestamp);
        if(t != null) {
            out.writeLong(t.getTransactionId());
            out.writeLong(t.getLastLSN());
        } else {
            out.writeLong(0);
            out.writeLong(0);
        }
        out.writeShort(logEntryType);
        out.writeInt((int)logEntryLength);
    }

    /**
     * Return a complete header, including the LogEntry.  TODO:  this can easily move into the actual log
     * manager that uses it... it may easily be different for every log manager
     */
    public LogEntryHeader readLogEntryHeader(DataInput in) throws IOException, SQLException {
        long lsn = in.readLong();
        long prevLsn = in.readLong();
        long timestamp = in.readLong();
        long transactionId = in.readLong();
        long prevTranactionLsn = in.readLong();

        short logEntryType = in.readShort();
        long logEntryLength = in.readInt();
        logEntryLength &= 0xffffffff;   //  mask out the negative numberness

        long nextLsn = lsn + getHeaderSize() + logEntryLength;

        LogEntry logE = logEntryManager.readLogEntry(logEntryType, logEntryLength, in);
        LogEntryHeader header = new LogEntryHeader(lsn, prevLsn, nextLsn, timestamp, transactionId, prevTranactionLsn, logE);

        return header;
    }

    /**
     * For now returns a constant, later will be more flexible
     */
    public int getHeaderSize() {
        return 46;
    }
}
