/*
 * Copyright JJ Furman
 * User: user
 * Date: Mar 17, 2002
 * Time: 10:47:41 AM
 */
package freebase.log.impl;

import freebase.api.LogManager;
import freebase.engine.Transaction;
import freebase.log.DefaultLogEntryCodes;
import freebase.log.LogEntry;
import freebase.log.LogEntryHeader;
import freebase.log.LogEntryManager;
import freebase.log.tm.CommitLogEntry;
import freebase.util.FileHelper;
import freebase.util.Props;

import java.io.*;
import java.sql.SQLException;

/**
 * Uses a file as it's persistence mechanism.  This is a first attempt at a log implementation.  Even
 * the next most basic file implementation will eventually need multiple files (or will it? perhaps an offline
 * process could chop/copy the log and reset the positions?)<BR><BR>
 *
 * This is by no means intended to be high performance or concurrent.<BR><BR>
 *
 * TODO:  create a configuration paradim that can be initialized from an xml file and get rid of
 * the custom constructor.
 */
public class SimpleFileLogManager implements LogManager {
    final public static String LOG_FILENAME_PROPERTY = "LOG_FILENAME";

    RandomAccessFile logFile;

    /**
     * map from codes to logentries
     */
    LogEntryManager logEntryManager;

    String filename;

    long lastLSN = 0;

    /**
     * curPos of the file (ie, the next LSN)
     */
    long nextLSN = 0;

    Object logTailLatch = new Object();

    //  TODO:  add highest LSN flushed, for later the question isFlushTo(LSN)

    /**
     * No need to seek if we are at the end
     */
    boolean lastOpWasWrite = false;

    public SimpleFileLogManager() throws IOException {
        String logFilename = Props.get(LOG_FILENAME_PROPERTY);

        synchronized(logTailLatch) {
            File f = new File(logFilename);
            FileHelper.ensureParentDirExists(f);
            logFile = new RandomAccessFile(f, "rw");

            filename = logFilename;

            nextLSN = logFile.length();

            if(nextLSN  != 0) {
                //  TODO:  read the file in?  or just wait for the calls to come?
                //  must find out what the last LSN number is (unless the end of the log always
                //     has that number written)
            }
        }
    }

    public long writeEntry(Transaction t, LogEntry entry, boolean flush) throws SQLException {
        try{
        synchronized(logTailLatch) {
            if(entry instanceof CommitLogEntry) {
                if(t.getTransactionId() == 0)
                    new Exception("STACK TRACE").printStackTrace();
            }

            if(entry.getCode() == DefaultLogEntryCodes.ABSTRACT_LOG_CODE) {
                System.out.println("ERROR writing log entry:  type " + entry.getClass() + " does not have it's own log code.");
            }

              //  This line may slow down the code, but for now is necessary
            if(! lastOpWasWrite)
                logFile.seek(nextLSN);

              //  write out the basic logheader stuff
            writeLogEntryHeader(logFile, nextLSN, lastLSN, System.currentTimeMillis(), t,
                    entry.getCode(), entry.getSize());

            entry.writeOut(logFile);

            lastLSN = nextLSN;
            nextLSN += entry.getSize() + getHeaderSize();

              //  a high performance version of this would not hold the tail lock while writing out!
              //  but I'm not sure how RandomAccessFile was implemented, most likely not multithreaded
            if(flush) {
                logFile.getFD().sync();
            }

            lastOpWasWrite = true;

            return lastLSN;
        }
        }catch(IOException ioe) {
            throw new SQLException(ioe.getMessage());
        }
    }

    public LogEntryHeader getLogEntry(long LSN) throws SQLException {
        try{
        synchronized(logTailLatch) {
              //  when reached the end, return null.
              //  TODO:  I don't like the meaning of this...
            if(LSN == nextLSN)
                return null;

            logFile.seek(LSN);

            lastOpWasWrite = false;

            return readLogEntryHeader(logFile);
        }
        }catch(IOException ioe) {
            ioe.printStackTrace();
            throw new SQLException(ioe.getMessage());
        }
    }

    /**
     * Currently, nothing stored in memory so no problem...
     */
    public void raiseLowWaterMark(long hardLSN, long softLSN) {

    }


    public void setLogEntryManager(LogEntryManager logEntryManager) {
        this.logEntryManager = logEntryManager;
    }

    /********************
     * Header stuff
     */

    /**
     * In the case of writing, the data written does not even
     * need to come from a
     * TODO:  consider in this implementation not writing out the lsn and prevLSN, since it is a bit
     * redundant.
     */
    public void writeLogEntryHeader(DataOutput out, long lsn, long prevLsn, long timestamp, Transaction t,
                                    short logEntryType, long logEntryLength) throws IOException {
        out.writeLong(lsn);
        out.writeLong(prevLsn);
        out.writeLong(timestamp);
        if(t != null) {
            out.writeLong(t.getTransactionId());
            out.writeLong(t.getLastLSN());
        } else {
            out.writeLong(0);
            out.writeLong(0);
        }
        out.writeShort(logEntryType);
        out.writeInt((int)logEntryLength);
    }

    /**
     * Return a complete header, including the LogEntry.  TODO:  this can easily move into the actual log
     * manager that uses it... it may easily be different for every log manager
     */
    public LogEntryHeader readLogEntryHeader(DataInput in) throws IOException, SQLException {
        long lsn = in.readLong();
        long prevLsn = in.readLong();
        long timestamp = in.readLong();
        long transactionId = in.readLong();
        long prevTranactionLsn = in.readLong();

        short logEntryType = in.readShort();
        long logEntryLength = in.readInt();
        logEntryLength &= 0xffffffff;   //  mask out the negative numberness

        long nextLsn = lsn + getHeaderSize() + logEntryLength;

        LogEntry logE = logEntryManager.readLogEntry(logEntryType, logEntryLength, in);
        LogEntryHeader header = new LogEntryHeader(lsn, prevLsn, nextLsn, timestamp, transactionId, prevTranactionLsn, logE);

        return header;
    }

    /**
     * For now returns a constant, later will be more flexible
     */
    public int getHeaderSize() {
        return 46;
    }
}
