/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 1, 2002
 * Time: 10:59:03 AM
 */
package freebase.log.tm;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * Just some ideas:
 *
 * When a rollback occurs first a rollback entry is written to the log.  Then all of the changes
 * made by this transaction are undone, by rummaging back through the log and applying the reverse
 * of each entry.  The last change of each unit of state will point to a unique Rollback entry, so that
 * there is a complete forward history. <BR><BR>
 *
 * The point of all this complexity is to try to avoid rewriting all of the undo messages to the log
 * again the way Transaction Processing suggests.  I feel that this can be avoided.<BR><BR>
 *
 * All rollback entries must be followed by a RollbackComplete entry.  This would allow the system
 * to know if all of the changes to rollback have been undone.  This becomes critical when dealing
 * with checkpointing.  Consider the following sequence of log message:<BR><BR>
 *
 * 1  begin T4
 * 2  begin T5
 * 3  T5:  write 8 to A
 * 4  T5:  write 7 to B
 * 5  rollback T5
 * 6  T4:  write 9 to B
 * 7  start checkpoint: T5 still active
 * 8  end checkpoint<BR><BR>
 *
 * Now consider the recovery process.  It cannot know if the rollback operations for T5 (log entry 3)
 * have been completely undone.  So we introduce the RollbackComplete log message;  If a rollback complete
 * message is seen before a start checkpoint message, then all of the undo's are guarenteed to be written
 * out to the checkpoint.  Otherwise the rollback will need to be redone.
 */
public class RollbackCompleteLogEntry extends TMLogEntry {
    /**
     *  0 means back everything back to the beginning and abort the transaction
     */
    long toLSN;

    public RollbackCompleteLogEntry() {
    }

    public RollbackCompleteLogEntry(long toLSN) {
        this.toLSN = toLSN;
    }

    /**
     * Enough size for the LSN
     */
    public int getSize() {
        return super.getSize() + 8;
    }

    public short getCode() {
        return ROLLBACK_COMPLETE_LOG_CODE;
    }

    public long getToLSN() {
        return toLSN;
    }

    public void readIn(DataInput in) throws IOException {
        super.readIn(in);

        toLSN = in.readLong();
    }

    public void writeOut(DataOutput out) throws IOException {
        super.writeOut(out);

        out.writeLong(toLSN);
    }

    public String toString() {
        return "Rollback Complete to LSN:" + toLSN;
    }
}
