/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Nov 11, 2001
 * Time: 8:09:32 AM
 * To change template for new class use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb;

import freebase.api.DeadlockableResource;
import freebase.api.TableLockManager;
import freebase.engine.Transaction;
import freebase.lock.LockWait;
import freebase.lock.WaitLockHolder;
import freebase.util.LongHashMap;
import org.apache.log4j.Category;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedList;

/**
 * Implement all of the locks in memory.  This should get migrated as the first class implementation of lock manager.
 *
 * TODO:  Should probably have a cleanup thread that wakes up
 * every so often and does some work.  Would clean out old
 * transactions in rowLocks and waitsPerTransaction
 *
 * TODO:  grab working code from LockManager
 */
public class MemoryLockManager implements TableLockManager {
    static Category cat = Category.getInstance(MemoryLockManager.class.getName());

    MemoryTableStorage table;

    Object tableLockLatch = new Object();

    /**
     * locks per row..
     * Currently map from rowId -> Transaction -or- WaitLock
     * If it maps to a Transaction, then there is no other Transactions waiting for this row lock.  If it is
     * a WaitLock object, then there is at least one other object waiting for
     */
    LongHashMap rowLocks = new LongHashMap();

    /**
     * Map from transactionId -> LongHashMap acting as a set of rowNums
     * that have waiters for this transaction.  This is a required
     * structure for making commits fast... we don't want to rummage
     * through the log!
     */
    LongHashMap waitsPerTransaction = new LongHashMap();

    /**
     * For each table lock type, this keeps track of exactly
     * who holds them.
     */
    HashSet[] tableLocks = new HashSet[NUM_TABLE_LOCKS];

    /**
     * Corresponding to the above are the table waits per type, but
     * this is a Set of TableLockWaits
     */
    HashSet[] tableWaits = new HashSet[NUM_TABLE_LOCKS];

    /**
     * This is used to guarentee progress.  The first
     * in the list has privelidge over the others in that it can
     * grab an available lock even if it doesn't have a lock already.
     */
    LinkedList tableLockWaitQueue = new LinkedList();


    public MemoryLockManager(MemoryTableStorage table) {
        this.table = table;

        for(int i = 0; i < NUM_TABLE_LOCKS; i++) {
            tableLocks[i] = new HashSet();
            tableWaits[i] = new HashSet();
        }
    }

    /**
     * Locks the table in the requested mode.   I would much prefer an algorithm that always woke
     * up each waiter explicitly in order; that way there would be no need to over notify all the time.
     * @return if the lock was newly granted
     */
    public boolean lockTable(Transaction t, byte requestedMode, boolean noWait) throws SQLException {
        //System.out.println("Lock table: " + t + " " + requestedMode);
        if(cat.isInfoEnabled()) {
            cat.info("lockTable " + t + "," + requestedMode);
        }

          //  I think this is safe from the DCL problem because of the
          //  synchronized body inside of the call, but I don't really know.
          //  should get a vm expert to check this out
        boolean[] ourTableLocks = t.getTableLocks(this);

          //  do we already have the lock?  This is the key to high performance
          //  table locks, in my current thinking
        if(ourTableLocks[requestedMode])
            return false;

//System.out.println("ENTERING the table lock latch " + t + " mode " + requestedMode);
          //  only one thread should be obtaining a table lock at a time,
          //  even in a distributed environment!
        synchronized(tableLockLatch) {
            boolean firstAttempt = true;
            TableLockWaiter ourLockWait = null;
            //System.out.println("LOCKS FOR " + table.getName() + " has locks " + StringUtil.dump(tableLocks));

            while(true) {
                //  ODD case:  what if multiple threads are getting a row ex
                    //  lock at the exact same time.  Should we double check
                    //  in here?  For now disallow multiple update/select for update
                    //  threads.  This is important to remember though!!  Seems
                    //  similar to DCL but it is really double synchronized...
                    //  I don't know if that solves it.

                  //  we don't keep this value cached because then we would just
                  //  need to recompute it on release:: is this garbage?
                int highestHeldLockBySomeoneElse = -1;
                for(int i = NUM_TABLE_LOCKS - 1; i >= 0; i--) {
                    if((tableLocks[i].size() > 0 && ! ourTableLocks[i]) ||
                        tableLocks[i].size() > 1 && ourTableLocks[i]) {
                        highestHeldLockBySomeoneElse = i;
                        break;
                    }
                }

                boolean grantLock = false;
                switch(requestedMode) {
                    case ROW_SHARE_LOCK:
                        if(! isLockHeldByAnother(EXCLUSIVE_LOCK, ourTableLocks)) {
                            grantLock = true;
                        }
                        break;

                    case ROW_EXCLUSIVE_LOCK:
                        if(highestHeldLockBySomeoneElse <= ROW_EXCLUSIVE_LOCK) {
                            grantLock = true;
                        }
                        break;

                    case SHARE_LOCK:
                        if(highestHeldLockBySomeoneElse <= SHARE_LOCK &&
                           ! isLockHeldByAnother(ROW_EXCLUSIVE_LOCK, ourTableLocks)) {
                            grantLock = true;
                        }
                        break;

                    case SHARE_ROW_EXCLUSIVE_LOCK:
                        if(highestHeldLockBySomeoneElse <= ROW_SHARE_LOCK) {
                            grantLock = true;
                        }
                        break;

                    case EXCLUSIVE_LOCK:
                        if(highestHeldLockBySomeoneElse == -1) {
                            grantLock = true;
                        }
                        break;

                    default: throw new SQLException("TABLE LOCK MODE '" + requestedMode + "' not recognized");
                }

 //System.out.println("INITIAL DECISION IS " + grantLock + "   " + t);
                  //  double check that it is legal for us to grab
                if(grantLock && tableLockWaitQueue.size() > 0) {
                    if(tableLockWaitQueue.getFirst() != ourLockWait) {
                        boolean alreadyHaveAnyLock = false;
                        for(int i = 0; i < NUM_TABLE_LOCKS; i++) {
                            if(ourTableLocks[i]) {
                                alreadyHaveAnyLock = true;
                                break;
                            }
                        }

                          //  if we don't have a single lock, and we aren't the first
                          //  on the list then we gots to wake up
                        if(! alreadyHaveAnyLock)
                            grantLock = false;
                    }
                }

//System.out.println("AFTER ANALYSIS, DECISION IS " + grantLock + "   " + t);
                  // is cool to grab the lockito?
                if(grantLock) {
                    tableLocks[requestedMode].add(t);
                    ourTableLocks[requestedMode] = true;

                    //  no longer calling this from here
                    //t.writeLogEntry(new LockTableEntry(table.getTableId(), requestedMode));

  //                  System.out.println("Left Lock counts @ " + StringUtil.dump(tableLocks));

                    if(! firstAttempt) {
                        tableLockWaitQueue.remove(ourLockWait);
                        tableWaits[requestedMode].remove(ourLockWait);
                    }

                    /*  NOTE:  this body has been commented out but will probably be useful at some point
                               in the future.  See Footnote 2
                      //  for all incompatible table waits with us...  add DD edges
                      //  to repair the new graph.  hmmm
                    for(int i = 0; i < incompatibleLocks[requestedMode].length; i++) {
                        HashSet waitSet = tableWaits[incompatibleLocks[requestedMode][i]];

                        //System.out.println("conflicting wait Set of type " + incompatibleLocks[requestedMode][i] + " contains " + waitSet);
                        Iterator it = waitSet.iterator();
                        while(it.hasNext()) {
                            TableLockWaiter otherWaiter = (TableLockWaiter)it.next();

                            //System.out.println("  adding to DD: " + otherWaiter.getTransaction() + " -> " + t + "  res: " + otherWaiter);
                            t.getTM().getDeadlockDetector().
                                    addWaitingEdge(otherWaiter.getTransaction(),
                                        t, otherWaiter);
                        }
                    }
                    */

                      //  see Method Footnote 2 for explanation of why this is needed
                    tableLockLatch.notifyAll();
                    t.registerDoneCallback(this, null);

                    return true;
                }

                  //  if got here, lock is not available...do whatever

                if(noWait) {
                    throw new SQLException("Resource in mode " + tableLockNames[requestedMode] + " not available and requested in NOWAIT mode");
                }

                if(firstAttempt) {
                    firstAttempt = false;
                    ourLockWait = new TableLockWaiter(t, tableLockLatch);
                    tableLockWaitQueue.addLast(ourLockWait);
                    tableWaits[requestedMode].add(ourLockWait);
                }

                  //  add deadlock edges: establish all of our waits
                for(int i = 0; i < incompatibleLocks[requestedMode].length; i++) {
                    HashSet lockSet = tableLocks[incompatibleLocks[requestedMode][i]];
                    if(lockSet.size() > 0) {
                        //System.out.println("Adding waiting edges for " +
                        t.getTM().getDeadlockDetector().addWaitingEdges(t,
                            lockSet, ourLockWait);
                    }
                }

                try{
                    //System.out.println("sleeping....  " + t);
                    tableLockLatch.wait();
                    //System.out.println("WAKING  ....  " + t);
                }catch(InterruptedException ie) {
                      //  remove ourself from the local waits and the priority Queue
                      //  -and- remove waiting deadlock edges
                    tableLockWaitQueue.remove(ourLockWait);
                    tableWaits[requestedMode].remove(ourLockWait);
                    t.getTM().getDeadlockDetector().removeAllWaitingEdges(t);

                    throw new SQLException("Interrupted");
                }

                  //  TODO:  do we really need to remove all the edges, or just
                if(ourLockWait.isInDeadlock()) {
                      //  remove ourself from the local waits and the priority Queue
                    tableLockWaitQueue.remove(ourLockWait);
                    tableWaits[requestedMode].remove(ourLockWait);

                    //  no need to remove it from the deadlock detector, it already did
                    throw new SQLException("DEADLOCK DETECTED while waiting for resource.");
                } else {
                    t.getTM().getDeadlockDetector().removeAllWaitingEdges(t);
                }
            }
        }
    }
    /**  METHOD FOOTNOTE 1:
     *  In oracle if T1 grabs row exclusive, then
     *               T2 tries to grab share (and blocks)
     *  when T3 tries to grab a ROW SHARE lock, then she blocks despite
     *  the fact that ROW SHARE is not in conflict with either any actual
     *  or waiting locks.  We will have the same behavior, just in case
     *  they did it that way for some reason!!!  Maybe easier to keep track
     *  ok deadlocks or something?
     *
     * Another sequence is  T1: S,   T2: X,   T3: RS    T3 blocks!
     *
     * <BR><BR>
     * FOOTNOTE 2:  There are two reasons for the notifyAll() above.
     * The first reason:
     *   If an additional lock is obtained from a transaction that would represent
     *   a legitimate wait edge (would preven the lock) in another already waiting
     *   transaction, then that waiting transaction must have an edge added to reflect
     *   this to accurately detect the transaction.  The code that is commented out
     *   may correctly do this, but because the notifyAll is needed anyway because of
     *   the second reason as well, it is currently disabled.
     *
     *   An example of the above case is this:  Tables A and B
     *   T1          T2          T3          Notes
     *   ------      ------      ------      -------
     *   X on A
     *               RX on B
     *                           RS on B
     *   S on B                              this only blocks on T2
     *                           RX on B     in conflict with T1, but allowed because T3
     *                                       already owned a lock; T1 must now wait on T3 also
     *                           RX on A     deadlock with T1
     *
     *
     * The second reason:
     *   in the case 3 transactions T1, T2, and T3 grab three locks in the order:
     *   T1: RX,   T2: S,   T3: RS,    T1:  commit  <BR><BR>
     *
     * when T1 is done, it will wakeup the other waiters T2 and T3 for their turn to check if they
     * can legally obtain the lock they want.  If T3 wakes up first, it will correctly conclude
     * that it cannot obtain it's lock because it is not the first in the wait queue.  So it will
     * go back to sleep.  Then T2 will run, and grab the lock.  What should happen is that both T2 and
     * T3 should ALWAYS both get to go.  The correct way to make this happen is to just notify
     * all of the waiters in exact queue order.  notifyAll does not provide this facility!
     * It may be reasonable to rewrite this a different way, to be better performance and more
     * correct.<BR><BR>
     *
     * I would implement this by:<BR>
     * 1.   move the wait outsize of the synch(tableLockLatch) and into it's own area,
     *      synchronizing on the actual TableLockWaiter object "ourLockWait"
     * 2.   add a state to the TableLockWaiter to keep track of whether it has been
     *      allowed to recheck yet.  This will bridge the region between releasing the
     *      tableLockLatch and obtaining the lock on ourLockWait (a classic java issue)
     * 3.   change TableLockWaiter to notify only itself only upon deadlock found.  Likewise,
     *      the lock grabbing thread will need to check deadlock before waiting, in case the deadlock
     *      detector got to work fast (and found a deadlock before the bridge was crossed, as as above)
     *      or during some subsequent wakup a deadlock was found.  A wakeup can happen any time
     *      due to other transactions with locks commiting or whatnot.
     * 4.   isolate the wakeup sequence code to one area, and grab the tableLockLatch before
     *      waking up the first guy (find the first guy in the queue).  If that guy grabs a
     *      lock he can notify the next guy, in a chain style.  <BR><BR>
     *
     * IE, it is pretty complex and will probably only end up being faster in the case with a long
     * chain of requestors, each grabbing a lock that is not incompatible with the previous, but
     * each not the first.  This case will probably appear when a share lock is obtained on a table
     * and a batch of updates bunch up behind it waiting for it to finish.  Maybe truely worth it!
     */

    //private void cleanupLocalWaiter

    /**
     * Returns true if anyone besides us holds the given lock mode.
     */
    private boolean isLockHeldByAnother(byte lockMode, boolean[] ourLock) {
        if(tableLocks[lockMode].size() == 0)
            return false;
        if(tableLocks[lockMode].size() == 1 && ourLock[lockMode])
            return false;
        return true;
    }


    public void unlockTable(Transaction t, byte mode) throws SQLException {
       // System.out.println("Unlock table: " + t + " " + mode);

        boolean[] ourHeldLocks = t.getTableLocks(this);
        if(! ourHeldLocks[mode]) {
            cat.warn("Tried to undo table lock from " + t + " on " +
                    table.getTableId() + " mode: " + mode + " but didn't have the lock");
            return;
        }

        synchronized(tableLockLatch) {
            ourHeldLocks[mode] = false;
            tableLocks[mode].remove(t);
            tableLockLatch.notifyAll();
            //System.out.println("NOTIFY EVERYONE, current lock state is " + StringUtil.dump(tableLocks));
        }
    }

    /**
     * when the transaction commits, unlock everything
     * @param t
     */
    public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
        if(finalStatus == Transaction.COMMITED) {
            unlockAll(t);
        }
    }

    /**
     * Before entering, it is assumed that the transaction is already been committed.
     * The state must be COMMITED so that more waiters don't get added (see addToTransactionWaits)
     * Rollback will call each unlock explicitly
     */
    public void unlockAll(Transaction t) {
        if(cat.isInfoEnabled()) {
            cat.info("MemoryLockManager: unlockAll " + t);
        }

        //  NOTE:  It might be faster if we unlocked row locks first it might be faster, because then the
        //  things that might collide with our rows will be free to get everything by then

          //  handle table locks
        synchronized(tableLockLatch) {
            boolean[] ourHeldLocks = t.getTableLocks(this);
            for(int i = 0; i < ourHeldLocks.length; i++) {
                if(ourHeldLocks[i]) {
                    ourHeldLocks[i] = false;
                    tableLocks[i].remove(t);
                }
            }

            tableLockLatch.notifyAll();
        }

          //  eventually the following code could be in transaction.
        ArrayList allWaits = new ArrayList();
        LongHashMap allWaiters = null;
        synchronized(waitsPerTransaction) {
            long transId = t.getTransactionId();
            allWaiters = (LongHashMap)waitsPerTransaction.remove(transId);
        }

          //  unlock all waiters that exist
        if(allWaiters != null) {
            Enumeration e = allWaiters.keys();
            while(e.hasMoreElements()) {
                allWaits.add(e.nextElement());
            }
        }

          //  now, for each one unlockRow...
        for(int i = 0; i < allWaits.size(); i++) {
            Long row = (Long)allWaits.get(i);
            unlockRow(row.longValue());
        }

        //  here is where it might be faster to have WaitLockHolders in a separate
        //  LongHashMap or something.  If keeping both, then reduce WaitLockHolder
        //  to just a boolean holder -> isActive
    }

    /**
     * unlock the given row; I guess t had the lock...
     * Might be nice TODOuble check!!
     */
    public void unlockRow(long rowId) {
        if(cat.isInfoEnabled()) {
            cat.info("MemoryLockManager: unlockRow: " + rowId);
        }

        WaitLockHolder waitHolder = null;
        synchronized(rowLocks) {
            Object o = rowLocks.remove(rowId);

              //  by definition there are no waiters
            if(o instanceof Transaction)
                return;

            waitHolder = (WaitLockHolder)o;
        }

          //  since there is a waitHolder, then remove it from the transaction too
        synchronized(waitsPerTransaction) {
            // System.out.println("Lock released: " + waitHolder);
            long transId = waitHolder.getTransaction().getTransactionId();
            LongHashMap allWaiters = (LongHashMap)waitsPerTransaction.get(transId);

            if(allWaiters != null) {
                allWaiters.remove(rowId);
            }
        }

          //  now wake up those who are waiting for this lock
        synchronized(waitHolder) {
              //  in case another "A" grabbed this just before we removed it,
              //  and we grab this lock just before A gets it, thus
              //  it might go blindly waiting.  It needs to test the condition
              //  first
            waitHolder.setTransaction(null);
            waitHolder.notifyAll();
        }
    }

    /**
     * @return the penultimate lock holder
     * TODO:  old:  possibly rewrite this and get rid of the primary lock hash.  Use it
     *    *ONLY* for the waiters.  Much simpler.  (new:  I guess the idea was to
     */
    public LockWait lockRow(Transaction t, Transaction expectedHolder, long rowId) throws SQLException {
        return null;
        /*
        if(cat.isInfoEnabled()) {
            cat.info("MemoryLockManager: lockRow: " + rowId + " for " + t);
        }

        //  the following line was here before, and is definitely not the right idea
        //  it is done somewhere above
        //lockTable(t, tableRowExclusive ? ROW_EXCLUSIVE_LOCK : ROW_SHARE_LOCK, false);

          //  keep trying until we get it
        while(true) {
            WaitLockHolder waitLock = null;

              //  grab master lock (currently only one)
            synchronized(rowLocks) {
                Object o = rowLocks.get(rowId);

                  //  if there is no lock held at all
                if(o == null) {
                      //  mark us as the current transaction holding this lock
                    rowLocks.put(rowId, t);
                    return null; //changeCurrentTransaction(t, rowId);
                }

                  //  if we are the holder already, then great... just return
                if(o == t) {
                    return t;
                }

                  //  if we are the first entity to wait for this
                if(o instanceof Transaction) {
                    Transaction weWaitFor = (Transaction)o;
                    if(! weWaitFor.isActive()) {
                        rowLocks.put(rowId, t);
                        return weWaitFor; //changeCurrentTransaction(t, rowId);
                    }

                      //  promote the Transaction to be a WaitLockHolder
                    waitLock = new WaitLockHolder();
                    waitLock.setTransaction(weWaitFor);
                    rowLocks.put(rowId, waitLock);
                } else {
                    waitLock = (WaitLockHolder)o;
                    if(waitLock.getTransaction() == t)
                        return t;
                }
            }

              //  deals with it's own concurrency (NO IT DOESN'T!!!)
            addToTransactionWaits(waitLock.getTransaction(), rowId);

              //  while we are in this region, another thread may step in and
              //  change the state of our waitLock that we just put in.  So we
              //  will need to check again

              //  deal with deadlock
            synchronized(waitLock) {
                  //  if the transaction is now null, go back and try again to grab the lock
                if(waitLock.getTransaction() != null) {
                    if(waitLock.getTransaction().isActive()) {
                        DeadlockDetector detector = t.getTM().getDeadlockDetector();

                          //  TODO:  pool these, grab it in the main lock from an ArrayList
                        DeadlockDetectorDelegate delegate = new DeadlockDetectorDelegate(waitLock);
                        detector.addWaitingEdge(t, waitLock.getTransaction(), delegate);

                          //  record this because it will get nulled out in the waitLock if things go well...
                        Transaction whoWeWaitFor = waitLock.getTransaction();

                        try{
                        waitLock.wait();
                        }catch(InterruptedException in) {
                            detector.removeWaitingEdge(t, whoWeWaitFor, delegate);

                            throw new SQLException("interrupted");
                            // OLD:  Might this be thrown in the Distributed Deadlock case?
                            //  still probably not, just notify and set failed flag
                        }

                        if(delegate.isInDeadlock()) {
                            throw new SQLException("DEADLOCK DETECTED with : " + waitLock.getTransaction() + " (during row lock)");
                        } else {
                            detector.removeWaitingEdge(t, whoWeWaitFor, delegate);
                        }
                    }
                }
            }
        }*/
    }


    /*********************************************
     * local methods
     */

    /**
     *  3/28/2002 this was deprecated.  There was a split state style race condition with this
     * line of thinking.  The "recentTransaction" should not be set without also setting the log message
     *
     *  Updates the actual table storage, currently specific to memory tables
     * TODO:  get rid of this by rewriting the above lock manager totally.  It needs to eventually
     *  directly access the lock information held in the table storage
     *  returns old holder.
     *
    private Transaction changeCurrentTransaction(Transaction t, long rowId) {
        if(cat.isDebugEnabled())
            cat.debug("LOCK OBTAINED: " + t + " -> " +rowId);

        synchronized(table.rows) {
            MemoryRowData data = (MemoryRowData)table.rows.get((int)rowId);
            Transaction last = data.getRecentTransaction();
            data.setRecentTransaction(t);
            return last;
        }
    }

     */

    /**
     * Eventually the per transaction waits can be stored in
     * the transaction itself, lend to better concurrancy
     * and simpler lock code.  Or course, this breaks down
     * the nice layering but who cares.  One could
     * keep the nice layering by only exposing a waitLock
     * on the transaction object, but still keep everything in
     * a hashmap here.
     *
     * Correctness Proof:
     * This sync body adds the row to the list of waiters for transaction t
     * It locks out the t.commit process from fetching the list.  It makes
     * adding a new entry to the list and fetching the entire list atomic.
     * call then addEntry and fetchList
     *
     * A potential problem is that a waiter gets added to the list after
     * the commited transaction has already fetched it, leaving the waiter
     * to wait forever.  This cannot happen because:
     *
     * If addEntry is before fetchList, it will be included thus correct.
     * If fetchList happens before addEntry, then the transaction has already
     *     already commited.  Therefore, the state is COMMITED.
     *
     * Therefore, to prevent this problem, make sure we check the state one more time...
     */
    protected void addToTransactionWaits(Transaction t, long rowId) {
        synchronized(waitsPerTransaction) {
            if(! t.isActive())
                return;

            long transId = t.getTransactionId();

            LongHashMap map = (LongHashMap)waitsPerTransaction.get(transId);
            if(map == null) {
                map = new LongHashMap();
                waitsPerTransaction.put(transId, map);
            }

            map.put(rowId, map);
        }
    }

    /***********************************************************************************
     * Cleanup method, removes unnecessary references to Committed transactions.
     * NOTE:  slow...TOO BAD about this implementation.  A rewriting of
     * LongHashMap might offer an improvement if we use the deletable Iterator pattern
     */
    public void cleanup() {
        synchronized(rowLocks) {
            ArrayList onesToRemove = null;

            Enumeration en = rowLocks.keys();
            while(en.hasMoreElements()) {
                Long rowId = (Long)en.nextElement();
                Object o = rowLocks.get(rowId.longValue());
                if(o instanceof Transaction) {
                    Transaction t = (Transaction)o;
                    if(! t.isActive()) {
                        if(onesToRemove == null) {
                            onesToRemove = new ArrayList();
                        }

                        onesToRemove.add(rowId);
                    }
                }
            }

            if(onesToRemove != null) {
                for(int i = 0; i < onesToRemove.size(); i++) {
                    Long row = (Long)onesToRemove.get(i);
                    rowLocks.remove(row.longValue());
                }
            }
        }
    }

    /***********************************************************************************
     * debug methods
     */

    /**
     * dumpRowLocks dumps the row lock information to the prompt.  Should eventually consider a
     * virtual table or something cute
     */
    public void dumpLocks() {

        synchronized(rowLocks) {
            Enumeration en = rowLocks.keys();
            while(en.hasMoreElements()) {
                Long rowId = (Long)en.nextElement();

                System.out.print("Row Locked :" + rowId + "   ");
                Object o = rowLocks.get(rowId.longValue());
                if(o instanceof Transaction) {
                    Transaction t = (Transaction)o;

                    System.out.println("Transaction is " + t.getStatusReadable());
                } else {
                    System.out.println("Holder is " + o + "  a  " + o.getClass());
                }
            }
        }
    }
}




/**
 *  Pick a better name
 */


/**
 * Created uniquely representing a particular transaction's wait for a table lock.
 * Edges are formed with this to all transactions that have locks of competing lock
 * types with the one we want.
 */
class TableLockWaiter implements DeadlockableResource {
      //  note this is a different use of a transaction variable in a waiter
      //  object.  This is so future lock obtainers can find out both about
      //  the transaction that is waiting and it's TableLockWaiter
    Transaction t;
    boolean isInDeadlock = false;
    Object whoToNotify;

    public TableLockWaiter(Transaction t, Object whoToNotifyOfDeadlock) {
        whoToNotify = whoToNotifyOfDeadlock;
        this.t = t;
    }

    public void notifyOfDeadlock() {
        //System.out.println("BEING NOTIFIED OF DEADLOCK");

          //  probably not the ideal implementation but I
          //  suspect that deadlock will not be common
        synchronized(whoToNotify) {
            isInDeadlock = true;
            whoToNotify.notifyAll();
        }
    }

    public Transaction getTransaction() {
        return t;
    }

      //  does this need to be synchronized for the same reason
      //  that DCL doesn't work?
    public boolean isInDeadlock() {
        return isInDeadlock;
    }
}
