/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Dec 12, 2001
 * Time: 7:10:16 PM
 * To change template for new class use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb;

import freebase.api.*;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.log.DefaultLogEntryCodes;
import freebase.log.LogEntry;
import freebase.log.UndoLogEntry;
import freebase.log.data.TableLogEntry;
import freebase.log.index.IndexLogEntry;
import freebase.memorydb.index.MemoryIndexStorage;
import freebase.memorydb.log.MemorySequenceIncLogEntry;
import freebase.memorydb.log.MemorySequenceStructureLogEntry;
import freebase.memorydb.log.MemoryStorageLogEntry;
import freebase.util.ListHelper;

import java.math.BigInteger;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 *
 */
public class MemoryStorageManager implements StorageManager {
    /**
     * Table storage by tableId
     */
    ArrayList tables = new ArrayList();

    /**
     * Index storage by indexId
     */
    ArrayList indexes = new ArrayList();

    /**
     * Sequence storage by sequenceId
     */
    ArrayList sequences = new ArrayList();

    TransactionManager tm;

    /**
     * Background thread removing unnecessary references to committed transactions
     * and other stuff
     */
    CleanupThread cleanupThread;

      //  TODO:  put this in the config file
    int cleanupPeriodSeconds = 60;

    /**
     * Constructor.  Starts the cleanup thread.
     */
    public MemoryStorageManager() {
        cleanupThread = new CleanupThread();
        cleanupThread.start();
    }

    /**
     * Called at initialization time only.
     */
    public void setTransactionManager(TransactionManager tm) throws SQLException {
        this.tm = tm;
    }

    /**
     * TODO:  when tables get deleted, make sure holes are reused.
     *    Use Hole list
     */
    public TableStorage createTableStorage(Transaction t) throws SQLException {
        synchronized(tables) {
            TableStorage ts = _createTableStorage(tables.size());
            tables.add(ts);

            t.writeLogEntry(new MemoryStorageLogEntry(ts.getTableId(),
                            MemoryStorageLogEntry.CREATE_STORAGE,
                            MemoryStorageLogEntry.TABLE_STORAGE, false));
            return ts;
        }
    }

    /**
     * Eventually this should take parameters
     */
    private MemoryTableStorage _createTableStorage(int tableId) {
        return new MemoryTableStorage(tm, tableId);
    }

    public TableStorage getTableStorage(int tableId) throws SQLException {
        synchronized(tables) {
            return (TableStorage)tables.get(tableId);
        }
    }

    public IndexStorage getIndexStorage(int indexId) throws SQLException {
        synchronized(indexes) {
            return (IndexStorage)indexes.get(indexId);
        }
    }

    public IndexStorage createIndexStorage(Transaction t, boolean unique) throws SQLException {
        synchronized(indexes) {
            IndexStorage is = _createIndexStorage(indexes.size(), unique);
            indexes.add(is);

            t.writeLogEntry(new MemoryStorageLogEntry(is.getIndexId(),
                            MemoryStorageLogEntry.CREATE_STORAGE,
                            MemoryStorageLogEntry.INDEX_STORAGE, unique));

            return is;
        }
    }

    /**
     * create the index storage but don't
     * @param indexId
     * @param unique
     */
    private IndexStorage _createIndexStorage(int indexId, boolean unique) {
        return new MemoryIndexStorage(tm, indexId, unique);
    }

    /**
     * Writes the log entry and removes the reference.  Table is allowed to be cleaned up later
     * when no more readers are in there (via a call to TableStorage.dropTable())
     * TODO:  Currently doesn't drop any corresponding indexes!!!!!!!
     * @param tableId
     * @throws SQLException
     */
    public void dropTableStorage(Transaction t, int tableId) throws SQLException {
        _dropTable(t, tableId);

        t.writeLogEntry(new MemoryStorageLogEntry(tableId,
                            MemoryStorageLogEntry.DELETE_STORAGE,
                            MemoryStorageLogEntry.TABLE_STORAGE));
    }

    /**
     * Do everything but don't log the message.  Called from above, and the log message.
     */
    private void _dropTable(Transaction t, int tableId) {
        Object table = null;
        synchronized(tables) {
            table = tables.set(tableId, null);
        }

        if(t != null) {
            t.registerDoneCallback(new DropTableCallback(), table);
        }
    }

    public void dropIndexStorage(Transaction t, int indexId) throws SQLException {
        _dropIndex(t, indexId);

        t.writeLogEntry(new MemoryStorageLogEntry(indexId,
                            MemoryStorageLogEntry.DELETE_STORAGE,
                            MemoryStorageLogEntry.INDEX_STORAGE));
    }

    /**
     * Do everything but don't log the message.  Called from above, and the log message.
     * @param t the transaction dropping this index.  pass in null when calling from a rollback of create!
     */
    private void _dropIndex(Transaction t, int indexId) {
        Object index = null;
        synchronized(indexes) {
            index = indexes.set(indexId, null);
        }

          //  register for a callback if this is an original operation
        if(t != null) {
            t.registerDoneCallback(new DropIndexCallback(), index);
        }
    }

    /******************************************************************************************
     *
     */


    /**
     * @param sequenceId
     * @return
     * @throws SQLException
     */
    public SequenceStorage getSequenceStorage(int sequenceId) throws SQLException {
        synchronized(sequences) {
            return (SequenceStorage)sequences.get(sequenceId);
        }
    }

    public SequenceStorage createSequenceStorage(Transaction t, BigInteger nextVal) throws SQLException {
        synchronized(sequences) {
            SequenceStorage ss = _createSequenceStorage(sequences.size(), nextVal);
            sequences.add(ss);

            t.writeLogEntry(new MemorySequenceStructureLogEntry(ss.getSequenceId(),
                            MemorySequenceStructureLogEntry.CREATE_STORAGE,
                            nextVal));

            return ss;
        }
    }

    public void dropSequenceStorage(Transaction t, int sequenceId) throws SQLException {
        _dropSequence(t, sequenceId);

        t.writeLogEntry(new MemorySequenceStructureLogEntry(sequenceId,
                            MemorySequenceStructureLogEntry.DELETE_STORAGE, null));
    }

    private MemorySequenceStorage _createSequenceStorage(int sequenceId, BigInteger nextVal) {
        return new MemorySequenceStorage(tm, sequenceId, nextVal);
    }

    /**
     * Do everything but don't log the message.  Called from above, and the log message.
     */
    private void _dropSequence(Transaction t, int sequenceId) {
        Object seq = null;
        synchronized(sequences) {
            seq = sequences.set(sequenceId, null);
        }

        if(t != null) {
            t.registerDoneCallback(new DropSequenceCallback(), seq);
        }
    }

    /**
     * Called during startup
     */
    public int getMaxTableId() throws SQLException {
        return tables.size() - 1;
    }

    public int getMaxIndexId() throws SQLException {
        return indexes.size() - 1;
    }

    /**
     * TODO:  should the table and index specific entries be dispatched in
     * a higher level, generic to all Storage implementations?
     */
    public void processLogEntry(int applyMode, long lsn, Transaction t, LogEntry logEntry)
            throws SQLException {
        if(logEntry instanceof UndoLogEntry ) {
            System.out.println("undo " + logEntry);
        }

        if(logEntry instanceof MemoryStorageLogEntry) {
            MemoryStorageLogEntry storeEntry = (MemoryStorageLogEntry)logEntry;

            if(applyMode == LogEntryProcessor.REDO_ENTRY) {
                if(storeEntry.getOperation() == MemoryStorageLogEntry.CREATE_STORAGE) {
                    if(storeEntry.getStorageType() == MemoryStorageLogEntry.TABLE_STORAGE) {
                        ListHelper.ensureSize(tables, storeEntry.getEntityId() + 1);

                        tables.set(storeEntry.getEntityId(), _createTableStorage(storeEntry.getEntityId()));
                    } else {
                        ListHelper.ensureSize(indexes, storeEntry.getEntityId() + 1);
                        indexes.set(storeEntry.getEntityId(),
                                _createIndexStorage(storeEntry.getEntityId(), storeEntry.isUnique()));
                    }
                } else if(storeEntry.getOperation() == MemoryStorageLogEntry.TRUNCATE_STORAGE) {
                    if(storeEntry.getStorageType() == MemoryStorageLogEntry.TABLE_STORAGE) {
                        ((MemoryTableStorage)getTableStorage(storeEntry.getEntityId()))._truncateTable();
                    } else {
                        ((MemoryIndexStorage)getIndexStorage(storeEntry.getEntityId()))._truncateIndex();
                    }
                } else if(storeEntry.getOperation() == MemoryStorageLogEntry.DELETE_STORAGE) {
                    if(storeEntry.getStorageType() == MemoryStorageLogEntry.TABLE_STORAGE) {
                        _dropTable(t, storeEntry.getEntityId());
                    } else {
                        _dropIndex(t, storeEntry.getEntityId());
                    }
                }
            } else {  //  handle UNDO case:
                if(storeEntry.getOperation() == MemoryStorageLogEntry.CREATE_STORAGE) {
                    if(storeEntry.getStorageType() == MemoryStorageLogEntry.TABLE_STORAGE) {
                          //  undo create by nuking the data.
                          //  TODO:  does this need to be dropped?  could any readers be there already?  what if they are?
                        //tables.set(storeEntry.getEntityId(), null);

                          //  better:
                        _dropTable(null, storeEntry.getEntityId());
//                        ensureSize(tables, storeEntry.getEntityId() + 1);

//                        tables.set(storeEntry.getEntityId(), _createTableStorage(storeEntry.getEntityId()));
                    } else {
                        _dropIndex(null, storeEntry.getEntityId());
//                        ensureSize(indexes, storeEntry.getEntityId() + 1);
//                        indexes.set(storeEntry.getEntityId(),
//                                _createIndexStorage(storeEntry.getEntityId(), storeEntry.isUnique()));
                    }
                } else if(storeEntry.getOperation() == MemoryStorageLogEntry.TRUNCATE_STORAGE) {
                    //  These are dealt with in callbacks, so there is no need for them here
                } else if(storeEntry.getOperation() == MemoryStorageLogEntry.DELETE_STORAGE) {
                      //  These are dealt with in callbacks, so there is no need for them here
                }
            }

        } else if(logEntry.getCode() == DefaultLogEntryCodes.MEMORY_SEQUENCE_STRUCTURE_LOG_CODE) {
            MemorySequenceStructureLogEntry seqLogEntry = (MemorySequenceStructureLogEntry)logEntry;
            if(applyMode == LogEntryProcessor.REDO_ENTRY) {
                if(seqLogEntry.getOpCode() == MemorySequenceStructureLogEntry.CREATE_STORAGE) {
                    ListHelper.ensureSize(sequences, seqLogEntry.getEntityId() + 1);
                    SequenceStorage s = _createSequenceStorage(seqLogEntry.getEntityId(), seqLogEntry.getNextVal());
                    sequences.set(seqLogEntry.getEntityId(), s);
                } else {
                    _dropSequence(t, seqLogEntry.getEntityId());
                }
            } else {  //  undo (rollback) operation
                if(seqLogEntry.getOpCode() == MemorySequenceStructureLogEntry.CREATE_STORAGE) {
                    _dropSequence(null, seqLogEntry.getEntityId());
                } else {
                    //  the drop will get undone by the callback so do nothing
                }
            }
        } else if(logEntry.getCode() == DefaultLogEntryCodes.MEMORY_SEQUENCE_INC_LOG_CODE) {
            MemorySequenceIncLogEntry inc = (MemorySequenceIncLogEntry)logEntry;
            getSequenceStorage(inc.getEntityId()).processLogEntry(applyMode, lsn, t, logEntry);
        } else if(logEntry instanceof TableLogEntry) {
            TableStorage tab = (TableStorage)tables.get(((TableLogEntry)logEntry).getTableId());

            /**
             * NOTE:  the only way tab can be null is in a very rare timing problem... if somehow
             * a drop occurs just as another waiter queues up, then when the drop is complete the waiter
             * wakes up and then fails to do the operation, thus rolling back the lock.  The alternative
             * solution is to add something in the table locks code that notifies the table lock
             * manager when the table has been dropped, thus not allowing any other lock that might be waiting
             * to be granted...  That might be more elegant and clear, but this alternative is pretty simple.
             * TODO:  see if this is still something that can happen
             */
            if(tab != null) {
                tab.processLogEntry(applyMode, lsn, t, logEntry);
            }
        } else if(logEntry instanceof IndexLogEntry){
            IndexStorage indexStorage = (IndexStorage)indexes.get(((IndexLogEntry)logEntry).getIndexId());

            indexStorage.processLogEntry(applyMode, lsn, t, logEntry);
        } else {
            System.out.println("MemoryStorageManager.processLogEntry ignored log entry: [" + logEntry + "]");
        }
    }

    /**
     * Do nothing, since a memory-only implementation.
     * @throws SQLException
     */
    public void checkpoint() throws SQLException {
        //  Do nothing.  Also, the corresponding LogAnchor will need to ignore updates to the
        //  current log position.
    }

    /**
     * Cleanup method, for explicitly removing memory structures.  Normally
     * these will get cleaned up eventually anyway
     */
    public void cleanup() throws SQLException {
        synchronized(cleanupThread) {
            cleanupThread.notify();
        }
    }

    /**
     * Does the actual work of cleanup.  Should *only* be called from the Cleanup Thread.
     * Carefully locks the tables structure to assure no corruption.<BR><BR>
     *
     * TODO:  possibly cleanup indexes as well.  Less of a good idea if we need to hit disk, since
     * we will get to those eventually and locality plays a bigger part on those than hashing (ie,
     * got bucket, will travel)
     */
    private void performCleanup() throws SQLException {
        int i = 0;
        TableStorage t = null;
        while(true) {
              //  TODO:  why do we grab the lock twice (here and inside getTableStorage)
            synchronized(tables) {
                if(i >= tables.size())
                    return;

                t = getTableStorage(i);
            }

            if(t != null) {
                //((MemoryLockManager)t.getLockManager()).cleanup();
            }

            i++;
        }
    }

    /***************************************************************************
     * Transaction callbacks
     */

    /**
     * registered when a drop table is requested.  Note that because of guareneteed ordering
     * of these callbacks, this will get called strictly before the table is unlocked.
     * (And in the case of a rollback,
     */
    class DropTableCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            TableStorage ts = (TableStorage)callbackData;

            if(finalStatus == Transaction.ROLLED_BACK) {
                synchronized(tables) {
                    tables.set(ts.getTableId(), ts);
                }
            } else {
                ts.dropTable();
            }
        }
    }

    class DropIndexCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            MemoryIndexStorage index = (MemoryIndexStorage)callbackData;

            if(finalStatus == Transaction.ROLLED_BACK) {
                synchronized(indexes) {
                    indexes.set(index.getIndexId(), index);
                }
            } else {
                index.dropIndex();
            }
        }
    }

    class DropSequenceCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            MemorySequenceStorage seq = (MemorySequenceStorage)callbackData;

            if(finalStatus == Transaction.ROLLED_BACK) {
                synchronized(sequences) {
                    indexes.set(seq.getSequenceId(), seq);
                }
            } else {
                //seq.dropIndex();
            }
        }
    }

    /***************************************************************
     * Cleanup worker thread.
     */
    class CleanupThread extends Thread {
        public CleanupThread() {
            super("MemoryStorageManager Cleanup Thread");
        }

        public synchronized void run() {
            while(true) {
                try{
                    this.wait(cleanupPeriodSeconds * 1000);

                    //System.out.println("Cleaning up locks...");
                    performCleanup();
                }catch(Exception ie) {
                    //  TODO:  log critical error
                    ie.printStackTrace();
                }
            }
        }
    }
}


