/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Nov 11, 2001
 * Time: 3:37:27 PM
 * To change template for new class use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb;

import freebase.api.TableCursor;
import freebase.engine.ConsistentView;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.lock.LockWait;
import freebase.log.LogEntryHeader;
import freebase.memorydb.log.MemoryRowUpdateLogEntry;
import org.apache.log4j.Category;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Could periodically have a thread going through and cleaning out
 * old garbage.
 *
 * TODO:  Why does the Transaction have to keep being passed in?  Get rid of it everywhere
 * TODO:  rewrite to have multiple locks for various subgroups of rows (like perhaps odd and even?)
 * TODO:  can curData ever be not populated?  It doesn't appear so.  It not, remove the flag
 */
public class MemoryTableCursor extends TableCursor {
    static Category cat = Category.getInstance(MemoryTableCursor.class.getName());

    TransactionManager tm;
    MemoryTableStorage table;

    /**
     * we fetch the actual data at construct time
     */
    ArrayList rows = null;

    int cursorPos = -1;

    boolean curDataHasBeenPopulated = false;

    /**
     * What we see! (or what we just wrote)
     */
    Object[] curData = null;

    /**
     * Represents the most recent version of this row.
     * We definitely might NOT see this, but it is cached here for the moment to
     * test some implementations
     */
    Object[] mostRecentRowData = null;

    /**
     * Maybe an unnecessary optimization, but this is an attempt to
     * not require extra log messages for row locks when row updates
     * occur.  An explicit row lock would require an extra 30 bytes
     * or more each, a cost i think is worth striving to save.
     *
     * The data below holds pending lock info.  If a write occurs,
     * then this info can be cleared.  Otherwise an explicit lock
     * will need to be made, as in "select for update".
     * Both the pieces of data below apply to the current cursor position.
     *
     * lockPending is true when a new lock has been obtained temporarily on this row
     * when a row lock was not owned before
     */
    boolean lockPending = false;

      //  true when the current cursor can automatically see the most recent version of this row
    boolean currentRowVisible = false;

      //  who previously
    //Transaction prevTransaction;

    /**
     * The id of the writer who wrote to this entity before us.  This will only be set
     * when a lock is obtained, to determine if the lock should be released before returning
     */
    //long prevTransactionID;

    /**
     * This object is the snapshot of the cursor
     */
    ConsistentView viewSnapshot;

    /**
     * This is only ONLY used when we are updating the data.
     * This saves us the trouble of going back to the system and finding the data.
     */
    MemoryRowData curDataHolder;

    /**
     * We will overwrite the lastWriterTransactio ID in the data holder above so we need to
     * record the old one to make sure our log message is correct.  This is set after
     * the row is locked
     */
    long originalPrevTransactionID;

    /**
     * The object corresponding, currently used only to get the SCN before locking in SERIALIZABLE
     * or SELECT FOR UPDATE mode
     */
    Transaction originalPrevTransaction;

    /**
     * Transaction is used to register the cursor with the connection and
     * to extract a consistent view
     */
    public MemoryTableCursor(Transaction t, MemoryTableStorage storage) throws SQLException {
        super(t);
        this.tm = t.getTM();
        this.table = storage;

        rows = storage.getRowsAndIncNumCursors();

          //  TODO:  move view snapshot up to the abstract class
          //  TODO:  or it can go into TableTupleIterator
        viewSnapshot = t.getCurrentConsistentView();
        viewSnapshot.incrementUseCount();

        //  debug code
        //if(table.getTableId() == 8 && first) {
        //    first = false;
        //    Helper.sleep(20000);
        //}
    }
    //static boolean first = true;

    /**
     * Moves the cursor to the front
     */
    public void reset() throws SQLException {
        curDataHasBeenPopulated = false;
        currentRowVisible = false;
        curDataHolder = null;

        cursorPos = -1;
    }

    /**
     * Scans for the next row visibly by the given transaction.
     */
    public boolean next(Transaction t) throws SQLException {
        curDataHasBeenPopulated = false;
        currentRowVisible = false;

          //  must we hold on to this lock while we rebuild the log entry?
        while(true) {
            synchronized(rows) {
                cursorPos++;
                if(cursorPos >= rows.size()) { //  could use the intial row count here, but no big difference
                    return false;
                }

                curDataHolder = (MemoryRowData)rows.get(cursorPos);
            }

            populateCurData(t);

              //  if curData is null, then the row has been deleted
            if(curData != null) {
                return true;
            }
        }
    }

    public long getCurrentRowNum() throws SQLException {
        return cursorPos;
    }

    /**
     * Move the cursor to the given rowId
     */
    public boolean setCurrentRowNum(Transaction t, long newRowNum) throws SQLException {
        //  TODO:  what about lockPending?  This should send a log message
        curDataHasBeenPopulated = false;
        currentRowVisible = false;
        //curDataHolder = null;

        synchronized(rows) {
            curDataHolder = (MemoryRowData)rows.get((int)newRowNum);
            if(curDataHolder == null) {
                throw new SQLException("Illegal row num");
            }

            cursorPos = (int)newRowNum;
        }

        populateCurData(t);

        return curData != null;
    }

    /**
     * The lock for the row is assumed to be already obtained.
     * @param data the caller expects data to be copied locally
     */
    public void setRowData(Transaction t, Object[] data) throws SQLException {
          //  deals with it's own concurrency
        setNewRowState(cursorPos, curDataHolder, t, data);
    }

    /**
     * Gets data for the current row.  If the row does not exist, then returns false.
     */
    public boolean getRowData(Transaction t, Object[] copy) throws SQLException {
        if(! curDataHasBeenPopulated) {
            populateCurData(t);
        }

          //  oops, the data has been deleted.
        if(curData == null) {
            return false;
        }

        //System.out.println("about to copy data: " + StringUtil.dump(curData) +  "  to " + StringUtil.dump(copy));
        for(int i = 0; i < curData.length; i++) {
              // TODO:  remove these, since they should never happend ever ever every
            if(copy == null) {
                System.out.println("ERROR: copy IS NULL");
            }

            if(curData == null) {
                System.out.println("ERROR: curData IS NULL");
            }

            copy[i] = curData[i];
        }

          //  fill in the rest with nulls.  This only happens if the table was altered
        if(curData.length < copy.length) {
            for(int i = curData.length; i < copy.length; i++) {
                copy[i] = null;
            }
        }

        return true;
    }

    /**
     * Currently has no effect on the cursor.  Return the rowNum created
     */
    public long insertRow(Transaction t, Object[] data) throws SQLException {
        long newRowId = -1;

        boolean putAtEnd = false;
        boolean foundAHoleToUse = false;

          //  is there ever a time when rows, the holes is obtained (deadlock potential?)

        synchronized(table.holes) {
            if(table.holes.size() > 0) {
                for(int index = table.holes.size() - 1; index >= 0; index--) {
                    Long hole = (Long)table.holes.get(index);
                    newRowId = hole.longValue();

                      //  now do the change
                    MemoryRowData wasThere = null;
                    synchronized(rows) {
                        wasThere = (MemoryRowData)rows.get((int)newRowId);
                    }

                      //  check the space, is it still occupied?
                    Transaction prevTrans = tm.getTransaction(wasThere.getRecentTransactionId());
                    if((prevTrans == null || ! prevTrans.isActive()) && wasThere.isDeleted()) {
                        originalPrevTransactionID = wasThere.getRecentTransactionId();
                        //  do the whole hole thing...
                        table.removeHoleAt(index);   //  this is only ok if we abandon the iterator
                        setNewRowState(newRowId, wasThere, t, data);
                        foundAHoleToUse = true;
                        break;
                    }
                }
            }
        }

        synchronized(rows) {
            if(! foundAHoleToUse) {
                  //  totally new row in new posistion
                newRowId = rows.size();
                Object[] copy = duplicate(data);

                long curLSN = t.writeLogEntry(new MemoryRowUpdateLogEntry(0, 0,
                        table.getTableId(), newRowId, null, copy));
                MemoryRowData newData = new MemoryRowData(t.getTransactionId(), copy, curLSN);
                rows.add(newData);
            }
        }

        return newRowId;
    }

    /**
     * The lock for the row is assumed to be already obtained.
     */
    public void deleteRow(Transaction t) throws SQLException {
        /*
        MemoryRowData data = null;
        synchronized(rows) {
            data = (MemoryRowData)rows.get(cursorPos);
        }*/

        table.addHole(cursorPos);
        setNewRowState(cursorPos, curDataHolder, t, null);
        curDataHolder = null;

        //  preserve cursorPos for indexes and for position stability
    }

    /**
     * Locks the current row in memory.  It doesn't immediately write a lock row
     * log record for two reasons:  sometimes the lock is not really needed permanently,
     * in which case it is unlocked (a lot of constraint checks work like that)
     *
     * Also,
     * @param t redundant, we should already know who the transaction is.
     */
    public void reserveRowLock(Transaction t, boolean tableRowExclusive) throws SQLException {
          //  because of this sync body, we are doing direct calls into
        LockWait lockWait = null;

        if(curDataHolder == null) {
            System.out.println("DEBUG.reserveRowLock: " + curData + " " + t + "  rownum:" + cursorPos);
        }

        synchronized(curDataHolder) {
            if(curDataHolder.lastWriterTid != t.getTransactionId()) {
                Transaction lastWriterTrans = tm.getTransaction(curDataHolder.pendingLockTid);
                /**
                 * Note that in some cases lastWriterTrans will not actually be the holder of the lock.
                 * in a rare race, lastWriterTrans may have just finished and woken up another transaction
                 * that we actually beat to the synchronized curDataHolder below (see HERE).
                 * The value lock already handles this case... it takes lastWriterTrans as a hint
                 * of who to wait on if no wait queue is found for the row
                 */
                lockWait = table.getLockManager().lockRow(t, lastWriterTrans, cursorPos);

                  //  do we even need to wait?
                if(lockWait == null) {
                    _establishBeingLockHolder(t.getTransactionId());
                }
            }
        }

        if(lockWait != null) {
            lockWait.await();
            synchronized(curDataHolder) {   //  HERE
                _establishBeingLockHolder(t.getTransactionId());
            }
        }

        currentRowVisible = true;
    }

    /**
     * called from above, the curDataHolder monitor MUST be held!!!
     */
    private void _establishBeingLockHolder(long transactionId) {
          //  this works because before writing a log entry, this always gets called.
        //originalPrevTransactionID = curDataHolder.lastWriterTid;

        curDataHolder.pendingLockTid = transactionId;  //  has lock already

          //  note we are not setting the LSN... yet.  It will be set soon enough
        lockPending = true;  //  this flag tells the system to undo the lock if the row no longer fits criteria
        curDataHasBeenPopulated = false;
    }

    /**
     * Unlocks the current row. This is ONLY called when an update finally gets a lock
     * and realizes that the update condition is no longer true.
     * By virtue of requiring the lock and finding the data has changed,
     * this is required to unlock the row.
     *
     * It is like a very mini rollback, but since a log entry was never made
     * one is not revoked
     */
    public void releaseRowLock() throws SQLException {
        currentRowVisible = false;
        if(! lockPending) {
            System.out.println("WARNING:  unlock row called but no lock pending");
            return;
        }

          //  leave the data structure in a good state.  Can only be null if deleted
        if(curDataHolder != null) {
            synchronized(curDataHolder) {
                curDataHolder.pendingLockTid = curDataHolder.lastWriterTid;
            }
            //curDataHolder.setlastWriterTid(originalPrevTransactionID);
        }

        lockPending = false;
        table.getLockManager().unlockRow(cursorPos);
    }

    /**
     * used in serializable mode
     * @return the SCN of the last transaction that wrote to this, or 0 if it is sufficiently old
     * (no active view pins it in memory)
     * @throws SQLException
     */
    public long getLastSCN() throws SQLException {
        if(originalPrevTransaction == null) {
            return 0;
        } else {
            return originalPrevTransaction.getFinishTime();
        }
    }

    /*****************************************************
     * internal methods
     */

    /**
     * Populates the currentData variable.  Called from
     * next, setrow and getdata etc.  This walks the log entries back to reconstuct a good
     * version of the row.
     */
    private void populateCurData(Transaction t) throws SQLException {
        //System.out.println("POPULATE row " + cursorPos + " t:" + t + "  lock:" + this.lockPending);

        if(curDataHolder == null) {
            curData = null;
            throw new SQLException("assert failed FOUND NULL DATA IN MemoryTableCursor");  //  should only happen in a hole
        }

        synchronized(curDataHolder) {
            originalPrevTransactionID = curDataHolder.getRecentTransactionId();
            originalPrevTransaction = tm.getTransaction(originalPrevTransactionID);
            Transaction recentTransaction = originalPrevTransaction;
            long nextLSNtoGrab = curDataHolder.getCurStateLSN();

              //  TODO:  in the case that the Object[] is created every time, this doesn't need to
              //  be converted until last minute.  Add a special flag boolean first or something
            Object[] foundData = curDataHolder.getData();
            mostRecentRowData = foundData;  //  cached for putting in the log message if overwritten

            while(true) {
                //if(show) System.out.print("Comparing us: " + t + " and t:" + recentTransaction + " row: " + cursorPos);

                  //  since recent transaction is the
                if(currentRowVisible ||
                        viewSnapshot.canSeeChangesFrom(recentTransaction, nextLSNtoGrab)) {
                    //if(show) System.out.println(" see  " + StringUtil.dump(curData));

                      //  no longer need to duplicate the data
                    curData = foundData;
                    curDataHasBeenPopulated = true;

                    /*
                    if(foundData != null) {
                        DebugLog.log(t.getConnection(), "CAN SEE: lsn: " + nextLSNtoGrab + " (t:" + recentTransaction +
                            ")  data:" + StringUtil.dump(foundData) + "  our consistent view: " + viewSnapshot +
                            "\n    Row #" + cursorPos + "  start lsn:" + curDataHolder.getCurStateLSN() + "  us: " + t + "  ");
                    }
                    */

                    return;
                }

                //if(show) System.out.println(" no see");

                  //  keep building the old state
                LogEntryHeader header = t.getLogEntry(nextLSNtoGrab);
                if(header == null) {
                    throw new SQLException("Snapshot too old.  Log data no longer available: lsn: " + nextLSNtoGrab);
                    /*
                      //  TODO:  make non-debugging version of this exceptioN!
                    StringBuffer sb = new StringBuffer("Snapshot too old.  Log data no longer available: lsn: " + nextLSNtoGrab +
                            " (t:" + recentTransaction + ")  data:" + StringUtil.dump(foundData) + "\n");
                    if(recentTransaction != null) {
                        sb.append("   status:").append(recentTransaction.getStatus());
                        sb.append("  scn:").append(recentTransaction.getFinishTime());
                    }

                    sb.append("  our consistent view: ").append(t.getCurrentConsistentView() + "  cached: " + viewSnapshot);
                    sb.append("\nRow #" + cursorPos + "  start lsn:" + curDataHolder.getCurStateLSN() + "  is locked:" + lockPending);
                    sb.append("  most recent writer t: " + originalPrevTransactionID);
                    throw new SQLException(sb.toString());
                    */
                }

                /**
                 * TODO:  this got a bad cast once:  RollbackLogEntry
                 */
                MemoryRowUpdateLogEntry logEntry = (MemoryRowUpdateLogEntry)
                            header.getLogEntry();

                //  TODO:  see if the next trans is the same as the last.
                //    if so, just keep looking back through the log
                //if(show) System.out.println("previous transaction is id " + logEntry.getPreviousTransactionId());
                recentTransaction = tm.getTransaction(logEntry.getPreviousTransactionId());
                nextLSNtoGrab = logEntry.getPrevRowUpdateLSN();
                foundData = logEntry.getOldData();
            }
        }
    }

    /**
     * For now just checks to see if there is a pending lock log
     * entry to send.  Later this will actually do the data flushing too.
     */
    private void prepareForMove() {
        if(lockPending) {
            cat.error("STILL NEED TO IMPLEMENT LOCK LOG message");
        }
    }

    /**
     * TODO:  see if this is being called at the right places,
     *   and only at the right places!
     *
     * TODO:  does not need to clone each
     *     subobject if they are Immutable
     */
    private Object[] duplicate(Object[] data) {
        if(data == null)
            return null;

        Object[] copy = new Object[data.length];

        for(int i = 0; i < data.length; i++) {
            copy[i] = data[i];
        }

        return copy;
    }

    /**
     * Could eventually be called by all of the above methods.
     */
    private void setNewRowState(long rowNum, MemoryRowData row, Transaction t,
                                Object[] newData) throws SQLException {
        //cat.assert(lockPending, "We expected a lock to be pending but it wasn't");

       // System.out.println("writing as previous transaction:" + originalPrevTransactionID);

          //  NOTE:  not protected, can it be corrupted by concurrent writers?  No because we are the
          //  exclusive lock holder for this row and we necessarily went through a memory barrier to get here
        long lastLSN = row.getCurStateLSN();

        Object[] newDataCopy = duplicate(newData);

        long curLSN = t.writeLogEntry(new MemoryRowUpdateLogEntry(
                        originalPrevTransactionID, lastLSN, table.getTableId(),
                        rowNum, mostRecentRowData, newDataCopy));

        lockPending = false;
        row.setNewState(t.getTransactionId(), newDataCopy, curLSN);

        curData = newDataCopy;
    }

    /**
     * TODO: change this to closeCursorSpecific() or cursorSpecificClose()
     */
    public void close() throws SQLException {
        super.close();

        rows = null;

        table.decNumCursors();

          //  TODO: maybe move view snapshot up to the abstract class
        viewSnapshot.decrementUseCount();
    }
}
