/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Dec 12, 2001
 * Time: 7:14:50 PM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb;

import freebase.api.*;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.lock.TableLockManagerImpl;
import freebase.log.LogEntry;
import freebase.log.data.LockTableEntry;
import freebase.memorydb.log.MemoryRowUpdateLogEntry;
import freebase.memorydb.log.MemoryStorageLogEntry;
import freebase.util.ErrorCode;
import freebase.util.GlobalProps;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * First pass at an in memory storage implementation
 * TODO: replace rows with multiple ones for better concurrency.  put methods:
 *        Object getGroupMonitor(long rowId)  (rather than all locking same)
 *        MemoryRowData getRowData(long rowId)
 */
public class MemoryTableStorage implements TableStorage {
    /**
     * A debugging flag for turning off hole recording (and thus the entire feature).
     * Setting to false is sure to get you OutOfMemory, so don't unless you think holes is causing a bug.
     */
    final public static boolean ENABLE_HOLES = true;

    TransactionManager tm;
    int tableId;
    String name;

    TableLockManager locks;

    /**
     * Holds the list of holes in the ArrayList.  Eventually they get plugged.
     * Eventually implement this as a HashSet of Integers or LongList.
     * See MemoryTableCursor.insertRow for implementation details.
     */
    ArrayList holes = new ArrayList();

    /**
     * The row data.
     */
    private ArrayList rows = new ArrayList();

    /**
     * Used to make either getting a cursor and dropping or truncating the table atomic.  Unfortunately this
     * needs to be implemented in all implementations of this class because I couldn't come up with a nice design
     * putting it into Table (it was there).  The problem comes in with the fact that dropping the table should
     * return quickly, but read cursors on the failover node should still be legal.  Thus, to handle this properly,
     * all the code currently needs to be at this level.  Otherwise we would need to route drop table events through
     * the Table object corresponding to this, not part of the current design.
     * <P>
     * NOTE:  so far, something like this appears to still be needed to ensure atomic inc and dec of # cursors
     */
    Object cursorLock = new Object();

    /**
     * only when this is zero can the data actually be deleted.  Later consider dealing with
     * this at a higher level, perhaps in the Table data structure.
     */
    int numOutstandingCursors;

    /**
     * true when the table has been dropped
     */
    boolean isDropped = false;

    public MemoryTableStorage(TransactionManager tm, int tableId) {
        this.tm = tm;
        this.tableId = tableId;

        //locks = new MemoryLockManager(this);
        locks = new TableLockManagerImpl(this);
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    public int getTableId() {
        return tableId;
    }

    /**
     * Called ONLY from TableTupleIterator
     */
    public void decNumCursors() {
        synchronized(cursorLock) {
            numOutstandingCursors--;

            //  TODO:  maybe notify all?  and, have all indexes waiting on this

            if(isDropped && numOutstandingCursors == 0) {
                _dropTable();
            }
        }
    }

    /**
     * Called when a new cursor is forming from MemoryTableCursor.
     * Will throw an exception if the table has not been dropped!!!
     */
    public ArrayList getRowsAndIncNumCursors() throws SQLException {
        synchronized(cursorLock) {
            if(isDropped) {
                throw new SQLException("Table does not exist.");
            }

            if(numOutstandingCursors == GlobalProps.getMaxCursors()) {
                ErrorCode.throwException(ErrorCode.TOO_MANY_CURSORS, "Table " + getName());
            }

            numOutstandingCursors++;

            return rows;
        }
    }

    /**
     * Basically a token no-op.  Since this will frequently be called from the close of someones
     * cursor, errors need to be directed elsewhere...  what to do?
     */
    public void dropTable() {
        synchronized(cursorLock) {
            isDropped = true;
            if(numOutstandingCursors == 0) {
                _dropTable();
            }
        }
    }


    /**
     * do the actual dropping.
     * back into the Table object
     */
    private void _dropTable() {
        holes = null;
        rows = null;
        locks = null;
        tm = null;
    }

    /**
     * Truncate the table.  Reset the data.  Since only write cursors use holes, there is no need to copy it aside.<BR>
     * This works so simply because an EXCLUSIVE lock has already been obtained at the higher level,
     * thus no transaction has any pending work to commit or rollback...  thus holes and everything else
     * can just go away!
     * @param t if is not null, then log a truncate log message.  Otherwise, don't.
     * @throws SQLException
     */
    public void truncateTable(Transaction t) throws SQLException {
        t.writeLogEntry(new MemoryStorageLogEntry(tableId,
                        MemoryStorageLogEntry.TRUNCATE_STORAGE,
                        MemoryStorageLogEntry.TABLE_STORAGE));

        t.registerDoneCallback(new TruncateTableCallback(), null);
    }

    /**
     * does the actual work of truncating
     */
    public void _truncateTable() {
        synchronized(cursorLock) {
            rows = new ArrayList();
            holes = new ArrayList();
        }
    }

    /**
     * Grab a cursor to do a table scan.
     * @param t
     * @param readWriteMode
     * @return
     */
    public TableCursor getTableCursor(Transaction t, boolean readWriteMode) throws SQLException {
        return new MemoryTableCursor(t, this);
    }

    /**
     * Apply the log entries either forward or back.  Current heuristic is
     * that if the transaction that is referenced is not found then don't actually
     * grab the lock.  This should only be the case if it is decided to not actually
     * create the real transactions.
     */
    public void processLogEntry(int applyMode, long lsn, Transaction t, LogEntry logEntry) throws SQLException {
        if(logEntry instanceof MemoryRowUpdateLogEntry) {
            MemoryRowUpdateLogEntry logE = (MemoryRowUpdateLogEntry)logEntry;

              //  find the row first
            Transaction prev = tm.getTransaction(logE.getPreviousTransactionId());

            MemoryRowData data = null;
            int rowId = (int)logE.getRowId();
            synchronized(rows) {
                while(rowId >= rows.size()) {
                    rows.add(null);
                }
                data = (MemoryRowData)rows.get(rowId);
            }

            if(applyMode == LogEntryProcessor.UNDO_ENTRY) {
                  /**
                   * The new state has the LSN that it had before.  This is stable against
                   * reads (since nothing else could have changed it since the prev state)
                   * TODO:  but will need to implement something like a dirty flag or something for
                   * disk writeout
                   */
                data.setNewState(logE.getPreviousTransactionId(), logE.getOldData(), logE.getPrevRowUpdateLSN());
                //System.out.println("ROLLBACK ROW: " + rowId);

                  //  undoing delete operation (thus insert
                if(logE.getNewData() == null) {
                    removeHole(rowId);
                }

                  //  insert operation, remove hole just in case
                if(logE.getOldData() == null) {
                    addHole(rowId);
                }

                if(t != prev && t != null) {
                    locks.unlockRow(rowId);
                }
            } else {  //  DO:  not redo, since it was not
                if(data == null) {
                    data = new MemoryRowData(t.getTransactionId(), logE.getNewData(), lsn);
                    synchronized(rows) {
                        rows.set(rowId, data);
                    }
                } else {
                    data.setNewState(t.getTransactionId(), logE.getNewData(), lsn);
                }

                  //  delete operation
                if(logE.getNewData() == null) {
                    addHole(rowId);
                }

                  //  insert operation, remove hole just in case
                if(logE.getOldData() == null) {
                    removeHole(rowId);
                }

                if(t != prev && t != null) {
                      //  redo:  don't expect any owner.
                      //  TODO:  what are we doing locking the data anyway!!!
                    locks.lockRow(t, null, rowId);
                }
            }
        } else if(logEntry instanceof LockTableEntry) {
            LockTableEntry lockTable = (LockTableEntry)logEntry;

            if(applyMode == LogEntryProcessor.UNDO_ENTRY) {
                getLockManager().unlockTable(t, lockTable.getLockMode());
            } else {
                if(t != null) {
                    getLockManager().lockTable(t, lockTable.getLockMode(), false);
                }
            }
        } else {
            System.out.println("MemoryCatalog.processLogEntry could not process: " + logEntry);
        }
    }

    public TableLockManager getLockManager() throws SQLException {
        return locks;
    }

    public void addHole(int hole) {
        if(! ENABLE_HOLES) {
            return;
        }

        synchronized(holes) {
            holes.add(new Long(hole));
        }
    }

    public void removeHole(int hole) {
        synchronized(holes) {
            for(int i = 0; i < holes.size(); i++) {
                Long h = (Long)holes.get(i);
                if(h.longValue() == hole) {
                    holes.remove(i);
                    return;
                }
            }

              //  this will get called often when there is no real hole being filled
            //throw new OhShitException("Tried to remove hole " + hole + " but not found in : " + holes);
        }
    }

    /**
     * Remove the hole recorded at the position in the list (not value), since its being filled.  No longer synchronized
     * because the caller synchronizes the same thing.
     * @param offsetInList
     */
    public void removeHoleAt(int offsetInList) {
        //synchronized(holes) {
            holes.remove(offsetInList);
        //}
    }

    /***************************************************************************
     * Transaction callbacks
     */

    /**
     * Do the actual truncation now that the log message has successfully been flushed (and
     * before the UseLock is released)
     */
    class TruncateTableCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            if(finalStatus == Transaction.COMMITED) {
                _truncateTable();
            }
        }
    }
}
