/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Dec 28, 2001
 * Time: 8:35:58 PM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb.index;

import freebase.api.IndexCursor;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;

/**
 * Represents a node in an index.  How does a changed node know when it is no longer
 * changed thus not needing TODO a lookup on the actual status of the data?
 */
public class IndexNode {
    /**
     * This probably should not be used to make judgements about whether the reader can definitely
     * not see the value unless the actual transaction itself cannot be found anywhere.  This is
     * because we may have recently grabbed a lock that the index itself does not know about, indeed
     * it can't know about it because the lock information itself is perhaps on disk.<BR><BR>
     *
     * Note, if we just kept a pointer to the transaction it would actually save space in memory!
     * just make sure we look the thing up via the transaction id to see if it is still in scope,
     * ie store the object, but lookup to see if there is still a reference to it in the Trans table.
     */
    long lastWriterTransactionId;

    byte status = IndexCursor.INSERTED_FLAG;
    long rowId;

    IndexNode next = null;
    IndexNode prev = null;

    /**
     * When the lock count is non-zero, the node cannot be removed.
     * TODO:  2004 now, the node can be removed, but it is left in a way that it can
     *    still traverse into the chain of legal nodes (or some even still illegal)
     * TODO:  get rid of this eventually as long as things continue to work
     *
     * Now uses an explicit mark of REMOVED_FLAG for concurrent placement on nodes that
     * have been removed.  a REMOVED node will not be removed again.  Its forward pointer
     * remains a (semi) legal path, so a chain of removed nodes can be traversed.
     *
     * When rewriting this as an ArrayList (for speed), keep in mind that something similar
     * (a counter in the header) will record how many are currently referencing this list
     *   and thus, we must keep the numbers preserved
     */
    int lockCount = 0;

    /**
     * All new nodes are in INSERTED state
     */
    public IndexNode(long rowId, long transactionId) {
        this.rowId = rowId;
        this.lastWriterTransactionId = transactionId;
    }

    /**
     * The inc and dec lock assume that they are already in a protected region.
     */
    public void incrementLock() {
        lockCount++;
    }

    public void decrementLock() {
        lockCount--;
    }

    public long getRowId() {
        return rowId;
    }

    public byte getStatus() {
        return status;
    }

    public void setStatus(byte newStatus, long transactionId) {
        //System.out.println("Writing entry with status: " + newStatus + " tid:"+ transactionId);
        status = newStatus;
        lastWriterTransactionId = transactionId;
    }

    /**
     * For determining if this dddata is visible or not
     */
    public long getLastWriterTransactionId() {
        return lastWriterTransactionId;
    }

    /**
     * Returns true if this node is valid as a possible entry in some view.
     *  It is false only if the node was deleted by a transaction
     * that is sufficiently old that it's object is no longer findable in the
     * transaction manager.  The caller is then responsible to remove the node
     * from the index.<BR><BR>
     *
     * This method serves the secondary function of getting rid of any old dirty state
     * by detecting when the transaction is no longer.  It will cleanup a freshly inserted
     * value by clearing the status to NORMAL and erasing the transaction id.
     */
    public boolean isValid(TransactionManager tm) {
        //System.out.println("Validating index node row:" + rowId + " status: " + status + " lockcount: " + lockCount);

        if(status == IndexCursor.NORMAL_FLAG) {  // || lockCount != 0) // -or- lastWriterTransactionId == 0)
            return true;
        }

        Transaction t = tm.getTransaction(lastWriterTransactionId);

          //  if we can still see the transaction, then this should keep its status
        if(t != null) {
            return true;
        }

          //  now t == null.
        if(status == IndexCursor.DELETED_FLAG) {
            //System.out.println("getting rid of node!!!");
            return false;
        } else {
            lastWriterTransactionId = 0;
            status = IndexCursor.NORMAL_FLAG;  //TODO:  uncomment this back!!! debug only
            return true;
        }
    }

    /***********************************************************
     * link traversal methods
     */

    /**
     *
     */
    public IndexNode getNext() {
        return next;
    }

    public void setNext(IndexNode next) {
        this.next = next;
    }

    public IndexNode getPrev() {
        return prev;
    }

    public void setPrev(IndexNode prev) {
        this.prev = prev;
    }

    public String toString() {
        return "status:" + IndexCursor.statusNames[status] + " rid:" + rowId +
                " tid:" + lastWriterTransactionId + " p:" + super.toString();
    }

    /**
     * For debug only, traverses chain back, then forward and records the state.
     * @param forward for the first call, make it false!
     */
    public void debugBuildChain(StringBuffer sb, boolean forward) {
        if(! forward) {
            if(prev != null) {
                sb.append("[ going back from " + this + "] ");
                prev.debugBuildChain(sb, false);
            } else {
                debugBuildChain(sb, true);
            }
        } else {
            sb.append(this);
            if(next != null) {
                sb.append(", ");
                next.debugBuildChain(sb, true);
            }
        }
    }

    /**
     * return the entire chain that this belongs to, for debug only
     */
    public String debugChain() {
        StringBuffer sb = new StringBuffer("Chain: ");
        debugBuildChain(sb, false);
        return sb.toString();
    }
}
