/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Dec 28, 2001
 * Time: 3:21:23 PM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb.index;

import freebase.api.IndexCursor;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.memorydb.log.MemoryIndexLogEntry;
import freebase.util.SqlTypes;

import java.sql.SQLException;
import java.util.Iterator;
import java.util.TreeMap;


/**
 * MemoryIndexCursor is intended to be a preliminary study of the requirements for
 * an index so that the apis can be well tuned.  The implementation of this is not
 * to be considered hightly concurrent. (correct, but not high performance)
 *
 * TODO:  need to add log messages for rollback and redo
 * TODO:  If another thread somehow eliminates the current node somehow, how does this fair?
 *   Although it won't be included in either view, will it be removed correctly, especially if
 *   the next node is bad too?
 */
public class MemoryIndexCursor implements IndexCursor {
    TransactionManager tm;  //  is really needed?
    Transaction t;
    MemoryIndexStorage storage;

    Comparable curKey = null;
      //  called xcurNode so that it doesn't get referenced directly
    IndexNode xcurNode = null;

      //  data pulled out from the cur node before returning (in setCurNode())
    long curRowId;
    byte curFlags;


    final TreeMap tree;

    /**
     * Construct a MemoryIndexCursor ready to go
     */
    public MemoryIndexCursor(TransactionManager tm, Transaction t, MemoryIndexStorage store) throws SQLException {
        this.tm = tm;
        this.t = t;
        this.storage = store;

        this.tree = store.getTreeAndIncNumCursors();
    }

    /**
     * Insert the pair (value, rowid), or more specifically make sure it is there.
     */
    public void insertEntry(Object value, long rowId) throws SQLException {
          //  go ahead and use the storage implementation, reuse good
        storage.insertEntry(value, rowId, t.getTransactionId(), t, true);
    }

    /**
     * Mark the current position as DELETED, and the new position as inserted?
     * @deprecated update is dealt with in the Index object, no need for it here too
     */
    public void updateEntry(Object newValue) throws SQLException {
          //  NOTE:  the following makes this operation totally atomic.  Is it needed?
          //  if so, how TODO it in the final block case?
        synchronized(tree) {
            getCurNode().setStatus(DELETED_FLAG, t.getTransactionId());
            insertEntry(newValue, getCurNode().getRowId());
        }
    }

    /**
     * Its a bit of a mess how insert is implemented in MemoryIndexStorage but delete is implemented here.
     * I don't like this breaking of the layers, but working is first goal.
     * @throws SQLException
     */
    public void deleteEntry() throws SQLException {
        Object curKeyValue = curKey;
        if(curKey instanceof IndexKey) {
            curKeyValue = ((IndexKey)curKey).getValue();
        }

          //  TODO:  finsih this and put in the update case above
        t.writeLogEntry(new MemoryIndexLogEntry(storage.getIndexId(), curKeyValue,
                getCurNode().getRowId(), MemoryIndexLogEntry.DELETE, getCurNode().getStatus(),
                getCurNode().getLastWriterTransactionId()));

          //  this is here only for the sake of thread visibility
        synchronized(tree) {
            getCurNode().setStatus(DELETED_FLAG, t.getTransactionId());

            /**
             * try:
             * First:  update test set id=id + 1;
             * Then:   update test set id=id - 1 where condition results in collision of one guy
             */

            //  TODO:  check that there exists a reverse
            boolean foundPeerAfter = false;
            IndexNode cur = getCurNode().next;
            searching_for_peer_with_same_transaction:
            while(cur != null) {
                if(cur.getLastWriterTransactionId() == t.getTransactionId()) {
                    foundPeerAfter = true;
                    break searching_for_peer_with_same_transaction;
                }
                cur = cur.next;
            }

              //  if there is another entry after ours in the list that means that the unique
              //  contention count was increased.  Decrease it to assure correct balance.
            if(storage.isUniqueIndex && foundPeerAfter) {
                t.decNumUniqueContentions();
            }
        }
    }

    /**
     * Jump the cursor to this position.  Should be able to support partial position if this is
     * an index of multiple values but only one is supplied.  Ie start at "A"
     * TODO:  support null in secondary component representing "first"
     */
    public boolean setPos(Object value) throws SQLException {
          //  null means "the beginning"
        if(value == null) {
            reset();
            return next();
        }

        curKey = IndexKey.makeKey(value);
        return scanCurKeyToNextLegalNode();
    }

    /**
     * Move the cursor to the beginning of the tree
     */
    public void reset() throws SQLException {
        curKey = null;
        setCurNode(null);
    }

    /**
     *  Find the next entry in this index.  An alternate way of implementing this would
     * be to write a private fetchNext and deal with the validation in the main next only
     */
    public boolean next() throws SQLException {
        synchronized(tree) {
              //  beginning of the tree case, just find the first node if available
            if(curKey == null) {
                curKey = (Comparable)tree.firstKey();
                return setPos(curKey);
            }

              //  normally, just loop until you find a valid node

              //  check the "next" node, the core next code
            IndexNode node = getCurNode().getNext();
            setCurNode(getFirstValidNode(node, curKey));
            if(getCurNode() != null) {
                return true;
            }

              //  we ran to the end, so now fetch the first.  We can't use the iterator
              //  because it is not stable against updates
              //  TODO:  coalese this code with setPos above, nearly the same and should be submethod
            curKey = getNextKeyAfter(curKey);
            return scanCurKeyToNextLegalNode();
        }
    }

    /**
     * All these methods are so similar, it would be nice to extract just one or two smart versions!!
     * When curKey is null, return.  Keep using curKey as the cursor position
     * @return
     */
    private boolean scanCurKeyToNextLegalNode() {
        while(true) {
            if(curKey == null)
                return false;

            setCurNode(getFirstValidNode(curKey));
            if(getCurNode() != null) {
                return true;
            }
            curKey = getNextKeyAfter(curKey);
        }
    }

    /**
     * Given the key, this finds the first valid node.  Any invalid nodes found up to
     * the valid node are removed.  If all of the nodes are deemed invalid, the key itself
     * is removed.
     * @return null if there is no valid node for the given key
     */
    private IndexNode getFirstValidNode(Comparable key) {
        IndexNode node = (IndexNode)tree.get(key);
        return getFirstValidNode(node, key);
    }

    /**
     * returns the first valid node starting with node.  Will return node if valid
     * @param node the first node to start looking for a valid one
     * @param key the key this node is under, used to remove nodes in certain cases
     * @return null if the entire key chain is invalid (or empty
     */
    private IndexNode getFirstValidNode(IndexNode node, Comparable key) {
        while(true) {
            if(node == null || node.isValid(tm)) {
                return node;
            } else {
                IndexNode nextNode = node.getNext();
                removeNode(key, node);
                node = nextNode;
            }
        }
    }

    /**
     * Sets the current node.  Deals with incrementing protection locks etc
     */
    private void setCurNode(IndexNode node) {
        if(xcurNode != null) {
            xcurNode.decrementLock();
        }

        xcurNode = node;
        if(node != null) {
            curFlags = node.getStatus();
            curRowId = node.getRowId();
        }

        if(xcurNode != null) {
            xcurNode.incrementLock();
        }
    }

    public String debugChain;

    private IndexNode getCurNode() {
        return xcurNode;
    }

    /**
     * Removes the given node from the index tree if another concurrent thread hasn't already done the same.
     * (very rare situation, but can happen when multiple nodes in a chain are all removed in a row)
     * NOTE:  leave the node.next pointing the same so that the chain can still be traversed if anyone is in it.
     * Currently protected by the tree monitor.
     */
    private void removeNode(Comparable key, IndexNode node) {
          //  if already removed, just follow the chain on...
        if(node.getStatus() == REMOVED_FLAG) {
            return;
        }

        if(node.getPrev() == null) {
            if(node.getNext() == null) {
                tree.remove(key);   //  alone in the tree, remove the entry itself
            } else {
                tree.put(key, node.getNext());
                node.getNext().setPrev(null);
            }
        } else {
            node.getPrev().setNext(node.getNext());
            if(node.getNext() != null) {
                node.getNext().setPrev(node.getPrev());
            }
        }

        node.status = REMOVED_FLAG;
    }

    /**
     * Returns the next key after the given key.  If the given key does not exist
     * the
     */
    private Comparable getNextKeyAfter(Comparable key) {
          //  I agree, this is nasty.  TreeMap should have a next key function
        Iterator it = tree.tailMap(key).keySet().iterator();

          //  get the first in the list.  If it is not == key then, it is good
        if(it.hasNext()) {
            Comparable first = (Comparable) it.next();
            if(SqlTypes.compare(first, key) != 0) {
                return first;
            }
        } else {
            return null;
        }

          //  Otherwise, get the second in the list... thus the next after
        if(it.hasNext()) {
            return (Comparable)it.next();
        } else {
            return null;
        }
    }

    //  returns everything as Comparable for now
    public Comparable getCurValue(int pos) throws SQLException {
        if(curKey instanceof IndexKey) {
            return ((IndexKey)curKey).getValue(pos);
        } else {
            //assert pos == 0;
            return curKey;
        }
    }

    public long getCurRowId() throws SQLException {
        return curRowId;
    }

    public byte getCurFlags() throws SQLException {
        return curFlags;
    }


    /**
     * called by IndexScan
     * @return
     * @throws SQLException
     */
    public boolean isPossiblyNotVisible() throws SQLException {
        //System.out.println("  " + xcurNode);
        return getCurFlags() != NORMAL_FLAG;
    }

    public void close() throws SQLException {
          //  release the last grip on this guy
        setCurNode(null);

        storage.decNumCursors();

        //  TODO:  nuke the consistent view or whatever (if not part of a scan that also has a table cursor, & thus a view)
    }
}
