/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Dec 28, 2001
 * Time: 3:20:58 PM
 * To change template for new class use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.memorydb.index;

import freebase.api.IndexCursor;
import freebase.api.IndexStorage;
import freebase.api.LogEntryProcessor;
import freebase.api.TransactionDoneCallback;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.lock.LockWait;
import freebase.lock.ValueLock;
import freebase.log.LogEntry;
import freebase.memorydb.log.MemoryIndexLogEntry;
import freebase.memorydb.log.MemoryStorageLogEntry;
import freebase.util.ErrorCode;
import freebase.util.GlobalProps;
import freebase.util.IndexComparator;

import java.sql.SQLException;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * The MemoryIndexStorage uses a TreeMap to a linked list of values.
 * This is probably not ideal since the thing does not have a simple nextKey, but
 * we will make due with slower alternates.
 *
 * Unique Keys:  we can only continue when we are the front active transaction in the list
 * for a given key.  If there are any inserted values in front of us in commited transactions
 * we throw unique key violation.  Otherwise we wait on an enqueue.
 *
 * TODO:  add a background thread that walks the index and does cleanup and promotion.
 * A deleted entry can only be removed finally when the transaction view falls below
 * the online low water mark, thus guarenteeing that an index scan will not need to see
 * that value.
 *
 * TODO:  consider rewriting this with ArrayList's instead of LinkedLists.  Would be simpler logic,
 *     faster, and probably less errors.  Also *may* more closely match what will be in a b-tree.
 *     Also:  concurrency could be much better because only the first lookup will keep the tree lock
 *     NOTE:  would allocate ArrayLists of size 2 or 3 for unique keys, maybe same initial for others
 *       but would try to aggressively reduce the real size frequently
 *    The cursors would have to keep their last pos and object node referenced, and when moving
 *    would have to ensure that the node hasn't moved.  Could store a count of how many cursors are in there
 *
 * This is probably one of the most complex concepts, and made worse by some of the
 * data structures I used here. 
 */
public class MemoryIndexStorage implements IndexStorage {
    TransactionManager tm;
    int indexId;

      //  for debugging
    String name;

    /**
     * The actual index data
     */
    private TreeMap tree;

    /**
     * true if this is a unique index
     */
    boolean isUniqueIndex;
    ValueLock keyLocks;

    /**
     * Same pattern as in MemoryTableStorage
     */
    Object cursorLock = new Object();

    /**
     * only what this is zero can the data actually be deleted.  Later consider dealing with
     * this at a higher level, perhaps in the Table data structure.
     */
    int numOutstandingCursors;

    /**
     * true when the index has been dropped
     */
    boolean isDropped = false;

    public MemoryIndexStorage(TransactionManager tm, int indexId, boolean unique) {
        this.tm = tm;
        this.indexId = indexId;
        this.isUniqueIndex = unique;

        //  TODO:  probably only need to construct this if unique
        if(isUniqueIndex) {
            keyLocks = new ValueLock();
        }

        if(GlobalProps.isStringsCaseInsensitive()) {
            tree = new TreeMap(new IndexComparator());
        } else {
            tree = new TreeMap();
        }
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    public IndexCursor getIndexCursor(Transaction t, boolean readWriteMode) throws SQLException {
        return new MemoryIndexCursor(tm, t, this);
    }

    public int getIndexId() {
        return indexId;
    }

    /**
     * Called ONLY from TableTupleIterator
     */
    public void decNumCursors() {
        synchronized(cursorLock) {
            numOutstandingCursors--;

            //  TODO:  maybe notify all?  and, have all indexes waiting on this

            if(isDropped && numOutstandingCursors == 0) {
                _dropIndex();
            }
        }
    }

    /**
     * Called when a new cursor is forming from MemoryIndexCursor.
     * Will throw an exception if the table has not been dropped!!!
     */
    public TreeMap getTreeAndIncNumCursors() throws SQLException {
        synchronized(cursorLock) {
            if(isDropped) {
                throw new SQLException("Index does not exist.");
            }

            if(numOutstandingCursors == GlobalProps.getMaxCursors()) {
                ErrorCode.throwException(ErrorCode.TOO_MANY_CURSORS, "Index " + getName());
            }

            numOutstandingCursors++;

            return tree;
        }
    }

    /**
     * Truncate the data in the index, much like done in the table
     * @param t
     * @throws SQLException
     */
    public void truncateIndex(Transaction t) throws SQLException {
        t.writeLogEntry(new MemoryStorageLogEntry(indexId,
                        MemoryStorageLogEntry.TRUNCATE_STORAGE,
                        MemoryStorageLogEntry.INDEX_STORAGE));

        t.registerDoneCallback(new TruncateIndexCallback(), null);
    }

    /**
     * called in both post-commit callback and on replay recovery
     */
    public void _truncateIndex() {
        synchronized(cursorLock) {
            tree = new TreeMap();
        }
    }

    /**
     * Basically a token no-op.  Since this will frequently be called from the close of someones
     * cursor, errors need to be directed elsewhere...  what to do?
     * TODO:  rename delete?  along with sequence and table
     */
    public void dropIndex() {
        synchronized(cursorLock) {
            isDropped = true;
            if(numOutstandingCursors == 0) {
                _dropIndex();
            }
        }
    }

    /**
     * do the actual dropping.
     */
    private void _dropIndex() {
        tree = null;
    }

    //  yields insight into the capability of this index
    public boolean canFindWith(String field)  { // []  {
        return false;
    }

    public boolean canScanWith(String field)  { //  [] {
        return false;
    }

    public ValueLock getLockManager() {
        return keyLocks;
    }

    /*
    public boolean canUseIndex(SqlCondition condition) throws SQLException {

    }
    */

    public void dumpContents() {
        synchronized(tree) {
        //Object key = tree.firstKey();

            Iterator it = tree.keySet().iterator();
            while(it.hasNext()) {
                Object key = it.next();
                IndexNode cur = (IndexNode)tree.get(key);
                System.out.println("Entry maps " + key + " -> ");
                while(cur != null) {
                    System.out.println("    " + cur);
                    cur = cur.getNext();
                }
            }
        }
    }

    /******************************************************************
     * methods for doing and undoing
     */

    /**
     * Inserts an entry into the list corresponding to the key value.  Normal behavior is to put the entry
     * at the end of the list, to guarentee a certain order of processing of unique key checks.  If the index
     * is unique and "this" transaction detects that it already has one entry associated with the value, it should
     * put the new value immediately after the existing one (before any other waiters).  This is because it cannot
     * wait on itself to finish (meaningless) and if it puts its entry after the existing waiter that waiter,
     * upon waking up, may think that it has a right to put its value in the list (since it is the front active).
     *
     * Another justification of putting the new row for the same transaction in front of existing waiters is that
     * the current transaction essentially has the "lock" on this index value, and just can do what it likes with
     * what exact row is the one that will point to the value.
     *
     * Transaction t is only required if writeLogEntry is true.  writeLogEntry is redundant but
     * easier to read<BR><BR>
     *
     * TODO:  handle including the undo information in the log entry
     * @param t will be null only when rolling back
     */
    protected void insertEntry(Object value, long rowId, long transactionId, Transaction t, boolean writeLogEntry)
            throws SQLException {
        Comparable key = IndexKey.makeKey(value);
        IndexNode newNode = null;
        LockWait uniqueWait = null;

        IndexNode frontActive = null;  //  out here for debugging purpose
        Transaction frontTransaction = null;
        IndexNode wasFirstForDebugOnly = null;
        boolean neededToWaitOnFront = false;

          //  currently locks at the top of the tree.  Later must lock on key --JJ  1/2/2004
        synchronized(tree) {
            wasFirstForDebugOnly = (IndexNode)tree.get(key);
            IndexNode node = wasFirstForDebugOnly;
            IndexNode prevNode = null;

              //  unique key implementation fields
              //  exclusively: frontActive will be set or
              //  alreadyInc will be true, not both (neither ok)

            boolean alreadyIncContentionCount = false;
            IndexNode firstWithSameTransaction = null;

              //  follow chain to the end
            while(node != null) {
                  //  cache it in case we permantently delete/move this node
                IndexNode nextNode = node.getNext();

                  //  determine if this is front active
                if(isUniqueIndex && frontActive == null && ! alreadyIncContentionCount) {
                    if(node.getStatus() == IndexCursor.NORMAL_FLAG) {
                          //  t will be null if rolling back.  Rollback case works, but is messy through here :(
                        if(t != null) {
                            t.incNumUniqueContentions();
                        }

                        alreadyIncContentionCount = true;
                    } else {
                        Transaction lastWriterTrans = tm.getTransaction(node.getLastWriterTransactionId());
                          //  if the writer is either long gone, or us...  increment the count
                        if(lastWriterTrans == null || node.getLastWriterTransactionId() == transactionId) {  //  whatever the operation, its done!
                            if(node.getStatus() == IndexCursor.INSERTED_FLAG) {  //  if inserted, then violation
                                if(t != null) {
                                    t.incNumUniqueContentions();
                                }
                                alreadyIncContentionCount = true;
                            } else if(node.getStatus() == IndexCursor.DELETED_FLAG &&
                                    node.getLastWriterTransactionId() == transactionId) {
                                frontActive = node;
                                frontTransaction = lastWriterTrans;
                            }
                        } else {
                            frontActive = node;
                            frontTransaction = lastWriterTrans;
                        }
                    }
                }

                  //  group all locks for same transaction together to prevent ordering errors
                  //  if we are updating a crazy id=id+1, our new entry will be legal and should
                  //  end up in front of any waiters that might be waiting on the value
                if(firstWithSameTransaction == null && node.getLastWriterTransactionId() == transactionId) {
                    firstWithSameTransaction = node;
                }

                  //  only update the value in place if we were the last writing transaction
                  //  or if this is a rollback (t == null).  IE, always put it at the back!
                if(node.getRowId() == rowId) {
                      //  NOTE:  this code was inside (see HERE!)  If problems, put back
                    if(node.getStatus() != IndexCursor.DELETED_FLAG) {
                        throw new SQLException("CRITICAL ERROR:  While inserting value " + value + ", row " + rowId + ": found the exact entry already present");
                    }

                    if(node.getLastWriterTransactionId() == transactionId || t == null) {
                          //  HERE!
                        if(writeLogEntry) {
                            t.writeLogEntry(new MemoryIndexLogEntry(getIndexId(), value, rowId, MemoryIndexLogEntry.INSERT,
                                node.getStatus(), node.getLastWriterTransactionId()));
                        }

                          //  we are no longer the last writer here, thus unlock  (even in rollback situation)
                        if(node == frontActive && node.getLastWriterTransactionId() != transactionId) {
                            //System.out.println("unlock " + key);
                            keyLocks.unlock(key);
                        }

                          //  TODO:  we can probably leave the node in the NORMAL state!  (only when transactions are != ?)
                          // since how else did we change the value associated with this row?  (enuf to roll it back?)
                        node.setStatus(IndexCursor.INSERTED_FLAG, transactionId);
                        return;
                    } else {
                          //  since we are writing a new value for this row, then we have the row lock.  Thus,
                          //  remove the node.  It will be put in at the end, just a fine dandy place for it
                        if(prevNode != null) {
                            prevNode.setNext(node.getNext());
                        } else {
                            tree.put(key, node.getNext());
                        }

                        if(node.getNext() != null) {
                            node.getNext().setPrev(prevNode);
                        }

                          //  make node = prevNode because we are deleting this (thus moving cursor virtually back)
                        node = prevNode;
                    }
                }

                prevNode = node;
                node = nextNode;
            }

            if(writeLogEntry) {
                t.writeLogEntry(new MemoryIndexLogEntry(getIndexId(), value, rowId, MemoryIndexLogEntry.INSERT,
                            IndexCursor.NONE_FLAG, 0));
            }

            newNode = new IndexNode(rowId, transactionId);

              //  if none existed
            if(prevNode == null) {
                tree.put(key, newNode);
            } else {
                if(firstWithSameTransaction != null) {
                      //  we go after our first transaction brother.
                      //  TODO: pretty sure this is dangerous in the case of multiple values in the same
                      //  transaction/statement. Should probably be inserted last in our group.
                    if(firstWithSameTransaction.getNext() != null) {
                        firstWithSameTransaction.getNext().setPrev(newNode);
                    }
                    newNode.setNext(firstWithSameTransaction.getNext());
                    newNode.setPrev(firstWithSameTransaction);
                    firstWithSameTransaction.setNext(newNode);
                } else {
                      //  normal case, add to end.  Nulls for newNode.next and
                    prevNode.setNext(newNode);
                    newNode.setPrev(prevNode);
                }
            }

              //  NOTE: Normal use of enqueues would require us to wait on the value anyway
              //  even though the transaction is no longer active.  But unique keys are a special
              //  case because if the value is there, we are probably going to throw a unique key violation
              //  as will all other waiters so it doesn't matter if we race ahead of them with our error.
              //  if front active is null, then there is no need to wait.  If we were the front transaction, there
              //  is also no need to wait because that can only be possible if the front active node is deleted
            if(isUniqueIndex && frontActive != null && t != frontTransaction && t != null) {
                neededToWaitOnFront = true;
                uniqueWait = keyLocks.getLockWait(key, frontTransaction, t);
            }
        }

          //  wait and see if we end up being the legal frontActive transaction
        /*was:

        if(uniqueWait != null) {
            uniqueWait.await();
            if(! weAreFrontActive(key, newNode)) {
                t.incNumUniqueContentions();
            }
        }*/

        if(isUniqueIndex) {
              //  if we need to wait, do so
            if(uniqueWait != null) {
                uniqueWait.await();
            }

              //  see if this is ok...
            if(t != null && neededToWaitOnFront && ! weAreFrontActive(key, newNode)) {
                t.incNumUniqueContentions();
            }
        }

        //Thread.yield();
        /*  debug stuff
        if(t.getNumUniqueContentions() == 0) {
            System.out.print("We leave our object " + value + " wait: " + uniqueWait);
            if(frontActive != null) {
                System.out.print("  front node: " + IndexCursor.statusNames[frontActive.getStatus()]);
            }
            if(frontTransaction != null) {
                System.out.print("  trans: " + Transaction.statusName[frontTransaction.getStatus()]);
            }
            System.out.println("we are front active: "+ weAreFrontActive(key, newNode));
        }
        //*/
    }

    /**
     *  There should be a way to collapse some of this code into the main body above for the common case
     * ie when there is no other waiter.
     *
     * TOO bad, this is repeat code of above
     */
    protected boolean weAreFrontActive(Comparable key, IndexNode us) {
        synchronized(tree) {
            IndexNode node = (IndexNode)tree.get(key);

            while(node != null) {
                if(node == us) {
                    return true;
                }

                if(node.getStatus() == IndexCursor.NORMAL_FLAG) {
                    return false;
                }

                Transaction lastWriterTrans = tm.getTransaction(node.getLastWriterTransactionId());
                if(lastWriterTrans == null || ! lastWriterTrans.isActive()) {  //  whatever the operation, its done!
                    if(node.getStatus() == IndexCursor.INSERTED_FLAG) {  //  if inserted, then violation
                        return false;
                    }
                }

                /*  TODO:  turn this on to see when it gets called.  Put the comment here as to what legit
                   circumstance makes this happen
                System.out.println("Transaction " + lastWriterTrans +
                        " is still active though we got the lock state: " +
                        Transaction.statusName[lastWriterTrans.getStatus()] +
                        "  node:" + IndexCursor.statusNames[node.getStatus()]);
*/
                node = node.next;
            }
        }

        return false;
    }

    /**
     * Only called from rollback.
     * Mark the entry named as deleted.  NOTE:  the log writing etc is done in cursor for delete!!!
     */
    protected void deleteEntry(Object value, long rowId, long transactionId, long prevTransactionId) throws SQLException {
          //  TODO:  get rid of this.  Use the value as itself, nothing more or less.  MultiValue for
          //  multivalues!
        Comparable key = IndexKey.makeKey(value);

        synchronized(tree) {
            IndexNode node = (IndexNode)tree.get(key);
            IndexNode prevNode = null;
            IndexNode frontActive = null;

            while(node != null) {
                if(isUniqueIndex && frontActive == null) {
                    Transaction lastWriterTrans = tm.getTransaction(node.getLastWriterTransactionId());
                    if(lastWriterTrans != null && lastWriterTrans.isActive()) {  //  whatever the operation, its done!
                        frontActive = node;
                    }
                }

                if(node.getRowId() == rowId) {
                    if(node.getStatus() == IndexCursor.DELETED_FLAG || node.getStatus() == IndexCursor.REMOVED_FLAG) {
                        throw new SQLException("CRITICAL ERROR:  While deleting value " + value + ", for row " + rowId + " found the entry already deleted");
                    }

                    /**
                     * If we try to insert a value on top of one we already inserted, we are not the front
                     * active.  If another waiter is already waiting doing an unlock is dangerous!
                     */
                    if(isUniqueIndex && node.getLastWriterTransactionId() != transactionId && frontActive == node) {
                          //  if we are no longer going to be the last writer for this rowID
                        keyLocks.unlock(key);
                    }

                    node.setStatus(IndexCursor.DELETED_FLAG, transactionId);

                    /*  THIS CODE CAUSES OCCASSIONAL PROBLEMS in ParallelInc  */
                      //  if new transId is 0, then this is a rollback of a freshly inserted value... thus
                      //  remove it immediately
                    if(transactionId == 0) {
                        /*if(prevNode == null) {
                            if(node.next != null) {
                                node.next.prev = null;
                            }
                            tree.put(key, node.getNext());
                        } /*else */
                        if(prevNode != null) {

                            //if(node.lockCount == 0) {
                                prevNode.next = node.getNext();
                                if(node.next != null) {
                                    node.next.prev = prevNode;
                                }

                            node.status = IndexCursor.REMOVED_FLAG;
                            //}
                        }
                    }
                    //*/

                    return;
                }

                prevNode = node;
                node = node.getNext();
            }

            throw new SQLException("entry to delete not found: value " + value + ", row " + rowId);
        }
    }

    /**
     * Apply the log entry, forward or reverse
     */
    public void processLogEntry(int applyMode, long lsn, Transaction t, LogEntry logEntry)
            throws SQLException {
        MemoryIndexLogEntry logE = (MemoryIndexLogEntry)logEntry;

        if(applyMode == LogEntryProcessor.REDO_ENTRY) {
            if(logE.getOperation() == MemoryIndexLogEntry.INSERT) {
                insertEntry(logE.getValue(), logE.getRowId(), t.getTransactionId(), null, false);
            } else {
                deleteEntry(logE.getValue(), logE.getRowId(), t.getTransactionId(), logE.getPrevTransaction());
            }
        } else {  //  UNDO section:

            if(logE.getOperation() == MemoryIndexLogEntry.INSERT) {
                deleteEntry(logE.getValue(), logE.getRowId(), logE.getPrevTransaction(), t.getTransactionId());
            } else {
                insertEntry(logE.getValue(), logE.getRowId(), logE.getPrevTransaction(), null, false);
                  //  todo:  probably notify the waiters in the IndexLock.  Since this is undoing a delete or something,
                  //  it should get reverted back to the 'NORMAL' status.  One case breaks this:
                  //  if the row is insertted, then deleted, and only the deleted is undone, then this should
                  //  reset the value back to inserted (not normal).  Thus, the index log entry must contain
                  //  the old value
            }
        }

    }

    /***************************************************************************
     * Transaction callbacks
     */

    /**
     * Do the actual truncation now that the log message has successfully been flushed (and
     * before the UseLock is released)
     */
    class TruncateIndexCallback implements TransactionDoneCallback {
        public void transactionDone(Transaction t, int finalStatus, Object callbackData) {
            if(finalStatus == Transaction.COMMITED) {
                _truncateIndex();
            }
        }
    }
}
