/*
 * Copyright JJ Furman
 * User: user
 * Date: Mar 13, 2002
 * Time: 11:12:25 AM
 */
package freebase.memorydb.log;

import freebase.log.index.IndexLogEntry;
import freebase.util.ObjectIO;
import freebase.util.OhShitException;
import freebase.util.StringUtil;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;


/**
 * Represents a change in a memory index.  This is a very simple record, far simpler
 * than index records to come.<BR><BR>
 *
 * For now this record is only the simplest possible.  Some issues still pending include:
 *   1.  Does the rollback order matter?
 *   2.  What about the new status of the entry... should that be included in this?  probably
 */
public class MemoryIndexLogEntry extends IndexLogEntry {
    final public static byte DELETE = 0;
    final public static byte INSERT = 1;

    /**
     *  value is really the position in the index
     */
    Object value;

    /**
     * The row that was modified
     */
    long rowId;

    /**
     * one of the above
     */
    byte operation;

    /** the way this entry was before we changed it
     *  if the entry was recently inserted, then the status should be NONE and the transaction should be 0
     * NOTE:  not sure what good prevStatus is.  Right now, on rollback the status gets set to
     * INSERTED or DELETED, and based on the prevTransaction will get promoted soon after
     */
    byte prevStatus;
    long prevTransaction;

    /**
     * Called from the system upon recovery
     */
    public MemoryIndexLogEntry() {}

    public MemoryIndexLogEntry(int indexId, Object value, long rowId, byte operation, byte prevStatus,
                               long prevTransaction) {
        super(indexId);
        this.value = value;
        this.rowId = rowId;
        this.operation = operation;

        this.prevStatus = prevStatus;
        this.prevTransaction = prevTransaction;
    }

    public String toString() {
        String valString;
        if(value instanceof Object[]) {
            valString = StringUtil.dump((Object[])value);
        } else {
            valString = value.toString();
        }

        return "MemoryIndexLogEntry: value: " + valString + " rowId: " + rowId + " op: " + operation;
    }

    public Object getValue() {
        return value;
    }

    public long getRowId() {
        return rowId;
    }

    public byte getOperation() {
        return operation;
    }

    public byte getPrevStatus() {
        return prevStatus;
    }

    public long getPrevTransaction() {
        return prevTransaction;
    }
    /************************************************************
     * basic log methods that are required
     */

    /**
     * @return MEMORY_INDEX_LOG_CODE
     */
    public short getCode() {
        return MEMORY_INDEX_LOG_CODE;
    }

    /**
     * TODO:  implement all of these so that the thing can live
     */

    /**
     *  Format of the written data is:
     * 8 for rowid
     * 1 for operation
     * 1 for prevStatus
     * 8 for prevTransaction
     * 1 for size (limit of 255 entries)
     * variable for actual data
     */
    public void readIn(DataInput in) throws IOException {
        super.readIn(in);

        rowId = in.readLong();
        operation = in.readByte();
        prevStatus = in.readByte();
        prevTransaction = in.readLong();

        short size = (short)((int)in.readByte() & 0x00ff);

        if(size == 1) {
            value = ObjectIO.readObject(in);
        } else {
            Object[] val = new Object[size];
            for(int i = 0; i < size; i++) {
                val[i] = ObjectIO.readObject(in);
            }
            value = val;
        }
    }

    public void writeOut(DataOutput out) throws IOException {
        super.writeOut(out);

        out.writeLong(rowId);
        out.writeByte(operation);
        out.writeByte(prevStatus);
        out.writeLong(prevTransaction);

        if(value instanceof Object[]) {
            Object[] val = (Object[])value;

              //  assert
            if(val.length > 255)
                throw new OhShitException("Oh no, length of " + val.length + " too big");

            out.writeByte((byte)val.length);

            for(int i = 0; i < val.length; i++) {
                ObjectIO.writeObject(out, val[i]);
            }
        } else {
            out.writeByte((byte)1);
            ObjectIO.writeObject(out, value);
        }

    }

    public int getSize() {
        if(value instanceof Object[]) {
            int sumObjSize = 0;
            Object[] val = (Object[])value;

            for(int i = 0; i < val.length; i++) {
                sumObjSize += ObjectIO.getSize(val[i]);
            }

            return super.getSize() + 19 + sumObjSize;
        } else {
            return super.getSize() + 19 + ObjectIO.getSize(value);
        }
    }
}
