package freebase.server;

import freebase.api.TupleIterator;
import freebase.sql.SelectStatement;
import freebase.sql.SqlStatement;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Represents a connection to the database.  The actual implementation lives on one
 * of the nodes.  Any node should be able to handle any query (no matter what configuration)
 * but certain nodes will be better than others for certain queries.
 */
public interface DBConnection extends Remote {
    /**
     * The new external method, to replace executeQuery and executeAction.
     * Eventually there will also be a prepared statement version of
     * this.
     * @return either a TupleIterator or a ActionResult, depending
     *      upon what is appropriate
     *
     * TODO:  change this to be a byte[], to remove unnecessary copying
     */
    Object executeStatement(String statement, boolean autoCommit)
            throws RemoteException, SQLException;

    /**
     * The prepared statement version of the above
     */
    Object executeStatement(String statement, ArrayList parameters, boolean autoCommit)
            throws RemoteException, SQLException;

    /**
     * Handles a batch of prepared statements.  All statements are executed atomically.
     */
    BatchResult executeBatch(String statement, ArrayList parameters, boolean autoCommit)
            throws RemoteException, SQLException;

    /**
     * Handles a batch of statements.  All statements are executed atomically.
     */
    BatchResult executeBatch(String[] statements, boolean autoCommit)
            throws RemoteException, SQLException;

    /**
     * executes a query to the DB.
     */
    TupleIterator executeQuery(SelectStatement statement) throws RemoteException, SQLException;

    /**
     * Perform an action in the DB, represented by the SqlStatement.  Simplistic, eventually
     * the query parsing should probably go on in the kernal, or at least the original string
     * should be passed along with the parse tree for quicker lookup.
     */
    ActionResult executeAction(SqlStatement statement) throws RemoteException, SQLException;
    
    //  TODO:  create prepared statement hooks for ?,?,? to be passable all the
    //       way into the db
    
    //  TODO:  needs a way to specify the last update info
    //      so that subsequent read operations can bounce to the secondary (and for failover)
    
    //  needs a way of communicating it's knowledge of the partition maps.  (and the nodes)
    //  don't know how often but should

    /**
     * Closes the connection.  If commit is true, then this tries to commit
     * any transaction still active
     */
    void close(boolean commit) throws RemoteException, SQLException;

    /**
     * Commits the current transation
     */
    void commit() throws RemoteException, SQLException;

    /**
     * rolls back the current transation to savepoint, or all the way if null.
     */
    void rollback(String savepoint) throws RemoteException, SQLException;
}
