package freebase.server;

import freebase.api.TupleIterator;
import freebase.api.ViewInconsistentException;
import freebase.catalog.Catalog;
import freebase.catalog.Table;
import freebase.engine.Transaction;
import freebase.parser.ParseException;
import freebase.parser.SqlParser;
import freebase.sql.*;
import freebase.sql.plan.visitor.SequenceBinder;
import freebase.util.ErrorCode;
import freebase.util.StringInputStream;
import freebase.util.StringUtil;
import org.apache.log4j.Category;

import java.rmi.RemoteException;
import java.rmi.server.Unreferenced;
import java.sql.BatchUpdateException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;

/**
 *  Initial implementation will be with RMI:  it will get done so much faster this way
 *  that there is no real alternative.  now if I could
 *  just convince someone to help me with the app level
 *  networking....<BR><BR>
 *
 * all externally called methods should
 */
public class DBConnectionImpl implements DBConnection, Unreferenced {
    static Category cat = Category.getInstance(DBConnectionImpl.class.getName());

    DatabaseNode localNode;

    /**
     * Transaction active for this connection.  Called Current because it will change.
     */
    Transaction curTrans = null;

    /**
     * the name of the database user who created this connection
     */
    String user;

    /**
     * is this connection coming in off the network
     */
    boolean isRemoteConnection;

    /**
     * The default
     */
    int sessionIsolationLevel = Connection.TRANSACTION_READ_COMMITTED;

    /**
     * connection lost variables.  isBusy is true whenever there is a call active in this connection.
     */
    volatile boolean isBusy = false;

    /**
     * Set to the last return time.  NOTE:  This is probably not being used, remove it
     */
    long lastCallTime = System.currentTimeMillis();

    Object busyLatch = new Object();

    /**
     * Since only one thread will be using this at a time, there should be no concurrency problems.
     * But does each thread see the correct state?
     */
    SqlParser parser = null;

    /**
     * Currently only one thread is allowed to be a client of a connection at a time.
     * This is the pointer to the one that is currently active.
     */
    Object activeThreadLock = new Object();
    int activeCount = 0;
    Thread activeThread = null;
    boolean isClosed = false;  //  checked when the system has the activeThreads lock

    /**
     * Keeps track of all active results being traversed.  When this connection is closed, all of these
     * should be killed.
     */
    HashSet activeResults = new HashSet();

    /**
     * Construct a new DBConnection
     */
    protected DBConnectionImpl(DatabaseNode localNode, String user, boolean remoteConnection) {
        //new Exception("stack trace").printStackTrace();

        this.localNode = localNode;
        this.user = user;
        this.isRemoteConnection = remoteConnection;
    }

    public String getUser() {
        return user;
    }

    public Catalog getCatalog() {
        return localNode.getCatalog();
    }

    /**
     * This will be used by JDBC drivers and whatever else.  Will also need a couple of prepared statement
     * ones.
     */
    public Object executeStatement(String statement, boolean autoCommit) throws RemoteException,
            SQLException {
        return executeStatement(statement, null, autoCommit);
    }

    /**
     * Prepared Statement version.  Currently this does not do any tricks to use that
     * fact that a prepared statement is being used to save time.<BR><RB>
     *
     * Eventually it will be necessary to rework the system to store a representation
     * of the analysed query/update/delete/insert in a common area, and bind the parameters
     * to it (or a duplicate of it)
     */
    public Object executeStatement(String statement, ArrayList parameters,
                                   boolean autoCommit) throws RemoteException, SQLException {
        enterActiveState();
        try{
              //  parse the statement
            SqlStatement st = parseStatement(statement);

              //  bind all variables
            if(parameters != null) {
                Comparable[] paramArray = new Comparable[parameters.size()];
                parameters.toArray(paramArray);

                int highestBound = st.bindLiterals(paramArray, 0);
                if(highestBound < paramArray.length) {
                    throw new SQLException("Too many value in bind set: " + StringUtil.dump(paramArray) + " for statement " + statement);
                }
            }

            Object returnObject = null;
            if(st instanceof SelectStatement) {
                returnObject = executeQuery((SelectStatement)st);
            } else {
                returnObject = executeAction(st);
            }

            //  journal that the statement was made...?

            if(autoCommit) {
                commit();
            }

            return returnObject;
        } finally {
            exitActiveState();
        }
    }

    /**
     * Handles a batch of prepared statements.
     *   All statements are executed in one atomic unit all or nothing
     * @param statement the sql statement
     * @param parameters ArrayList of Comparable[] for each row
     */
    public BatchResult executeBatch(String statement, ArrayList parameters,
                                    boolean autoCommit)
            throws SQLException {
        //  TODO:  ok to reuse the update/delete/insert parse, but will need to reparse selects...

        boolean commandCompleted = false;

        enterActiveState();

        try{
            prepTransaction();

            try{
                  //  parse the statement
                SqlStatement st = parseStatement(statement);
                st.bindSequences(new SequenceBinder(curTrans, getCatalog()));

                BatchResult results = new BatchResult(parameters.size());

                for(int i = 0; i < parameters.size(); i++) {
                    Comparable[] row = (Comparable[])parameters.get(i);
                    int highestBound = st.bindLiterals(row, 0);
                    if(highestBound < row.length) {
                        throw new SQLException("Too many value in bind set: " + StringUtil.dump(row) + " for statement " + statement);
                    }

                    //Object returnObject = null;
                    if(st instanceof SelectStatement) {
                        TupleIterator ti = _executeQuery((SelectStatement)st);
                        results.addQueryResult(ti);
                    } else if(st instanceof DMLStatement) {
                        ActionResult a = _executeDML(st);

                          //  TODO:  detect if the long was chopped down
                        results.addUpdateResult((int)a.getNumRowsEffected());
                    } else {
                        throw new SQLException("Only queries and DML statements are supported:  illegal '" +
                                statement + "'");
                    }
                }

                commandCompleted = true;

                return results;
            } finally {
                endStatement();

                  //  if auto commit is on, then commit the transaction
                if(autoCommit && commandCompleted) {
                    commit();
                }
            }
        } catch(Throwable s) {
            logError(s);

            try{
                if(! curTrans.rollbackStatement()) {
                    curTrans = null;
                }
            }catch(SQLException ohNo) {
                //  TODO:  deal with this double failure in some better way
                ohNo.printStackTrace();
            }

              //  rethrow a Batch failed...
            throw new BatchUpdateException(s.getMessage(), null);
        } finally {
            exitActiveState();
        }

    }

    /**
     * Handles a batch of statements.  All statements are executed atomically.
     */
    public BatchResult executeBatch(String[] statements, boolean autoCommit) throws SQLException {
        boolean commandCompleted = false;

        enterActiveState();

        try{
            prepTransaction();

            try{
                BatchResult results = new BatchResult(statements.length);

                for(int i = 0; i < statements.length; i++) {
                      //  parse the statement
                    SqlStatement st = parseStatement(statements[i]);
                    st.bindSequences(new SequenceBinder(curTrans, getCatalog()));

                    //Object returnObject = null;
                    if(st instanceof SelectStatement) {
                        TupleIterator ti = _executeQuery((SelectStatement)st);
                        results.addQueryResult(ti);
                    } else if(st instanceof DMLStatement) {
                        ActionResult a = _executeDML(st);

                          //  TODO:  detect if the long was chopped down
                        results.addUpdateResult((int)a.getNumRowsEffected());
                    } else {
                        throw new SQLException("Only queries and DML statements are supported:  illegal '" +
                                statements[i] + "'");
                    }
                }

                commandCompleted = true;
                return results;
            } finally {
                endStatement();

                  //  if auto commit is on, then commit the transaction
                if(autoCommit && commandCompleted) {
                    commit();
                }
            }
        } catch(Throwable t) {
            logError(t);

            try{
                if(! curTrans.rollbackStatement()) {
                    curTrans = null;
                }
            }catch(SQLException ohNo) {
                //  TODO:  deal with this double failure in some better way
                ohNo.printStackTrace();
            }

              //  rethrow a Batch failed...
            throw new BatchUpdateException(t.getMessage(), null);
        } finally {
            exitActiveState();
        }
    }

    /**************************************************************************
     *
     */

    public TupleIterator executeQuery(SelectStatement s)
            throws SQLException {
        enterActiveState();
        try{
              //  ensure that there is a transaction.  Of course, if the transaction did not exist,
              //  it is best to close the transaction at the end of this (unless in serializable mode)
            prepTransaction();

            if(cat.isInfoEnabled()) {
                cat.info("Executing Query for " + curTrans + ": " + s);
            }

            s.bindSequences(new SequenceBinder(curTrans, getCatalog()));

            try{
                return _executeQuery(s);
            } finally {
                endStatement();
            }
        } finally {
            exitActiveState();
        }
    }

    /**
     * Executes the query without demarking the statement boundary... called from within
     * executeQuery and executeBatch
     */
    private TupleIterator _executeQuery(SelectStatement s) throws SQLException {
        TupleIterator ti = localNode.getQueryProcessor().executeQuery(curTrans, user, s);
        addActiveResult(ti);

          //  either ti or a proxy will be added to the
        if(isRemoteConnection) {
            ti = new QueryResult(this, ti);
        } else {
            ti = new LocalQueryResult(this, ti);
        }

        return ti;
    }

    /**
     * Process this action.  Return information if appropriate, otherwise
     * null is ok.
     */
    public ActionResult executeAction(SqlStatement statement)
            throws SQLException {
        enterActiveState();

        try{
            if(cat.isInfoEnabled()) {
                cat.info("Executing Action for " + curTrans + ": " + statement);
            }

            if(statement instanceof CommitStatement) {
                commit();
                return null;
            }

            if(statement instanceof RollbackStatement) {
                RollbackStatement rs = (RollbackStatement)statement;
                rollback(rs.getPointToRollbackTo());
                return null;
            }

            if(statement instanceof SavepointStatement) {
                if(curTrans != null) {
                    SavepointStatement ss = (SavepointStatement)statement;
                    curTrans.savepoint(ss.getSavepoint());
                }
                return null;
            }

            if(curTrans != null) {
                curTrans.clearNumUniqueContentions();
            }

            if(statement instanceof DDLStatement) {
                executeDDL(statement);
                return null;
            } else if(statement instanceof DMLStatement) {
                ActionResult result = executeDML(statement);
                return result;
            }

            if(executeSessionStatement(statement)) {
                return null;
            }

            System.out.println("Action not handled: " + statement);
            return null;
        }catch(Throwable t) {
            logError(t);
            if(! curTrans.rollbackStatement()) {
                curTrans = null;
            }

            if(t instanceof SQLException) {
                throw (SQLException)t;
            } else {
                ErrorCode.throwException(ErrorCode.INTERNAL_ERROR, t.getMessage());
            }

              //  never gets here!
            return null;
        } finally {
            exitActiveState();
        }
    }

    /**
     * All DML's have something in common...  They all need to be preped apparently!  This can probaby collapsed above
     */
    private ActionResult executeDML(SqlStatement statement) throws SQLException {
        prepTransaction();

        statement.bindSequences(new SequenceBinder(curTrans, getCatalog()));

        try{
            return _executeDML(statement);
        } finally {
            endStatement();
        }
    }

    /**
     * Executes the Action without demarking the beginning and the end of the statement.  This is
     * called from both the executeDML and from the executeBatch method.
     * TODO:  this class needs to be cleaned out.  To do so the process should be:
     *      1.  For each statement type, figure out where it is executed in the underlying system
     *      2.  Figure out what other things are set and tested before and after the calls.
     */
    private ActionResult _executeDML(SqlStatement statement) throws SQLException {
        curTrans.beginSubStatement();

        while(true) {
            try{
                if(statement instanceof LockTableStatement) {
                    LockTableStatement lts = (LockTableStatement)statement;
                    lts.getTable(0).applyDefaultSchema(user);
                    Table table = localNode.getCatalog().getTable(lts.getTable(0), curTrans.getUser());
                    table.lockTable(curTrans, lts.getMode(), lts.isNoWait());

                      //  required because eventually it will have a log position associated with it
                    return new ActionResult(-1, 0);
                }

                if(statement instanceof InsertStatement) {
                    InsertStatement st = (InsertStatement)statement;

                    Table table = localNode.getCatalog().getTable(st.getTable(), user);

                      //  handle INSERT INTO SELECT statement
                    if(st.getSelectStatement() != null) {
                        TupleIterator queryResult =
                                localNode.getQueryProcessor().executeQuery(curTrans, user, st.getSelectStatement());
                        try{
                            return new ActionResult(table.processInsertIntoSelectStatement(curTrans, st, queryResult), 0);
                        }finally{
                            queryResult.close();
                        }
                    } else {
                        table.processInsertStatement(curTrans, st);
                        return new ActionResult(1, 0);
                    }

                    /*  TODO:  for replication:  fetch the update code from the log (Log Position)
                     *      mark end of statement log pos in transaction
                     */
                }

                if(statement instanceof UpdateStatement) {
                    //  TODO:  validate update
                    long numRows = localNode.getQueryProcessor().executeUpdate(curTrans, user,
                            (UpdateStatement)statement);

                    return new ActionResult(numRows, 0);

                    //  TODO:  finish this garbage
                    //TupleIterator r = localNode.getQueryProcessor().createIterator(t, schema
                            //t, schema, "table", u.getCondition(), true);

                    //  if success, then may be possible to mark t as such
                    //  so it can lump the work into the other savepoint, save epoch
                    //  or era objects
                }

                if(statement instanceof DeleteStatement) {
                    long numRows = localNode.getQueryProcessor().executeDelete(curTrans, user,
                            (DeleteStatement)statement);
                    return new ActionResult(numRows, 0);
                }
                throw new SQLException("STATEMENT " + statement + " was not recognized");
            } catch(ViewInconsistentException v) {
                curTrans.resetSubStatement();
            }
        }
    }

    /**
     * All DDL should commit pending transaction and continue
     */
    private void executeDDL(SqlStatement statement) throws SQLException {
        if(curTrans != null) {
              //  if this throws and exception, should not process the next DDL
            commit();
        }

        prepTransaction();
        try{
            if(statement instanceof CreateTableStatement) {
                CreateTableStatement st = (CreateTableStatement)statement;
                String schema = st.getSchema();
                if(schema == null)
                    st.setSchema(user);
                localNode.getCatalog().createTable(curTrans, st);
            } else if(statement instanceof DropTableStatement) {
                DropTableStatement drop = (DropTableStatement)statement;
                TableRef ref = drop.getTable();
                if(ref.getSchema() == null) {
                    ref.setSchema(user);
                }

                localNode.getCatalog().dropTable(curTrans, ref);
            } else if(statement instanceof CreateIndexStatement) {
                CreateIndexStatement cis = (CreateIndexStatement)statement;
                if(cis.getTable().getSchema() == null) {
                    cis.getTable().setSchema(user);
                }

                localNode.getCatalog().createIndex(curTrans, cis);
            } else if(statement instanceof TruncateStatement)  {
                TruncateStatement st = (TruncateStatement)statement;
                localNode.getCatalog().truncateTable(curTrans, st.getTable());
            } else if(statement instanceof DropIndexStatement) {
                DropIndexStatement st = (DropIndexStatement)statement;
                localNode.getCatalog().dropIndex(curTrans, st.getSchema(user), st.getIndexName());
            } else if(statement instanceof CreateOrAlterSequenceStatement) {
                CreateOrAlterSequenceStatement seq = (CreateOrAlterSequenceStatement)statement;

                if(seq.isCreate()) {
                    localNode.getCatalog().createSequence(curTrans, user, seq);
                } else {
                    localNode.getCatalog().alterSequence(curTrans, user, seq);
                }
            } else if(statement instanceof DropSequenceStatement) {
                DropSequenceStatement drop = (DropSequenceStatement)statement;
                localNode.getCatalog().dropSequence(curTrans, user, drop);
            } else if(statement instanceof AlterTableStatement) {
                AlterTableStatement alter = (AlterTableStatement)statement;
                localNode.getCatalog().alterTable(curTrans, alter);
            } else {
                throw new SQLException("DDL statement " + statement + " not currently implemented");
            }
        } finally {
            endStatement();
        }

          //  now commit all work that has been done
          //  TODO:  probably should abandon the transaction if this fails!
        commit();
    }

    /**
     * Currently for set transaction isolation level only.
     * @param statement
     * @throws SQLException
     */
    private boolean executeSessionStatement(SqlStatement statement) throws SQLException {
        prepTransaction();
        try{
            if(statement instanceof SetTransactionStatement) {
                //set transaction isolation level repeatable read;
                SetTransactionStatement setT = (SetTransactionStatement)statement;
                curTrans.setIsolationLevel(setT.getIsolationLevel());
                return true;
            }
        } finally {
            endStatement();
        }

        return false;
    }

    /**
     * categorize and log the error if appropriate
     * @param t
     */
    private void logError(Throwable t) {
          //  SQLException is common, and not noteworthy
        if(! (t instanceof SQLException)) {
            //TODO:  log the exception better than this!  log4j good
            System.out.print("Major Error: ");
            t.printStackTrace();
            cat.error("Nasty", t);
        }
    }

    /**
     * Commits the current transation
     */
    public void commit() throws SQLException {
        if(curTrans != null) {
            curTrans.commit();
            curTrans = null;
        }
    }

    /**
     * rollsback the current transation.  Null savepoint means all the way
     */
    public void rollback(String savepoint) throws SQLException {
        if(curTrans != null) {
            curTrans.rollback(savepoint);

            //  even if all was rolled back, is it proper to nullify
            //  the transaction?  Perhaps not in Serializable mode

              //  if rollback all, then clear the transaction
            if(savepoint == null) {
                curTrans = null;
            }
        }
    }

    /***************************************************************************
     * Internal and private methods
     */

    /**
     * Uses the parser, then
     */
    private SqlStatement parseStatement(String statement) throws SQLException {
       // SqlStatement st = null;
        try{
            if(parser == null) {
                parser = new SqlParser(new StringInputStream(statement));
            } else {
                parser.ReInit(new StringInputStream(statement));
            }

            return parser.sqlStatement();
        }catch(ParseException pe) {
            throw new SQLException("ParseError: " + pe.getMessage() + " @ " + pe.currentToken.image);
        }
    }

    /**
     * Called before a statement is executed to assure that a transaction is active and
     * that a new statement has been started.  The beginStatement spot anchors a view for
     * this statement
     */
    private void prepTransaction() throws SQLException {
        if(curTrans == null) {
            curTrans = localNode.getTransactionManager().beginTransaction(this, user, sessionIsolationLevel);
        } else {
              //  a new transaction always has a new statement ready to go
            curTrans.beginStatement();
        }
    }

    /**
     * Must be called after a prepTransaction!
     */
    private void endStatement() throws SQLException {
        curTrans.endStatement();
    }


    /***************************************************************************
     * state change management
     */

    /**
     * TODO:  call this from the close methd
     */
    public void addActiveResult(TupleIterator ti) {
        synchronized(activeResults) {
            activeResults.add(ti);
        }
    }

    /**
     * TODO:  call this from the close methd
     */
    public void removeActiveResult(TupleIterator ti) {
        synchronized(activeResults) {
            activeResults.remove(ti);
        }
    }

    /**
     * Sets the current thread to the set of active threads
     * for this transaction.  activeCount appears to be some kind of
     * counter for multiple executions of this in the same thread:  not
     * sure how  that happens right now.  (2 bottles of wine later)
     */
    public void enterActiveState() throws SQLException {
        synchronized(activeThreadLock) {
            Thread currentThread = Thread.currentThread();

            if(activeThread == currentThread) {
                activeCount++;
                return;
            }

            while(activeThread != null) {
                if(isClosed) {
                    throw new SQLException("Connection closed");
                }

                try{
                    activeThreadLock.wait();
                }catch(InterruptedException ie) {
                    System.out.println("CRITICAL ERROR:  got interrupted waiting for another");
                    ie.printStackTrace();
                }
            }

            if(isClosed) {
                throw new SQLException("Connection closed");
            }

            activeCount++;
            activeThread = Thread.currentThread();
        }
    }

    /**
     * Should always be called before entering.
     */
    public void exitActiveState() {
        synchronized(activeThreadLock) {
            activeCount--;

            /**
             * clear the interrupted flag, in case of poor timing.  In a Thread pooling
             * situation there is no limits to what a thread might go off and do... and
             * the interrupt flag should definitely not be left to screw up the next things
             */
            Thread.interrupted();  //  NOTE:  return ignored

            if(activeCount == 0) {
                activeThread = null;

                  //  if something is waiting to do something, let it in
                activeThreadLock.notify();
            }
        }
    }

    /**
     * Called by the connection manager to make an estimation of whether this connection has been lost.
     * Current thinking is to not depend on socket connection losses so that we can be resillient against
     * them.  The current time is passed in to save redundant calls to the System.currentTimeMillis(),
     * although this may be a silly optimization.
     */
    public long getIdleTime(long now) {
        synchronized(busyLatch) {
            if(isBusy)
                return 0;

            return now - lastCallTime;
        }
    }

    /**
     * Currently:
     * 1.  kills the current active thread.
     * 2.  grabs the active lock and sets the state of the connection to closed.  This
     *        prevents other threads from doing anything further
     * 3.  cleanup all open result sets
     * 4.  commit or rollback the current transaction if there is one,
     */
    public void close(boolean commit) throws SQLException {
        synchronized(activeThreadLock) {
              // the following is bad because it causes hanging if
            terminateActiveThread();

            isClosed = true;
        }

        cleanupOpenResults();

          //  rollback any data on the transaction if there is one
        if(curTrans != null) {
            if(commit) {
                curTrans.commit();
            } else {
                curTrans.rollback(null);
            }

            curTrans = null;
        }

        //  now let the connection get garbage collected
    }

    /**
     * If one exists, terminates the active thread
     */
    private void terminateActiveThread() {  // throws SQLException {
          //  interrupt threads that might be active
        synchronized(activeThreadLock) {
            if(isClosed) {
                return;
            }

              //  set this to closed so that future threads that try to enter
            isClosed = true;

              //  interrupt active thread if there is one
            if(activeThread != null) {
                  //  TODO:  ensure that no matter where the thread is waiting (or about to wait)
                  //  this call tells it that its time is up (done!!! dead!!!)
                  //  The only way a SQL thread can be interruped is here
                activeThread.interrupt();
            }

              //  wait for interrupted thread to finish so that we can go to work.  Only
              //  one thread is allowed to modify the transaction at a time
            while(activeThread != null) {
                try{
                    activeThreadLock.wait();
                }catch(InterruptedException ie) {
                    System.out.println("CRITICAL ERROR:  got interrupted waiting for another");
                    ie.printStackTrace();
                }
            }
        }
    }

    /**
     * abandons the connection.  Called from Unreferenced and Finalize!
     */
    private void abandon() throws SQLException {
        terminateActiveThread();

        try{
            cleanupOpenResults();
        }catch(SQLException e) {
            //  TODO:  add to some odd exception list or something
            e.printStackTrace();
        }

          //  rollback any data on the transaction if there is one
        if(curTrans != null) {
            curTrans.rollback(null);
        }
    }

    /**
     *  Closes all result sets that are still open.
     */
    private void cleanupOpenResults() throws SQLException {
        synchronized(activeResults) {
            Iterator it = activeResults.iterator();
            while(it.hasNext()) {
                TupleIterator ti = (TupleIterator)it.next();
                ti.close();
            }

            activeResults.clear();
        }
    }

    /*********************************************************************/
    /****  actual specific actions here, could go in different location  */
    /*********************************************************************/
    protected void finalize() throws Throwable {
        //System.out.println("DBConnectionImpl Finalizer called.  Connection for user '" + user + "'");

        try{
              //  TODO:  eventually this will not be needed because local connections will
              //  not be abandonable!!!  they will be for the server prompt and procedures only
            if(! isRemoteConnection) {
                abandon();
            }
        }catch(SQLException e) {
            e.printStackTrace();
            //  TODO:  journal perhaps in an internal virtual table or critical error log
        }
    }

    /**
     * Called "eventually" (ug) when a client has been dropped by the RMI server.
     * To speed up this process, set the property
     * -Dsun.rmi.dgc.checkInterval=1000 (or whatever)
     */
    public void unreferenced() {
        //System.out.println("DBConnection:  unreferenced @" + (new Date()));

        try{
            if(isRemoteConnection) {
                abandon();
            }
        }catch(SQLException e) {
            e.printStackTrace();
            //  TODO:  journal perhaps in an internal virtual table
        }

        //  now let the connection get garbage collected
    }
}
