package freebase.server;

import freebase.api.LogManager;
import freebase.api.StorageManager;
import freebase.catalog.Catalog;
import freebase.engine.Transaction;
import freebase.engine.TransactionManager;
import freebase.log.LogEntryManager;
import freebase.sql.*;
import freebase.sql.plan.QueryProcessor;
import freebase.util.Props;

import java.io.IOException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Properties;

/**
 * The main class representing this database node.  Eventually this
 * class should not contain any code to initialize specific implementation of.
 *
 * Current model lets each node represent one catalog, with multiple schemas.
 */
public class DatabaseNode {
    final public static String DATA_DIR_PROPERTY = "data_location";
    final public static String LOG_CLASS_PROPERTY = "LOG_CLASS";
    final public static String STORAGE_CLASS_PROPERTY = "STORAGE_CLASS";

    private static DatabaseNode localNode;

    TransactionManager tm = null;
    //CursorManager cursorManager = null;
    //EntityManager entityManager = null;  //  duh, what?  This should be part
                //   of the catalog or something else
    QueryProcessor queryProcessor = null;  //  load from the config file

    //  for initially getting the basics working, this will help
    Catalog catalog = null;

    StorageManager storage;
    LogManager log;

    /**
     * Build a local in memory database.<BR><BR>
     *
     * Eventually use another constructor to accept configuration
     * via a file.
     *
     * current hack.  Should load all this from a file.  Each separately
     * configurable component should have one entry here, and no more.
     */
    public DatabaseNode() throws SQLException, ClassNotFoundException {
        try{
            localNode = this;

            storage = (StorageManager)Props.getClass(STORAGE_CLASS_PROPERTY).newInstance();
            log = (LogManager)Props.getClass(LOG_CLASS_PROPERTY).newInstance();

            log.setLogEntryManager(new LogEntryManager());

            queryProcessor = new QueryProcessor();

              //  TM will recover right away
            tm = new TransactionManager(log, storage);
            catalog = new Catalog(tm, storage, queryProcessor);

            //  TODO:  this should be a question of the catalog... it should know the answer by
            //  knowing if there is the basic TABLES table
            if(! catalog.isInitialized()) {
                  //  TODO:  move this inside beginTransaction
                Transaction t = tm.beginTransaction(null, "SYS", Connection.TRANSACTION_READ_COMMITTED);
                catalog.initializeNew(t);
                t.commit();
            }


            queryProcessor.setCatalog(catalog);

            //catalog.setTransactionManager(tm);

            exportServer();
        }catch(Throwable ioe) {
            ioe.printStackTrace();
            throw new SQLException(ioe.getMessage());
        }
    }

    /**
     *  create the node with the properties file
     *  it contains all of the important information
     */
    public DatabaseNode(Properties p) throws IOException, SQLException {
        String dataDir = p.getProperty(DATA_DIR_PROPERTY);

        //  TODO:  get the other properties here

          //  find da data files
        //entityManager = new EntityManager(dataDir, 8192, 100);

        //cursorManager = new CursorManager(entityManager);

        //  load the log
        //  if we are in replication mode then publish our log entity
        //  and wait for the other guys, unless....

        //  if we are in 2-SAFE-DISK then we may be able to
        //  start independently but it would be nice to get the signal
        //  from the outside.  This is because one can never know when a disk
        //  can go bad.

        //  now recover completely from the new found happy log.

        //  load the local catalog

        //  lookup all peers.  Get their log objects so that we can pass them
        //  into the OldLog for cross reference


        //  create the log from all of the entity objects that
        //  it makes up.
        //
        //  read in the last checkpoint...  here?
        tm = new TransactionManager(null, null);

        //  at some point we need to recover and repair the log
        //

        localNode = this;
    }

    /**
     * Temporary method, eventually either:
     *  1.  each user will log into their (oracle)
     *  2.  each user will be associated with a schema
     * @throws SQLException
     */
    public void createDefaultSchema() throws SQLException {
        if(! getCatalog().schemaExists("JJ")) {
            Transaction t = getTransactionManager().beginTransaction(null, "SYS",
                    Connection.TRANSACTION_READ_COMMITTED);
            getCatalog().createSchema(t, "JJ");
            t.commit();
        }
    }

    /**
     * Create the JJ schema and whatever, this should go away very soon!
     * TODO:  put this into a test file
     */
    public void createTestSchema() throws SQLException {
        createDefaultSchema();
        
        try{
              //  create and insert into test
            CreateTableStatement createTable = new CreateTableStatement();
            createTable.setTableName(new TableRef("NAMES"));
            createTable.addColumn(new ColumnDefinition("ID", Types.INTEGER, 11, 0, true));
            createTable.addColumn(new ColumnDefinition("NAME", Types.VARCHAR, 255, 0, true));

            DBConnection con = DatabaseNode.getLocalDBConnection("JJ", "");
            try{
            con.executeAction(createTable);
            }catch(SQLException s) {
                return;
            }

            InsertStatement
            ins = new InsertStatement("NAMES", new String[] {"" + 1, "'JJ'"});
            con.executeAction(ins);
            ins = new InsertStatement("NAMES", new String[] {"" + 2, "'Pranav'"});
            con.executeAction(ins);
            ins = new InsertStatement("NAMES", new String[] {"" + 3, "'Tito'"});
            con.executeAction(ins);
            ins = new InsertStatement("NAMES", new String[] {"" + 4, "'Smoove B'"});
            con.executeAction(ins);

              //  create and insert into test2
            createTable = new CreateTableStatement();
            createTable.setTableName(new TableRef("NUMS"));
            createTable.addColumn(new ColumnDefinition("A", Types.INTEGER, 11, 0, true));
            createTable.addColumn(new ColumnDefinition("B", Types.INTEGER, 11, 0, true));
            con.executeAction(createTable);

            ins = new InsertStatement("NUMS", new String[] {"1", "1"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"1", "1"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"1", "2"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"null", "7"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"3", "4"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"4", "4"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"5", "5"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"null", "7"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"2", "2"});
            con.executeAction(ins);
            ins = new InsertStatement("NUMS", new String[] {"5", "5"});
            con.executeAction(ins);

            createTable = new CreateTableStatement();
            createTable.setTableName(new TableRef("UNIQUE_NUMS"));
            createTable.addColumn(new ColumnDefinition("A", Types.INTEGER, 11, 0, true));
            createTable.addColumn(new ColumnDefinition("B", Types.INTEGER, 11, 0, true));
            con.executeAction(createTable);

              //  create 2 indexes

            con.executeAction(new CreateIndexStatement("testIndex", "JJ", "NAMES", "ID", true));
            CreateIndexStatement createIndex = new CreateIndexStatement("testIndex2", "JJ", "UNIQUE_NUMS", "A", true);
            createIndex.addColumn("B");
            con.executeAction(createIndex);

              //
            createTable = new CreateTableStatement();
            createTable.setTableName(new TableRef("IND_NUMS"));
            createTable.addColumn(new ColumnDefinition("A", Types.INTEGER, 11, 0, true));
            createTable.addColumn(new ColumnDefinition("B", Types.INTEGER, 11, 0, true));
            createTable.addColumn(new ColumnDefinition("C", Types.VARCHAR, 200, 0, false));
            createTable.addColumn(new ColumnDefinition("D", Types.DATE, 0, 0, true));
            con.executeAction(createTable);

            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"1", "1", "'bob'"}));
            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"1", "2", "'Bob'"}));
            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"1", "4", "'Jim'"}));
            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"4", "6", "'James'"}));
            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"4", "7", "'Fred'"}));
            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"5", "8", "'Fred'"}));
            con.executeAction(new InsertStatement("IND_NUMS", new String[] {"6", "9", "'Fred'"}));

            con.executeAction(new CreateIndexStatement("IND_A", "JJ", "IND_NUMS", "A", false));
            con.executeAction(new CreateIndexStatement("IND_B", "JJ", "IND_NUMS", "B", true));
            con.executeAction(new CreateIndexStatement("IND_C", "JJ", "IND_NUMS", "C", false));


            createTable = new CreateTableStatement();
            createTable.setTableName(new TableRef("IND_NULL"));
            createTable.addColumn(new ColumnDefinition("A", Types.INTEGER, 11, 0, true));
            createTable.addColumn(new ColumnDefinition("B", Types.INTEGER, 11, 0, true));
            con.executeAction(createTable);
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"1", "1"}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"1", "2"}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"1", null}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"4", "6"}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"4", null}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {null, "6"}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"5", null}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"6", "3"}));
            con.executeAction(new InsertStatement("IND_NULL", new String[] {"4", "6"}));
            createIndex = new CreateIndexStatement("NULL_INDEX", "JJ", "IND_NULL", "A", false);
            createIndex.addColumn("B");
            con.executeAction(createIndex);

            con.close(true);
        }catch(RemoteException re) {
            throw new SQLException(re.getMessage());
        }
    }

    /**
     * Returns the current local node, or null if it doesn't exist.
     */
    public static DatabaseNode getLocalNode() {
        return localNode;
    }

    /**
     * Puts this server out there.
     */
    protected void exportServer() throws SQLException {
        try{
              //  register our RMI socket factory
            //RMISocketFactory factory = RMISocketFactory.getDefaultSocketFactory();
            //factory = new DBSocketFactory(factory);
            //RMISocketFactory.setSocketFactory(factory);

            /*
            Enumeration e = System.getProperties().propertyNames();
            while(e.hasMoreElements()) {
                Object key = e.nextElement();
                System.out.println(key + " = " + System.getProperties().get(key.toString()));
            }
            */

            if (System.getSecurityManager() == null) {
                //System.setSecurityManager(new RMISecurityManager());
            }

            try{
                LocateRegistry.createRegistry(1099);
            }catch(RemoteException re) {  }  //  ignore this stupid thing

            try {
                Naming.rebind("//localhost/f3_database", new LoginAgentImpl(this));
            } catch (Exception e) {
                e.printStackTrace();
            }

        }catch(Throwable t) {
            t.printStackTrace();
            throw new SQLException();
        }
    }

    /********
     * linux hacks
     * @return
     */

    /********************************************************************
     * Methods for getting internal objects
     */

    public TransactionManager getTransactionManager() {
        return tm;
    }

    public QueryProcessor getQueryProcessor() {
        return queryProcessor;
    }

    public Catalog getCatalog() {
        return catalog;
    }

    /*******************************************************************
     * Methods for getting connections
     */

    /**
     * Internal getConnection
     */
    public static DBConnection _getConnection(String user, String password,
                            boolean isRemote) throws SQLException {
        //  TODO:  check permissions

        DBConnectionImpl con = new DBConnectionImpl(localNode, user, isRemote);
        if(isRemote) {
            try{
            UnicastRemoteObject.exportObject(con);
            }catch(RemoteException re) {
                re.printStackTrace();
                throw new SQLException("Failed to export connection! " + re.getMessage());
            }
        }
        return con;
    }

    /**
     *  Current thinking is that there will only be one database no
     */
    public static DBConnection getLocalDBConnection(String user, String password)
            throws SQLException {
        if(localNode == null)
            throw new SQLException("no locally running node in this virtual machine!");

        return _getConnection(user, password, false);
    }

    /**
     * Call this to get a remote connection to the database.  If port is -1, then default
     * port is used
     */
    public static DBConnection getRemoteDBConnection(String ip, int port,
                          String db, String user, String password)
            throws SQLException {

        if (System.getSecurityManager() == null) {
            //System.setSecurityManager(new RMISecurityManager());
        }

        try{
            //Registry remoteRegistry = java.rmi.registry.LocateRegistry.getRegistry(ip);
            String name = "//" + ip + "/f3_database"; // + db;
            //LoginAgent loginAgent = (LoginAgent)  remoteRegistry.lookup(name);
            LoginAgent loginAgent = (LoginAgent) Naming.lookup(name);
            DBConnection con = loginAgent.getConnection(user, password);
            return con;
        }catch(Exception e) {
            e.printStackTrace();
            throw new SQLException(e.getMessage());
        }
    }
}
