package freebase.server;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.sql.SQLException;

/**
 * Serialized object that is returned as a query result, includes the first batch of
 * data in the curData spot.<BR><BR>
 *
 * All internal structures use c style arrays meaning the first
 *  entry is 0 and the last is len - 1.  Only at the last step
 *  do the results get shifted by one for the JDBC result set and
 *  all that.
 */
public class QueryResult implements TupleIterator, Serializable {
      //  the cursor object on the server from which to pull results
    DataStreamer dataSource;

    int numColumns;
    Comparable[][] curData;

    int curPos = -1;

    //String[] colNames;

    int[] colTypes;

    ColumnInfo[] columnInfo;

    /**
     * The number of rows to prefetch at a time
     */
    int fetchSize = 400;

    /**
     *  NOTE:  soon the more flag will go away and a cursor object
     *  may take it's place, depending of course if it we go with that or an id
     * TODO:  pass in fetchSize, used
     */
    public QueryResult(DBConnectionImpl con, TupleIterator data) throws SQLException {
        //  prefetch some data
        dataSource = new DataStreamerImpl(con, data);

        try{
            UnicastRemoteObject.exportObject(dataSource);
        }catch(RemoteException re) {
            re.printStackTrace();
            throw new SQLException("Failed to export object : " + re.getMessage());
        }

          //  fetch the columns
          //  TODO:  fetch their types!
        numColumns = data.getNumColumns();

        columnInfo = new ColumnInfo[numColumns];
        for(int i = 0; i < data.getNumColumns(); i++) {
            columnInfo[i] = data.getColumnInfo(i);
        }

        /*
        colNames = new String[data.getNumColumns()];
        for(int i = 0; i < data.getNumColumns(); i++) {
            colNames[i] = data.getColumnName(i);
        }
        */
    }

    /**********************************************************
     * TupleIterator methods
     */
    public boolean next() throws SQLException {
        curPos++;

        if(curData == null || curPos == curData.length) {
            curPos = 0;
            fetchData();
            //fetchDataAsRows();
        }

          //  this may be troublesome if data is kept streaming
        if(curData[curPos] == null)
            return false;

        return true;
    }

    public void setFetchSize(int fetchSize) throws SQLException {
        if(fetchSize <= 0) {
            throw new SQLException("fetch size must be at least 1");
        }

        this.fetchSize = fetchSize;
    }

    public int getFetchSize() {
        return fetchSize;
    }

    /**
     * Fetches more data, the size of the preFetch
     */
    private void fetchData() throws SQLException {
        try{
        curData = dataSource.getData(fetchSize);
        }catch(RemoteException re) {
            re.printStackTrace();
            throw new SQLException("Failed to fetch more data: "  + re.getMessage());
        }
    }

    /**
     * An alternate method, just a experiment for now.  Ug, seems slower
     */
    private void fetchDataAsRows() throws SQLException {
        try{
            RowData[] d = dataSource.getDataAsRows(fetchSize);
            curData = new Comparable[d.length][];
            for(int i = 0; i < d.length; i++) {
                curData[i] = d[i].getData();
            }
        }catch(RemoteException re) {
            re.printStackTrace();
            throw new SQLException("Failed to fetch more data: "  + re.getMessage());
        }
    }

    public void close() throws SQLException {

        try{
        dataSource.close();
        }catch(RemoteException re) {
            re.printStackTrace();
            throw new SQLException("Failed to close: "  + re.getMessage());
        }
    }

    public Comparable getValue(int i) throws SQLException {
        return curData[curPos][i];
        //return new SqlValue(-666, (Comparable)curData[curPos][i]);
        //return null;
    }

    public void setValue(int i, Comparable value) throws SQLException {
        throw new SQLException("SETTING DATA NOT SUPPORTED BY THIS now");
    }

    public int getNumColumns() throws SQLException {
        return columnInfo.length;
    }

    public String getColumnName(int i) throws SQLException {
        return columnInfo[i].getColumnName();
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return columnInfo[i];
    }

    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException {
        throw new SQLException("DELETING DATA NOT SUPPORTED BY THIS now");
    }

    public void reset() throws SQLException {
        throw new UnsupportedOperationException("Reset not currently supported from the client.  Easy to fix though");
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        try{
        return dataSource.getExecutionPlan();
        }catch(RemoteException re) {
            throw new SQLException("Network exception: " + re.getMessage());
        }
    }
}