/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 26, 2002
 * Time: 1:13:12 PM
 */
package freebase.server;

import freebase.api.TupleIterator;
import freebase.util.Bool;
import freebase.util.ObjectIOTypes;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.SQLException;
import java.sql.Timestamp;

/**
 * Smart class knows how to take data directly from the iterator and write it out
 * the door
 */
public class RowData implements Externalizable, ObjectIOTypes  {
    transient TupleIterator source;

    transient Comparable[] row;

    public RowData() {}

    public RowData(TupleIterator it) {
        source = it;
    }

    public Comparable[] getData() {
        return row;
    }

    public void writeExternal(ObjectOutput out) throws IOException {
        try{
        boolean hasMore = source.next();
        if(! hasMore) {
            out.writeInt(-1);
            return;
        }

        out.writeInt(source.getNumColumns());

        for(int i = 0; i < source.getNumColumns(); i++) {
            Object o = source.getValue(i);

            if(o == null) {
                out.writeByte(NULL);
            } else if(o instanceof Bool) {
                out.writeByte(BOOLEAN);
                out.writeBoolean(((Bool)o).getValue());
            } else if(o instanceof Byte) {
                out.writeByte(BYTE);
                out.writeByte(((Byte)o).byteValue());
            } else if(o instanceof Short) {
                out.writeByte(SHORT);
                out.writeShort(((Short)o).shortValue());
            } else if(o instanceof Integer) {
                out.writeByte(INT);
                out.writeInt(((Integer)o).intValue());
            } else if(o instanceof Long) {
                out.writeByte(LONG);
                out.writeLong(((Long)o).longValue());
            } else if(o instanceof Float) {
                out.writeByte(FLOAT);
                out.writeFloat(((Float)o).floatValue());
            } else if(o instanceof Double) {
                out.writeByte(DOUBLE);
                out.writeDouble(((Double)o).doubleValue());
            } else if(o instanceof String) {
                out.writeByte(STRING);
                out.writeInt(((String)o).length());
                out.writeBytes((String)o);
            } else if(o instanceof java.sql.Date) {
                out.writeByte(SQLDATE);
                out.writeLong(((java.sql.Date)o).getTime());
            } else if(o instanceof java.sql.Time) {
                out.writeByte(SQLTIME);
                out.writeLong(((java.sql.Time)o).getTime());
            } else if(o instanceof java.sql.Timestamp) {
                out.writeByte(SQLTIMESTAMP);
                Timestamp ts = (Timestamp)o;
                out.writeLong(ts.getTime());
                  //  always thinking about the future of Nanotech
                out.writeInt(ts.getNanos());
            } else if(o instanceof byte[]) {
                out.writeByte(BYTES);
                out.writeInt(((byte[])o).length);
                out.write((byte[])o);
            } else if(o instanceof BigInteger) {
                out.writeByte(BIG_INTEGER);
                writeBigInteger(out, (BigInteger)o);
            } else if(o instanceof BigDecimal) {
                out.writeByte(BIG_DECIMAL);
                BigDecimal d = (BigDecimal)o;
                out.writeByte((byte)d.scale());
                writeBigInteger(out, d.unscaledValue());
            } else {
                throw new RuntimeException("CRITICAL ERROR: don't support writing : " + o.getClass());
            }
        }
        }catch(SQLException e) {
            throw new IOException("SQL Exception: " + e.getMessage());
        }
    }

    /**
     *
     */
    private static BigInteger readBigInteger(ObjectInput in) throws IOException {
        int length = in.readByte() & 0x000000ff;
        byte[] buf = new byte[length];
        in.readFully(buf);
        return new BigInteger(buf);
    }



    /**
     *
     */
    private static void writeBigInteger(ObjectOutput out, BigInteger i) throws IOException {
        out.writeByte(getBigIntLength(i));
        out.write(i.toByteArray());
    }

    /**
     * Returns the number of bytes required to store this big integer.
     */
    private static byte getBigIntLength(BigInteger i) {
        return (byte)(i.bitLength() / 8 + 1);
    }

    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        int size = in.readInt();
        if(size == -1) {
            row = null;
            return;
        }

        row = new Comparable[size];

        for(int i = 0; i < size; i++) {
            byte type = in.readByte();

            switch(type) {
                case NULL:     row[i] = null;  continue;
                case BOOLEAN:  row[i] = Bool.getBool(in.readBoolean());  continue;
                case BYTE:     row[i] = new Byte(in.readByte());     continue;
                case SHORT:    row[i] = new Short(in.readShort());   continue;
                case INT:      row[i] = new Integer(in.readInt());   continue;
                case LONG:     row[i] = new Long(in.readLong());     continue;
                case FLOAT:    row[i] = new Float(in.readFloat());   continue;
                case DOUBLE:   row[i] = new Double(in.readDouble()); continue;
                case STRING:   int len = in.readInt();
                               byte[] data = new byte[len];
                               in.readFully(data);
                               row[i] = new String(data);    continue;
                case SQLDATE:  row[i] = new java.sql.Date(in.readLong());  continue;
                case SQLTIME:  row[i] = new java.sql.Time(in.readLong());  continue;
                case SQLTIMESTAMP:
                    Timestamp ts = new Timestamp(in.readLong());
                    ts.setNanos(in.readInt());
                    row[i] = ts;   continue;

                case BYTES:
                    throw new RuntimeException("DON'T SUPPORT BYTES BECAUSE OF Comparable problem");
                    /*
                    len = in.readInt();
                    data = new byte[len];
                    in.readFully(data);
                    row[i] = data;
                    */
                case BIG_INTEGER:
                    row[i] = readBigInteger(in);    continue;

                case BIG_DECIMAL:
                    int scale = in.readByte() & 0x000000ff;
                    BigInteger unscaledValue = readBigInteger(in);
                    row[i] = new BigDecimal(unscaledValue, scale);   continue;

                default:
                        throw new RuntimeException("CRITICAL ERROR: don't support reading type : " + type);
            }
        }
    }
}
