package freebase.sql;


import freebase.sql.expression.ExpressionVisitor;
import freebase.util.Bool;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Represents a Query.
 */
public class SelectStatement extends SqlStatement {
    final public static int TOP_MODE_ALL = 0;
    final public static int TOP_MODE_NUM = 1;
    final public static int TOP_MODE_PERCENT = 2;

    boolean distinct = false;
    boolean star = false;

    /**
     * One entry for everything in the select clause.  Each entry is either
     * an SqlExpression or a String.  The string is for {TABLE}.* pattern
     */
    ArrayList selectList = new ArrayList();

    /**
     * One to one with the select List, null entries
     * for everything that is either not definined or doesn't make sense
     * such as "select test.* as b ..." is meaningless
     */
    ArrayList columnAliases = new ArrayList();

    ArrayList tables = new ArrayList();
    ArrayList tableAlias = new ArrayList();

    /**
     * Where clause condition:
     */
    SqlCondition condition = null;

    /**
     * Group stuffs
     */
    ArrayList groupByExpressions = null;
    SqlCondition havingCondition = null;

    ArrayList orderByExpressions = null;
    ArrayList orderByDirections = null;  //  true is ascending, false descending


    int topMode = TOP_MODE_ALL;
    BigDecimal topPercent;
    long topNum;
    boolean withTies = false;

    public SelectStatement() { }

    /**
     * Binds the literals in order that they appear in the select statement...
     * select A from ...
     */
    public int bindLiterals(Comparable[] values, int curPos) throws SQLException {
          //  bind select items
        for(int i = 0; i < getNumSelectItems(); i++) {
            curPos = getSelectItem(i).bindLiterals(values, curPos);
        }

        //  When there are subqueries in the from clause, bind them too!

          //  bind where clause
        if(getCondition() != null) {
            curPos = getCondition().bindLiterals(values, curPos);
        }

          //  bind group by clause
        for(int i = 0; i < getNumGroupByExprs(); i++) {
            curPos = getGroupByExpr(i).bindLiterals(values, curPos);
        }

          //  bind having clause
        if(getHavingCondition() != null) {
            curPos = getHavingCondition().bindLiterals(values, curPos);
        }

          //  bind order by clause
        for(int i = 0; i < getNumOrderByClauses(); i++) {
            curPos = getOrderByExpression(i).bindLiterals(values, curPos);
        }

        return curPos;
    }

    /**
     * bind sequences for all select items!
     * @throws SQLException
     */
    public void bindSequences(ExpressionVisitor seqBinder) throws SQLException {
          //  bind select items
        for(int i = 0; i < getNumSelectItems(); i++) {
            getSelectItem(i).accept(seqBinder);
        }
    }

    public void setWithTies(boolean wt) {
        withTies = wt;
    }

    public boolean getWithTies() {
        return withTies;
    }

    /**
     * Constructs a select * from {table}
     */
    public SelectStatement(String table) {
        addTable(new TableRef(table), null);
    }

    public void setDistinct(boolean val) {
        distinct = val;
    }

    /**
     * Probably an unnecessary shortcut to make "select *" faster
     */
    public void setStar(boolean val) {
        star = val;
    }

    /**
     * Either a ColumnRef or an Expression.
     *
     *  TODO:  currently: select a.* pattern does not work.  TODO so, the a.* stuff will
     * need to get expanded at some point either in place before the query gets analyzed or
     * in the external maps that get derived from this object.
     */
    public void addAllTableColumns(String tableName) {
        selectList.add(tableName);
        columnAliases.add(null);
    }

    public void addExpressionSelectItem(SqlExpression x, String alias) {
        selectList.add(x);
        columnAliases.add(alias);
    }

    /**
     * TODO:  will need to rework this, based on LEFT OUTER JOIN
     * sematics of sql 92
     */
    public void addTable(TableRef table, String alias) {
        tables.add(table);
        tableAlias.add(alias);
    }

    public void setCondition(SqlCondition condition) {
        this.condition = condition;
    }

    public void setGroupByItems(ArrayList groupExpressions) {
        this.groupByExpressions = groupExpressions;

    }

    public void setHavingCondition(SqlCondition havingClause) {
        this.havingCondition = havingClause;
    }

    public void addOrderByClause(SqlExpression x, boolean ascending) {
        if(orderByExpressions == null) {
            orderByExpressions = new ArrayList();
            orderByDirections = new ArrayList();
        }

        orderByExpressions.add(x);
        orderByDirections.add(Bool.getBool(ascending));
    }

    /************************************************************
     * Getters
     */

    /**
     * True if the select results should have duplicates removed
     */
    public boolean isDistinct() {
        return distinct;
    }

    /**
     * True if "select *" is all there is
     */
    public boolean isStar() {
        return star;
    }

    /**
     * Return the condition, or null if there are none
     */
    public SqlCondition getCondition() {
        return condition;
    }

    /**
     * TODO:  eventually go to a
     */
    public ArrayList getTables() {
        return tables;
    }

    /**
     * Returns the number of select items, or zero if * was specified
     */
    public int getNumSelectItems() {
        if(selectList == null)
            return 0;

        return selectList.size();
    }

    public ArrayList getSelectList() {
        return selectList;
    }

    public SqlExpression getSelectItem(int i) {
        return (SqlExpression)selectList.get(i);
    }

    public ArrayList getSelectItemAliases() {
        return columnAliases;
    }

    public int getNumGroupByExprs() {
        if(groupByExpressions == null) {
            return 0;
        } else {
            return groupByExpressions.size();
        }
    }

    public SqlExpression getGroupByExpr(int i) {
        return (SqlExpression)groupByExpressions.get(i);
    }

    public ArrayList getGroupByExpressions() {
        return groupByExpressions;
    }

    public SqlCondition getHavingCondition() {
        return havingCondition;
    }

    public String getTableAlias(int tableOffset) {
        return (String)tableAlias.get(tableOffset);
    }

    public int getNumOrderByClauses() {
        if(orderByExpressions == null) {
            return 0;
        } else {
            return orderByExpressions.size();
        }
    }

    public SqlExpression getOrderByExpression(int i) {
        return (SqlExpression)orderByExpressions.get(i);
    }

    /**
     * True is ascending, false is descending
     */
    public boolean getOrderByDirection(int i) {
        return ((Bool)orderByDirections.get(i)).getValue();
    }

    /**
     * Called during Query Analysis and planning
     * @return TOP_MODE_ALL, TOP_MODE_NUM, or TOP_MODE_PERCENT
     */
    public int getTopMode() {
        return topMode;
    }

    public long getTopNum() {
        //assert topMode == TOP_MODE_NUM;

        return topNum;
    }

    public BigDecimal getTopPercent() {
        //assert topMode == TOP_MODE_PERCENT;

        return topPercent;
    }

    /**
     *  set TOP parameters.  Current TOP # is limited to a long value.  If you want more that the top
     * 8,000,000,000,000,000,000 (+,-) then you are S.O.L.
     */
    public void setTopInfo(int mode, String quantity) {
        topMode = mode;

        if(mode == TOP_MODE_NUM) {
            topNum = Long.parseLong(quantity);
        } else {
            topPercent = new BigDecimal(quantity);
        }
    }
}