/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Nov 18, 2001
 * Time: 11:06:53 AM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.sql;

import freebase.api.ColumnInfo;
import freebase.api.Tuple;
import freebase.sql.expression.ExpressionVisitor;
import freebase.sql.expression.GroupByItem;
import freebase.sql.plan.QueryAnalysis;

import java.io.Serializable;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * An expression that either evaluates to a truth value or a real value.<BR><BR>
 *
 * TODO:  would like to see a phase where constants get resolved, for the where and having
 *    clause.  This would be so that the sysdate and others could be assigned all at once,
 *    not per call.  This could be done a number of ways, perhaps with a virtual tuple, or
 *    with the buildColumnList pass.
 *
 * At the time of resolution it could be determined if a particular refernce is legal or not
 * (such as seq.nextval in a where clause)
 *
 * The process should be called binding
 * Another pass might be nice to detect which subexpressions are associated with only
 * a single table
 */
public abstract class SqlExpression implements Serializable {
    /**
     * eventually this will have modes etc
     */
    abstract public void writeToBuffer(StringBuffer sb);

    /**
     * Evaluate the expression, possibly based on tuple values.
     * Later the values will come from a QueryContext or something
     * that represents all of the values of variables etc
     */
    abstract public Comparable evaluate(Tuple t, int[] offsetMap) throws SQLException;

    /**
     * Called when a column info is needed from an expression.  Should do validation.
     * In the case of conditions, validation is essentially all it does
     */
    abstract public ColumnInfo getColumnInfo(Tuple t, int[] offsetMap) throws SQLException;

    /**
     * Allows this visitor to visit us and then all our children.  Will only visit children
     * if the visit to use returns true.  <BR>In theory, most of the other methods
     * below could be rewritten to use this pattern; not a bad idea really.<BR>
     *
     * Children are guarenteed to be visited in left to right order for easing implementation
     * of prepared statements (for example)
     * @param v
     */
    abstract public void accept(ExpressionVisitor v) throws SQLException;

    /**
     *
     *
     * TODO:  since all of the columns have been extracted already
     *   perhaps this could simply be part of the actual bind pass.
     *   The basic theory there would be that each condition fragment
     *   can only be exist at one location in the query at a time
     * TODO:  perfect easy candidate for replacement with VisitorPattern
     *    Change ArrayList to ColumnList, and rewrite this function
     *     to not call the children.  Then remove the children when necessary
     */
    //final public void buildColumnLists(ArrayList columnList) {}

    /**
     * Called on all prepared statement fragments immediately before execution.
     * The first call is made with zero.  Returned in the value to pass in the next.
     * TODO:  if the bindable expressions have already been extracted, is this needed?
     *
     */
    abstract public int bindLiterals(Comparable[] values, int curPos) throws SQLException;

    /**
     * Called at the beginning of a query, this does things like extract
     * equi join clauses, single table conditions etc and separates them
     * so that they can be used most effectively.  The results are accumulated
     * in the QueryAnalysis.<BR><BR>
     *
     * Upon return columns contains a list of single columns referenced in this
     * query EXCEPT columns that are referenced inside of an aggregate function.
     * Aggregate functions are placed in the variable called groupFunctions.  If
     * groupFunctions is null, then the analysis assumes that they are illegal in
     * the context and thus throws an exception. (ie, updates & deletes?)<BR><BR>
     *
     * TODO:  possibly combine this and performHavingAnalysis
     */
    abstract public void performAnalysis(QueryAnalysis qa, boolean seekingEquiJoins,
                                         ArrayList columns, ArrayList groupFunctions) throws SQLException;

    /**
     * Recursively traverses the entire expression tree and replaces high level expression bodies that
     * represent group by clause values and replaces them with objects of type GroupByReference.
     * This is called on the Having condition.  <BR><BR>
     *
     * The implementations of this are safe to use ArrayList.indexOf(x) on each sub expression
     * because ArrayList uses only the equals method to determine equality, not hashCode.
     * This is important because we currently do not implement hashCode()<BR><BR>
     *
     * groupByClauses is IN ONLY
     * groupFunctions is built
     */
    abstract public void performHavingAnalysis(QueryAnalysis qa, ArrayList groupFunctions,
                                               ArrayList groupByClauses) throws SQLException;

    /**
     * Recursively finds all columns' tableOffsets.  Uses QueryAnalysis TODO so.
     * There is a slight overlap in functionality here, in that performAnalysis already
     * does this.  Maybe someday we can separate the two passes, but not a huge deal for
     * now.  This is called by findGroupByItem below...
     */
    abstract public void verifyColumns(QueryAnalysis qa) throws SQLException;

    /**
     * Helper function for above performHavingAnalysis,
     * If appropriate, returns the group by item represented by this expression.  Otherwise, it
     * calls performHavingAnalysis.  Using this method drastically reduces the complexity of
     * the analysis code.
     */
    protected SqlExpression findGroupByItem(SqlExpression x, QueryAnalysis qa,
                                            ArrayList groupFunctions, ArrayList groupByClauses) throws SQLException {
        int pos = -1;
        x.verifyColumns(qa);

        if(groupByClauses != null) {
            //System.out.println("GROUP CLAUSES: " + groupByClauses + "  X: " +x + " equal: " + x.getClass() + " vs " + groupByClauses.get(0).getClass());
            pos = groupByClauses.indexOf(x);
        }

        if(pos != -1) {
            // System.out.println("Found group by expression: " + x);
            return new GroupByItem(pos);
        } else {
            x.performHavingAnalysis(qa, groupFunctions, groupByClauses);
            return x;
        }
    }

    /**
     * For performance, uses the writeToBuffer routine
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        writeToBuffer(sb);
        return sb.toString();
    }
}
