/*
 * Created by IntelliJ IDEA.
 * User: administrator
 * Date: Nov 19, 2001
 * Time: 5:55:01 PM
 * To change template for new class use 
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.sql.expression;

import freebase.api.Tuple;
import freebase.catalog.ColumnStatisticProvider;
import freebase.sql.ColumnRef;
import freebase.sql.SqlCondition;
import freebase.sql.SqlExpression;
import freebase.sql.plan.QueryAnalysis;
import freebase.util.Helper;
import freebase.util.OhShitException;
import freebase.util.SqlTypes;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Support all basic conditions.
 */
public class BasicCondition extends SqlCondition {
    final public static int EQUALS = 0;             // "="
    final public static int NOT_EQUALS = 1;         // "!=" & "<>" & "^="
    final public static int GREATER_THAN = 2;       // ">"
    final public static int GREATER_THAN_OR_EQUALS = 3;  // ">="
    final public static int LESS_THAN = 4;          // "<"
    final public static int LESS_THAN_OR_EQUALS = 5;     // "<="

    //  these are loners, who don't consider the right portion of the query
    final public static int IS_NULL = 6;
    final public static int IS_NOT_NULL = 7;

      //  TODO:  what is the "#" comparator?

    final public String[] opSymbols = new String[] {"=", "!=", ">", ">=", "<", "<=", "IS NULL"};


    SqlExpression left;
    SqlExpression right;
    int operator = -1;

    //  list operators here

    public BasicCondition() { }

    /******************************************
     *  init time methods
     */
    public void setLeftExpr(SqlExpression left) {
        this.left = left;
    }

    public void setOperator(int op) {
        operator = op;
    }

    public void setRightExpr(SqlExpression right) {
        this.right = right;
    }

    /*****************************************************
     *
     */
    public void negate() {
        switch(operator) {
            case EQUALS:                  operator = NOT_EQUALS;             break;
            case NOT_EQUALS:              operator = EQUALS;                 break;
            case GREATER_THAN:            operator = LESS_THAN_OR_EQUALS;    break;
            case GREATER_THAN_OR_EQUALS:  operator = LESS_THAN;              break;
            case LESS_THAN:               operator = GREATER_THAN_OR_EQUALS; break;
            case LESS_THAN_OR_EQUALS:     operator = LESS_THAN;              break;
            case IS_NULL:                 operator = IS_NOT_NULL;            break;
            case IS_NOT_NULL:             operator = IS_NULL;                break;
            default:
                throw new OhShitException("Operator " + operator + " does not have an entry in the BasicCondition.negate() method");
        }
    }

    /**
     * The fundamental truth condition.
     * @param t
     * @param offsetMap used to lookup the 
     * @return
     * @throws SQLException
     */
    public boolean isTrue(Tuple t, int[] offsetMap) throws SQLException {
        Comparable leftVal = left.evaluate(t, offsetMap);

        if(operator == IS_NULL) {
            return leftVal == null;
        }

        if(operator == IS_NOT_NULL) {
            return leftVal != null;
        }

        Comparable rightVal = right.evaluate(t, offsetMap);

        if(leftVal == null || rightVal == null) {
            return false;
        }

        //System.out.println("LEFT: " + leftVal.getClass() + "  " + leftVal + " " + left);
        //System.out.println("RIGHT: " + rightVal.getClass() + "  " + rightVal + " " + right);

        int comp = SqlTypes.compare(leftVal, rightVal);
        //int comp = leftVal.compareTo(rightVal);
        //System.out.println("COMPARISON IS : " + comp);

        switch(operator) {
            case EQUALS:                 return comp == 0;
            case NOT_EQUALS:             return ! (comp == 0);
            case GREATER_THAN:           return comp > 0;
            case GREATER_THAN_OR_EQUALS: return comp >= 0;
            case LESS_THAN:              return comp < 0;
            case LESS_THAN_OR_EQUALS:    return comp <= 0;
            default:
                    throw new SQLException("We don't support operator " + operator);
        }
    }

    /**
     * Return the selectivity factor of this BasicCondition.  Uses only very simple
     * determination methods
     */
    public double getSelectivity(ColumnStatisticProvider[] tables) throws SQLException {
        switch(operator) {
            case EQUALS:
                ColumnRef leftCol = null;
                ColumnRef rightCol = null;
                double numLeftValues = 0;
                double numRightValues = 0;
                if(left instanceof ColumnExpression) {
                    leftCol = ((ColumnExpression)left).getColumn();
                    //System.out.println("COLUMN: " +leftCol);
                    numLeftValues = tables[leftCol.getTableOffset()].
                            getColumnStatistics(leftCol.getColumn()).getEstimatedNumDistinctValues();
                }

                if(right instanceof ColumnExpression) {
                    rightCol = ((ColumnExpression)right).getColumn();
                    numRightValues = tables[rightCol.getTableOffset()].
                            getColumnStatistics(rightCol.getColumn()).getEstimatedNumDistinctValues();
                }

                //  if it is an always false constant condition, then perhaps it
                //  should return very high selectivity  ;)  For now, this
                //  will do
                if(leftCol == null && rightCol == null) {
                    return 1;  //  total guess
                }

                  //  equality within the same table... hmmm
                if(leftCol != null && rightCol != null) {
                    return Math.max(numLeftValues, numRightValues);
                }

                if(leftCol != null) {
                    return  numLeftValues;
                }

                return numRightValues;

            case NOT_EQUALS:  return 1;    //  Page 369
            case GREATER_THAN:
            case GREATER_THAN_OR_EQUALS:
            case LESS_THAN:
            case LESS_THAN_OR_EQUALS: return 3;  //  Page 369
            case IS_NULL:   return 1.25;    //  WHAT TODO ABOUT THESE?  Don't want too
            case IS_NOT_NULL:  return 1.25;  //  much selectivity for these
            default:
                throw new OhShitException("Operator " + operator + " does not have an entry in the BasicCondition.getSelectivity() method");
        }
    }

    /**
     * Visitor Pattern.  Visit us, and if true visit children.
     * @param v
     */
    public void accept(ExpressionVisitor v) throws SQLException {
        if(v.forCond(this)) {
            left.accept(v);
            if(right != null) {
                right.accept(v);
            }
        }
    }

    /**
     * If this is an equijoin clause, then cool.  Otherwise,
     *
     * NOTE:  given the way this was written it would not be impossible to
     *    include the ability to use smart joins for conditions like a.id + b.id=0
     *    or whatever.  Of course these will be incredibly rare, but just a thought
     *
     * @param seekingEquiJoins same as appropriateSubConditions
     */
    public void performAnalysis(QueryAnalysis qa, boolean seekingEquiJoins,
                                ArrayList columns, ArrayList groupFunctions) throws SQLException {
        int curPos = columns.size();

        left.performAnalysis(qa, seekingEquiJoins, columns, groupFunctions);
        if(right != null)
            right.performAnalysis(qa, seekingEquiJoins, columns, groupFunctions);

        if(seekingEquiJoins) {
            if((columns.size() - curPos) == 2 && operator == EQUALS &&
                    (left instanceof ColumnExpression) &&
                    (right instanceof ColumnExpression)) {
                ColumnRef col1 = (ColumnRef)columns.get(curPos);
                ColumnRef col2  = (ColumnRef)columns.get(curPos + 1);
                int tableOffset1 = col1.getTableOffset();
                int tableOffset2 = col2.getTableOffset();

                if(tableOffset1 != tableOffset2) {
                      //  reorder so that the lower id is always to the left
                    if(tableOffset1 > tableOffset2) {
                        SqlExpression x = left;
                        left = right;
                        right = x;
                    }

                      //  order of col1 and col2 doesn't matter
                    qa.addEquiJoinCondition(this, col1, col2);
                } else /* oops, duh */ {
                    qa.addSingleTableCondition(tableOffset1, this);
                }

                return;
            }

            analyzeColumnList(qa, columns, curPos);
        }
    }

    public void verifyColumns(QueryAnalysis qa) throws SQLException {
        left.verifyColumns(qa);
        if(right != null) {
            right.verifyColumns(qa);
        }
    }

    public void performHavingAnalysis(QueryAnalysis qa, ArrayList groupFunctions,
                                      ArrayList groupByClauses) throws SQLException {
        left = findGroupByItem(left, qa, groupFunctions, groupByClauses);
        if(right != null)
            right = findGroupByItem(right, qa, groupFunctions, groupByClauses);
    }

    /*
    public void buildColumnList(ArrayList columnList) {
        left.buildColumnList(columnList);
        if(right != null)
            right.buildColumnList(columnList);
    }*/

    public int bindLiterals(Comparable[] values, int curPos) throws SQLException {
        curPos = left.bindLiterals(values, curPos);
        if(right != null)
            curPos = right.bindLiterals(values, curPos);
        return curPos;
    }

    public void writeToBuffer(StringBuffer sb) {
        left.writeToBuffer(sb);
        sb.append(opSymbols[operator]);
        if(right != null)
            right.writeToBuffer(sb);
    }

    /************************************************************
     * Getters
     */

    /**
     * Get the left expression
     */
    public SqlExpression getLeftExpression() {
        return left;
    }

    public SqlExpression getRightExpression() {
        return right;
    }

    public int getOperator() {
        return operator;
    }


    public boolean equals(Object obj) {
        if(! (obj instanceof BasicCondition)) {
            return false;
        }

        BasicCondition other = (BasicCondition)obj;
        if(operator != other.operator) {
            return false;
        }

        if(! left.equals(other.left)) {
            return false;
        }

          //  null values ok
        if(! Helper.areEqual(right, other.right)) {
            return false;
        }

        return true;
    }
}
