/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 11, 2002
 * Time: 6:32:59 PM
 */
package freebase.sql.expression;

import freebase.api.Tuple;
import freebase.catalog.ColumnStatisticProvider;
import freebase.sql.SqlCondition;
import freebase.sql.SqlExpression;
import freebase.sql.plan.QueryAnalysis;

import java.sql.SQLException;
import java.util.ArrayList;


/**
 *  Since I'm not sure about:<BR><BR>
 *
 * 1.  if it will be legal to have the exact same expression object hooked into two
 *     places in the tree
 * 2.  if there will be something that can detect similarity between different parts of
 *     the expression<BR><BR>
 *
 * Someday I may change this to a simple ( left >= min) AND (left <= max
 * Also, not is currently built in to directly encapsulate the value of this expression.  I
 * realize this is redundant.
 *
 * TODO:  implement .equals()
 */
public class BetweenCondition extends SqlCondition {
    SqlExpression val;   //  val BETWEEN min AND max
    SqlExpression min;
    SqlExpression max;

    /**
     * Is this condition negated?  TODO:  can't this be handled somehow generically?
     */
    boolean not;

    /*********************************************
     * parse time methods
     */
    public void setInfo(SqlExpression leftValue, SqlExpression min, SqlExpression max, boolean not) {
        this.val = leftValue;
        this.min = min;
        this.max = max;

        this.not = not;
    }

    /*********************************************
     * runtime methods
     */
    public void negate() {
        not = ! not;
    }

    public SqlExpression getLeft() {
        return val;
    }

    public SqlExpression getMin() {
        return min;
    }

    public SqlExpression getMax() {
        return max;
    }

    /**
     * Determines if this statement is true.  Similar to Oracle, when negated, this
     * method negates each portion of the between test separately (<= and >=)
     */
    public boolean isTrue(Tuple t, int[] offsetMap)
            throws SQLException {
        Comparable theVal = val.evaluate(t, offsetMap);
        if(theVal == null)
            return false;

        if(not) {  //  if negated
            Comparable minVal = min.evaluate(t, offsetMap);
            if(minVal != null && theVal.compareTo(minVal) < 0)
                return true;

            Comparable maxVal = max.evaluate(t, offsetMap);
            if(maxVal != null && theVal.compareTo(maxVal) > 0)
                return true;

            return false;
        } else /* no not, ie normal */ {
            Comparable minVal = min.evaluate(t, offsetMap);
            if(minVal == null || theVal.compareTo(minVal) < 0)
                return false;

            Comparable maxVal = max.evaluate(t, offsetMap);
            if(maxVal == null || theVal.compareTo(maxVal) > 0)
                return false;

            return true;
        }
    }

    public double getSelectivity(ColumnStatisticProvider[] tables) throws SQLException {
        return 9;  // 3 * 3, whether it is negated or not (har har)
    }

    public void writeToBuffer(StringBuffer sb) {
        sb.append(val);
        sb.append(" BETWEEN ");
        sb.append(min);
        sb.append(" AND ");
        sb.append(max);
    }

    /**
     * Visitor Pattern.  Visit us, and if true visit children.
     * @param v
     */
    public void accept(ExpressionVisitor v) throws SQLException {
        if(v.forCond(this)) {
            val.accept(v);
            min.accept(v);
            max.accept(v);
        }
    }

    public void performAnalysis(QueryAnalysis qa, boolean appropriateSubConditions,
                                ArrayList columns, ArrayList groupFunctions) throws SQLException {
        int curPos = columns.size();
        val.performAnalysis(qa, appropriateSubConditions, columns, groupFunctions);
        min.performAnalysis(qa, appropriateSubConditions, columns, groupFunctions);
        max.performAnalysis(qa, appropriateSubConditions, columns, groupFunctions);

          //  NOTE: the similarity between this and in BasicCondition
          //  perhaps put the code in a single common area in the SqlCondition base class
        if(appropriateSubConditions) {
            analyzeColumnList(qa, columns, curPos);
        }
    }

    public void performHavingAnalysis(QueryAnalysis qa, ArrayList groupFunctions,
                                      ArrayList groupByClauses) throws SQLException {
        val = findGroupByItem(val, qa, groupFunctions, groupByClauses);
        min = findGroupByItem(min, qa, groupFunctions, groupByClauses);
        max = findGroupByItem(max, qa, groupFunctions, groupByClauses);
    }

    public void verifyColumns(QueryAnalysis qa) throws SQLException {
        val.verifyColumns(qa);
        min.verifyColumns(qa);
        max.verifyColumns(qa);
    }

    public int bindLiterals(Comparable[] values, int curPos) throws SQLException {
        curPos = val.bindLiterals(values, curPos);
        curPos = min.bindLiterals(values, curPos);
        curPos = max.bindLiterals(values, curPos);
        return curPos;
    }

    /*
    public void buildColumnList(ArrayList columnList) {
        val.buildColumnList(columnList);
        min.buildColumnList(columnList);
        max.buildColumnList(columnList);
    }
    */
}
