/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 12, 2002
 * Time: 8:43:30 AM
 */
package freebase.sql.expression;

import freebase.api.ColumnInfo;
import freebase.api.Tuple;
import freebase.sql.SqlExpression;
import freebase.sql.function.Function;
import freebase.sql.function.FunctionManager;
import freebase.sql.plan.QueryAnalysis;
import freebase.util.Helper;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Represents a functional expression, like sin(x) or count(distinct *)
 */
public class FunctionalExpression extends SqlExpression {
    String user = null;
    String ppackage = null;
    String functionName = null;

    /**
     * A list of SqlExpressions
     */
    ArrayList parameters;

    /**
     * Both of these only apply to group functions.  If either is true, and the function is not
     * a group function then oh no!
     */
    boolean distinct = false;
    boolean star = false;

    Function function;

    int relativeOffset = -1;

    /**
     * NOTE:  currently only one parameter space is created for the whole function evaluation.
     *        If this becomes multithreaded this will need to change, though surely each branch
     *        will get its own instance of all parts of the execution plan.
     * NOTE:  this is not necessary.  The GC does good work
     */
    Comparable[] parameterSpace;

    /*************************************************************
     * parse time methods
     */
    public void setUser(String user) {
        this.user = user;
    }

    public void setPackage(String ppackage) {
        this.ppackage = ppackage;
    }

    public void setFunction(String function) {
        this.functionName = function;
    }

    public void setFullFunction(String user, String ppackage, String function) {
        setUser(user);
        setPackage(ppackage);
        setFunction(function);
    }

    public void setParameters(ArrayList parameters) {
        this.parameters = parameters;
    }

    public void setDistinct(boolean distinct) {
        this.distinct = distinct;
    }

    public void setStar(boolean star) {
        this.star = star;
    }

    /**************************************************************
     * getters
     */

    public String getFunctionName() {
        return functionName;
    }

    public int getNumArguments() {
        if(parameters == null)
            return 0;
        return parameters.size();
    }

    public SqlExpression getArgument(int i) {
        return (SqlExpression)parameters.get(i);
    }

    public Function getFunction() {
        return function;
    }

    public boolean isStar() {
        return star;
    }

    public boolean isDistinct() {
        return distinct;
    }

    public void setRelativeOffset(int relativeOffset) {
        this.relativeOffset = relativeOffset;
    }

    /*************************************************************
     * Runtime methods
     */

    public void accept(ExpressionVisitor v) throws SQLException {
        if(v.forFunc(this)) {
            for(int i = 0; i < getNumArguments(); i++) {
                getArgument(i).accept(v);
            }
        }
    }

    public void writeToBuffer(StringBuffer sb) {
        if(user != null) {
            sb.append(user);
            sb.append('.');
        }

        if(ppackage != null) {
            sb.append(ppackage);
            sb.append('.');
        }

        sb.append(functionName);
        sb.append('(');
        if(distinct) {
            sb.append("DISTINCT ");
        }
        if(star) {
            sb.append("*");
        } else {
            if(parameters != null) {
                for(int i = 0; i < parameters.size(); i++) {
                    if(i != 0)
                        sb.append(',');
                    ((SqlExpression)parameters.get(i)).writeToBuffer(sb);
                }
            }
        }

        sb.append(')');
    }

    /**
     * NOTE:  for standard functions, since this repeatedly uses the same array each time to
     *    pass in the input parameters when evaluating the function, some thought needs to be
     *    put in when going to parallel execution
     */
    public Comparable evaluate(Tuple t, int[] offsetMap) throws SQLException {
        if(function.isAggregation()) {
            return t.getValue(offsetMap[relativeOffset]);
        }

        //Comparable[] parameterSpace = new Comparable[getNumArguments()];  not much slower, ~.02 sec on a million
          //  evaluate the standard function
        for(int i = 0; i < getNumArguments(); i++) {
            parameterSpace[i] = getArgument(i).evaluate(t, offsetMap);
        }

        return function.getStandardFunction().evaluate(parameterSpace);
    }

    /**
     * TODO:  this looks like its going to produce some null pointer exceptions.  If
     * it is called only once, then just allocate an array with the tuple entries and pass it in
     * @param t
     * @param offsetMap
     * @return
     * @throws SQLException
     */
    public ColumnInfo getColumnInfo(Tuple t, int[] offsetMap) throws SQLException {
        return ColumnInfo.rename(function.getColumnInfo(null), toString());
    }

    /**
     * Sub columns of this function only get added to the columns list
     * if this function is a non aggregation function.  Otherwise, they
     * need to be
     */
    public void performAnalysis(QueryAnalysis qa, boolean appropriateSubConditions,
                                ArrayList columns, ArrayList groupFunctions) throws SQLException {
        bindFunction(qa.getFunctionManager());
        //System.out.println("function found: " + function);

        if(function.isAggregation()) {
            if(groupFunctions == null)
                throw new SQLException("Aggregation function '" + functionName + "' not legal here");

            if(star && ! function.getAggregationFunction().acceptsFullTuples()) {
                throw new SQLException("Function " + functionName + " does not accept full tuples");
            }

            groupFunctions.add(this);
              //  NOTE:  sub expression will get analyzed in a separate pass
        } else {
            for(int i = 0; i < getNumArguments(); i++) {
                getArgument(i).performAnalysis(qa, appropriateSubConditions, columns, groupFunctions);
            }
        }
    }

    /**
     * Since this is called at the exclusion of performAnalysis, we need to bind the function
     * the same as above.
     */
    public void performHavingAnalysis(QueryAnalysis qa, ArrayList groupFunctions,
                                      ArrayList groupByClauses) throws SQLException {
        bindFunction(qa.getFunctionManager());

        if(function.isAggregation()) {
            groupFunctions.add(this);
        } else {
            for(int i = 0; i < getNumArguments(); i++) {
                getArgument(i).performHavingAnalysis(qa, groupFunctions, groupByClauses);
            }
        }
    }

    /**
     * Binds the local function via the name.  Also does basic checks.
     */
    private void bindFunction(FunctionManager fm) throws SQLException {
        function = fm.getFunction(functionName);

        if(! function.isAggregation()) {
            if(distinct)
                throw new SQLException("DISTINCT not a legal modifier in the non aggregation function: " + functionName);

            if(star)
                throw new SQLException("* not a legal in the non aggregation function: " + functionName);
        }

        parameterSpace = new Comparable[getNumArguments()];
    }

    public void verifyColumns(QueryAnalysis qa) throws SQLException {
        for(int i = 0; i < getNumArguments(); i++) {
            getArgument(i).verifyColumns(qa);
        }
    }

    /**
     * We add ourself to the column list if we are a group by function, otherwise we add all
     * of our sub columns.
     *
    public void buildColumnList(ArrayList columnList) {
        if(function.isAggregation()) {
              //  record where we should get our value!!!
            relativeOffset = columnList.size();
            columnList.add(this);
        } else {
            for(int i = 0; i < getNumArguments(); i++) {
                getArgument(i).buildColumnList(columnList);
            }
        }
    }*/

    public int bindLiterals(Comparable[] values, int curPos) throws SQLException {
        for(int i = 0; i < getNumArguments(); i++) {
            curPos = getArgument(i).bindLiterals(values, curPos);
        }
        return curPos;
    }

    public boolean equals(Object obj) {
        if(! (obj instanceof FunctionalExpression)) {
            return false;
        }

        FunctionalExpression other = (FunctionalExpression)obj;
          //  TODO:  deal with system package or whatever
        if(! (Helper.areEqual(user, other.user) &&
              Helper.areEqual(ppackage, other.ppackage) &&
              Helper.areEqual(functionName, other.functionName))) {
            return false;
        }

        if(! (star == other.star && distinct == other.distinct)) {
            return false;
        }

        if(! Helper.areListsEqual(parameters, other.parameters)) {
            return false;
        }

        return true;
    }
}
