/*
 * Copyright JJ Furman
 * User: user
 * Date: May 15, 2002
 * Time: 1:26:03 PM
 */
package freebase.sql.expression;

import freebase.api.Tuple;
import freebase.catalog.ColumnStatisticProvider;
import freebase.sql.SqlCondition;
import freebase.sql.SqlExpression;
import freebase.sql.plan.QueryAnalysis;
import freebase.util.GlobalProps;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Encapsulates (or will) both operators LIKE and REGEX.
 */
public class LikeCondition extends SqlCondition {
    boolean isNegated = false;

    /**
     * When true, this is a real REGEX, not just a LIKE
     */
    boolean isRegexp = false;

    /**
     * When true, then
     */
    boolean patternIsConstant = false;

    /**
     * x LIKE y
     */
    SqlExpression x;
    SqlExpression y;

    /**
     * the compiled pattern
     */
    Pattern pattern = null;


    public void setLeft(SqlExpression left) {
        x = left;
    }

    public void setRight(SqlExpression right) {
        y = right;

          //  simplistic detection of a constant
        if(y instanceof LiteralExpression ||
                y instanceof BindableLiteral) {
            patternIsConstant = true;
        }
    }

    public void setRegexp(boolean reg) {
        isRegexp = reg;
    }

    /*****************************************************************************
     * SqlCondition methods...
     */
    public int bindLiterals(Comparable[] values, int curPos) throws SQLException {
        curPos = x.bindLiterals(values, curPos);
        return y.bindLiterals(values, curPos);
    }

    /*
    public void buildColumnList(ArrayList columnList) {
        x.buildColumnList(columnList);
        y.buildColumnList(columnList);
    }*/

    public double getSelectivity(ColumnStatisticProvider[] tables) throws SQLException {
        return 4;
    }

    public boolean isTrue(Tuple t, int[] offsetMap)
            throws SQLException {
        Object val = x.evaluate(t, offsetMap);
        if(val == null) {
            return false;
        }

        if(pattern == null || ! patternIsConstant) {
            buildPattern(t, offsetMap);

              //  when the pattern was null... always false
            if(pattern == null) {
                return false;
            }
        }

        if(! (val instanceof String)) {
            throw new SQLException("value comparing to the pattern must be a string but it is a " + val.getClass());
        }

        Matcher m = pattern.matcher((String)val);
        boolean isLike = m.matches();

        if(isNegated) {
            return ! isLike;
        } else {
            return isLike;
        }
    }

    /**
     * Visitor Pattern.  Visit us, and if true visit children.
     * @param v
     */
    public void accept(ExpressionVisitor v) throws SQLException {
        if(v.forCond(this)) {
            x.accept(v);
            y.accept(v);
        }
    }

    /**
     * Builds the appropriate Pattern object for matching against.
     */
    private void buildPattern(Tuple t, int[] offsetMap) throws SQLException {
        Object val = y.evaluate(t, offsetMap);

        if(val == null) {
            pattern = null;
            return;
        }

        if(! (val instanceof String)) {
            throw new SQLException("Pattern : " + val + " must be a string but is a " + val.getClass());
        }

        String like = (String)val;

        if(isRegexp) {
            if(GlobalProps.isStringsCaseInsensitive()) {
                pattern = Pattern.compile(like, Pattern.CASE_INSENSITIVE);
            } else {
                pattern = Pattern.compile(like);
            }

            return;
        }

        StringBuffer pat = new StringBuffer();
        pat.append('^');

        int curPos = 0;
        while(curPos < like.length()) {
            char c = like.charAt(curPos);

              //  probably does not work with all characters
            switch(c) {
                  //  supported features
                case '%':  pat.append(".*");  break;
                case '_':  pat.append(".");  break;

                  //  things to make regexp implementation work
                case '*':  pat.append("\\*");  break;
                case '.':  pat.append("\\.");  break;
                case '?':  pat.append("\\?");  break;
                default:
                    pat.append(c);
            }

            curPos++;
        }
        pat.append('$');

        //System.out.println("PATTERN IS: " + pat);
        if(GlobalProps.isStringsCaseInsensitive()) {
            pattern = Pattern.compile(pat.toString(), Pattern.CASE_INSENSITIVE);
        } else {
            pattern = Pattern.compile(pat.toString());
        }
    }

    public void negate() {
        isNegated = ! isNegated;
    }

    public void performAnalysis(QueryAnalysis qa, boolean seekingEquiJoins,
                                ArrayList columns, ArrayList groupFunctions) throws SQLException {
        int curPos = columns.size();

        x.performAnalysis(qa, seekingEquiJoins, columns, groupFunctions);
        y.performAnalysis(qa, seekingEquiJoins, columns, groupFunctions);

        analyzeColumnList(qa, columns, curPos);
    }

    public void performHavingAnalysis(QueryAnalysis qa, ArrayList groupFunctions,
                                      ArrayList groupByClauses) throws SQLException {
        x = findGroupByItem(x, qa, groupFunctions, groupByClauses);
        y = findGroupByItem(y, qa, groupFunctions, groupByClauses);
    }

    public void verifyColumns(QueryAnalysis qa) throws SQLException {
        x.verifyColumns(qa);
        y.verifyColumns(qa);
    }

    public void writeToBuffer(StringBuffer sb) {
        x.writeToBuffer(sb);

        if(isNegated) {
            sb.append(" NOT LIKE ");
        } else {
            sb.append(" LIKE ");
        }

        y.writeToBuffer(sb);
    }
}
