/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 17, 2002
 * Time: 2:53:01 PM
 */
package freebase.sql.function;

import freebase.api.ColumnInfo;

import java.sql.SQLException;

/**
 * Represents the interface to an aggregation function like
 * COUNT, MIN, MAX, AVG, SUM.  In the future it might
 * be nice to have a HISTOGRAM, but whatever.
 *
 * TODO:  support multiple valued operations, for example
 *     choose_max(val, a, b, c) that for the max val, would produce a,b,c
 */
public abstract class AggregationFunction implements Function {
    /**
     * By default, the same as the first input column (one only)
     * @param inputs
     * @return Don't worry, this will get renamed anyway
     * @throws SQLException
     */
    public ColumnInfo getColumnInfo(ColumnInfo[] inputs) throws SQLException {
        return inputs[0];
    }

    /**
     *  It is assumed that null is never passed in
     */
    abstract public void addValue(Comparable val) throws SQLException;

    /**
     * Only called if the aggregation function needs more than one value, thus will
     * throw an exception here and not elsewhere
     * @param val
     * @throws SQLException
     */
    public void addValues(Comparable[] val) throws SQLException {
        throw new SQLException("Function " + this.getClass().getName() + " is not supposed to accept more than one parameter!!");
    }

    /**
     * Gets the current result.
     */
    abstract public Comparable getResult() throws SQLException;

    /**
     * Prepares this aggregation function for another run (sub group)
     */
    abstract public void reset() throws SQLException;

    /**
     * Says whether this accepts * as an argument.  Currently
     * only count does.
     */
    abstract public boolean acceptsFullTuples() throws SQLException;

    /**
     * TODO:  could throw the exception in here to give better error message
     */
    public boolean supportsValues(int number) throws SQLException {
        return number == 1;
    }

    /*************************************
     * function methods
     */
    public boolean isAggregation() {
        return true;
    }

    public AggregationFunction getAggregationFunction() {
        return this;
    }

    /**
     * only one of getAgg or getStandard need to be implemented, so this returns null.
     * @return
     */
    public StandardFunction getStandardFunction() {
        return null;
    }
}
