/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 24, 2002
 * Time: 3:00:53 PM
 */
package freebase.sql.plan;

import freebase.api.TupleIterator;
import freebase.sql.ColumnRef;
import freebase.sql.SelectStatement;
import freebase.sql.SqlExpression;
import freebase.sql.expression.ColumnExpression;
import freebase.sql.expression.LiteralExpression;
import freebase.sql.plan.visitor.ColumnList;
import freebase.sql.query.Projection;
import freebase.sql.query.SimpleSortPass;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Helps process the order by clause.  It builds the sortColumns and ascendingFlags
 * as well as add any necessary columns to the
 */
public class OrderByHandler {
    /**
     * Passed Data
     */
    SelectStatement select;
      //  Both these:  only needed in the non group by sort
    String[] perColumnNames;
    int[] perColumnTableOffsets;

    /**
     * Derived data
     */
    int originalSize;
    int[] sortColumns = null;  // indexes into penultimate projection
    boolean[] ascendingFlags = null;

    /**
     * Construct this order by handler to have all of the information needed to enable the caller
     * to easily sort the information.  Just construct it and grab the info required
     */
    public OrderByHandler(SelectStatement select, String[] perColumnNames, int[] perColumnTableOffsets)
            throws SQLException {
        this.select = select;
        this.perColumnNames = perColumnNames;
        this.perColumnTableOffsets = perColumnTableOffsets;

        processInformation();
    }

    /**********************************************************************************
     * The main workhorse, takes the information passed in the constructor and
     * does all of the changes needed.<BR>
     *
     * What it does is scans through all of the
     */
    private void processInformation() throws SQLException {
        ArrayList finalProjections = select.getSelectList();
        originalSize = select.getSelectList().size();  //  only needed because finalProj grows
        int maxLegalColumn = select.isStar() ? perColumnNames.length : originalSize;
       // int numExtraSortClauses = 0;

          //  determine all order by references.  TODO:  move this to another place
        if(select.getNumOrderByClauses() != 0) {
            boolean foundNonSort = false;
            sortColumns = new int[select.getNumOrderByClauses()];
            ascendingFlags = new boolean[select.getNumOrderByClauses()];

            //boolean foundNull = false;  //  if this is true, a rewrite is needed, dumb
            nextOrderByItem:
            for(int i = 0; i < select.getNumOrderByClauses(); i++) {
                ascendingFlags[i] = select.getOrderByDirection(i);
                SqlExpression x = select.getOrderByExpression(i);

                if(x instanceof LiteralExpression) {
                    sortColumns[i] = extractSortColumnNumber(x, maxLegalColumn);
                    if(sortColumns[i] == -1)
                        foundNonSort = true;
                    continue nextOrderByItem;
                } else if(x instanceof ColumnExpression && select.isStar()) {
                    ColumnExpression cx = (ColumnExpression)x;
                    sortColumns[i] = QueryAnalysis.findColumnOffset(cx.getColumn(),
                            perColumnTableOffsets, perColumnNames);
                    continue nextOrderByItem;
                }

                for(int j = 0; j < originalSize; j++) {
                    if(finalProjections.get(j).equals(x)) {
                        sortColumns[i] = j;
                        continue nextOrderByItem;
                    }
                }

                  //  expression sort:  add another column to the end, only to remove it later
                sortColumns[i] = finalProjections.size();
                finalProjections.add(x);
                if(select.isStar()) {
                    sortColumns[i] += perColumnTableOffsets.length;
                } else if(select.isDistinct()) {
                      //  for the case:   select distinct a from test order by a+b;
                    //verifyTest(x, select.getSelectList(), originalSize);
                    verifyOrderByExpressionInDistinctList(x, select.getSelectList(), originalSize);
                }
            }

            //  Oracle compatibility:  if any of the sort entries were meaningles
            //  indeed, in oracle check the plan of:  select * from test order by 'fred';
            if(foundNonSort) {
                sortColumns = copyOnlyLegalSortEntries(sortColumns, ascendingFlags);
                if(sortColumns == null) {
                    ascendingFlags = null;
                }
            }
        }
    }

    /****************************
     * internal helper methods
     */
    /**
     * If you execute a query like:<BR>
     * select distinct a from test2 order by b; <BR><BR>
     *
     * Then the system does not have enough information to determine the "b" part of
     * the order by clause (cuz b info has been eliminated) thus must abort.<BR><BR>
     *
     * TODO:  this probably ought to be implemented via a recursive method, passing
     *   the list of legal expressions down...
     *   this should work:  select distinct a, abs(b) from test2 order by a+abs(b);
     *
     * TODO:  tempted to put this testing logic in the analyzeOrderBy
     */
    private void verifyOrderByExpressionInDistinctList(SqlExpression x, ArrayList selectItems,
                                                       int originalSize) throws SQLException {
        //  for each column in the expression, verify that it is present in the select items
        ColumnList columns = new ColumnList();
        //x.buildColumnList(columns);
        x.accept(columns);

        scanningcolumns:
        for(int i = 0; i < columns.size(); i++) {
            ColumnRef ref = null;
            Object obj = columns.get(i);
            if(obj instanceof ColumnRef) {
                ref = (ColumnRef)obj;
            }

            for(int j = 0; j < originalSize; j++) {
                if(ref != null) {
                    if(selectItems.get(j) instanceof ColumnExpression) {
                        ColumnExpression cx = (ColumnExpression)selectItems.get(j);
                        if(cx.getColumn().isEquivalentTo(ref)) {
                            continue scanningcolumns;
                        }
                    }
                } else {
                    if(selectItems.get(j).equals(obj)) {
                        continue scanningcolumns;
                    }
                }
            }

            throw new SQLException("Illegal ORDER BY item:  select items do not contain " + obj);
        }
    }


    /**
     * eliminates all -1 entries and creates a new array.  If there are no remaining legal
     * entries, it will return null.  It also modifies the ascending buffer in such a way
     * that the first output.length entries have the collapsed entries of the
     */
    private int[] copyOnlyLegalSortEntries(int[] input, boolean[] ascending) {
          //  count non -1's
        int numOkSorts = 0;
        for(int i = 0; i < input.length; i++) {
            if(input[i] != -1) {
                numOkSorts++;
            }
        }

        if(numOkSorts == 0) {
            return null;
        } else {
              //  copy the good ones
            int[] output = new int[numOkSorts];
            int pos = 0;
            for(int i = 0; i < input.length; i++) {
                if(input[i] != -1) {
                    ascending[pos] = ascending[i];
                    output[pos++] = input[i];
                }
            }

            return output;
        }
    }

    /**
     *
     */
    protected int extractSortColumnNumber(SqlExpression x, int maxLegalColumn) throws SQLException {
        LiteralExpression le = (LiteralExpression)x;
        if(le.isNumeric()) {  //  then this should be a column position reference
            //  extract integer portion of number!
            BigDecimal num = (BigDecimal)le.getValue();
            int pos = num.intValue();
            if(pos < 1 || pos > maxLegalColumn) {
                throw new SQLException("order by item must be position of a select item");
            }
            return pos - 1;
        } else {
            return -1;
        }
    }

    /*****************************************************************************
     * Getters:  information available after processInformation has been called
     */

   // public ArrayList getFinalSelectList() {
    //    return select.getSelectList();
   // }

    public int[] getSortColumns() {
        return sortColumns;
    }

    public boolean[] getAscendingFlags() {
        return ascendingFlags;
    }

    public int getOriginalNumSelectItems() {
        return originalSize;
    }

    public TupleIterator sort(TupleIterator data) throws SQLException {
        if(getSortColumns() == null) {
            return data;
        }

        return new SimpleSortPass(data, getSortColumns(), getAscendingFlags());
    }

    /**
     * If sorting is needed, then do it and retruncate the query back down to size
     *
     *
    public TupleIterator sortAndChopIfNecessary(TupleIterator data) throws SQLException {
        data = sort(data);
        return chopIfNecessary(data);
    }
     */

    /**
     * Trim down the set of columns to just those requested
     * @param rs
     * @return
     * @throws SQLException
     */
    public TupleIterator chopIfNecessary(TupleIterator rs) throws SQLException {
        if(select.getSelectList().size() > getOriginalNumSelectItems()) { //  deals with star and everything
            int difference = select.getSelectList().size() - getOriginalNumSelectItems();
            rs = new Projection(rs, rs.getNumColumns() - difference);
        }

        return rs;
    }

    /*******************************************************************************
     * methods below are special helper methods, some only called in the group by case,
     * others called in the case of sorting standard results
     */
}
