/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 15, 2002
 * Time: 10:13:54 AM
 */
package freebase.sql.plan;

import freebase.api.TupleIterator;
import freebase.catalog.Catalog;
import freebase.catalog.Index;
import freebase.catalog.Table;
import freebase.engine.IndexScan;
import freebase.engine.Transaction;
import freebase.lock.UseLock;
import freebase.sql.*;
import freebase.sql.expression.*;
import freebase.sql.function.FunctionManager;
import freebase.sql.plan.visitor.ColumnList;
import freebase.sql.query.*;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * QueryAnalysis is a first attempt to implement a reasonable full featured query execution.
 * It encapsulates the both the analysis phase and the analysis results.  Currently it only
 * support select statements.
 * <p>
 * What a monolithic beast this is.
 * <P>
 * TODO:  must start thinking about the issue of the table being changed while we are doing
 * the anaysis, or between sessions.
 *
 * TODO:  later, deal with saving these for the future prepared statements
 *
 * TODO:  optimize this for the single or two table case, so that a lot of extra internal
 *    structures do not get created all of the time (joinedYet)
 *
 * TODO:  add heuristic for joins with large number of tables, say above 10 or 15, TODO
 *      a more linear search, not this n^2 thing
 *
 * TODO:  eventually do sorting with a smaller set, earlier.  Use the counting method to reduce
 *    the projection set as soon as possible
 *
 * TODO:  to clean this up, easiest to encapsulate data objects and their functions out
 *    to other classes.  This way the functions can be changed easily to experiment with various
 *    performance tradeoffs.  Functions such as lookup.
 *
 * TODO:  could cleanup select items in aggregation queries by replacing group by item expressions
 *    with GroupByItem objects as in the having clause.  Would certainly be simpler code, though maybe
 *    a teensy bit slower
 */
public class QueryAnalysis {
    final public static Integer MINUS_ONE_INTEGER = new Integer(-1);

    Catalog catalog;
    FunctionManager fm;
    SelectStatement select;
    String defaultSchema;

    ArrayList tableRefs;

    /**
     * the map of names and aliases to offsets.  If the offset is -1 that means
     * that the name is ambiguous.  Eventually this will only be active if there
     * are, say perhaps 3 or more tables.<BR><BR>
     *
     * NOTE: To save a little time this only stores offsets under 'schema.table' instead
     * of just 'table' only when:  1.  the schema is explicitly mentioned and 2. the schema
     * is not the default schema.  Compare with Oracle by writing the query:
     *  select * from test where {schema}.test.id = 1;   --  will throw an error, same as ours
     */
    HashMap tableOffsets = new HashMap();
    Table[] tables;
    SqlCondition[] tableConditions;
    boolean[] joinedAlready;
    int[] joinSequence;
    int curSequencePos = 2;  //  start it essentially after the initial join
    long curCost = -1;
    boolean selectClauseHasSomeWholeGroupByExpressionReferences = false;

    SqlCondition postJoinCondition = null;

    ArrayList tableUseLocks = new ArrayList();

    /**
     * expressions like 1=0 and silly 'Bob' like 'B%'.  Execute once, either true or false!
     */
    SqlCondition globalConstantCondition = null;

    /**
     * estimates for the size of the results after the single table filters
     */
    long[] sizeEstimates;

    /**
     * per join (even the first non join) estimates for how big the results will be.
     * This is useful for deciding between using an index to pre-filter data or to
     * use it TODO value lookup on join condition
     */
    long[] postJoinSizeEstimates;

    /**
     * Is size n * (n - 1) / 2, a triangle shape, between each pair of tables
     */
    ArrayList[] joinClauses;

    /**
     * This is size n - 1, one list for each new table in the sequence, excluding the first
     */
    ArrayList[] joinClausesInSequence;

    // results from select and group analysis---------------------------
    /**
     * if null, no outer grouping is needed.  Eventually must support nested aggregation, rename
     * this outerGroupFunctions  <BR>
     * TODO:  instead of an ArrayList, use an explicit class that has build in search functions
     */
    ArrayList aggregations = null;  //  null if none were found

    /**
     * Build the analysis and prep it for work.  Will only do work that does not require a shared lock on the
     * tables involved
     * @param defaultSchema the current user name
     */
    public QueryAnalysis(Catalog catalog, FunctionManager fm, String defaultSchema, SelectStatement select) throws SQLException {
        this.catalog = catalog;
        this.fm = fm;
        this.select = select;
        this.defaultSchema = defaultSchema;

        this.tableRefs = select.getTables();
        tables = new Table[tableRefs.size()];
        tableConditions = new SqlCondition[tables.length];
        joinedAlready = new boolean[tables.length];
        sizeEstimates = new long[tables.length];

        //  TODO:  The following are created every time.  Eventually rewrite this to optimize
        //    for the one table case, since that could save a decent number of object creations.
        //if(tables.length > 1) {
            joinSequence = new int[tables.length];
            joinClausesInSequence = new ArrayList[tables.length - 1];
            postJoinSizeEstimates = new long[tables.length];
        //}

        joinClauses = new ArrayList[getTriangleSize(tableRefs.size() - 1)];

          //  build up the list of tables & their use locks
        for(int i = 0; i < tableRefs.size(); i++) {
            TableRef tabRef = (TableRef)tableRefs.get(i);
            tables[i] = catalog.getTable(tabRef, defaultSchema);
            tableUseLocks.add(tables[i].getUseLock());

            Integer offset = new Integer(i);
            String fullName = tabRef.getTable();
              //  expand fullname with schema ONLY if not the default schema
            if(tabRef.getSchema() != null && ! tabRef.getSchema().equals(defaultSchema)) {
                fullName = tabRef.getSchema() + '.' + tabRef.getTable();
            }

              //  add a binding for the table name alone
            Object prev = tableOffsets.put(fullName, offset);
            if(prev != null) {
                  //  this allows the operation, but will fail later if columns refer to it!
                tableOffsets.put(fullName, MINUS_ONE_INTEGER);
            }

              //  add a binding for the alias too  (unlike Oracle, both can work...
              //  "select * from part a where part.part_id=1"
            String alias = select.getTableAlias(i);
            if(alias != null) {
                prev = tableOffsets.put(alias, offset);
                if(prev != null && ! prev.equals(offset)) {
                    throw new SQLException("Alias '" + alias + "' multiply defined.");
                }
            }
        }
    }

    /**
     * lock all required resources in shared mode
     */
    public void lockResources() throws SQLException {
        UseLock.lockAllShared(tableUseLocks);

          //  TODO:  check to ensure all resources exist (always check)
    }

    /**
     * unlock all required resources in shared mode
     */
    public void unlockResources() throws SQLException {
        UseLock.unlockAllShared(tableUseLocks);
    }

    /**
     * Analyse the entire query to validate it and determine the best plan of
     * attack.
     */
    public void analyzeQuery() throws SQLException {
        //System.out.println("analyzing Query: " + select);
          //  Analyze 'group by' first because later we need to compare selects against it
        analyzeGroupByClause();
        analyzeSelectClause();    //  select is currently first because it defines the aggregations variable
        analyzeHavingClause();
        analyzeOrderByClause();

        ArrayList conditionColumns = new ArrayList();
        if(select.getCondition() != null)
            select.getCondition().performAnalysis(this, true, conditionColumns, null);

        //  TODO:  use the conditionColumns for something
        //  TODO:  future:  also use existing join information in the from clause

        determineBasicCosts();
        findFirstJoin();
        completeJoinSequence();

        //System.out.println(StringUtil.dump(joinSequence));
        //  analyze group bys and scan for aggregate functions in select list

        //  nullify all data structures that are no longer needed.
    }

    /**
     * Executes the query with a physical plan decided upon internally.  Currently
     * it builds the Physical plan each time, but in the future it could re-used
     * the plan built (and the space allocated). <BR><BR>
     *
     * TODO:  right now it puts physical plan information right here in the same class.
     * There would be problems in multiple threads tried to access the data at the same
     * time.
     *
     * TODO:  this should be broken down into subparts, either external methods or
     *      putting search type capability into other classes
     */
    public TupleIterator executeQuery(Transaction t) throws SQLException {
        TupleIterator rs = null;

        /*  Put all potential columns from all tables into a list.  Thus, each column
          from each table has an offset.  Just because each column has an offset in some
          layer of the execution does not mean that all the values are actually materialized...
          TODO:  could move this to a separate section and make them global
          */
          //  count all columns, not necessarily in order
        int totalColumns = 0;
        for(int i = 0; i < tables.length; i++) {
            totalColumns += tables[i].getNumColumns();
        }

        int[] perColumnTableOffsets = new int[totalColumns];  //  could be (num columns grabbed)
        String[] perColumnNames = new String[totalColumns];

          //  build list in sequence
        int curPos = 0;
        for(int i = 0; i < joinSequence.length; i++) {
            Table tab = tables[joinSequence[i]];
            for(int j = 0; j < tab.getNumColumns(); j++) {
                perColumnTableOffsets[curPos + j] = joinSequence[i];
                perColumnNames[curPos + j] = tab.getColumnInfo(j).getName();
            }

            curPos += tab.getNumColumns();
        }
        //System.out.println("Columns are : "+ StringUtil.dump(perColumnNames));

          //  perform joins
        try{  //  any failure should close this set, eventually this will be building an external entity
        rs = getSingleTableQuery(t, joinSequence[0]);

        for(int i = 1; i < tables.length; i++) {
            ArrayList equiJoins = joinClausesInSequence[i - 1];
            joinClausesInSequence[i - 1] = null;

            if(equiJoins == null) {  //  cartesian product, best join is raw NestedLoops
                rs = new NestedLoopJoin(rs, getSingleTableQuery(t, joinSequence[i]));
            } else /* build up column filter list */ {
                rs = performBestJoin(t, rs, equiJoins, joinSequence[i], perColumnTableOffsets, perColumnNames);

                  //  further filter anything that has not been addressed by the "best join" (indeed!)
                if(equiJoins.size() > 0) {
                    SqlCondition allEquis = collapseConditions(equiJoins);
                      //  build a column list from it (HERE, see below)
                    ColumnList columns = new ColumnList();
                    allEquis.accept(columns);
                    int[] offsets = createOffsetMap(columns, perColumnTableOffsets, perColumnNames);

                    rs = new RowFilterIterator(rs, allEquis, offsets, false);
                }
            }
        }

          //  handle post join condition
        if(postJoinCondition != null) {
              //  NOTE:  the following code is the same as the above "HERE"
              //  when the technique is finalized, the equivalent code
              //  can be centralized to one area
            ColumnList columns = new ColumnList();
            postJoinCondition.accept(columns);
            int[] offsets = createOffsetMap(columns, perColumnTableOffsets, perColumnNames);

            rs = new RowFilterIterator(rs, postJoinCondition, offsets, false);
        }


        //  TODO:  this would be a good time to reduce the query
        //     down to just those columns that might be needed in the final passes
        //--------------------------------------------------------------------
        OrderByHandler orderHelper = null;

          //  handle aggregation:  either if there is a group by or lone aggregation functions
        if(aggregations != null || select.getGroupByExpressions() != null) {
            int[] groupByColumns = null;
            if(select.getGroupByExpressions() != null) {  //  need to project and sort
                groupByColumns = new int[select.getNumGroupByExprs()];
                ArrayList groupExpressions = null;
                for(int i = 0; i < select.getNumGroupByExprs(); i++) {
                    SqlExpression x = select.getGroupByExpr(i);

                    if(x instanceof ColumnExpression) {
                        ColumnRef col = ((ColumnExpression)x).getColumn();
                        groupByColumns[i] = findColumnOffset(col, perColumnTableOffsets, perColumnNames);
                    } else {
                        //  TODO:  we already figured out which expressions are not single columns.
                        //    we should recycle this information
                        if(groupExpressions == null)
                            groupExpressions = new ArrayList();

                          //  set the group position
                        groupByColumns[i] = perColumnNames.length + groupExpressions.size();
                        groupExpressions.add(x);
                    }
                }

                  //  if we need to group by an expression, then materialize the value of the
                  // expression up front and append it via the Projection
                if(groupExpressions != null) {
                    int[] offsets = createOffsetMapFromExpressions(groupExpressions,
                            perColumnTableOffsets, perColumnNames);
                    rs = new Projection(rs, groupExpressions, offsets, true, null);
                }

                rs = new SimpleSortPass(rs, groupByColumns);
            }

            FunctionalExpression[] aggFunctions = null;
            int[] offsetMap = null;
            if(aggregations != null) {
                aggFunctions = new FunctionalExpression[aggregations.size()];
                aggregations.toArray(aggFunctions);

                ColumnList columns = new ColumnList();
                for(int i = 0; i < aggFunctions.length; i++) {
                    if(! aggFunctions[i].isStar()) {
                        for(int j = 0; j < aggFunctions[i].getNumArguments(); j++) {
                            aggFunctions[i].getArgument(j).accept(columns);
                        }
                    }
                }
                offsetMap = createOffsetMap(columns, perColumnTableOffsets, perColumnNames);
            }
            rs = new Aggregation(rs, groupByColumns, aggFunctions, offsetMap);

              //  handle HAVING clause
            if(select.getHavingCondition() != null) {
                ColumnList columns = new ColumnList();
                select.getHavingCondition().accept(columns);
                int[] offset = new int[columns.size()];
                for(int i = 0; i < offset.length; i++) {
                    if(columns.get(i) instanceof GroupByItem) {
                        offset[i] = ((GroupByItem)columns.get(i)).getGroupByItem();
                    } else {
                        offset[i] = aggregations.indexOf(columns.get(i));
                        if(groupByColumns != null)
                            offset[i] += groupByColumns.length;
                    }
                }

                rs = new RowFilterIterator(rs, select.getHavingCondition(), offset, false);
            }

              //  handle order by
            if(select.getNumOrderByClauses() != 0) {
                orderHelper = new OrderByHandler(select, perColumnNames, perColumnTableOffsets);;
                //System.out.println(StringUtil.dump(orderHelper.getSortColumns()));
                //System.out.println(StringUtil.dump(orderHelper.());
            }

              //  do pen-ultimate projection down to desired results
              //  include sort columns needed
            int[] finalColumns = new int[select.getNumSelectItems()];  // + extra sort size
            ColumnList columnsAndAggregates = new ColumnList();
            for(int i = 0; i < select.getNumSelectItems(); i++) {
                SqlExpression x = select.getSelectItem(i);

                //  TODO:  here detect if the whole expression is a group by
                //    if so, DON'T build up the column list, instead mark it explicitly
                boolean foundWholeExpression = false;
                if(selectClauseHasSomeWholeGroupByExpressionReferences) {
                    for(int j = 0; j < select.getNumGroupByExprs(); j++) {
                        if(x.equals(select.getGroupByExpr(j))) {
                              //  oh goodie, mark an exact passthrough to the Aggregation above
                            foundWholeExpression = true;
                            finalColumns[i] = j;
                        }
                    }
                }

                if(! foundWholeExpression) {
                    finalColumns[i] = -(1 + i);
                    x.accept(columnsAndAggregates);
                }
            }

            //  TODO:  add

              //  match each column or aggregation with the source in the previous result set
            offsetMap = new int[columnsAndAggregates.size()];
            for(int i = 0; i < offsetMap.length; i++) {
                offsetMap[i] = -1;
                  //  figure out what this is, a group by column or a function.  The whole group by expressions should have been delt with above
                if(columnsAndAggregates.get(i) instanceof ColumnRef) {
                    ColumnRef col = (ColumnRef)columnsAndAggregates.get(i);
                    for(int j = 0; j < select.getNumGroupByExprs(); j++) {
                        SqlExpression x = select.getGroupByExpr(j);
                        if(x instanceof ColumnExpression) {
                            ColumnExpression cx = (ColumnExpression)x;
                            if(cx.getColumn().isEquivalentTo(col)) {
                                offsetMap[i] = j;
                                break;
                            }
                        }
                    }
                } else {
                      //  columnsAndAggregates.get(i) is a FunctionalExpression, but no need to
                      //  do any casting or anything
                    for(int j = 0; j < aggregations.size(); j++) {
                        if(columnsAndAggregates.get(i).equals(aggregations.get(j))) {
                            offsetMap[i] = j;
                            if(groupByColumns != null) {
                                offsetMap[i] += groupByColumns.length;
                            }
                        }
                    }
                }

                if(offsetMap[i] == -1) {
                    throw new SQLException("INTERNAL ERROR:  Failed to find column representing : " + columnsAndAggregates.get(i));
                }
            }

              //  project down to penultimate results.
            rs = new Projection(rs, finalColumns, select.getSelectList(), offsetMap,
                    select.getSelectItemAliases());
        } else /*  unaggregated query, do final projection */ {
            if(select.getNumOrderByClauses() != 0) {
                orderHelper = new OrderByHandler(select, perColumnNames, perColumnTableOffsets);;
            }

            int[] offsetMap = createOffsetMapFromExpressions(select.getSelectList(), perColumnTableOffsets,
                    perColumnNames);

              //  if isStar() is true, then append extra sorting values at the end
            if(select.isStar()) {
                if(tables.length > 1) {  //  reorder the stuff around
                    int[] rearrangeMap = buildRearrangeMap(perColumnTableOffsets, select.getSelectList());
                    rs = new Projection(rs, rearrangeMap, select.getSelectList(), offsetMap, null);
                } else {
                      //  if there are any, append them.  Otherwise, It is fine.
                    if(select.getSelectList().size() > 0) {
                        rs = new Projection(rs, select.getSelectList(), offsetMap, true, null);
                    }
                }
            } else {
                rs = new Projection(rs, select.getSelectList(), offsetMap, false, select.getSelectItemAliases());
            }
        }

          //  TODO:  if order by a subset of group by, no need for further sort
          //  TODO:  if one of earlier sort columns is unique then lower sorts aren't needed

          //  eliminate duplicates before sorting for the final time
        if(select.isDistinct()) {
            if(orderHelper != null) {
                  //  does it by sorting
                rs = eliminateDuplicates(rs, orderHelper.getSortColumns(), orderHelper.getAscendingFlags());
            } else {
                rs = eliminateDuplicates(rs, null, null);
            }
        } else if(orderHelper != null) {
            rs = orderHelper.sort(rs);
        }

          //  Handle TOP n & TOP n PERCENT
        switch(select.getTopMode()) {
            case SelectStatement.TOP_MODE_ALL:    break;
            case SelectStatement.TOP_MODE_PERCENT:  throw new SQLException("TOP n PERCENT not implemented yet.  TODO:  clone initial query and pass in count to FirstRows (make mode in FirstRows to use tuple result");
            case SelectStatement.TOP_MODE_NUM:
                if(select.getTopNum() < 0)
                    throw new SQLException("Reza Behforooz suggests positive TOP numbers since negatives don't really make much sense (now do they?)");

                int[] withTiesSortColumns = null;
                if(select.getWithTies()) {
                    if(orderHelper == null) {
                        throw new SQLException("WITH TIES requires an order by clause");
                    }
                    withTiesSortColumns = orderHelper.getSortColumns();
                }
                rs = new FirstRows(rs, select.getTopNum(), withTiesSortColumns);
        }

          //  prune out unwanted sort columns
        if(orderHelper != null) {
            rs = orderHelper.chopIfNecessary(rs);
        }

          //  check global constant conditions like 1=0
        if(globalConstantCondition != null) {
              //  null, null because there is no needed tuple or offsets!
            if(! globalConstantCondition.isTrue(null, null)) {
                rs = new NoRowsMetaOnly(rs, globalConstantCondition);
            }
        }

        return rs;
        }catch(SQLException sqle) {
            if(rs != null) {
                rs.close();
            }
            throw sqle;
        }
    }

    /*******************************************************************
     * Sub steps of executeQuery
     */

    /**
     * Determines the best join plan possible.  Pretty dumb right now, just finds the very first index.
     * @param leftRs the left result set, or driving result set.  Joins are done to the right, left deep
     * @param equiJoins this method will remove any equiJoin condition that it uses in any index scan.
     * @return
     * @throws SQLException
     */
    protected TupleIterator performBestJoin(Transaction t, TupleIterator leftRs, ArrayList equiJoins,
                                            int rightTableOffset, int[] perColumnTableOffsets,
                                            String[] perColumnNames) throws SQLException {
        if(QueryProcessor.USE_INDEXES) {
              //  all equiJoins must be BasicConditions (for now)
            for(int i = 0; i < equiJoins.size(); i++) {
                BasicCondition bc = (BasicCondition)equiJoins.get(i);

                ColumnExpression targetCol = (ColumnExpression)bc.getLeftExpression();
                ColumnExpression sourceCol = (ColumnExpression)bc.getRightExpression();
                if(targetCol.getColumn().getTableOffset() != rightTableOffset) {
                    ColumnExpression tmp = targetCol;
                    targetCol = sourceCol;
                    sourceCol = tmp;
                }

                ArrayList indexes = tables[rightTableOffset].getIndexesFor(targetCol.getColumn().getColumn());
                if(indexes != null) {
                      //  TODO:  note: stupidly takes first on, doesn't even try to match the second column yet!!!
                    Index index = (Index)indexes.get(0);
                    //System.out.println("Found index " + index + " for Target column: " + targetCol);
                    equiJoins.remove(i);  //  remove so that extra work isn't done

                    IndexScan rightScan = index.getIndexScan(t, false, null);

                    int leftColOffset = findColumnOffset(sourceCol.getColumn(), perColumnTableOffsets, perColumnNames);

                      //  Do Index Loop:  currently only does a one column index join, easy to best fit more!!
                    TupleIterator rs = new NestedLoopJoin(leftRs, rightScan, new int[] {leftColOffset});

                      //  no single table filters?  If so, return
                    SqlCondition tableFilter = tableConditions[rightTableOffset];
                    if(tableFilter == null) {
                        return rs;
                    }

                      //  determine offset map, but in global space (join order space)
                    ColumnList columnList = new ColumnList();
                    tableFilter.accept(columnList);
                    int[] offsets = new int[columnList.size()];
                    for(int j = 0; j < columnList.size(); j++) {
                        ColumnRef col = (ColumnRef)columnList.get(j);
                        offsets[i] = findColumnOffset(col, perColumnTableOffsets, perColumnNames);
                    }

                    return new RowFilterIterator(rs, tableFilter, offsets, false);
                }
            }
        }

          //  give up, just do brute force nested loops
        return new NestedLoopJoin(leftRs, getSingleTableQuery(t, rightTableOffset));
    }

    /**
     * Use this to test:
     * select distinct count(*) from test2 group by a order by count(*) desc;
     *
     * TODO:  when we go to building templates rather than the real result set
     *    we will need to pass in the real size
     */
    protected TupleIterator eliminateDuplicates(TupleIterator data, int[] sortColumns,
                                                boolean[] ascendingFlags) throws SQLException {
          //  sort by EVERYTHING!!!
        int[] finalSortColumns = new int[data.getNumColumns()];
        for(int i = 0; i < finalSortColumns.length; i++) {
            finalSortColumns[i] = i;
        }

          //  but put explicit user requests first to deal with it all at once
        if(sortColumns != null) {
            for(int i = 0; i < sortColumns.length; i++) {
                  //  do a swap if need be (no tmp needed, look carefully)
                if(finalSortColumns[i] != sortColumns[i]) {
                    finalSortColumns[sortColumns[i]] = finalSortColumns[i];
                    finalSortColumns[i] = sortColumns[i];
                }
            }
        }

        TupleIterator rs = new SimpleSortPass(data, finalSortColumns, ascendingFlags);
        return new EliminateDuplicates(rs);
    }

    /**
     * Builds the map to revert the order of the data from join order back into normal select
     * statement definition order,
     */
    private int[] buildRearrangeMap(int[] perColumnTableOffsets, ArrayList finalProjections) {
          //  need to reorder the columns back to the select ordering, not join ordering
        int pos = 0;
        int[] rearrangeMap = new int[perColumnTableOffsets.length + finalProjections.size()];
        for(int i = 0; i < tables.length; i++) {
            int first = 0;
            for(int j = 0; j < joinSequence.length; j++) {
                if(joinSequence[j] == i) {
                    break;
                }
                first += tables[joinSequence[j]].getNumColumns();
            }

            for(int j = 0; j < tables[i].getNumColumns(); j++) {
                rearrangeMap[pos + j] = first + j;
            }

            pos += tables[i].getNumColumns();
        }

        if(finalProjections.size() > 0) {
            for(int i = 0; i < finalProjections.size(); i++) {
                rearrangeMap[perColumnTableOffsets.length + i] = -(1 + i);
            }
        }

        return rearrangeMap;
    }

    /**
     * For each column, finds an appropriate entry in the perColumnXYZ arrays passed in.  These arrays
     * are correlated
     * TODO:  extend this to support aggregationFunction values too
     */
    protected int[] createOffsetMap(ArrayList columns, int[] perColumnTableOffsets,
                                    String[] perColumnNames) throws SQLException {
        int[] offsets = new int[columns.size()];

        for(int j = 0; j < columns.size(); j++) {
            ColumnRef col = (ColumnRef)columns.get(j);
            offsets[j] = findColumnOffset(col, perColumnTableOffsets, perColumnNames);
        }

        return offsets;
    }

    protected int[] createOffsetMapFromExpressions(ArrayList expressions, int[] perColumnTableOffsets,
                                    String[] perColumnNames) throws SQLException {
        ColumnList columns = new ColumnList();
        for(int i = 0; i < expressions.size(); i++) {
            SqlExpression x = (SqlExpression)expressions.get(i);
            x.accept(columns);
        }

        return createOffsetMap(columns, perColumnTableOffsets, perColumnNames);
    }

    /**
     * Called by createOffsetMap, and directly from the aggregation section.  given the columns
     * passed in, finds the global offset into the fully joined list of tables.  Since joins are done
     * left deep, the offset is also the local offset into the list of all columns.
     */
    public static int findColumnOffset(ColumnRef col, int[] perColumnTableOffsets,
                                    String[] perColumnNames) throws SQLException {
        for(int k = 0; k < perColumnTableOffsets.length; k++) {
            //System.out.println(col.getColumn()
            if(perColumnTableOffsets[k] == col.getTableOffset() &&
               perColumnNames[k].equals(col.getColumn())) {
                return k;
            }
        }

        throw new SQLException("Failed to find column : " + col + " (offset: " +
                    col.getTableOffset() + ") in per column lists");
    }

    protected SqlCondition collapseConditions(ArrayList conditions) {
          //  build a single big join condition
        if(conditions.size() == 1) {
            return (SqlCondition)conditions.get(0);
        } else {
            CompoundCondition group = new CompoundCondition(CompoundCondition.AND);
            for(int j = 0; j < conditions.size(); j++) {
                group.addSubExpr((SqlCondition)conditions.get(j));
            }

            return group;
        }
    }

    /**
     * Gets the simplest of single table queries, for now.  It does a decent
     * job of forming a plan for one table, but given that this will be blindly
     * nested to form a join it isn't ideal for later
     * @param t the transaction.  Since this object is created anew for each query, not sure
     *     why the transaction needs to be passed in.  Anyway, will change later when prepared
     *    statements are done right
     * @param tableOffset the id of the table being queried
     */
    protected TupleIterator getSingleTableQuery(Transaction t, int tableOffset) throws SQLException {
        //  TODO:  improve this with an index analysis
        //    and review some alternatives

        if(tableConditions[tableOffset] == null)
            return tables[tableOffset].getTableTupleIterator(t, false);

          //  builds a local column list, since this is a one table focus anyway
        ColumnList columnList = new ColumnList();
        tableConditions[tableOffset].accept(columnList);
        int[] offsets = new int[columnList.size()];
        for(int i = 0; i < columnList.size(); i++) {
            ColumnRef col = (ColumnRef)columnList.get(i);
            offsets[i] = tables[tableOffset].getColumnOffset(col.getColumn());
            //System.out.println("LOCAL OFFSET IS : " + offsets[i]);
        }

        return QueryProcessor.getBestSingleTablePlan(t, tableConditions[tableOffset],
                tables[tableOffset], offsets, false);  //  false at least so far, no update

        //return new RowFilterIterator(basicTableScan, tableConditions[tableOffset], offsets, false);
    }


    //protected TupleIterator

    /*******************************************************************
     * Sub steps of the main analyse step
     */

    protected void analyzeGroupByClause() throws SQLException {
        if(select.getGroupByExpressions() == null) {
            return;
        }

        for(int i = 0; i < select.getGroupByExpressions().size(); i++) {
            SqlExpression x = (SqlExpression)select.getGroupByExpressions().get(i);

            ArrayList groupColumns = new ArrayList();
            ArrayList groupFunctions = new ArrayList();
            x.performAnalysis(this, false, groupColumns, groupFunctions);
        }
    }

    protected void analyzeHavingClause() throws SQLException {
        //System.out.println("Aggregations before : " + aggregations);
        //System.out.println("Having is : " + select.getHavingCondition());

        if(select.getHavingCondition() == null) {
            return;
        }

        if(aggregations == null) {
            aggregations = new ArrayList();
        }

        int prevPos = aggregations.size();
        select.getHavingCondition().performHavingAnalysis(this, aggregations,
                select.getGroupByExpressions());

          //  if there are new group conditions, must analyze them so that their variables get bound
        if(aggregations.size() > prevPos) {
            ArrayList nestedColumns = new ArrayList();
            for(int i = prevPos; i < aggregations.size(); i++) {
                FunctionalExpression fx = (FunctionalExpression)aggregations.get(i);
                if(shouldAnalyzeGroupSubExpression(fx)) {
                    fx.getArgument(0).performAnalysis(this, false, nestedColumns, null);
                }
            }
        }
    }

    protected void analyzeOrderByClause() throws SQLException {
        if(select.getNumOrderByClauses() == 0)
            return;

        ArrayList columns = new ArrayList();
        ArrayList groupFunctions = new ArrayList();
        for(int i = 0; i < select.getNumOrderByClauses(); i++) {
            SqlExpression item = select.getOrderByExpression(i);
            item.performAnalysis(this, false, columns, groupFunctions);
        }

          //  TODO:  rewrite aggregations as a new class type, perhaps a subclass of a list type
        if(! groupFunctions.isEmpty()) {
            if(aggregations == null) {
                aggregations = new ArrayList();
            }

            ArrayList nestedColumns = null;
            for(int i = 0; i < groupFunctions.size(); i++) {
                FunctionalExpression fx = (FunctionalExpression)groupFunctions.get(i);
                aggregations.add(fx);
                if(shouldAnalyzeGroupSubExpression(fx)) {
                    if(nestedColumns == null) {
                        nestedColumns = new ArrayList();
                    }
                    fx.getArgument(0).performAnalysis(this, false, nestedColumns, null);
                }
            }
        }

          //  TODO:  do something about the columns for sure
    }

    /**
     * Scans through the select clauses.  Throws an error if there are any
     * expressions that have any lone unaggregated variables (or expressions)
     * that are not in the group by clause, -or- that the whole expression is
     * not a group by element<BR><BR>
     *
     * In the future there may be a case to support some extentions to this
     * they way oracle does.  Oracle for example supports:<BR>
     * select sin(a+b) from test group by a,b;<BR>
     * Because it finds the group by expression A+B.  One implementation of this
     * would be to add a pass that explicitly searched for legal group functions
     * by passing in the list of legal expressions and failing when if fails to find
     * a good match.  The same function would also search for group functions in the
     * same pass.
     */
    protected void analyzeSelectClause() throws SQLException {
        if(select.isStar()) {
            if(select.getGroupByExpressions() != null) {
                throw new SQLException("Cannot 'select *' when group by is present");
            }

              //  otherwise, fine
            return;
        }

        //  TODO:  deal with the SELECT TABLE1.*, TABLE2.* stuff here.  It should
        //  be easy enough to just expand in place. just scan for any such constructions and
        //  pass in a map of table/alias paired with a list of columns in order

          //  scan through the select items to verify that they are all ok
        ArrayList columns = new ArrayList();
        ArrayList groupFunctions = new ArrayList();
        int prevColumnPos = 0;
        int prevGroupPos = 0;
        for(int i = 0; i < select.getNumSelectItems(); i++) {
            SqlExpression item = select.getSelectItem(i);
            item.performAnalysis(this, false, columns, groupFunctions);

              //  if there is a group by clause, then check the single columns
            if(select.getGroupByExpressions() != null) {
                  //  check all of the columns
                boolean foundBadOne = false;
                ColumnRef badColumn = null;
                for(int j = prevColumnPos; j < columns.size(); j++) {
                    ColumnRef col = (ColumnRef)columns.get(j);
                    if(! isColumnInGroupByClause(col)) {
                        badColumn = col;
                        foundBadOne = true;
                        break;
                    }
                }

                  //  this is the part that only searches for exact group by equality,
                  //  but instead it could do a more rigorous search
                if(foundBadOne) {
                    boolean foundWholeExpression = false;
                    for(int j = 0; j < select.getGroupByExpressions().size(); j++) {
                        Object groupByX = select.getGroupByExpressions().get(j);
                        if(groupByX.equals(item)) {
                            //  TODO:  strongly consider explicitly marking or modifying
                            //  the item for quicker lookup later
                            // System.out.println("Found a whole expression match: " + groupByX);
                            foundWholeExpression  = true;
                            selectClauseHasSomeWholeGroupByExpressionReferences = true;
                            break;
                        }
                    }

                    if(! foundWholeExpression) {
                        throw new SQLException("Expression " + item + " not a legal group by");
                    }
                }
            }

            prevColumnPos = columns.size();
            prevGroupPos = groupFunctions.size();
        }

        //  TODO:  for each column, add it to the minimum list
        //System.out.println("Still need to add these to the list of columns: " + columns);

        if(groupFunctions.size() > 0) {
            //System.out.println("Found group functions : " + groupFunctions);

            this.aggregations = groupFunctions;
              //  TODO:  remove duplicates in the aggregations list!!!  like multiple count(*)

            if(columns.size() > 0 && select.getGroupByExpressions() == null) {
                throw new SQLException("Columns " + columns + " not group by expressions");
            }

            //  analyse the inside of the group expression
            //     add all internal variable to the list of required
            //     see if there are any nested
            //      - can skip if all are requested by passing null
            //      but must determine if there are subgroup functions

            //  build up the list of columns needed from all of the group functions found.
            ArrayList nestedColumns = new ArrayList();
            ArrayList nestedGroupFunctions = new ArrayList();
            for(int i = 0; i < groupFunctions.size(); i++) {
                FunctionalExpression fe = (FunctionalExpression)groupFunctions.get(i);
                if(shouldAnalyzeGroupSubExpression(fe)) {
                    for(int j = 0; j < fe.getNumArguments(); j++) {
                        SqlExpression nestedItem = fe.getArgument(j);
                        nestedItem.performAnalysis(this, false, nestedColumns, nestedGroupFunctions);
                    }
                }
            }

            //  TODO:  this is where we would add nestedColumns to the list of used columns
            //System.out.println("Still need to add these to the list of columns: " + nestedColumns);

            if(nestedGroupFunctions.size() > 0) {
                //  TODO:  when we do support these, we need to assume check that all of the previous
                //  group by's are all either nested on a group by or a double aggregation function
                throw new SQLException("Nested aggregation functions not currently supported.  Not hard, just lazy.");
            }
        }

        //  if there were nested group functions, check again to make sure there are no MORE
        //  and at least add the new columns to the ones required
    }


    private boolean shouldAnalyzeGroupSubExpression(FunctionalExpression fe) throws SQLException {
        /*WAS:
        if(fe.getNumArguments() == 0) {
            return false;
        } else if(fe.getNumArguments() != 1) {
            throw new SQLException("Group function '" + fe.getFunctionName() + "' does not support multiple arguments.");
        }
        */

        if(fe.getNumArguments() == 0) {
            return false;
        } else if(! fe.getFunction().getAggregationFunction().supportsValues(fe.getNumArguments())) {
            throw new SQLException("Group function '" + fe.getFunctionName() + "' does not support multiple arguments.");
        }

        return true;
    }

    /**
     * True if the column is mentioned exactly alone in the group by clause.
     * This is moved off into a method in case the method for determining changes.<BR>
     * This method uses the ArrayList directly to save an extra cast.
     */
    private boolean isColumnInGroupByClause(ColumnRef col) {
        for(int i = 0; i < select.getGroupByExpressions().size(); i++) {
            Object o = select.getGroupByExpressions().get(i);
            if(o instanceof ColumnExpression) {
                ColumnExpression cx = (ColumnExpression)o;
                if(cx.getColumn().getTableOffset() == col.getTableOffset() &&
                   cx.getColumn().getColumn().equals(col.getColumn())) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Will Determine basic costs for all tables given the singleTableQueries.  This will
     * be doing only very basic analysis.
     */
    protected void determineBasicCosts() throws SQLException {
        for(int i = 0; i < tables.length; i++) {
            long numRows = tables[i].getEstimatedNumRows();

            double selectivity = 1;
            if(tableConditions[i] != null) {
                selectivity = tableConditions[i].getSelectivity(tables);
            }

            if(selectivity > numRows) {
                sizeEstimates[i] = 1;
            } else {
                sizeEstimates[i] = (long)((double)numRows / selectivity);
            }

            //System.out.println("size estimated to " + sizeEstimates[i] + " because of " + tableConditions[i]);
        }
    }

    /**
     * The strategy for determining join order is take from "Database System
     * Implementation"  Garcia-Molina, Ullman, Widom.  It does not currently
     * have all of the bells and whistles that I would like, but it might one day.
     */
    protected void findFirstJoin() throws SQLException {
        long bestCost = Long.MAX_VALUE;
        int a = -1;
        int b = -1;

        if(tables.length == 1) {
            //  TODO:  FROM HERE to /HERE (below) will probably need to be removed when cleanup happends
            joinSequence[0] = 0;

            //  /HERE
            return;
        }

         /* //  optimization for the simple case  Doesn't fill in costs, so it is mothballed for now
        if(tables.length == 2) {
            if(sizeEstimates[0] < sizeEstimates[1]) {
                joinSequence[0] = 0;
                joinSequence[1] = 1;
            } else {
                joinSequence[0] = 1;
                joinSequence[1] = 0;
            }

            addFirstJoinGroupToSequence(0, 1);
            return;
        }
        */

        //  for each pair of tables determine cost
        for(int i = 1; i < tables.length; i++) {
            for(int j = 0; j < i; j++) {
                ArrayList list = getEquiJoins(i, j);

                  //  even close enough in the case of a non equi join
                long cost = sizeEstimates[i] * sizeEstimates[j];
                if(list != null) {
                    long denominator = 1;
                    for(int k = 0; k < list.size(); k++) {
                        BasicCondition c = (BasicCondition)list.get(k);

                        ColumnRef leftCol = ((ColumnExpression)c.getLeftExpression()).getColumn();
                        ColumnRef rightCol = ((ColumnExpression)c.getLeftExpression()).getColumn();

                        denominator *= Math.max(getNumColumnValues(leftCol), getNumColumnValues(rightCol));
                    }

                    if(denominator > cost) {
                        cost = 1;
                    } else {
                        cost = cost / denominator;
                    }
                }

                if(cost < bestCost) {
                    bestCost = cost;
                    a = i;
                    b = j;
                }
            }
        }

        if(sizeEstimates[a] < sizeEstimates[b]) {
            joinSequence[0] = a;
            joinSequence[1] = b;
        } else {
            joinSequence[0] = b;
            joinSequence[1] = a;
        }

        joinedAlready[a] = true;
        joinedAlready[b] = true;

        curCost = bestCost;

        postJoinSizeEstimates[0] = sizeEstimates[joinSequence[0]];
        postJoinSizeEstimates[1] = curCost;

        addFirstJoinGroupToSequence(a, b);

        //System.out.println("We have decided to start join with " + joinSequence[0] + " to " + joinSequence[1] + " at a cost of " + curCost);
    }

    private void addFirstJoinGroupToSequence(int a, int b) {
          //  copy the equi join clauses for faster finding later
        ArrayList list = getEquiJoins(a, b);
        if(list != null) {
            joinClausesInSequence[0] = new ArrayList();
            for(int i = 0; i < list.size(); i++) {
                joinClausesInSequence[0].add(list.get(i));
            }
        }
    }

    private long getNumColumnValues(ColumnRef col) throws SQLException {
        return tables[col.getTableOffset()].getColumnStatistics(col.getColumn()).
                getEstimatedNumDistinctValues();
    }

    /**
     * Fill in the rest of the join sequence, iteratively.
     *
     * TODO:  add something that immediately finds the last one when appropriate,
     * ie, scan when there is only one damn table left!
     */
    protected void completeJoinSequence() throws SQLException {
        while(curSequencePos < tables.length) {
            long bestCost = Long.MAX_VALUE;
            int nextToJoin = -1;

              //  for each unjoined table, see how good it is
            for(int i = 0; i < tables.length; i++) {
                if(joinedAlready[i])
                    continue;

                long cost = curCost * sizeEstimates[i];
                long denominator = 1;

                  //  find all clauses between i and the full joined set, and multiply
                  //  the denominator up
                for(int j = 0; j < tables.length; j++) {
                    if(! joinedAlready[j])
                        continue;

                    ArrayList list = getEquiJoins(i, j);
                    if(list == null)
                        continue;

                    for(int k = 0; k < list.size(); k++) {
                        BasicCondition c = (BasicCondition)list.get(k);
                        //System.out.println(" testing weight of " + c);

                        ColumnRef leftCol = ((ColumnExpression)c.getLeftExpression()).getColumn();
                        ColumnRef rightCol = ((ColumnExpression)c.getLeftExpression()).getColumn();

                        denominator *= Math.max(getNumColumnValues(leftCol), getNumColumnValues(rightCol));
                    }
                }

                if(denominator > cost) {
                    cost = 1;
                } else {
                    cost = cost / denominator;
                }

                if(cost < bestCost) {
                    bestCost = cost;
                    nextToJoin = i;
                }
            }

              //  build the next entry in the ArrayList list = getEquiJoins(a, b);
            for(int i = 0; i < tables.length; i++) {
                if(! joinedAlready[i])
                    continue;

                ArrayList list = getEquiJoins(i, nextToJoin);
                if(list == null)
                    continue;

                if(joinClausesInSequence[curSequencePos - 1] == null) {
                    joinClausesInSequence[curSequencePos - 1] = new ArrayList();
                }

                for(int k = 0; k < list.size(); k++) {
                    joinClausesInSequence[curSequencePos - 1].add(list.get(k));
                }
            }
            //System.out.println("Build next join list in sequence: " + joinClausesInSequence[curSequencePos - 1]);

              //  set the variables
            joinedAlready[nextToJoin] = true;
            postJoinSizeEstimates[curSequencePos] = bestCost;
            joinSequence[curSequencePos++] = nextToJoin;
            curCost = bestCost;
            //System.out.println("NEXT IS " + nextToJoin);
        }
    }

    /*******************************************************************
     * Methods called from externally.  (And their private helpers)
     */

    public FunctionManager getFunctionManager() throws SQLException {
        return fm;
    }

    /**
     * Called from the analysis phase of the condition, assigns a table offset to this
     * column reference.  It also verifies that the column exists.
     *
     * @throws SQLException if the column or table does not exist or the column is ambiguous
     */
    public void assignTableOffset(ColumnExpression columnExpr) throws SQLException {
        ColumnRef columnRef = columnExpr.getColumn();

          //  if the table is not null, verify it and the column
        if(columnRef.getTable() != null) {
            String fullName = columnRef.getTable();
            if(columnRef.getSchema() != null && ! columnRef.getSchema().equals(defaultSchema)) {
                fullName = columnRef.getSchema() + '.' + columnRef.getTable();
            }

            Integer offset = (Integer)tableOffsets.get(fullName);

            if(offset == null) {
                if(columnExpr.isBoundToSequence()) {
                    return;
                }

                throw new SQLException("Table '" + columnRef.getTable() + "' not in query");
            }

            if(offset.intValue() == -1) {
                throw new SQLException("Column " + columnRef + " ambiguously defined");
            }

            if(tables[offset.intValue()].getColumnOffset(columnRef.getColumn()) == -1) {
                throw new SQLException("Column " + columnRef + " invalid for table " +
                        select.getTables().get(offset.intValue()));
            }

            columnRef.setTableOffset(offset.intValue());
        } else /*  search for the column!  */ {
            boolean beenAssigned = false;
            int assignedTableOffset = -1;

            for(int i = 0; i < getNumTables(); i++) {
                boolean hasColumn = getTable(i).hasColumnNamed(columnRef.getColumn());
                if(hasColumn) {
                    if(beenAssigned) {
                        throw new SQLException("Column " + columnRef + " ambiguous between tables " +
                                tableRefs.get(assignedTableOffset) + " and " +
                                tableRefs.get(i));
                    }

                    beenAssigned = true;
                    assignedTableOffset = i;
                }
            }

            if(! beenAssigned) {
                throw new SQLException("Column \"" + columnRef + "\" invalid");
            }

            columnRef.setTableOffset(assignedTableOffset);
        }
    }


    /**
     * Called from inside the condition.performAnalysis.  For now, stores the SqlCondition
     * but uses the a and b refs immediately for sparing another lookup.  <I>The equi join clause
     * is expected to have been orders such that the left clause has the lower tableOffset.</I>
     */
    public void addEquiJoinCondition(BasicCondition joinCondition, ColumnRef a, ColumnRef b) {
        //System.out.println("addEquiJoinCondition called w/ " + a + " = " + b);
        addEquiJoins(a.getTableOffset(), b.getTableOffset(), joinCondition);
    }

    /**
     * This top level AND condition was found to have no dependent columns, and so for now is treated as a
     * constant boolean filter.  Later we may treat these differently, (ie a < RANDOM() but for now this is
     *   going to be fine)
     * @param constant
     */
    public void addConstantGlobalCondition(SqlCondition constant) {
        //System.out.println("addEquiJoinCondition called w/ " + a + " = " + b);
        globalConstantCondition = gatherAnd(globalConstantCondition, constant);
    }

    /**
     * Helper method, essentially returns a new condition for (base AND next) in an efficient way
     * @param base
     * @param next
     * @return
     */
    public static SqlCondition gatherAnd(SqlCondition base, SqlCondition next) {
        if(base == null) {
            return next;
        }

        if(base instanceof CompoundCondition) {
            CompoundCondition compound = (CompoundCondition)base;
            if(compound.getCompoundOp() == CompoundCondition.AND) {
                compound.addSubExpr(next);
                return base;
            }
        }

        CompoundCondition compound = new CompoundCondition(CompoundCondition.AND, base);
        compound.addSubExpr(next);
        base = compound;
        return base;
    }

    /**
     * The condition c passed is AND'ed with the previous ones for this tables.
     * This might be a good time to decide on the best single table plan to use,
     * or at least keep the best one.
     */
    public void addSingleTableCondition(int tableOffset, SqlCondition c) {
        tableConditions[tableOffset] = gatherAnd(tableConditions[tableOffset], c);
    }

    /**
     * TODO:  this could pass in an int[] of tables effected so that
     *      the filtering could happen any time after the joining of
     *      the tables involved.  Could drastically save performance time
     *      in the future
     */
    public void addPostJoinCondition(SqlCondition c) {
        postJoinCondition = gatherAnd(postJoinCondition, c);
    }

    /**************************************************************8
     * access methods
     */

    /**
     * Return all the equi joins between offsets a and b
     */
    private ArrayList getEquiJoins(int a, int b) {
          //  make sure a > b
        if(a < b) {
            int x = a;
            a = b;
            b = x;
        }

        return joinClauses[getTriangleSize(a - 1) + b];
    }

    private void addEquiJoins(int a, int b, BasicCondition equiCond) {
          //  make sure a > b
        if(a < b) {
            int x = a;
            a = b;
            b = x;
        }

        int pos = getTriangleSize(a - 1) + b;

        if(joinClauses[pos] == null) {
            joinClauses[pos] = new ArrayList();
        }

        joinClauses[pos].add(equiCond);
    }

    /**
     * Determines the number of entries in the triangle
     */
    protected final int getTriangleSize(int i) {
        return i * (i + 1) / 2;
    }

    /**
     *
     */
    private int getNumTables() {
        return tableRefs.size();
    }

    private Table getTable(int offset) {
        return tables[offset];
    }
}

//  What are you doing reading down here?