/*
 * Created by IntelliJ IDEA.
 * User: Administrator
 * Date: Sep 9, 2001
 * Time: 5:25:30 PM
 * To change template for new interface use
 * Code Style | Class Templates options (Tools | IDE Options).
 */
package freebase.sql.plan;

import freebase.api.ColumnInfo;
import freebase.api.LockableTupleIterator;
import freebase.api.TableLockManager;
import freebase.api.TupleIterator;
import freebase.catalog.Catalog;
import freebase.catalog.Index;
import freebase.catalog.Table;
import freebase.engine.ScanRegionList;
import freebase.engine.TableTupleIterator;
import freebase.engine.Transaction;
import freebase.server.ExecutionPlan;
import freebase.sql.*;
import freebase.sql.function.FunctionManager;
import freebase.sql.plan.visitor.ColumnList;
import freebase.sql.query.RowFilterIterator;
import freebase.util.ErrorCode;
import freebase.util.SqlTypes;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * Takes the query and produces a physical plan.  This might become an abstract class,
 * containing the core functionality of converting sql -> logical plans -> physical plans.
 * All of the various steps are fairly specific to a database, but there is a lot of common
 * functionality too.  For sure, put hooks for subclass calls.<BR><BR>
 *
 * probably should also have a method for creating plans for updates and deletes
 * TODO:  rename StatementProcessor ?
 */
public class QueryProcessor {
    /**
     * Way to quickly turn off indexes for basic unit testing, and sanity checking
     * TODO:  expose this through the config files for sure!!
     */
    public static boolean USE_INDEXES = true;
    public static boolean MAKE_INDEXES_SELECTIVE = true;  //  if false, all indexes are full scan

    Catalog catalog;

    //  TODO:  set this exernally
    FunctionManager fm = new FunctionManager();

    /**
     * Create a Physical exection plan
     */
    public TupleIterator executeQuery(Transaction t, String defaultSchema,
                            SelectStatement query) throws SQLException {
        validateQuery(t, query);
        //  convert to logical plan
        //  optimize logical plan

        //  convert to physical plan  (possibly the only part that is database specific)

          //  NOTE:  appears that this code doesn't work with: SELECT count(*)
        //  TODO:  dangerous, doesn't lock at all
        if(query.getTables().size() == 0) {  //  Used to be 1!!!  This means it never gets called
            Table table = catalog.getTable((TableRef)query.getTables().get(0), defaultSchema);

              //  very very basic first attempt to implement queries
            if(query.getCondition() != null) {
                int[] offsetMap = buildOffsetMap(query.getCondition(), table);

                if(USE_INDEXES) {
                    TupleIterator indexPlan = getBestSingleTablePlan(t, query.getCondition(), table, offsetMap, false);
                    if(indexPlan != null) {
                        System.out.println("Using index");
                        return indexPlan;
                    }

                    //System.out.println(query.getCondition() + "    Class:" + query.getCondition().getClass());
                }

                return new RowFilterIterator(table.getTableTupleIterator(t, false),
                    query.getCondition(), offsetMap, false);
            }

            return new TableTupleIterator(t, table, false);
        } else /* run the query analysis stuff..  eventually all will use it */  {
            QueryAnalysis qa = new QueryAnalysis(catalog, fm, defaultSchema, query);

            qa.lockResources();
            try{
                qa.analyzeQuery();

                return qa.executeQuery(t);
            } finally {
                qa.unlockResources();
            }

            //System.out.println("QueryProcessor: currently does not support multi table queries [" + query.getTables() + "]");
            //throw new SQLException("NOT SUPPORTED");
        }

        //return null;
    }

    /**
     * If returns null, then use the plan.  This method is a very first attempt.
     * It uses brute force to search for a good index.
     * <p>
     * NOTE:   currently used by QueryAnalysis as well.  Not sure about that long term
     * @param cond currently requires this cond to be on one table only
     */
    static TupleIterator getBestSingleTablePlan(Transaction t, SqlCondition cond, Table table,
                                                 int[] offsetMap, boolean update) throws SQLException {
        LockableTupleIterator basePass = null;

        if(USE_INDEXES) {
            //System.out.println("trying to determine if index is good");

            ArrayList indexes = table.getAllIndexes();
            for(int i = 0; i < indexes.size(); i++) {
                Index index = (Index)indexes.get(i);
                ScanRegionList scanRegions = index.findScanRegions(cond);
                if(scanRegions != null) {
                    if(! MAKE_INDEXES_SELECTIVE) {
                        scanRegions = null;
                    }

                    //  TODO:  implement something that extracts at least the min/max from the condition,
                    //  but preferably a set of regions to scan
                    basePass = index.getIndexScan(t, update, scanRegions);
                    break;
                }
            }
        }

        if(basePass == null) {
            basePass = table.getTableTupleIterator(t, update);
        } else {
            //System.out.println("found one!");
        }

        return new RowFilterIterator(basePass, cond, offsetMap, update);
    }

    /**
     * TODO:  implement this.  A lot of it is already ok but consider the
     * types.  Could happen in an analysis phase perhaps
     */
    private void validateQuery(Transaction t, SelectStatement query) {
        //  check relation use (pg 336 of DB implementation)
        //  check and resolve attribute uses
        //  check types
    }


    /**
     * Builds an appropriate offset map that maps the given condition
     * into the given single table (singles only for now)
     */
    private int[] buildOffsetMap(SqlExpression expr, Table table)
            throws SQLException {
        if(expr == null)
            return new int[0];

        ColumnList columns = new ColumnList();
        expr.accept(columns);

        int[] offsetMap = new int[columns.size()];
        for(int i = 0; i < offsetMap.length; i++) {
            int offset = table.getColumnOffset(((ColumnRef)columns.get(i)).getColumn());
            if(offset == -1)
                throw new SQLException("Column '" + columns.get(i) + "' not recognized"); //  TODO, assign a code
            offsetMap[i] = offset;
        }

        return offsetMap;
    }

    /**
     * Performs the update.  Currently does so with a table scan only.
     */
    public long executeUpdate(Transaction t, String defaultSchema,
                              UpdateStatement update) throws SQLException {
        t.clearNumUniqueContentions();

        String schema = update.getSchema();
        if(schema == null)
            schema = defaultSchema;

        Table tab = catalog.getTable(schema, update.getTableName());
        TupleIterator it = null;

        int[][] setExprOffsetMaps = new int[update.getNumSets()][];
        int[] setCols = new int[update.getNumSets()];
        ColumnInfo[] setColInfo = new ColumnInfo[update.getNumSets()];
        int[] condOffsetMap;
        try{
              //  use lock needed until done dealing with columns
            tab.getUseLock().lockShared();
            try{
                for(int i = 0; i < update.getNumSets(); i++) {
                    ColumnRef col = update.getSetColumn(i);
                    SqlExpression expr = update.getSetExpr(i);

                    /*  TODO:  add this checking back in and make it work!
                    if(col.getTable().equals(update.getTableName()))
                        throw new SQLException("Table in '" + col + "' not valid for this query");
                    */

                    setExprOffsetMaps[i] = buildOffsetMap(expr, tab);
                    int oi = tab.getColumnOffset(col.getColumn());
                    setColInfo[i] = tab.getColumnInfo(oi);
                    if(oi == -1) {
                        throw new SQLException("Column '" + col.getColumn() + "' not found in table " + update.getTableName());
                    }
                    setCols[i] = oi;
                }

                  //  check for duplicates  (make an external call to
                  //  and array helper
                for(int i = 0; i < setCols.length - 1; i++) {
                    for(int j = (i + 1); j < setCols.length; j++) {
                        if(setCols[i] == setCols[j]) {
                            throw new SQLException("Duplicate columns: " +
                                update.getSetColumn(i) + " and " + update.getSetColumn(j));
                        }
                    }
                }

                condOffsetMap = buildOffsetMap(update.getCondition(), tab);
                it = getBestSingleTablePlan(t, update.getCondition(), tab, condOffsetMap, true);
                //it = new TableTupleIterator(t, tab, true);  //  <-- will fail if already deleted
            } finally {
                tab.getUseLock().unlockShared();
            }

              //  validate all expressions -> columns and conditions before letting go of lock
              //  note, is here because the set columns are protected, ie they are copied out
            for(int i = 0; i < update.getNumSets(); i++) {
                  //  validate column set
                SqlExpression expr = update.getSetExpr(i);
                ColumnInfo exprType = expr.getColumnInfo(it, setExprOffsetMaps[i]);
                ColumnInfo columnType = setColInfo[i];
                SqlTypes.isCompatibleType(exprType, columnType);
            }

              //  TODO:  validate conditions (which currently don't happen)

            //it = new RowFilterIterator(it, update.getCondition(), offsetMap, true);
            tab.lockTable(t, TableLockManager.ROW_EXCLUSIVE_LOCK, false);

              //  perform the actual updates
            long count = 0;
            while(it.next()) {
                count++;

                for(int i = 0; i < setCols.length; i++) {
                      //  TODO:  rework the values
                    Comparable val = update.getSetExpr(i).evaluate(it, setExprOffsetMaps[i]);
                    val = SqlTypes.convertValue(val, setColInfo[i]);
                    it.setValue(setCols[i], val); //new SqlValue(-1, (Comparable)val));
                }
            }

            if(it != null) {
                it.close();
                it = null;
            }

            if(t.getNumUniqueContentions() != 0) {
                ErrorCode.throwException(ErrorCode.UNIQUE_KEY_VIOLATION);
            }

            return count;

        }finally{
            if(it != null) {
                it.close();
            }
        }

    }

    /**
     * TODO:  move this to Table object?  Anyway, consider it!
     * @param t
     * @param defaultSchema
     * @param delete
     * @return
     * @throws SQLException
     */
    public long executeDelete(Transaction t, String defaultSchema,
                              DeleteStatement delete) throws SQLException {
        String schema = delete.getSchema();
        if(schema == null) {
            schema = defaultSchema;
        }
        Table tab = catalog.getTable(schema, delete.getTableName());

        tab.getUseLock().lockShared();  //  TODO:  check that the table is not deleted (part of the use lock automatically?)
        TupleIterator it = null;
        int[] offsetMap = null;
        try{
            offsetMap = buildOffsetMap(delete.getCondition(), tab);

            it = getBestSingleTablePlan(t, delete.getCondition(), tab, offsetMap, true);
            //it = new TableTupleIterator(t, tab, true);
        }finally {
            tab.getUseLock().unlockShared();
        }
        try{
            //it = new RowFilterIterator(it, delete.getCondition(), offsetMap, true);

            tab.lockTable(t, TableLockManager.ROW_EXCLUSIVE_LOCK, false);

            long count = 0;
            while(it.next()) {
                count++;
                it.deleteTuple();
            }

            return count;
        }finally{
            it.close();
        }
    }

    public void setCatalog(Catalog cat) {
        this.catalog = cat;
    }
}


class EmptyTupleIterator implements TupleIterator {
    public void reset() throws SQLException { }

    public boolean next() throws SQLException {
        return false;
    }

    public Comparable getValue(int i) throws SQLException {
        return null;
    }

    public void close() throws SQLException { }

    public void setValue(int i, Comparable value) throws SQLException { }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        return new ExecutionPlan("Explictly Empty Set [For Debugging]");
    }

    public int getNumColumns() throws SQLException {
        return 0;
    }

    public String getColumnName(int i) throws SQLException {
        return null;
    }

    /**
     * NOTE:  this will probably not work!
     */
    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return null;
    }

    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException { }
}