/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 18, 2002
 * Time: 4:50:20 PM
 */
package freebase.sql.query;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;
import freebase.server.ExecutionPlan;
import freebase.sql.expression.FunctionalExpression;
import freebase.sql.function.AggregationFunction;

import java.sql.SQLException;

/**
 * Does the pass over the data to get aggregation results.  Currently does not
 * do any projection itself, but may at some point if it is deemed to be much faster.
 * In the case of group by, it is assumed that the data is presorted coming in.<BR><BR>
 *
 * This will do the distinct values internally.  The results of this pass is the group by
 * columns as the first group, then followed by the group function results.<BR><BR>
 *
 * The current implementation is aware of the nested internal expressions<BR><BR>
 *
 * results are:  all group columns first, then all aggregation column values
 */
public class Aggregation implements TupleIterator {
    /**
     * The value we pass into count(*) because null would give us zero entries
     */
    final static Integer DUMMY_VALUE = new Integer(0);

    TupleIterator source;
    int[] groupByColumns;
    int numGroupColumns = 0;

    FunctionalExpression[] aggregateFunctions;
    int[] subExprOffsetMap;
    int numAggFunctions = 0;

    Comparable[] latestValues = null;
    boolean hasStarted = false;
    boolean hasMore = true;

    /**
     * Created if there are any distinct values, is always as long as aggregateFunctions[]
     * but only has objects where needed.
     */
    SingleValueUniqueness[] distinctBuffers;

    /**
     * toAggregate is the underlying tuple set to aggregate, groupColumns
     * is the list of columns to aggregate by.
     *
     * TODO:  deal with aliasing, here or elsewhere
     */
    public Aggregation(TupleIterator toAggregate, int[] groupByColumns,
                       FunctionalExpression[] aggregateFunctions, int[] subExprOffsetMap)
            throws SQLException {
        this.source = toAggregate;
        this.groupByColumns = groupByColumns;
        this.aggregateFunctions = aggregateFunctions;
        this.subExprOffsetMap = subExprOffsetMap;

        if(groupByColumns != null) {
            numGroupColumns = groupByColumns.length;
        }

        if(aggregateFunctions != null) {
            numAggFunctions = aggregateFunctions.length;
            for(int i = 0; i < aggregateFunctions.length; i++) {
                if(aggregateFunctions[i].isDistinct()) {
                    if(distinctBuffers == null) {
                        distinctBuffers = new SingleValueUniqueness[aggregateFunctions.length];
                    }

                    distinctBuffers[i] = new SingleValueUniqueness();
                }
            }
        }

        latestValues = new Comparable[numGroupColumns + numAggFunctions];
    }

    /*******************************************************
     * Tuple Iterator operations
     */

    /**
     * We are going to go back over the data!
     */
    public void reset() throws SQLException {
        source.reset();
        hasStarted = false;
    }

    public Comparable getValue(int i) throws SQLException {
        return latestValues[i];
    }

    /**
     * Fussy, but to optimize this might be to take out the extra castings from the
     * sole sub expression inside each functional expression.
     */
    public boolean next() throws SQLException {
        if(! hasMore) {
            return false;
        }

        if(! hasStarted) {
            hasMore = source.next();
            if(! hasMore)
                return false;
        }

          //  set the local values to be equal to the current values
        for(int i = 0; i < numGroupColumns; i++) {
            latestValues[i] = source.getValue(groupByColumns[i]);
        }

        while(true) {
            if(! hasStarted) {
                hasStarted = true;
            } else {
                hasMore = source.next();
            }

            boolean groupIsDone = ! hasMore;  //  if no more, then the group is done
            if(hasMore) {
              //  if needed, check that the values have changed

                for(int i = 0; i < numGroupColumns; i++) {
                    Object val = source.getValue(groupByColumns[i]);
                    if(latestValues[i] == null) {
                        if(val != null) {
                            groupIsDone = true;
                            break;
                        }
                    } else if(! latestValues[i].equals(val)) {
                        groupIsDone = true;
                        break;
                    }
                }
            }

            if(groupIsDone) {
                  //  set the values of the current row
                for(int i = 0; i < numAggFunctions; i++) {
                    AggregationFunction f = aggregateFunctions[i].getFunction().getAggregationFunction();

                      //  in the distinct case, need to process the data here and now!
                    if(aggregateFunctions[i].isDistinct()) {
                        while(distinctBuffers[i].nextUnique()) {
                            f.addValue(distinctBuffers[i].getValue());
                        }
                        distinctBuffers[i].reset();
                    }

                    latestValues[i + numGroupColumns] = f.getResult();
                    f.reset();
                }
            }

            if(hasMore) {
                  //  apply aggregation functions here
                for(int i = 0; i < numAggFunctions; i++) {
                    AggregationFunction f = aggregateFunctions[i].getFunction().getAggregationFunction();
                    if(aggregateFunctions[i].isDistinct()) {
                        Comparable val = aggregateFunctions[i].getArgument(0).evaluate(source, subExprOffsetMap);
                        if(val != null)  //  <-- note that we are changing our tune here slightly, probably ok
                            distinctBuffers[i].addValue(val);
                    } else if(aggregateFunctions[i].isStar()) {
                        f.addValue(DUMMY_VALUE);  //  whatever, just a dummy
                    } else {
                        if(aggregateFunctions[i].getNumArguments() == 1) {
                            f.addValue(aggregateFunctions[i].getArgument(0).evaluate(source, subExprOffsetMap));
                        } else {
                              //  handle the multi-value input case
                            Comparable[] c = new Comparable[aggregateFunctions[i].getNumArguments()];
                            for(int j = 0; j < c.length; j++) {
                                c[j] = aggregateFunctions[i].getArgument(j).evaluate(source, subExprOffsetMap);
                            }
                            f.addValues(c);
                        }
                    }
                }
            }

            if(groupIsDone)
                return true;
        }
    }

    /**
     * Close any open distinctness buffers as well as the underlying data source
     */
    public void close() throws SQLException {
        source.close();

        if(distinctBuffers != null) {
            for(int i = 0; i < distinctBuffers.length; i++) {
                if(distinctBuffers[i] != null) {
                    distinctBuffers[i].close();
                }
            }
        }
    }

    public int getNumColumns() throws SQLException {
        return numGroupColumns + numAggFunctions;
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        StringBuffer sb = new StringBuffer("Aggregation [");
        for(int i = 0; i < aggregateFunctions.length; i++) {
            if(i != 0) {
                sb.append(',');
            }
            sb.append(aggregateFunctions[i]);
        }

        if(numGroupColumns != 0) {
            sb.append("]  Group By [");
            for(int i = 0; i < numGroupColumns; i++) {
                if(i != 0)
                    sb.append(',');
                sb.append(source.getColumnName(groupByColumns[i]));
            }
            sb.append(']');
        } else {
            sb.append("]  All");
        }


        return new ExecutionPlan(sb.toString(), source.getExecutionPlan());
    }

    public String getColumnName(int i) throws SQLException {
        if(i < numGroupColumns) {
            return source.getColumnName(groupByColumns[i]);
        } else {
            return aggregateFunctions[i - numGroupColumns].toString();
        }
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        if(i < numGroupColumns) {
            return source.getColumnInfo(groupByColumns[i]);
        } else {
            return aggregateFunctions[i - numGroupColumns].getColumnInfo(source, subExprOffsetMap);
            //return aggregateFunctions[i - numGroupColumns].getColumnInfo() toString();  //  TODO:  send to function
        }
    }
    /***********************************************************
     * Illegal operations
     */

    /**
     *
     */
    public void setValue(int i, Comparable value) throws SQLException {
        throw new UnsupportedOperationException("Not legal to update though an aggregation");
    }

    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException {
        throw new UnsupportedOperationException("Not legal to delete though an aggregation");
    }
}
