/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 23, 2002
 * Time: 6:20:17 PM
 */
package freebase.sql.query;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;
import freebase.server.ExecutionPlan;

import java.sql.SQLException;

/**
 * Takes a stream sorted by all columns and eliminates duplicate tuples.
 */
public class EliminateDuplicates implements TupleIterator {
    TupleIterator source;

    Comparable[] lastData = null;

    boolean hasStarted = false;

    public EliminateDuplicates(TupleIterator source) throws SQLException {
        this.source = source;
        lastData = new Comparable[source.getNumColumns()];
    }

    public void reset() throws SQLException {
        source.reset();
        hasStarted = true;
    }

    public Comparable getValue(int i) throws SQLException {
        return lastData[i];
    }

    public boolean next() throws SQLException {
        while(source.next()) {
            if(! hasStarted) {
                copySourceLocally();
                hasStarted = true;
                return true;
            }

            boolean isDifferent = false;
            for(int i = 0; i < lastData.length; i++) {
                if(lastData[i] == null) {
                    if(source.getValue(i) != null) {
                        isDifferent = true;
                        break;
                    }
                } else {
                    if(! lastData[i].equals(source.getValue(i))) {
                        isDifferent = true;
                        break;
                    }
                }
            }

            if(isDifferent) {
                copySourceLocally();
                return true;
            }
        }

        return false;
    }

    /**
     *  Makes the current image the same as the source
     */
    private void copySourceLocally() throws SQLException {
        for(int i = 0; i < lastData.length; i++) {
            lastData[i] = source.getValue(i);
        }
    }

    public void close() throws SQLException {
        source.close();
    }

    public int getNumColumns() throws SQLException {
        return source.getNumColumns();
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        return new ExecutionPlan("Eliminate Duplicates", source.getExecutionPlan());
    }

    public String getColumnName(int i) throws SQLException {
        return source.getColumnName(i);
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return source.getColumnInfo(i);
    }
    /*************************************************************
     * update functions that are not legal
     */

    /**
     *
     */
    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException {
        throw new UnsupportedOperationException("deleteTuple not supported on EliminateDuplicates");
    }

    public void setValue(int i, Comparable value) throws SQLException {
        throw new UnsupportedOperationException("setValue not supported on EliminateDuplicates");
    }
}
