package freebase.sql.query;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;
import freebase.server.ExecutionPlan;
import freebase.util.CompareUtil;

import java.sql.SQLException;

/**
 * Pulls the first n rows from the underlying TupleIterator and throws the rest away.  TODO:  later add a mode
 * that can execute another TupleIterator (count(*) * PERCENT from same query) and then use that as the value.
 * NOTE:  can't just pass in one source for that one because count might take a completely different physical
 * plan. (index scan for example)
 * Copyright 2004 JJ Furman
 * Time: 9:17:42 AM
 */
public class FirstRows implements TupleIterator {
    TupleIterator source;

    long topNum;
    long numSeen = 0;

    int[] withTiesColumns = null;

    Comparable[] cutoffValues;

    /**
     *
     * @param source
     * @param topNum
     * @param withTiesSortColumns null if not WITH TIES
     */
    public FirstRows(TupleIterator source, long topNum, int[] withTiesSortColumns) {
        this.source = source;
        this.topNum = topNum;
        this.withTiesColumns = withTiesSortColumns;
        if(withTiesSortColumns != null) {
            cutoffValues = new Comparable[withTiesSortColumns.length];
        }
    }


    public void reset() throws SQLException {
        numSeen = 0;
        source.reset();
    }

    public boolean next() throws SQLException {
        numSeen++;
        if(numSeen > topNum) {
            if(withTiesColumns != null) {
                return checkTies();
            }

            return false;
        } else if(numSeen == topNum) {
            if(withTiesColumns != null) {
                boolean areMore = source.next();
                if(areMore) {
                    recordCutoffValues();
                }
                return areMore;
            }
        }

        return source.next();
    }

    /**
     * Checks the current columns with the final values before the cuttoff to see if there
     * are still ties
     * @return true if there are more rows in source AND the sort columns data matches the last.
     */
    private boolean checkTies() throws SQLException {
        boolean areMore = source.next();
        if(! areMore) {
            return false;
        }

        for(int i = 0; i < withTiesColumns.length; i++) {
            if(! CompareUtil.isEqual(cutoffValues[i], getValue(withTiesColumns[i]))) {
                return false;
            }
        }

        return true;
    }

    private void recordCutoffValues() throws SQLException {
        for(int i = 0; i < withTiesColumns.length; i++) {
            cutoffValues[i] = getValue(withTiesColumns[i]);
        }
    }

    public void close() throws SQLException {
        source.close();
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        return new ExecutionPlan("FirstRows " + topNum, source.getExecutionPlan());
    }

    public Comparable getValue(int i) throws SQLException {
        return source.getValue(i);
    }

    public void setValue(int i, Comparable value) throws SQLException {
        source.setValue(i, value);
    }

    public int getNumColumns() throws SQLException {
        return source.getNumColumns();
    }

    public String getColumnName(int i) throws SQLException {
        return source.getColumnName(i);
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return source.getColumnInfo(i);
    }

    public boolean isUpdateable() throws SQLException {
        return source.isUpdateable();
    }

    public void deleteTuple() throws SQLException {
        source.deleteTuple();
    }
}
