/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 16, 2002
 * Time: 10:04:52 AM
 */
package freebase.sql.query;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;
import freebase.engine.IndexScan;
import freebase.server.ExecutionPlan;

import java.sql.SQLException;

/**
 * Represents a cartesian product or Index lookup.
 *
 * TODO:  handle the efficient case of pulling batches of data at a time (prefetch)
 */
public class NestedLoopJoin implements TupleIterator {
    TupleIterator left;
    TupleIterator right;

      //  only set when this is an index nested loop scan
    IndexScan scan;
    int[] leftIndexAttributes;
    Comparable[] curKey;

    int leftSize;
    int rightSize;

    boolean initialized = false;

    boolean rightIsEmptyDueToNull = false;

    /**
     * Nested join does a foreach on the left, and then a foreach on the right.
     * Eventually this code will be rewritten to be slightly optimized to perform
     * groups of operations on the right.
     */
    public NestedLoopJoin(TupleIterator left, TupleIterator right) throws SQLException {
        this.left = left;
        this.right = right;

        leftSize = left.getNumColumns();
        rightSize = right.getNumColumns();

        //  Should grab a marker for the right, since we will be going over it
        //  again and again
    }

    /**
     * Construct a NestedLoopJoin as an IndexedLoopJoin.  This could just as easily been a separate class
     * but the similarity was pretty strong and there is a tendency to make every little slight variant another
     * class, so in this case I just collapsed them.  Plus, they have the same name "NestedLoops" for now.
     * @param left the driving outer loop TupleIterator
     * @param right the index being driven.
     * @param leftIndexAttributes an array of the first n attribute offsets being used in this index lookup
     *     from the left hand tuple stream
     * @throws SQLException
     */
    public NestedLoopJoin(TupleIterator left, IndexScan right, int[] leftIndexAttributes) throws SQLException {
        this(left, right);

        this.scan = right;
        this.leftIndexAttributes = leftIndexAttributes;
        curKey = new Comparable[leftIndexAttributes.length];
    }

    public void reset() throws SQLException {
        initialized = false;
    }

    public boolean next() throws SQLException {
        if(! initialized) {
            initialized = true;
            return stepLeft();
        }

        boolean hasMore = nextRight();
        if(! hasMore) {
            return stepLeft();
        }

        return true;
    }

    /**
     * Moves the left side up one.  Will continue to scan if there is an index and that particular
     */
    private boolean stepLeft() throws SQLException {
        while(true) {
            boolean leftHasMore = left.next();
            if(! leftHasMore)
                return false;

            resetRight();
            boolean rightHasMore = nextRight();
            if(rightHasMore || scan == null)
                return rightHasMore;
        }
    }

    /********************************************************************
     * For handling both index and cartesian product join
     */

    private void resetRight() throws SQLException {
        rightIsEmptyDueToNull = false;

        if(scan != null) {  //  index loop
              //  TODO:  never been tests
            for(int i = 0; i < curKey.length; i++) {
                curKey[i] = left.getValue(leftIndexAttributes[i]);

                  //  if the source value is null, return false;
                if(curKey[i] == null) {
                    rightIsEmptyDueToNull = true;
                    return;
                }
            }

            scan.setSingleKeyRange(curKey);
        } else {
           right.reset();
        }
    }

    private boolean nextRight() throws SQLException {
        if(rightIsEmptyDueToNull) {
            return false;
        }

        if(scan != null) {  //  index loop
            return scan.next();
        } else {
            return right.next();
        }
    }

    /********************************************************************/

    /**
     *
     * @param i
     * @return
     * @throws SQLException
     */
    public Comparable getValue(int i) throws SQLException {
        if(i >= leftSize) {
            return right.getValue(i - leftSize);
        } else {
            return left.getValue(i);
        }
    }

    public void close() throws SQLException {
        left.close();
        right.close();
    }

    /**
     * Update operations not supported for now
     */
    public void setValue(int i, Comparable value) throws SQLException {
        throw new UnsupportedOperationException("Updates not supported");
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        StringBuffer details = new StringBuffer();
        if(scan == null) {
            details.append("(Cartesian Product)");
        } else {
            details.append("(Index Lookup From (");
            for(int i = 0; i < leftIndexAttributes.length; i++) {
                details.append(left.getColumnInfo(leftIndexAttributes[i]).getFullName());
            }
            details.append("))");
        }

        ExecutionPlan plan = new ExecutionPlan("Nested Loops " + details);
        plan.addSubPlan(left.getExecutionPlan());
        plan.addSubPlan(right.getExecutionPlan());
        return plan;
    }

    public int getNumColumns() throws SQLException {
        return leftSize + rightSize;
    }

    public String getColumnName(int i) throws SQLException {
        if(i >= leftSize) {
            return right.getColumnName(i - leftSize);
        } else {
            return left.getColumnName(i);
        }
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        if(i >= leftSize) {
            return right.getColumnInfo(i - leftSize);
        } else {
            return left.getColumnInfo(i);
        }
    }

    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException {
        throw new UnsupportedOperationException("Updates not supported yet");
    }
}
