/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 22, 2002
 * Time: 11:49:31 AM
 */
package freebase.sql.query;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;
import freebase.server.ExecutionPlan;
import freebase.sql.SqlExpression;

import java.sql.SQLException;
import java.util.ArrayList;

/**
 * A projection is a mapping and transformation of columns, including expressions that modify
 * the values.
 *
 * TODO:  support column renaming
 */
public class Projection implements TupleIterator {
    TupleIterator source;

    /**
     * If it exists, it is used to map outputs to input ports.  Otherwise, all output
     * ports get mapped directly to the exprs<BR><BR>
     *
     * Positive numbers go to the source.  Negative numbers get indexed into
     * exprs[- (1 + val)]  (ie, -1 -> 0, -2 -> 1, -3 -> 2.
     * Note that -(1 + i) is reversable, ie -(1 + -(1 + i)) = i
     */
    int[] map = null;

    /**
     * Projection expressions
     */
    ArrayList exprs;

    /**
     * Offset map
     */
    int[] offsets;

    /**
     * Select statement to use for aliases;  this saves on extra silly array from being created
     */
    ArrayList aliases = null;

    /**
     * Constructs a projection.  If leaveExisting is true, then this constructs a
     * projection that maps all of the source values through and then adds
     * expressions to the end.  Otherwise it only maps the expressions through.
     */
    public Projection(TupleIterator source, ArrayList expressions, int[] offsets,
                      boolean appendExpressions, ArrayList aliases) throws SQLException {
        this.source = source;
        this.exprs = expressions;
        this.offsets = offsets;
        this.aliases = aliases;

        if(appendExpressions) {
            map = new int[source.getNumColumns() + exprs.size()];
            for(int i = 0; i < source.getNumColumns(); i++) {
                map[i] = i;
            }

            for(int i = 0; i < exprs.size(); i++) {
                map[source.getNumColumns() + i] = - (i + 1);
            }
        }
    }

    /**
     * An alternative constructor that allows all columns to be remapped, either to
     * a source column or an expression.  The map (int[]) is expected to be filled in
     * with zero or positive numbers for source columns, and negative numbers for
     * indexes into the expressions list.<BR><BR>
     *
     * expressions & offsets can all be null as long as there are no references to them.
     * If aliases are null, then
     */
    public Projection(TupleIterator source, int[] map, ArrayList expressions, int[] offsets,
                      ArrayList aliases) throws SQLException {
        this.source = source;
        this.exprs = expressions;
        this.offsets = offsets;
        this.map = map;
        this.aliases = aliases;
    }

    /**
     * Create a truncation projection only showing the first (num) entries
     */
    public Projection(TupleIterator source, int num) {
        this.source = source;

        map = new int[num];
        for(int i = 0; i < map.length; i++) {
            map[i] = i;
        }
    }

    public void reset() throws SQLException {
        source.reset();
    }

    public Comparable getValue(int i) throws SQLException {
        if(map == null) {
            SqlExpression x = (SqlExpression)exprs.get(i);
            return x.evaluate(source, offsets);
        } else {
            if(map[i] >= 0) {
                return source.getValue(map[i]);
            } else {
                SqlExpression x = (SqlExpression)exprs.get(-(map[i] + 1));
                return x.evaluate(source, offsets);
            }
        }
    }

    public boolean next() throws SQLException {
        return source.next();
    }


    public void close() throws SQLException {
        source.close();
    }

    public int getNumColumns() throws SQLException {
        if(map == null) {
            return exprs.size();
        } else {
            return map.length;
        }
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        return new ExecutionPlan("Projection (no details yet)", source.getExecutionPlan());
    }

    /**
     * TODO:  do column renaming
     */
    public String getColumnName(int i) throws SQLException {
        if(aliases != null && aliases.size() > i && aliases.get(i) != null) {
            return (String)aliases.get(i);
        }

        if(map == null) {
            return ((SqlExpression)exprs.get(i)).toString();
        } else {
            if(map[i] >= 0) {
                return source.getColumnName(map[i]);
            } else {
                return exprs.get(-(map[i] + 1)).toString();
            }
        }
    }

    /**
     * TODO:  deal with aliases here
     */
    public ColumnInfo getColumnInfo(int i) throws SQLException {
          //  need to get all of the other info if needed
        ColumnInfo info = null;

        if(map == null) {
            return ((SqlExpression)exprs.get(i)).getColumnInfo(source, offsets);
        } else {
            if(map[i] >= 0) {
                info = source.getColumnInfo(map[i]);
            } else {
                  //  TODO:  execute the thing on the expression
                return ((SqlExpression)exprs.get(-(map[i] + 1))).getColumnInfo(source, offsets);
            }
        }

        return applyAlias(i, info);
    }

    /**
     * If the project is expected to rename the column then apply the alias.
     * It may create a new ColumnInfo in the process, it may not.
     */
    private ColumnInfo applyAlias(int i, ColumnInfo info) {
          //  if appropriate to apply alias
        if(aliases != null && aliases.size() > i && aliases.get(i) != null) {
            return ColumnInfo.rename(info, (String)aliases.get(i));
        }

        return info;
    }

    /*******************************************************************
     * illegal update methods (for now)
     */

    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException {
        throw new UnsupportedOperationException("update not legal through projection");
    }

    public void setValue(int i, Comparable value) throws SQLException {
        throw new UnsupportedOperationException("update not legal through projection");
    }
}
