/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 18, 2002
 * Time: 6:54:30 PM
 */
package freebase.sql.query;

import freebase.api.ColumnInfo;
import freebase.api.TupleIterator;
import freebase.server.ExecutionPlan;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * Sorts columns being passed in only.  For now uses immediate memory, but later should
 * use a sort buffer.
 *
 * TODO:  use an external sort buffer to do the sorting, this is horrendous
 */
public class SimpleSortPass implements TupleIterator {
    TupleIterator source;

    int[] sortColumns;
    boolean[] ascendingFlags;

    ArrayList data = null;
    int pos = -1;

    Comparable[] curData;

    /**
     *
     */
    public SimpleSortPass(TupleIterator source, int[] sortColumns) {
        this.source = source;
        this.sortColumns = sortColumns;
    }

    /**
     * The ascendingFlags array must be at least as long as the sortColumns, but it's size is not
     * used as a metric.  This allows a margin of slow to save an extra copy.
     */
    public SimpleSortPass(TupleIterator source, int[] sortColumns, boolean[] ascendingFlags) {
        this.source = source;
        this.sortColumns = sortColumns;
        this.ascendingFlags = ascendingFlags;
    }

    /**
     * Loads all of the data, then sorts it.
     */
    private void doSort() throws SQLException {
        data = new ArrayList();

        while(source.next()) {
            Comparable[] row = new Comparable[source.getNumColumns()];
            for(int i = 0; i < row.length; i++) {
                row[i] = source.getValue(i);
            }
            data.add(row);
        }

        Collections.sort(data, new SimpleComparator(sortColumns, ascendingFlags));
    }

    /**
     * Since on the first pass we grab all data and sort, this does
     * not need TODO anything really
     */
    public void reset() throws SQLException { }

    public Comparable getValue(int i) throws SQLException {
        return curData[i];
    }

    public boolean next() throws SQLException {
        if(data == null) {
            doSort();
        }

        pos++;
        if(pos >= data.size()) {
            return false;
        }

        curData = (Comparable[])data.get(pos);

        return true;
    }

    public void close() throws SQLException {
        source.close();
    }

    public int getNumColumns() throws SQLException {
        return source.getNumColumns();
    }

    public ExecutionPlan getExecutionPlan() throws SQLException {
        StringBuffer sb = new StringBuffer("Simple Sort [");
        for(int i = 0; i < sortColumns.length; i++) {
            if(i != 0) {
                sb.append(',');
            }
            sb.append(source.getColumnName(sortColumns[i]));
        }
        sb.append(']');

        return new ExecutionPlan(sb.toString(), source.getExecutionPlan());
    }

    public String getColumnName(int i) throws SQLException {
        return source.getColumnName(i);
    }

    public ColumnInfo getColumnInfo(int i) throws SQLException {
        return source.getColumnInfo(i);
    }
    /********************************************************
     *
     */
    /**
     *
     */
    public void setValue(int i, Comparable value) throws SQLException {
        throw new UnsupportedOperationException("Not legal to update though a sort");
    }

    public boolean isUpdateable() throws SQLException {
        return false;
    }

    public void deleteTuple() throws SQLException {
        throw new UnsupportedOperationException("Not legal to delete though a sort");
    }
}

/**
 * Just compares the two entries
 */
class SimpleComparator implements Comparator {
    int[] sortBy;

    boolean[] ascendingFlags;

    /**
     *  If ascending flags are null, then all are ascending, as in a standard group by
     */
    public SimpleComparator(int[] sortIndexes, boolean[] ascendingFlags) {
        this.sortBy = sortIndexes;
        this.ascendingFlags = ascendingFlags;
    }

    public int compare(Object o1, Object o2) {
        Comparable[] c1 = (Comparable[])o1;
        Comparable[] c2 = (Comparable[])o2;

        checkingeach:
        for(int i = 0; i < sortBy.length; i++) {
            if(c1[sortBy[i]] == null) {
                if(c2[sortBy[i]] == null)
                    continue checkingeach;
                return ascendingValue(i);
            }

            if(c2[sortBy[i]] == null) {
                return -ascendingValue(i);
            }

            int comp = c1[sortBy[i]].compareTo(c2[sortBy[i]]);
            if(comp != 0)
                return comp * ascendingValue(i);
        }

        return 0;  //  equal
    }

    /**
     * If there are no ascendingFlags, then they are all true.  If
     * the list is too short, then it treated as padded with all trues
     */
    private final byte ascendingValue(int pos) {
        if(ascendingFlags == null || pos >= ascendingFlags.length)
            return 1;
        if(ascendingFlags[pos])
            return 1;
        return -1;
    }
}