package freebase.store;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;
import java.util.Random;

/**
 *  One file per entity is the most basic model
 *  of course this prevents multi disk parallelism but this
 *  will come later once the model is all worked out
 *
 *  TODO:  implement read(Block b) and write(Block b) for slightly faster performance
 */
public class BasicFileEntity extends Entity {
    /**
     *  TODO:  this should be a per file (tablespace) parameter eventually and should
     *  be in a configuration table/file
     */
    final static int GROW_MINIMUM = 32;
    
    /**
     *  issue to be re-addressed later, growing/clearing should be logged!
     */
    final boolean CLEAR_NEW_SPACE = true;
    
    Random rand;
    
    String filename;
    RandomAccessFile f;
    int curBlock = 0;
    int blockSize;
    int entityId;
    
    int fileLen;  //  in blocks
    
    public BasicFileEntity(String filename, int blocksize, int entityId) throws IOException {
        this.filename = filename;
        this.blockSize = blocksize;
        this.entityId = entityId;
        
        rand = new Random();
        
        openFile();
    }
    

    /**
     *  read many in sequence starting with fromPage
     */
    public synchronized void read(int fromPage, Block[] b) throws IOException {
        if(fromPage != curBlock) {
            f.seek(convertBlockToPos(fromPage));
            curBlock = blockSize;
        }
        
        //  assure that the file is long enuf for the read...
        //  the alternative is to return the number of blocks available...
        //  probably ok, and have the fix command deal with growing the file.
        //  probably preferable.
        assureBlockLength(fromPage + b.length, b[0]);
        
        for(int i = 0; i < b.length; i++) {
            try{
            f.readFully(b[i].data);
            }catch(IOException ioe) {
                throw new BadBlockRead("Java IOException: \"" + ioe.getMessage() + "\"", entityId, entityId, filename, curBlock);
            }
            
              //  double check flip and flop
            if(b[i].getFlip() != b[i].getFlop()) {
                throw new BadBlockRead("Flip != Flop", entityId, entityId, filename, curBlock);
            }
            
            if(b[i].getFileNum() != entityId)
                throw new BadBlockRead("wrong file id: " + b[i].getFileNum(), entityId, entityId, filename, curBlock);
            
            if(b[i].getBlockNum() != curBlock)
                throw new BadBlockRead("wrong block num: " + b[i].getBlockNum(), entityId, entityId, filename, curBlock);
            
            //  TODO:  this will change once the entity and page number are stored in the
            //   actual block
            b[i].setEntityNum(entityId);
            b[i].setPageNum(curBlock);
            curBlock ++;
        }
    }
    
    /**
     *  write many, maybe not in direct sequence
     */
    public synchronized void write(Block[] b) throws IOException {
        for(int i = 0; i < b.length; i++) {
            if(b[i].getBlockNum() != curBlock) {
                f.seek(convertBlockToPos(b[i].getBlockNum()));
                curBlock = b[i].getBlockNum();
            }
            
              //  enable us to detect partial writes (especially if our block is bigger than the file system)
            b[i].toggleFlipFlop();
            
            f.write(b[i].data);
            
            curBlock++;
        }
    }
    
    /**
     *  returns the size in the number of bytes
     */
    public synchronized int getSize() throws IOException {
        return (int) f.length() / blockSize;
    }
    
    /**
     *  makes sure that all of the writes are secured on durable storage
     */
    public synchronized void sync() throws IOException {
        f.getFD().sync();
    }
    
    /**
     *  requests that the system grow by at least numBlocks number of
     *  additional blocks
     */
    public synchronized void grow(int numBlocks, Block bufferBlock) throws IOException {
        long growQty = convertBlockToPos(numBlocks);
        f.setLength(f.length() + growQty);
        
        if(CLEAR_NEW_SPACE) {
            // TODO:  since this will always be reading, it should use the block passed in
            //  for it's memory space
            f.seek(blockSize * fileLen);  //  seek to old end
            
            Arrays.fill(bufferBlock.data, (byte)0);
            
            for(int i = 0; i < numBlocks; i++) {
                  //  clue:  entityNum will not always be fileNum!!!!! nor block = page!
                initializeBlock(bufferBlock, entityId, i + fileLen);
                f.write(bufferBlock.data);
            }
        }
        
        sync();
        f.close();
        
        //  TODO:  try to throw a special no grow exception   so that the user can know specifically
        
        //  this may seem strange but windows2000 does not actually write all zeros at this time to
        //  the file, despite the sync (!) so we must close it and reopen it.  If you don't believe me
        //  try it yourself:  create a file, set the length to big and the shutdown your computer.  The
        //  file will have garbage in it
        openFile();
    }
    
    /**
     *  This correctly converts block num to a long so that you can get 4 billion blocks
     *  per file.
     */
    private final long convertBlockToPos(int block) {
        long pos = (long)block;
        pos = pos & 0x0000ffff;  //  get rid of the negative
        pos *= blockSize;
        return pos;
    }
    
    /**
     *  initialize a new block for this object
     */
    private void initializeBlock(Block b, int fileNum, int blockNum) throws IOException {
        b.setLSN(0);
        
        b.setFileNum((short)fileNum);
        b.setBlockNum(blockNum);
        
        int newFlipFlop = 0;
        while(newFlipFlop == 0x00000000 || newFlipFlop == 0xffffffff)
            newFlipFlop = rand.nextInt();
        
        b.setFlipFlop(newFlipFlop);
    }
    
    private void openFile() throws IOException {
       f = new RandomAccessFile(filename, "rw");
       curBlock = 0;
       f.seek(0);
       
       fileLen = fileLen = (int) f.length() / blockSize;
    }
    
    /**
     *  make sure the file is long enuf
     * @param bufferBlock is just space to use in memory for preparing the new
     *        block images
     */
    final private void assureBlockLength(int requiredLen, Block bufferBlock) throws IOException {
        //System.out.println("assuring length  needs:" + requiredLen + "   got:" + fileLen);

        if(requiredLen >= fileLen) {
            int growAmount = requiredLen - fileLen;
            growAmount = (growAmount + GROW_MINIMUM - 1) / GROW_MINIMUM;
            growAmount *= GROW_MINIMUM;
            
            System.out.println("assuring length " + requiredLen);
            grow(growAmount, bufferBlock);
        }
    }
}

