
package freebase.store;

import freebase.util.DataConverter;
import freebase.util.FauxShareSemaphore;
import freebase.util.ShareLock;

/**
 *  a generic disk block.  Use an adapter to interpret this into any
 *  more specific kind of block (index, data or otherwise)
 */
public class Block {
    final static int FLIP_POS = 0;
    final static int FILENUM_POS = 4;
    final static int BLOCKNUM_POS = 6;
    final static int LSN_POS = 10;
    final static int FLOP_POS = -4;
    
    /**
     *  space for flip, flop, filenum and blocknum
     */
    final public static int HEADER_SIZE = 18;
    final public static int FOOTER_SIZE = 4;
    
    byte[] data;
    
    boolean dirty;
    boolean logBlock;  //  this is here because it is currently considered
                      //  not ok to write out a log block ahead of time... unless version
                      //  times are used or something like that
    
    //  both of these could possibly be in byte arrays in the bufferpool
    //  for much faster cache coherency and lookup
    //  much less pointer dancing
    byte clockCount = 0;  //  used for the clock algorithm
    byte fixCount = 0;    //  the number of outstanding lock holders
    
    /**
     *  lock that enforced correct use of the block.
     *  Currently a faux share semaphore that always
     *  locks resource in exclusive mode.
     */
    public ShareLock accessLock = new FauxShareSemaphore();
    //  boolean bLastGrabbedExclusive = false;
    
    /**
     *  construct a block with the given size
     */
    public Block(int blocksize) {
        data = new byte[blocksize];
    }
    
    
    //  methods block are called by the entity only
    
    protected int getFlip() {
        return DataConverter.readInt(data, FLIP_POS);
    }
    
    protected int getFlop() {
        return DataConverter.readInt(data, data.length + FLOP_POS);
    }
    
    protected void toggleFlipFlop() {
        int f = getFlip() ^ 0xffffffff;
        setFlipFlop(f);
    }
    
    protected void setFlipFlop(int value) {
        DataConverter.writeInt(data, FLIP_POS, value);
        DataConverter.writeInt(data, data.length + FLOP_POS, value);
    }
    
    /**
     *  This is the disk file and the
     */
    protected short getFileNum() {
        return DataConverter.readShort(data, FILENUM_POS);
    }
    
    protected int getBlockNum() {
        return DataConverter.readInt(data, BLOCKNUM_POS);
    }
    
    protected void setFileNum(short fileNum) {
        DataConverter.writeShort(data, FILENUM_POS, fileNum);
    }
    
    
    protected void setBlockNum(int blockNum) {
        DataConverter.writeInt(data, BLOCKNUM_POS, blockNum);
    }
    
    public long getLSN() {
        return DataConverter.readLong(data, LSN_POS);
    }
    
    public void setLSN(long lsn) {
        DataConverter.writeLong(data, LSN_POS, lsn);
    }
    
    
    
    //   entity and page information
    //  this differs from file and block info.  eventually we will have a better store but
    //  for now we will do this.
    int entityId;
    int pageNum;
    
    public int getEntityNum() {
        return entityId;
    }
    
    public void setEntityNum(int e) {
        entityId = e;
    }
    
    public int getPageNum() {
        return pageNum;
    }
    
    public void setPageNum(int p) {
        pageNum = p;
    }
    
    /*  handy methods, maybe use them one day
    public void lock(boolean exclusive) {
        if(exclusive)
            
            
        bLastGrabbedExclusive = exclusive;
    }
    
    public void unlock() {
        if(bLastGrabbedExclusive)
            accessLock.releaseLockExclusive();
        else
            accessLock.releaseLockShared();
    }
    */
}