package freebase.store;

import freebase.engine.OldLog;
import freebase.util.Constants;
import freebase.util.LongHashMap;
import freebase.util.OhShitException;

import java.io.IOException;
import java.util.ArrayList;


/**
 *  Da Buffer Pool.  currently uses a simple clock algorithm to keep track
 *  of not so recently used pages.  This needs to be analyzed for performance.<BR><BR>
 *  
 *  Eventually it will be possible to have multiple buffer pools
 *  in the database, and be able to associate them with different
 *  objects.  But it will work with one.
 *
 *  TODO:  support online adding of files<BR><BR>
 *
 *  TODO:  this class needs to be rewritten to be highly sychronized.  One way
 *    for sure is to synchronized the hashmap on per entry in the array.
 *  also, use an int hash map instead... for blocks its ok and will save space<BR><BR>
 *
 * Need buffer fix commands for grabbing temporary blocks but not loading them, indeed clearing
 * them.  Sometimes temporary data will need to be loaded.
 */
public class BufferPool implements Constants {
    final static byte STANDARD_AGE_RATE = 16;
    final static byte STANDARD_RENEW_RATE = 16;
    
    //  note that this list is totally changable later.  Perhaps
    //  go directly to the Entity manager or something.
    EntityManager entityManager;
    LongHashMap[] pageMaps = null;
    
    Block[] blocks = null;
    Object walkLock = new Object();
    int lastGrabbedBlock = -1;
    
    int blockSize;
    
    /**
     *  This is to prevent needing to allocate literally billions of objects
     */
    ObjectPool lockPool = new ObjectPool();
    
    OldLog transactionLog;
    
    /**
     *  this is shoehorned in here in case there are multiple block sizes resident
     * in the system at any one time.  Ug
     */
    TablePageInterface tablePageInterface;
    
    /**
     *  construct a buffer pool to hold data
     */
    public BufferPool(int numBlocks, int blockSize, EntityManager em) {
        blocks = new Block[numBlocks];
        
        entityManager = em;
        pageMaps = new LongHashMap[entityManager.getNumEntities()];
        for(int i = 0; i < pageMaps.length; i++) {
            pageMaps[i] = new LongHashMap();
        }
        
        this.blockSize = blockSize;
        
        tablePageInterface = new TablePageInterface(blockSize);
    }
    
    
    /**
     *  because of a chicken and egg problem at startup, the log must be set after the
     *  bufferpool is created
     */
    public void setLog(OldLog systemLog) {
        transactionLog = systemLog;
    }
    
    public TablePageInterface getTablePageInterface(int entityId) {
        return tablePageInterface;
    }
    
    
    /**
     *  returns to the caller a block with the data in it
     *  for the given file and blockNum.  The block is guarenteed
     *  to not be recycled until an unfix command is issued
     *
     *  note that this differs from the standard fix command in
     *  that it does not issue any locks on the 
     *
     *  TODO:  re read this code to assure correct implementation
     *       also consider reimplementing it
     */
    public Block fix(int entity, int pageNum) throws IOException {
        Block b = null;
        Block candidate = null;
        Object inPlace;
        boolean weFoundHole = false;
        boolean creatingNewBlock = false;
        
        //  ouch! I hope this is fast enuf
        synchronized(pageMaps[entity]) {
            while(! weFoundHole) {
                inPlace = pageMaps[entity].get(pageNum);
                
                if(inPlace != null) {
                    if(inPlace instanceof Block) {
                        synchronized(inPlace) {
                            b = (Block)inPlace;
                            
                            if(b.fixCount != -1) {
                                b.fixCount++;
                                //  TODO:  should consider some alternate strategy
                                b.clockCount = STANDARD_RENEW_RATE;
                                return b;
                            } else {
                                b = null;
                                //  bad timing:  we found a block we though we wanted but it
                                //   is currently about to be retooled
                                inPlace = null;
                            }
                        }
                    }
                }
                
                if(inPlace == null) {
                    //  put a placeholder there so that another guy coming through does not
                    //  also try to load the block
                    inPlace = lockPool.get();
                    pageMaps[entity].put(pageNum, inPlace);
                    weFoundHole = true;
                }
                
                if(! weFoundHole) {
                    //  currently we wait on the whole hashtable.  This may be
                    // inefficient because any time anything is waiting on the hashtable
                    //  all waiting thread will be woken up!!!
                    try{
                    pageMaps[entity].wait();
                    //  STAT:  might be nice to keep track of the number of false
                    //  wakeups
                    }catch(InterruptedException ie) {
                        throw new IOException("Interrupted");
                    }
                }
            }
        }
        
        while(b == null) {
                //  grab next candidate
            synchronized(walkLock) {    
                lastGrabbedBlock = (lastGrabbedBlock + 1) % blocks.length;
                candidate = blocks[lastGrabbedBlock];
                if(candidate == null) {
                    creatingNewBlock = true;
                    candidate = new Block(blockSize);
                    blocks[lastGrabbedBlock] = candidate;
                }
            }
                
                //  reserve the block
            synchronized(candidate) {
                if(candidate.clockCount <= 0) {
                    if(candidate.fixCount == 0) {
                        b = candidate;
                        
                        //  prevent others using it until we unhook it
                        b.fixCount = -1;
                    } else {
                        //  let it ride
                        //  this is a weird situation since the fix lock should not be
                        //  held that long
                    }
                } else {
                    candidate.clockCount -= STANDARD_AGE_RATE;
                }
            }
        }
            
            //  remove the page from the map
        if(! creatingNewBlock) {
            synchronized(pageMaps[entity]) {
                Object removed = pageMaps[b.getEntityNum()].remove(b.getPageNum());
                    
                if(removed != b) {
                        //  oops shomething was really wanting that page
                        //  so they put something else there (an object, or even yet another block)
                    pageMaps[b.getEntityNum()].put(b.getPageNum(), removed);
                }
            }
        }
        
            //  recycle the block
            //  write out the blocks contents
        if(b.dirty && ! b.logBlock) {
            transactionLog.flush(b.getLSN(), 0);
                
            writeBlock(b);
                
            if(b.dirty)
                throw new OhShitException("block did not get written out to disk!!!!");
        }
            
        entityManager.getEntity(entity).read(pageNum, b);
            
        b.fixCount = 1;
        b.clockCount = STANDARD_RENEW_RATE;
            
        synchronized(pageMaps[entity]) {
            pageMaps[entity].put(pageNum, b);
                
            //  TODO:  for now the guy is going to wait on
            //  the actual hashtable object.  later something more
            //  highly concurrent can be put it this things place
            pageMaps[entity].notify();
        }
            
        return b;
    }
    
    /**
     *  returns the block to the pool
     */
    public void unfix(Block b) {
        synchronized(b) {
            b.fixCount--;
        }
    }
    
    /**
     *  Since this will be called by a writer after obtaining a read permission
     *  will that prevent deadlock?  what about with row chaining?
     *  seems not.  Consider writing a version of the ShareLock
     *  TODO deadlock detection.  Wow, that would be useful
     */
    public void upgradeLockExclusive(Block b) {
        //  perhaps use this to prevent others from grabbing the lock
        //  this is because our semaphores do not automatically know who 
        //  already has a share lock.
    }
    
    /**
     *  has the data written to disk, but does not guarentee a flush
     */
    public void writeData(int entity, int beginBlock, int endBlock) {
        //  should backup mode be dealt with at this level?
        
        //  if beginBlock == -1 then just do all with a duplicate enumeration
        
        for(int i = beginBlock; i < endBlock + 1; i++) {
            //  find all blocks and write them out
        }
    }
    
    /**
     *  writes the block out to the entity that it currently represents
     */
    private void writeBlock(Block b) throws IOException {
          //  prevent any more writes
          //  TODO:  consider impact of holding share lock on block for this long :(
        //try{
        b.accessLock.getLockShared();
        //}catch(InterruptedException ie) {
          //  throw new IOException("Interrupted");  //  TODO:  throw SQLException!!
        //}

        try{
        
          //  check dirty flag within the share lock, to assure that no writes are going on
        if(! b.dirty)
            return;
        
        if(! b.logBlock) {
            transactionLog.flush(b.getLSN(), 0);
            //  STAT:  keep track of this occurance (only if flush does something of course)
            //  it could be bad if writeblock was frequently forcing a log flush
        }
        
        entityManager.getEntity(b.getEntityNum()).write(b);
        } finally {
            b.accessLock.releaseLockShared();
        }
    }
    
    /**
     *  later we could write a parallel flush routine for more than
     *  one entity at a time.
     */
    public void syncData(int entity) throws IOException {
        entityManager.getEntity(entity).sync();
    }
    
    /**
     *  This should be used *only* when an OutOfMemory error occurs
     */
    public void shrink(int numBlocks) {
        throw new java.lang.UnsupportedOperationException();
    }
}


//  TODO:  should monitor this guy!
/**
 *  This may be either foolish or clever, I can't say now.
 *  but it avoids needing shared semaphores by just using this
 *  for a quick lock
 */
class ObjectPool {
    ArrayList a = new ArrayList();
    int outstandingLocks = 0;
    
    public Object get() {
        synchronized(a) {
            outstandingLocks ++;
            
            if(a.size() == 0)
                return new Object();
            else
                return a.remove(a.size() - 1);
        }
    }
    
    public void release(Object o) {
        synchronized(a) {
            outstandingLocks --;
            a.add(o);
        }
    }
}