/*
 * Copyright JJ Furman
 * User: user
 * Date: Apr 25, 2002
 * Time: 8:58:28 PM
 */
package freebase.test;

import freebase.StartDaemon;
import freebase.test.regression.RegressionTest;
import freebase.util.GlobalProps;
import freebase.util.Helper;

import java.sql.*;
import java.util.ArrayList;

/**
 * Runs lots of tests on queries.  This test differs from the others because it doens't use TestSessions,
 * but instead uses direct syncronous calls... a bit faster and simpler.<BR><BR>
 *
 * One of the primary tables it does queries on is NUMS, which represents a bunch of numbers.
 *
 * Later I will add more meaningful data, but this should be a great start
 * TODO:  make the failedQueries unified with the parent "errors" methodology, ie make errors support simple strings maybe
 * TODO:  make sure there are some with different number of columns
 * TODO:  test string operations, such as min, max, and the operators (case insensitive)
 *
 * TODO:  add test for choose_max, and for types of aggregation function results
 */
public class QueryTest extends RegressionTest {
    /**
     * for doing simple performance testing.
     */
    static int NUM_TIMES_REPEAT = 1;

    Connection con;

    /**
     * Holds the results of the most recent query
     */
    String sql = null;
    int numResultColumns = 0;
    ArrayList results = null;
    SQLException error = null;

    /**
     * keep track of queries
     */
    int totalQueries = 0;

    /**
     * Failed test information
     */
    ArrayList failedQueries = new ArrayList();
    ArrayList reasons = new ArrayList();

    /**
     * Create the local, in memory server
     */
    public QueryTest() throws SQLException, ClassNotFoundException {
        Class.forName("freebase.jdbc.F3Driver");
        con = DriverManager.getConnection("jdbc:freebase:local", "JJ", "");
    }


    /********************************************************************************
     * All the actual tests.  runTests is the parent of all other tests
     */
    public void testBasics() {
        executeQuery("select * from nums");
        checkValues(0, new int[] {1,1,1,2,3,4,5,5}, 2);

          //  test schema name in table
        executeQuery("select a from JJ.nums");
        checkValues(0, new int[] {1,1,1,2,3,4,5,5}, 2);

          //  test table name in column
        executeQuery("select NUMS.a, b, b from NUMS");
        checkValues(0, new int[] {1,1,1,2,3,4,5,5}, 2);

          //  test TOP n
        executeQuery("select TOP 5 NUMS.a, b, b from NUMS order by NUMS.a");
        checkValues(0, new int[] {1,1,1,2,3}, 0);

          //  test top 0
        executeQuery("select TOP 0 NUMS.a, b, b from NUMS order by NUMS.a");
        checkValues(0, new int[] {}, 0);

        executeQuery("select TOP -1 NUMS.a, b, b from NUMS order by NUMS.a");
        checkException();

          //  test with ties
        executeQuery("select top 2 with ties a, count(*) from nums group by a order by count(*)");
        checkCount(3);

          //  test with ties but zero
        executeQuery("select top 0 with ties a, count(*) from nums group by a order by count(*)");
        checkCount(0);

          //  test error with ties
        executeQuery("select top 3 with ties * from nums");
        checkException();

          //  test standard with ties
        executeQuery("select top 5 with ties a-b from nums order by a-b");
        checkCount(8);

          //  scan a non local schema table
        executeQuery("select * from sys.tables");
        checkNoException();

          //  assign the same alias twice, an error
        executeQuery("select * from nums err, nums err");
        checkException();
    }

    public void testOperators() {
          //  test =,  > and <
        executeQuery("select NUMS.a+b as \"blah blah\" from nums n where b=a");
        checkValues(0, new int[] {2,2,4,8,10,10}, 0);

        executeQuery("select a from nums where b>3");
        checkValues(0, new int[] {3,4,5,5}, 2);

        executeQuery("select a from nums where b<3");
        checkValues(0, new int[] {1,1,1,2}, 0);

        executeQuery("select a from nums where a >= 4");
        checkValues(0, new int[] {5,4,5}, 0);

        executeQuery("select a from nums n where a <= b-n.a");
        checkValues(0, new int[] {1}, 0);

        executeQuery("select a from nums n where (a=1 and b=2) or (a=3 and b >3)");
        checkValues(0, new int[] {1,3}, 0);

        executeQuery("select b from nums where a is null");
        checkValues(0, new int[] {7,7}, 0);

        executeQuery("select b from nums where a is not null and b < 4");
        checkValues(0, new int[] {2,2,1,1}, 0);

          //  test operator ordering
        executeQuery("select distinct 1 + 2 * a + b from nums");
        checkValues(0, new int[] {4,5,7,11,13,16}, 1);

          //  test many single table conditions
        executeQuery("select a from nums n where a is not null and b > -1 and b < 10 and a>1");
        checkValues(0, new int[] {3,4,5,2,5}, 0);

          //  test basic in condition
        executeQuery("select * from nums X where X.a in(2,3)");
        checkValues(1, new int[] {4,2}, 0);

          //  test multi table in condition
        executeQuery("select * from nums X, nums Y where x.a in (y.a)");
        checkCount(16);

          //  null in condition
        executeQuery("select * from nums where null in (a, b)");
        checkCount(0);

          //  another null
        executeQuery("select * from nums where a in (null)");
        checkCount(0);

          //  test reverse in condition
        executeQuery("select * from nums where 10 in (a+b, b)");
        checkCount(2);

          //  check false global condition
        executeQuery("select * from nums X, nums Y where x.a in (y.a) and 1=0");
        checkCount(0);

          //  check true global condition
        executeQuery("select * from nums X, nums Y where x.a in (y.a) and 1=1 and 'ddd'='ddd'");
        checkCount(16);

          //  check index lookup
        executeQuery("select * from ind_nums where c='fred'");
        if(GlobalProps.isStringsCaseInsensitive()) {
            checkCount(3);
        } else {
            checkCount(0);
        }

          //  index with gap
        executeQuery("select * from ind_nums where b > 5");
        checkCount(4);

        executeQuery("select * from ind_nums where b >= 6");
        checkCount(4);

          //  index between
        executeQuery("select * from ind_nums where b between -100 and 100");
        checkCount(7);

        executeQuery("select * from ind_nums where b between 3 and 6");
        checkCount(2);

          //  test index with nulls in second column
        executeQuery("select * from IND_NULL where a between 1 and 6");
        checkCount(8);

        executeQuery("select * from IND_NULL where a between 1 and 5");
        checkCount(7);

        executeQuery("select * from IND_NULL where b is null and a = 4");
        checkCount(1);
    }

    /**
     * NOTE:  to work right, this requires the test to be inprocess with the database
     */
    public void testLikeRegexpCaseInsensitive() {
          //  test like
        executeQuery("select * from names where name like 'J%'");
        checkCount(1);

          //  test case insensitive like
        executeQuery("select * from names where name like 'j%'");
        if(GlobalProps.isStringsCaseInsensitive()) {
            checkCount(1);
        } else {
            checkCount(0);
        }

          //  test = case insensitive
        executeQuery("select * from names where name='pRanaV'");
        if(GlobalProps.isStringsCaseInsensitive()) {
            checkCount(1);
        } else {
            checkCount(0);
        }

          //  test any letter
        executeQuery("select * from names where name LIKE 'Ti_o'");
        checkCount(1);

          //  test regexp compatibility
        executeQuery("select * from names where name NOT LIKE '.*'");
        checkCount(4);

          //  test regexp simple
        executeQuery("select * from names where name NOT REGEXP '.*'");
        checkCount(0);
    }

    public void testAggregation() {
          //  join/count
        executeQuery("select count(*) from nums x, nums y, nums z where x.a is null and y.a is not null and y.b = 5");
        checkValues(0, new int[] {40}, 0);

          //  count group by
        executeQuery("select b, count(*) from nums x group by b");
        checkValues(0, new int[] {1,2,4,5,7}, 0);
        checkValues(1, new int[] {2,2,2,2,2}, 0);

          //  bad select item
        executeQuery("select b, count(*) from nums x group by a");
        checkException();

        executeQuery("select b, count(*) from nums x group by a");
        checkException();

        //  count distinct
        executeQuery("select count(distinct a+b) from nums");
        checkValues(0, new int[] {6}, 0);

        executeQuery("select a, count(distinct b) from nums group by a order by count(*) desc, a");
        checkOrder(0, new int[] {1,5,2,3,4}, 1);
        checkOrder(1, new int[] {2,1,1,1,1,1}, 0);
        checkNumColumns(2);

        executeQuery("select a, count(distinct b) from nums group by a having count(*) > 2 order by count(*) desc, a");
        checkValues(1, new int[] {2}, 0);

        executeQuery("select a+1, count(*) from nums group by a + 1");
        checkValues(0, new int[] {2,3,4,5,6}, 1);

        executeQuery("select a-b, count(distinct b) from nums group by a-b");
        checkValues(1, new int[] {2,4,1}, 0);

        executeQuery("select a, count(distinct b) from nums group by a having a > 3");
        checkValues(0, new int[] {4,5}, 0);

        executeQuery("select a-b from nums group by a-b having count(b) > -1");
        checkValues(0, new int[] {-1,0}, 1);

          //  test MIN and MAX
        executeQuery("select a-b, min(b) from nums group by a-b order by a-b");
        checkOrder(0, new int[] {-1,0}, 1);
        checkOrder(1, new int[] {2,1,7}, 0);

        executeQuery("select min(a) from nums where b >3");
        checkValues(0, new int[] {3}, 0);

        executeQuery("select b, max(a+b) from nums group by b order by b");
        checkOrder(0, new int[] {1,2,4,5,7}, 0);
        checkOrder(1, new int[] {2,4,8,10}, 1);

        executeQuery("select b, max(a+b) from nums group by b order by b desc");
        checkOrder(0, new int[] {7,5,4,2,1}, 0);
        checkOrder(1, new int[] {10,8,4,2}, 1);

          //  test MIN and MAX with all nulls
        executeQuery("select min(a) from nums where a is null");
        checkValues(0, new int[] {}, 1);

        executeQuery("select max(a) from nums where a is null");
        checkValues(0, new int[] {}, 1);

          //  basic SUM test
        executeQuery("select sum(a) from nums");
        checkValues(0, new int[] {22}, 0);

        executeQuery("select sum(b) from nums");
        checkValues(0, new int[] {38}, 0);

        executeQuery("select sum(a+b) from nums");
        checkValues(0, new int[] {46}, 0);
    }

    public void testDistinct() {
        executeQuery("select distinct b from nums");
        checkValues(0, new int[] {1,2,4,5,7}, 0);

        executeQuery("select distinct * from nums order by a+b");
        checkValues(0, new int[] {1,1,2,3,4,5}, 1);
        checkNumColumns(2);

        executeQuery("select distinct a-b, a, b from nums");
        checkValues(0, new int[] {-1,-1,0,0,0,0}, 1);

        executeQuery("select distinct b from nums order by a");
        checkException();

        executeQuery("select distinct a-b, a, b from nums where a < 4 order by a-b, a-b+1 desc");
        checkOrder(0, new int[] {-1,-1,0,0}, 0);

        executeQuery("select distinct a-b, a, b from nums order by a desc, b");
        checkOrder(0, new int[] {0,0,-1,0,0,-1}, 1);

        executeQuery("select distinct a-b, a, b from nums order by -b");
        checkValues(0, new int[] {-1,-1,0,0,0,0}, 1);
        checkOrder(2, new int[] {7,5,4,4,2,2,1}, 0);

        executeQuery("select distinct a, count(*) from nums group by a order by -a");
        checkOrder(0, new int[] {5,4,3,2,1}, 1);
        checkOrder(1, new int[] {2,1,1,1,3,2}, 0);

        executeQuery("select distinct b, count(*) from nums where a<=5 group by b");
        checkValues(0, new int[] {1,2,4,5}, 0);
    }

    public void testSorting() {
        executeQuery("select * from Nums order by a-nums.b, a, b");
        checkOrder(1, new int[] {2,4,1,1,2,4,5,5,7,7}, 0);
        checkNumColumns(2);

        executeQuery("select a-b, a, b from NUMS n order by a-n.b, a, b");
        checkOrder(2, new int[] {2,4,1,1,2,4,5,5,7,7}, 0);
        checkNumColumns(3);

        executeQuery("select a-b, a, b from Nums n order by 1");
        checkOrder(0, new int[] {-1,-1,0,0,0,0,0,0,0,0}, 0);
        checkNumColumns(3);

        executeQuery("select a-b, a, b from nums n order by 'fred', null, a");
        checkOrder(1, new int[] {1,1,1,2,3,4,5,5}, 2);

        executeQuery("select * from nums x, nums y where y.a < 4 and x.a = y.b order by x.a+y.a");
        checkValues(1, new int[] {1,2,1,1,2,1,2,2,4}, 0);

        executeQuery("select a from nums x group by a order by count(b), a desc");
        checkOrder(0, new int[] {4,3,2,5,1}, 1);
        checkNumColumns(1);
    }


    public void testJoins() {
        executeQuery("select * from nums a, nums b, nums c");
        checkCount(1000);

        executeQuery("select * from nums a, nums b, nums c, nums d where a.a is null and c.a > 4 and b.a=d.b");
        checkCount(56);

        executeQuery("select * from nums, names where nums.a = names.id");
        checkCount(6);

        executeQuery("select * from nums, names nn where nums.a = nn.id");
        checkCount(6);

          //  test nested loop
        executeQuery("select * from nums a, names b, names c, names d where a.a=1 and a.a=b.id and b.id=c.id and c.id=d.id");
        checkCount(3);

        executeQuery("select * from IND_NUMS a, IND_NUMS b where a.c=b.c and b.c like 'b%'");
        checkCount(4);

          //  tests single table filters on all tables, despite index join
        executeQuery("select * from IND_NUMS a, IND_NUMS b where a.c=b.c and b.c like 'b%' and a.c like 'f%'");
        checkCount(0);

        executeQuery("select * from IND_nums a, ind_nums b where a.a = b.b");
        checkCount(6);

        executeQuery("select * from IND_nums a, ind_nums b where a.a = b.b and a.b > 5 and b.b < 5");
        checkCount(2);

        executeQuery("select * from ind_nums x, ind_nums y, ind_nums z where x.a = y.a and y.a = z.b and z.b = x.a and z.b = 1");
        checkCount(9);

          //  test string index lookup loop
        executeQuery("select * from IND_NUMS a, IND_NUMS b where a.c=b.c");
        if(GlobalProps.isStringsCaseInsensitive()) {
            checkCount(15);
        } else {
            checkCount(13);
        }

        executeQuery("select * from names nn,nums where nums.a = nn.id");
        checkCount(6);

          //  a is ambiguous, should fail
        executeQuery("select * from nums x, nums y where a is null");
        checkException();

          //  one join condition is redundantly defined, another post join condition
        executeQuery("select * from nums x, nums y where x.a=y.a and x.a=y.a and (x.a=x.b or x.a=y.b)");
        checkCount(14);

          //  force the optimizer to ignore the nice equi join
        executeQuery("select * from nums x, nums y where x.a+0=y.a+0 and x.b=5");
        checkValues(0, new int[] {5,5,5,5}, 0);
        checkValues(1, new int[] {5,5,5,5}, 0);
        checkValues(2, new int[] {5,5,5,5}, 0);

          //  non equi join clauses
        executeQuery("select count(distinct b.a) from nums a, nums b where b.a=1 and a.a > b.b group by a.a+b.a");
        checkValues(0, new int[] {1,1,1,1}, 0);
        checkNumColumns(1);

          //  equi all the way up
        executeQuery("select * from nums a, nums b, nums c, nums d where a.a = b.a and b.b = c.a and c.b = d.a and d.a <3 and d.b = a.b");
        checkCount(24);

        executeQuery("select * from nums x, nums y where x.a > y.b and x.b != y.b and (x.b+y.a between 6 and 6)");
        checkValues(1, new int[] {5,5,5,5,4,4,5,5}, 0);
        checkValues(2, new int[] {1,1,1,1,2,2,1,1}, 0);
    }


    /************************************************************************************
     * Execution and value testing methods
     ************************************************************************************/

    /**
     * Verifies that an exception was thrown (of type errorCode, eventually)
     */
    public void checkException(int errorCode) {
        if(error == null) {
            recordFailure("Expected an exception but got good data");
        }
    }

    /**
     * Same as above but no specific code.
     */
    public void checkException() {
        if(error == null) {
            recordFailure("Expected an exception but got good data");
        }
    }

    /**
     * Verifies that an exception was thrown (of type errorCode, eventually)
     */
    public void checkNoException() {
        if(error != null) {
            recordFailure("Expected no exception but got one! :" + error.getMessage());
        }
    }

    public void checkNumColumns(int numExpectedColumns) {
        if(error != null) {
            recordFailure("expected to check row count but got exception: " + error.getMessage());
            return;
        }

        if(numExpectedColumns != numResultColumns) {
            recordFailure("Explicitly expected " + numExpectedColumns + " columns but got " + numResultColumns);
        }
    }

    public void checkCount(int expectedRows) {
        if(error != null) {
            recordFailure("expected to check row count but got exception: " + error.getMessage());
            return;
        }

        if(expectedRows != results.size()) {
            recordFailure("Explicitly expected " + expectedRows + " rows but got " + results.size());
        }
    }


    /**
     * Note:  ignores nulls it finds for purposes of ordering
     */
    public void checkOrder(int column, int[] values, int numNullsExpected) {
        if(error != null) {
            recordFailure("expected to check number values but got exception: " + error.getMessage());
            return;
        }

        if((values.length + numNullsExpected) != results.size()) {
            recordFailure("expected " + (values.length + numNullsExpected) + " rows but got " + results.size());
            return;
        }

        int curPos = 0;
        for(int i = 0; i < results.size(); i++) {
            Number n = null;
            try{
            n = (Number)(((Object[])results.get(i))[column]);
            }catch(ClassCastException c) {
                recordFailure("trying to check the value of a number but data not a number: column # " + column);
                return;
            }

            if(n != null) {
                if(n.intValue() == values[curPos]) {
                    curPos++;
                } else {
                    recordFailure("Expected to find '" + values[curPos] + "' but instead found '" + n + "'  in column " + column);
                    return;
                }
            }
        }
    }

    /**
     * Checks the current data set against the values, ignoring the order.
     * Columns start at zero (0), not 1 like in JDBC!!!
     */
    public void checkValues(int column, int[] values, int numNullsExpected) {
        if(error != null) {
            recordFailure("expected to check number values but got exception: " + error.getMessage());
            return;
        }

        if((values.length + numNullsExpected) != results.size()) {
            recordFailure("expected " + (values.length + numNullsExpected) + " rows but got " + results.size());
            return;
        }

        boolean[] used = new boolean[values.length];
        int numNullsFound = 0;

        nextValue:
        for(int i = 0; i < results.size(); i++) {
            Number n = null;
            try{
            n = (Number)(((Object[])results.get(i))[column]);
            }catch(ClassCastException c) {
                recordFailure("trying to check the value of a number but data not a number: column # " + column);
                return;
            }

            if(n == null) {
                numNullsFound++;
                if(numNullsFound > numNullsExpected) {
                    recordFailure("too many null values found in column # " + column);
                    return;
                }

                continue;
            }

            for(int j = 0; j < used.length; j++) {
                if(used[j]) {
                    continue;
                }

                if(values[j] == n.intValue()) {
                    used[j] = true;
                    continue nextValue;
                }
            }

            recordFailure("Found value '" + n + "' not in list (or too many found) in column " + column);
            return;
        }
    }


    public void executeQuery(String sql) {
        this.sql = sql;
        results = null;
        error = null;
        totalQueries++;

        try{
            Statement st = con.createStatement();
            ResultSet rs = st.executeQuery(sql);
            ResultSetMetaData meta = rs.getMetaData();
            numResultColumns = meta.getColumnCount();

            results = new ArrayList();
            while(rs.next()) {
                Object[] row = new Object[numResultColumns];

                for(int i = 0; i < row.length; i++) {
                    row[i] = rs.getObject(i + 1);
                }

                results.add(row);
            }

            rs.close();
            st.close();
        }catch(SQLException e) {
            error = e;
        }
    }

    public void recordFailure(String info) {
        failedQueries.add(sql);
        reasons.add(info);
    }

    /******************************************************************
     * Static main and other methods
     */

    /**
     *
     */
    public static void main(String[] args) throws Throwable {
        StartDaemon.startDatabaseServer().createTestSchema();
        //PrintTracer.replaceSystemOutWithTracer("(2)");

        QueryTest t = new QueryTest();

        long start = System.currentTimeMillis();
        for(int i = 0; i < NUM_TIMES_REPEAT; i++) {
            t.runTests();
        }
        long timeTook = System.currentTimeMillis() - start;

        System.out.println();
        System.out.println("Queries tested: " + t.totalQueries);
        Helper.printElapsed(timeTook);
        System.out.println();

        //  generate report
        if(t.failedQueries.size() == 0 && t.getErrors().size() == 0 ) {
            System.out.println("All Tests -------> OK");
            System.exit(0);
        } else {

            Object lastFailed = null;
            for(int i = 0; i < t.failedQueries.size(); i++) {
                if(lastFailed != t.failedQueries.get(i)) {
                    System.out.println("Test Failed: " + t.failedQueries.get(i));
                    lastFailed = t.failedQueries.get(i);
                }
                System.out.println("     " + t.reasons.get(i));
            }

            lastFailed = null;
            for(int i = 0; i < t.getErrorMethods().size(); i++) {
                if(lastFailed != t.getErrorMethods().get(i)) {
                    System.out.println("Test Failed: " + t.getErrorMethods().get(i));
                    lastFailed = t.getErrorMethods().get(i);
                }
                Object error = t.getErrors().get(i);
                if(error instanceof Throwable) {
                    ((Throwable)error).printStackTrace();
                } else {
                    System.out.println("     " + t.getErrors().get(i));
                }
            }

            System.exit(1);
        }

        //  TODO:  perhaps do somekind of shutdown of system
    }
}
