/*
 * Copyright JJ Furman
 * User: user
 * Date: May 7, 2002
 * Time: 9:31:42 AM
 */
package freebase.test;

import freebase.util.CompareUtil;
import freebase.util.ErrorCode;
import freebase.util.Helper;

import java.sql.*;
import java.util.ArrayList;

/**
 * Represents a session with the database.  Used to test concurrent access to the system
 * the commands sent to this session generally run on their own Thread, though it is assumed
 * that commands sent to this are from a single test thread.  This is because it is not meaningful
 * to execute commands to a well defined test with arbitrary timing.
 *
 * Note:  this uses standard JDBC to talk to the database, thus could in prinicple be
 * used to test any database system.
 *
 * TODO:  handle unexpected hanging with TIMEOUT
 * TODO:  support immediate mode, no external thread
 */
public class TestSession implements Runnable {
    //final public static String DUMMY = "";

    /**
     * Command codes
     */
    final public static int COMMIT = 0;
    final public static int ROLLBACK = 1;
    final public static int SAVEPOINT = 2;
    final public static int STANDARD_SQL = 3;

      //  to implement still:
    final public static int BATCH_SQL = 4;
    final public static int PREPARED_SQL = 5;
    final public static int PREPARED_BATCH_SQL = 6;

    final public static int DEFAULT_TIMEOUT = 10000;
    final public static int DEFAULT_BLOCKED_CONFIDENCE_WAITTIME = 3000;

    /**
     * the length of time to wait to guarentee that a statement was very likely executed
     */
    int returnPause = 100;

    /**
     * If this amount of time passes waiting, then an error condition occurs
     */
    int timeoutMillis = DEFAULT_TIMEOUT;
    int blockWaitConfidence = DEFAULT_BLOCKED_CONFIDENCE_WAITTIME;

      //  use these when doing single step debugging, else you are in for a surprise
    //int timeoutMillis = 10000000;
    //int blockWaitConfidence = 10000000;

    /**
     * Our JDBC connection with the database
     */
    Connection con;


    Object executeLock = new Object();
    boolean closed = false;

    /**
     * The next statement to execute.  While this is n
     */
    String prevStatement = null;
    String nextStatement = null;
    int nextCode = -1;

    /**
     * Payload that is either:  <BR>
     * Batch:  list of sql Strings<BR>
     * PreparedStatement:  Comparable[]<BR>
     * PreparedBatch:  Comparable[][]
     */
    Object statementData;

    /**
     * The results of the last statement
     */
    boolean wasQuery = false;
    int numResultColumns = 0;
    int updateCount = 0;
    ArrayList results = null;
    SQLException error = null;

    /**
     * Construct a test session with the given JDBC connection.  It sets autoCommit to false.  It
     * runs the thread.
     */
    public TestSession(Connection con) throws SQLException {
        this.con = con;

        con.setAutoCommit(false);

        new Thread(this).start();
    }


    private void execute(String sql, int commandCode, Object data) {
        synchronized(executeLock) {
              //  wait for the previous statement to finish
            waitForCompletion();

            if(error != null) {
                throw new TestFailedException("SQL failed '" + prevStatement + "' : " + error.getMessage(), error);
            }

            nextStatement = sql;
            nextCode = commandCode;
            this.statementData = data;

            //System.out.println("executing " + nextStatement);
              //  notify the waiting exection thread that there is work to do.  Giddyup!
            executeLock.notify();

              //  wait long enough to try to guarentee that the statement got in
            try{
                executeLock.wait(returnPause);
            }catch(InterruptedException ie) {ie.printStackTrace();}
        }
    }

    /**
     * Executes a statement.  Returns when the statement is done, or when
     */
    public void execute(String sql) {
        //System.out.println("exec: " + sql);
        execute(sql, STANDARD_SQL, null);
    }

    public void executePrepared(String sql, Comparable[] row) {
        execute(sql, PREPARED_SQL, row);
    }

    public void executeBatch(String[] sqls) {
        execute("DUMMY_STATEMENT", BATCH_SQL, sqls);
    }

    public void executePreparedBatch(String sql, Comparable[][] rows) {
        execute(sql, PREPARED_BATCH_SQL, rows);
    }

    public void commit() {
        execute("commit", COMMIT, null);
    }

    public void rollback() {
        execute("rollback", ROLLBACK, null);
    }

    public void close() {
        //waitForCompletion();
        synchronized(executeLock) {
            closed = true;
            executeLock.notify();
        }
        try{
            con.close();
        }catch(SQLException e) {
            throw new TestFailedException("failed to close session", e);
        }
    }

    /*****************************************************************************
     * Results access methods
     */

    /**
     * most common method used below to wait for the statement to complete before doing the next thing.
     */
    private void waitForCompletion() {
        if(! waitForCompletionOrTimeout(timeoutMillis)) {
            throw new TestFailedException("Operation '" + nextStatement + "'   TIMED OUT");
        }
    }

    /**
     *
     * @param timeoutInMillis the amount of time to wait before considering it a timeout
     * @return true if completion, false if timeout
     */
    private boolean waitForCompletionOrTimeout(int timeoutInMillis) {
        long startTime = System.currentTimeMillis();
        synchronized(executeLock) {
            while(nextStatement != null) {
                long elapsedTime = System.currentTimeMillis() - startTime;

                if(elapsedTime >= timeoutInMillis) {
                    return false; //throw new TestFailedException("Operation '" + nextStatement + "'   TIMED OUT");
                }

                try{
                    executeLock.wait(timeoutInMillis + 50 - elapsedTime);  //  with time limit?
                }catch(InterruptedException ie) {ie.printStackTrace();}
            }
        }

        return true;
    }

    /**
     *  TODO:  could name checkIsBlocked to be more consistent or assert
     */
    public void ensureIsBlocked() {
        //System.out.println("ensureIsBlocked");
        if(waitForCompletionOrTimeout(blockWaitConfidence)) {
            throw new TestFailedException("Operation '" + prevStatement + "' isn't blocked as expected!!");
        }
    }

    public void ensureIsNotBlocked() {
        //System.out.println("ensureIsNotBlocked");
        if(! waitForCompletionOrTimeout(blockWaitConfidence)) {
              //  use nextStatement because it hasn't rolled to prev yet
            throw new TestFailedException("Operation '" + nextStatement + "' is blocked when not expected!!");
        }
    }

    /**
     * Verifies that an exception was thrown (of type errorCode, eventually)
     */
    public void checkException(int errorCode) {
        waitForCompletion();

        if(error == null) {
            throw new TestFailedException("Expected an exception but got good data");
        }

        if(error.getErrorCode() != errorCode) {
            throw new TestFailedException("Got exception but of wrong type:  \n  expected code " +
                    errorCode + " (" + ErrorCode.getStandardMessage(errorCode) + ")\n  but got: "
                    + error.getErrorCode() + " (" + ErrorCode.getStandardMessage(error.getErrorCode())
                    + ") details: " + error);
        }

          //  clear the error so that
        error = null;
    }

    /**
     * Same as above but no specific code.
     * TODO:  eventually it should be possible to drop this method, forcing an exact method
     */
    public void checkException() {
        waitForCompletion();

        if(error == null) {
            throw new TestFailedException("Expected an exception but got good data");
        }

        error = null;
    }

    /**
     * Verifies that an exception was thrown (of type errorCode, eventually)
     */
    public void checkNoException() {
        waitForCompletion();

        if(error != null) {
            throw new TestFailedException("Expected no exception but got one! :" + error.getMessage());
        }
    }

    public void checkNumColumns(int numExpectedColumns) {
        waitForCompletion();

        if(error != null) {
            throw new TestFailedException("expected to check row count but got exception: " + error.getMessage());
        }

        if(! wasQuery) {
            throw new TestFailedException("Prev statement '" + prevStatement + "' was not a query");
        }

        if(numExpectedColumns != numResultColumns) {
            throw new TestFailedException("Explicitly expected " + numExpectedColumns + " columns but got " + numResultColumns);
        }
    }

    /**
     * count the # result set rows
     */
    public void checkCount(int expectedRows) {
        waitForCompletion();

        if(! wasQuery) {
            throw new TestFailedException("Last operation was not a query as expected: " + prevStatement);
        }

        if(error != null) {
            throw new TestFailedException("expected to check row count but got exception: " + error.getMessage());
        }

        if(expectedRows != results.size()) {
            throw new TestFailedException("Explicitly expected " + expectedRows + " rows but got " + results.size());
        }
    }

    public void checkUpdateCount(int expectedRowsUpdated) {
        waitForCompletion();

        if(wasQuery) {
            throw new TestFailedException("Last operation was not an update as expected: " + prevStatement);
        }

        if(error != null) {
            throw new TestFailedException("expected to check update count but got exception: " + error.getMessage());
        }

        if(expectedRowsUpdated != updateCount) {
            throw new TestFailedException("Expected " + expectedRowsUpdated + " updated rows but got " + updateCount);
        }
    }


    /**
     * Note:  ignores nulls it finds
     */
    public void checkOrder(int column, int[] values, int numNullsExpected) {
        waitForCompletion();

        if(error != null) {
            throw new TestFailedException("expected to check number values but got exception: " + error.getMessage());
        }

        if(! wasQuery) {
            throw new TestFailedException("Prev statement '" + prevStatement + "' was not a query");
        }

        if((values.length + numNullsExpected) != results.size()) {
            throw new TestFailedException("expected " + (values.length + numNullsExpected) + " rows but got " + results.size());
        }

        int curPos = 0;
        for(int i = 0; i < results.size(); i++) {
            Number n = null;
            try{
            n = (Number)(((Object[])results.get(i))[column]);
            }catch(ClassCastException c) {
                throw new TestFailedException("trying to check the value of a number but data not a number: column # " + column);
            }

            if(n != null) {
                if(n.intValue() == values[curPos]) {
                    curPos++;
                } else {
                    throw new TestFailedException("Expected to find '" + values[curPos] + "' but instead found '" + n + "'  in column " + column);

                }
            }
        }
    }

    public void checkSingleReturnValue(int val) {
        checkValues(0, new int[] {val}, 0);
    }

    /**
     * Checks the current data set against the values, ignoring the order.
     * Columns start at zero (0), not 1 like in JDBC!!!
     */
    public void checkValues(int column, int[] values, int numNullsExpected) {
        waitForCompletion();

        if(error != null) {
            throw new TestFailedException("expected to check number values but got exception: " + error.getMessage());
        }

        if(! wasQuery) {
            throw new TestFailedException("Prev statement '" + prevStatement + "' was not a query");
        }

        if((values.length + numNullsExpected) != results.size()) {
            throw new TestFailedException("expected " + (values.length + numNullsExpected) + " rows but got " + results.size());
        }

        boolean[] used = new boolean[values.length];
        int numNullsFound = 0;

        nextValue:
        for(int i = 0; i < results.size(); i++) {
            Number n = null;
            try{
            n = (Number)(((Object[])results.get(i))[column]);
            }catch(ClassCastException c) {
                throw new TestFailedException("trying to check the value of a number but data not a number: column # " + column);
            }

            if(n == null) {
                numNullsFound++;
                if(numNullsFound > numNullsExpected) {
                    throw new TestFailedException("too many null values found in column # " + column);
                }

                continue;
            }

            for(int j = 0; j < used.length; j++) {
                if(used[j]) {
                    continue;
                }

                if(values[j] == n.intValue()) {
                    used[j] = true;
                    continue nextValue;
                }
            }

            throw new TestFailedException("Found value '" + n + "' not in list (or too many found) in column " + column);
        }
    }

    /**
     * Checks the current data set against the values, ignoring the order.
     * Columns start at zero (0), not 1 like in JDBC!!!
     * @param values null is not allowed in here!!!
     */
    public void checkValues(int column, Object[] values, int numNullsExpected) {
        waitForCompletion();

        if(error != null) {
            throw new TestFailedException("expected to check number values but got exception: " + error.getMessage());
        }

        if(! wasQuery) {
            throw new TestFailedException("Prev statement '" + prevStatement + "' was not a query");
        }

        if((values.length + numNullsExpected) != results.size()) {
            throw new TestFailedException("expected " + (values.length + numNullsExpected) + " rows but got " + results.size());
        }

        boolean[] used = new boolean[values.length];
        int numNullsFound = 0;

        nextValue:
        for(int i = 0; i < results.size(); i++) {
            Object val = ((Object[])results.get(i))[column];

            if(val == null) {
                numNullsFound++;
                if(numNullsFound > numNullsExpected) {
                    throw new TestFailedException("too many null values found in column # " + column);
                }

                continue;
            }

            for(int j = 0; j < used.length; j++) {
                if(used[j]) {
                    continue;
                }

                if(CompareUtil.isEqual(values[j], val)) {
                    used[j] = true;
                    continue nextValue;
                }
            }

            throw new TestFailedException("Found value '" + val + "' not in list (or too many found) in column " + column);
        }
    }


    /**
     * Debugging and test creation tool.  Print whatever results were last found.
     */
    public void printResult() {
        waitForCompletion();
        System.out.println("Results of last statement \"" + prevStatement + "\" are:");
        if(error != null) {
            error.printStackTrace();
            return;
        }

        if(wasQuery) {
            for(int i = 0; i < results.size(); i++) {
                Object[] row = (Object[])results.get(i);
                for(int j = 0; j < row.length; j++) {
                    if(j != 0) {
                        System.out.print(", ");
                    }
                    System.out.print(row[j]);
                }

                System.out.println();
            }

            System.out.println();
            System.out.println("Total Rows: " + results.size());
        } else {

        }
        //  TODO:  print # updates or whatever, or actual query results
    }

    /**************************************************************************
     * static methods for checking cross
     */

      //  perfect candidate for variable # arguments
    public static int checkOneThrewException(TestSession a, TestSession b) {
        return checkOneThrewException(new TestSession[] {a,b});
    }

    public static int checkOneThrewException(TestSession a, TestSession b, TestSession c) {
        return checkOneThrewException(new TestSession[] {a,b,c});
    }

    /**
     * One should block, while the other should go on.  This is what happens in the case
     * of deadlock.  Returns which one threw the exception
     */
    public static int checkOneThrewException(TestSession[] sessions) {
        int numExceptions = 0;

          //  loop until finding an exception
        long startTime = System.currentTimeMillis();
        while(true) {
            for(int i = 0; i < sessions.length; i++) {
                TestSession a = sessions[i];
                synchronized(a.executeLock) {
                    if(a.nextStatement == null) {
                        if(a.error == null) {
                            throw new TestFailedException("Got a non error result, but expected either an error or a hung session");
                        }

                        //System.out.println("DEBUG MESSAGE: " + a.error.getMessage());
                        a.error = null;
                        return i;
                    }
                }
            }

            if(System.currentTimeMillis() - startTime > DEFAULT_TIMEOUT) {
                throw new TestFailedException("Waiting for any session to throw an exception, timed out");
            }

            Helper.sleep(100);
        }
    }

    public boolean isClosed() {
        return closed;
    }

    /**************************************************************************
     * execution thread and helper methods
     */

    /**
     * This is the thread that does the actual work to the connection, locking
     * and all
     */
    public void run() {
        while(true) {
              //  grab command
            synchronized(executeLock) {
                while(nextStatement == null) {
                    try{
                        executeLock.wait();
                    }catch(InterruptedException ie) {ie.printStackTrace();}

                    if(closed)
                        return;
                }

                  //  clear previous results just in case
                results = null;
                error = null;
            }

              //  execute command
            Statement st = null;
            try{
                switch(nextCode) {
                    case STANDARD_SQL:
                        st = con.createStatement();

                        //System.out.println("executing: " + nextStatement + ", prev statement is " + prevStatement);
                        wasQuery = st.execute(nextStatement);

                        if(wasQuery) {
                            processResultSet(st);
                        } else {
                            updateCount = st.getUpdateCount();
                        }

                        st.close();
                        break;
                    case PREPARED_SQL:
                        PreparedStatement ps = con.prepareStatement(nextStatement);
                        Comparable[] data = (Comparable[])statementData;

                        for(int i = 0; i < data.length; i++) {
                            ps.setObject(i + 1, data[i]);
                        }

                        //System.out.println("executing: " + nextStatement + ", prev statement is " + prevStatement);
                        wasQuery = ps.execute();

                        if(wasQuery) {
                            processResultSet(ps);
                        } else {
                            updateCount = ps.getUpdateCount();
                        }

                        ps.close();
                        break;
                    case BATCH_SQL:
                        st = con.createStatement();
                        String[] queries = (String[])statementData;

                        for(int i = 0; i < queries.length; i++) {
                            st.addBatch(queries[i]);
                        }

                        int[] results = st.executeBatch();

                        //  NOTE:  essentially ignoring the results.  No reason to directly test them anyway!
                        st.close();
                        break;
                    case PREPARED_BATCH_SQL:
                        ps = con.prepareStatement(nextStatement);
                        Comparable[][] rows = (Comparable[][])statementData;

                        for(int i = 0; i < rows.length; i++) {
                            for(int j = 0; j < rows[i].length; j++) {
                                ps.setObject(j + 1, rows[i][j]);
                            }

                            ps.addBatch();
                        }

                        //System.out.println("executing: " + nextStatement + ", prev statement is " + prevStatement);
                        ps.executeBatch();

                        ps.close();
                        break;
                    case ROLLBACK:
                        wasQuery = false;
                        con.rollback();
                        break;
                    case COMMIT:
                        wasQuery = false;
                        con.commit();
                        break;

                }
            }catch(SQLException s) {
                error = s;
                if(st != null) {
                    try{
                        st.close();
                    }catch(SQLException e) {}
                }
            }

            synchronized(executeLock) {
                  //  make room for another statement to execute and wakeup either a waiting result checker or another
                prevStatement = nextStatement;
                nextStatement = null;
                executeLock.notify();
            }
        }
    }

    /**
     * Processes results
     */
    private void processResultSet(Statement st) throws SQLException {
        ResultSet rs = st.getResultSet();
        ResultSetMetaData meta = rs.getMetaData();
        numResultColumns = meta.getColumnCount();

        results = new ArrayList();
        while(rs.next()) {
            Object[] row = new Object[numResultColumns];

            for(int i = 0; i < row.length; i++) {
                row[i] = rs.getObject(i + 1);
            }

            results.add(row);
        }

        rs.close();
    }
}