/*
 * Copyright JJ Furman
 * User: user
 * Date: May 7, 2002
 * Time: 3:43:14 PM
 */
package freebase.test.regression;

import freebase.test.TestSession;

/**
 * Tests all write operations, rollback and commit, and tries to cover many basic concurrent operations.
 * It only modifies the NAMES test table, but will obtain locks on the NUMS table as well.  Note that
 * all sessions created with createSession() automatically get closed and released at the end of each test.<BR><BR
 *
 * TODO  here:  Add test of insert and delete of nums, a non indexed table
 *
 * add serializable tests
 */
public class ConcurrencyTest extends MultiSessionTest {
    /**
     * repeating the tests over and over is ok, but all of the rows get cleared and rebuilt.  This
     * number effects tests that internally know that they leave the system in the same state as they
     * started, thus can repeat.  Tests different aspects of the system.<BR><BR>
     *
     * in rigorous testing, higher numbers are better, but 5 or 10 is good.  Eventually it might be good to ask the test if
     * it leaves the system in a clean state
     */
    int NUM_INTERNAL_REPEATS = 5;  // higher is better.

    /**
     *
     */
    public void testBasics() {
        TestSession a = createSession();

        for(int i = 0; i < NUM_INTERNAL_REPEATS; i++) {
            a.execute("insert into names values (5, 'FRED')");
            a.checkUpdateCount(1);
            a.execute("select id from names");
            a.checkValues(0, new int[] {1,2,3,4,5}, 0);
            a.rollback();
            a.execute("select id from names");
            a.checkValues(0, new int[] {1,2,3,4}, 0);

            a.execute("insert into names values (5, 'FRED')");
            a.execute("select * from names");
            a.checkValues(0, new int[] {1,2,3,4,5}, 0);
            a.commit();
            a.execute("select id from names");
            a.checkValues(0, new int[] {1,2,3,4,5}, 0);

            a.execute("delete from names where id=2 or id=5");
            a.execute("select id from names");
            a.checkValues(0, new int[] {1,3,4}, 0);
            a.rollback();
            a.execute("delete from names where id=5");
            a.commit();
            a.execute("select id from names");
            a.checkValues(0, new int[] {1,2,3,4}, 0);
        }
    }

    /**
     * inserts some data, then makes sure it is not visible from other transactions
     */
    public void testBasicMVC() {
        TestSession a = createSession();
        TestSession b = createSession();

          //  update via session a
        a.execute("insert into names values (5, 'FRED')");
        a.execute("update names set id=id+9 where id<3");

          //  update via session b
        b.execute("insert into names values (7, 'billybob')");
        b.execute("update names set id=id+3 where id=3");

        a.execute("select id from NAMES");
        a.checkValues(0, new int[] {10, 11, 3, 4, 5}, 0);

        b.execute("select id from names");
        b.checkValues(0, new int[] {6,1,2,4,7}, 0);

        a.commit();
        b.commit();

        a.execute("select id from names");
        a.checkValues(0, new int[] {10, 11, 6, 4, 5, 7}, 0);
    }

    /**
     * Tests single table deadlock situation.
     * TODO:  repeat this one
     */
    public void testTableDeadlock() {
        TestSession a = createSession();
        TestSession b = createSession();

        a.execute("lock table names in share mode");
        //a.ensureIsNotBlocked();
            b.execute("lock table names in share mode");
        a.execute("lock table names in exclusive mode");
        //a.ensureIsBlocked();
            b.execute("lock table names in exclusive mode");

          //  Either a or b should have a deadlock, not both
        TestSession.checkOneThrewException(a, b);

          //  intentionally doesn't explicitly rollback, to see if the cleanup code
          //  gets it correctly
    }

    /**
     *   An example of the above case is this:<BR>
     *   Tables A (NAMES) and B (NUMS)<BR>
     *   T1          T2          T3          Notes
     *   ------      ------      ------      -------
     *   X on A
     *               RX on B
     *                           RS on B
     *   S on B                              this only blocks on T2
     *                           RX on B     in conflict with T1, but allowed because T3
     *                                       already owned a lock; T1 must now wait on T3 also
     *                           RX on A     deadlock with T1
     *
     * TODO:  test more of these, where the deadlock is subtle
     *        IE, if a transaction owns a lock already on a table it can get more, but the
     *          waiting transactions must add the new conflict to the list of deadlock potentials
     */
    public void testComplexTableDeadlock() {
        TestSession t1 = createSession();
        TestSession t2 = createSession();
        TestSession t3 = createSession();

        t1.execute("lock table names in exclusive mode");
            t2.execute("lock table nums in row exclusive mode");
                t3.execute("lock table nums in row share mode");
        t1.execute("lock table nums in share mode");
                t3.execute("lock table nums in row exclusive mode");
                t3.execute("lock table names in row exclusive mode");

        TestSession.checkOneThrewException(new TestSession[] {t1, t3});
    }

    /**
     * Tests row deadlock situation.  based on the assumption that rows are generally scanned in
     * linear order, this tests that the other changes made by a particular statement are also rolled
     * back along with it
     */
    public void testRowDeadlock() {
        TestSession a = createSession();
        TestSession b = createSession();

        for(int i  = 0; i < NUM_INTERNAL_REPEATS; i++) {
            a.execute("delete from names where id=4");
            b.execute("delete from names where id=3");
            a.execute("delete from names where id>1");
            b.execute("delete from names");

              //  Either a or b should have a deadlock, not both
            int deadlockSession = TestSession.checkOneThrewException(new TestSession[] {a, b});

            if(deadlockSession == 0) {
                a.execute("select id from names");
                a.checkValues(0, new int[] {1,2,3}, 0);
                a.rollback();
                b.rollback();
            } else {
                b.execute("select id from names");
                b.checkValues(0, new int[] {1,2,4}, 0);
                b.rollback();
                a.rollback();
            }

            a.execute("select id from names");
            a.checkValues(0, new int[] {1,2,3,4}, 0);

            b.execute("select id from names");
            b.checkValues(0, new int[] {1,2,3,4}, 0);
        }
    }

    /**
     * Just locks one, then
     */
    public void testTableLockRelease() {
        TestSession a = createSession();
        TestSession b = createSession();

        for(int i = 0; i < 2; i++) {
            a.execute("lock table names in exclusive mode");
            b.execute("update names set id = id + 10");

            if(i == 0) {
                a.commit();
            } else {
                a.rollback();
            }

            b.execute("select id from names");;
            b.checkValues(0, new int[] {11,12,13,14}, 0);
            b.rollback();
        }
    }


    public void testCircularDeadlock() {
        TestSession a = createSession();
        TestSession b = createSession();
        TestSession c = createSession();
        TestSession d = createSession();

          //  each gets a different lock
        a.execute("update names set id=id+10 where id=1");
            b.execute("update names set id=id+10 where id=2");
                c.execute("update names set id=id+10 where id=3");
                    d.execute("update names set id=id+10 where id=4");

          //  each gets another lock, in a different order
            b.execute("update names set id=id+10 where id=3");
        a.execute("update names set id=id+10 where id=2");
                c.execute("update names set id=id+10 where id=4");
                    d.execute("update names set id=id+10 where id=1");

        TestSession.checkOneThrewException(new TestSession[] {a, b, c, d});
    }

    public void testOutOfOrderDeadlock() {
        TestSession a = createSession();
        TestSession b = createSession();
        TestSession c = createSession();
        TestSession d = createSession();

          //  uses unique keys to contort deadlock
        a.execute("insert into names values (5, 'Paris')");
        b.execute("insert into names values (6, 'London')");
        c.execute("insert into names values (7, 'Milan')");
        d.execute("insert into names values (8, 'Santa Cruz')");

        a.execute("insert into names values (6, 'London 2')");   //  a -> b
        b.execute("insert into names values (7, 'Milan 2')");    //  b -> c
        c.execute("insert into names values (8, 'Santa Cruz')"); //  c -> d
        d.execute("insert into names values (7, 'wha!')");       //  d -> c

        TestSession.checkOneThrewException(new TestSession[] {c, d});
    }

    public void testPostLockFailedDeleteCondition() {
        TestSession a = createSession();
        TestSession b = createSession();

        a.execute("update names set id=15 where id=2");
            b.execute("delete from names where id < 3");
        a.commit();
            b.execute("select id from names");
            b.checkValues(0, new int[] {4, 3, 15}, 0);
    }

    public void testSeesCurrentValue() {
        TestSession a = createSession();
        TestSession b = createSession();

        a.execute("update names set id=15 where id=2");
        a.execute("insert into names values (11, 'Donna Haraway')");
            b.execute("update names set id=id+2");
        a.execute("insert into names values (10, 'Tim Leary')");
        a.commit();
            b.execute("select id from names");
            b.checkValues(0, new int[] {3,17,5,6,10,11}, 0);
    }

    /**
     * TODO:  make sure the commit cases of below are done as well.
     */
    public void testClosing() {
        TestSession a = createSession();
        TestSession b = createSession();
        TestSession c = createSession();

        a.execute("update names set id=10 where id=1");
            b.execute("update names set id=11 where id=1");
                c.execute("update names set id=12 where id=1");
            b.close();  //  <!--  note, closing while it is blocked... important
        a.rollback();
                c.ensureIsNotBlocked();
                c.rollback();

            b = createSession();
        a.execute("lock table names in exclusive mode");
            b.execute("lock table names in exclusive mode");
                c.execute("lock table names in exclusive mode");
            b.close();  //  <!--  note, closing while it is blocked... important
        a.rollback();
                c.ensureIsNotBlocked();
                c.rollback();
    }

    /*******************************************************************
     *
     */

    public void afterEachTestCleanup(TestSession cleanupSession) {
        cleanupSession.execute("DELETE FROM NAMES");
        cleanupSession.execute("INSERT INTO NAMES VALUES (1, 'JJ')");
        cleanupSession.execute("INSERT INTO NAMES VALUES (2, 'Pranav')");
        cleanupSession.execute("INSERT INTO NAMES VALUES (3, 'Tito')");
        cleanupSession.execute("INSERT INTO NAMES VALUES (4, 'Smoove B')");
        cleanupSession.commit();
    }

    /**
     * main entry point
     */
    public static void main(String[] args) throws Throwable {
        RegressionTest.runTestAndExit(ConcurrencyTest.class, null) ;//"testBasicMVC");
    }
}

