/*
 * Copyright JJ Furman
 * User: user
 * Date: May 7, 2002
 * Time: 3:38:49 PM
 */
package freebase.test.regression;

import freebase.StartDaemon;
import freebase.util.Helper;

import java.io.PrintStream;
import java.lang.reflect.Method;
import java.util.ArrayList;

/**
 * Base class of all regression tests
 * TODO: port to JUnit.
 * TODO:  add ability to run tests in different order, random with seed
 */
public class RegressionTest {
    final static boolean PRINT = true;  //  generally true, but sometimes false for debugging
    long testTime;
    int numTests = 0;

    ArrayList errors = new ArrayList();
    ArrayList errorMethod = new ArrayList();


    /**
     * Runs all of the tests in the class.  For each method called test{whatever} with no arguments do:
     * 1.  call setup
     * 2.  invoke
     * 3.  call teardown
     *
     */
    public void runTests() {
        runTests(null);
    }

    /**
     * Run all or one test
     * @param testName null if you want to run all
     */
    public void runTests(String testName) {
        if(PRINT) {
            System.out.println("Running tests:  " + getClass().getName());
        }

        long startTime = System.currentTimeMillis();
        Method[] methods = getClass().getDeclaredMethods();
        for(int j = 0; j < methods.length; j++) {
            if(! methods[j].getName().startsWith("test"))
                continue;

            if(methods[j].getParameterTypes().length != 0) {
                System.out.println("IGNORED method:" + methods[j].getName() + "() should take no parameters");
                continue;
            }

            if(testName != null && ! testName.equals(methods[j].getName())) {
                continue;
            }

            numTests++;

            boolean errorFound = false;

            try{
                setup();
                methods[j].invoke(this, null);
            }catch(Throwable t) {
                errorFound = true;
                errors.add(t);
                errorMethod.add(methods[j].getName());
            }

            try {
                teardown();
            }catch(Throwable t) {
                errorFound = true;
                errors.add(t);
                errorMethod.add(methods[j].getName() + " TEARDOWN");
            }

            if(PRINT) {
                if(errorFound) {
                    System.out.print("X");
                } else {
                    System.out.print(".");
                }
            }
        }

        if(PRINT) {
            System.out.println();
        }
        testTime = System.currentTimeMillis() - startTime;
    }

    /**
     * After tests are run with runTests, call this to dump the results to a particular stream, usually
     * System.out
     */
    public void displayResults(PrintStream out) {
        out.println("Tests Run: " + numTests);
        Helper.printElapsed(out, testTime);
        out.println();

        //  generate report
        if(errors.size() == 0) {
            out.println("All Tests -------> OK");
        } else {
            for(int i = 0; i < errors.size(); i++) {
                Throwable t = (Throwable)errors.get(i);
                System.out.println("----------------------------------------------------");
                System.out.println("Got Error in " + errorMethod.get(i) + " :");
                t.printStackTrace(out);
            }
        }
    }

    /**
     * Called before each test
     */
    public void setup() throws Throwable {}

    /**
     * Called after each test
     */
    public void teardown() throws Throwable {}


    public ArrayList getErrors() {
        return errors;
    }

    public ArrayList getErrorMethods() {
        return errorMethod;
    }
    /***********************************************************************************
     *
     * Helper static methods
     */

    /**
     * runs all tests in class, but no setup
     * @param testClass
     * @throws Throwable
     * @return true if everything passed
     */
    public static boolean runTest(Class testClass) throws Throwable {
        RegressionTest test = (RegressionTest)testClass.newInstance();
        test.runTests();
        test.displayResults(System.out);
          // no errors?
        return test.errors.size() == 0;
    }

    /**
     *
     * @param testClass
     */
    public static void runTestAndExit(Class testClass) throws Throwable {
        StartDaemon.startDatabaseServer().createTestSchema();

        if(runTest(testClass)) {
            System.exit(0);
        } else {
            System.exit(1);
        }
    }

    /**
     *
     * @param testClass
     * @param name null means run all
     * @throws Throwable
     */
    public static void runTestAndExit(Class testClass, String name) throws Throwable {
        StartDaemon.startDatabaseServer().createTestSchema();

        RegressionTest test = (RegressionTest)testClass.newInstance();
        test.runTests(name);
        test.displayResults(System.out);

        if(test.errors.size() == 0) {
            System.exit(0);
        } else {
            System.exit(1);
        }
    }
}
