package freebase.test.regression;

import freebase.test.TestSession;
import freebase.util.ErrorCode;

import java.math.BigDecimal;

/**
 * Test updates.
 */
public class UpdatesTest extends MultiSessionTest {

    int NUM_REPEATS = 3;

    /**
     * insert into select and savepoint
     */
    public void testInsertIntoSelect() {
        TestSession a = createSession();

        for(int i = 0; i < NUM_REPEATS; i++) {
            a.execute("insert into nums (a) select a from nums");
            a.execute("select * from nums");
            a.checkCount(20);
            a.execute("savepoint save_20");
            a.execute("insert into nums (a) select a from nums");
            a.execute("savepoint save_40");
            a.execute("insert into nums (a) select a from nums");
            a.execute("select * from nums");
            a.checkCount(80);
            a.execute("rollback to save_40");
            a.execute("select * from nums");
            a.checkCount(40);
            a.execute("rollback to save_40");
            a.checkNoException();
            a.execute("rollback to save_20");
            a.execute("select * from nums");
            a.checkCount(20);
            a.execute("rollback to save_40");
            a.checkException();
            a.rollback();
        }
    }

    /**
     * Monkey simple test
     */
    public void testInsert() {
        TestSession a = createSession();

        for(int i = 0; i < NUM_REPEATS; i++) {
            a.execute("insert into names values (5, 'ZZZ')");
            a.execute("insert into names values (6, 'ZZZ')");
            a.execute("select name from names where name like 'ZZ%'");
            a.checkValues(0, new String[] {"ZZZ", "ZZZ"}, 0);
            a.rollback();
        }

        a.execute("insert into names (NaMe, iD) values ('Billy', 10)");
        a.execute("select name from names where name like 'Billy%'");
        a.checkCount(1);
        a.execute("savepoint there");
        a.execute("update names set name='SHAHRUKH KHAN' where id=10");
        a.checkUpdateCount(1);
        a.execute("rollback to there");
        a.execute("select name from names where name like 'Billy%'");
        a.checkCount(1);
        a.commit();
        a.rollback();
        a.execute("select name from names where name like 'Billy%'");
        a.checkCount(1);
        a.execute("delete from names where id > 3");
        a.checkUpdateCount(2);
        a.execute("delete from names where id > 3");
        a.checkUpdateCount(0);
    }


    /*
    TO ADD:

    1.  multiple updaters for the same row, ensure that when the middle session is killed the
        third one still gets the lock.


    */

    /**
     * basic sequences test for now
     */
    public void testSequences() {
        TestSession a = createSession();

        a.execute("create sequence seq");
        a.execute("select seq.nextval, seq.currval from nums");
        a.checkOrder(0, new int[] {1,2,3,4,5,6,7,8,9,10}, 0);
        a.checkOrder(1, new int[] {1,2,3,4,5,6,7,8,9,10}, 0);
        a.execute("update nums set a = seq.nextval");
        a.execute("select sum(a) from nums");
        a.checkValues(0, new int[] {210 - 55}, 0);
        a.execute("insert into names (ID, NAME) select seq.nextval, name from names");
        a.checkUpdateCount(4);
        a.rollback();
        a.execute("drop sequence seq");
        a.execute("insert into names (ID, NAME) select seq.nextval, name from names");
        a.checkException();
    //    a.execute(drop sequence seq)  && try something again and expect failure
    }


    public void testRepeatableRead() {
        TestSession a = createSession();
        TestSession b = createSession();
        a.execute("set transaction isolation level repeatable read");
        runRepeatableReads(a,b);
    }

    public void testSerializableRead() {
        TestSession a = createSession();
        TestSession b = createSession();
        a.execute("set transaction isolation level serializable");
        runRepeatableReads(a,b);
    }

    /**
     * Used by the testRepeatableRead and testSerializableRead
     */
    private void runRepeatableReads(TestSession a, TestSession b) {
          //  establish the view
        checkSameView(a);

        b.execute("insert into names values (10, null)");
        b.execute("update names set id=5 where id=1");
        b.execute("delete from names where id=2");

        checkSameView(a);

        b.commit();

        checkSameView(a);

        b.execute("insert into names values (12, null)");
        b.execute("update names set id=6 where id=3");
        b.execute("delete from names where id=4");

        checkSameView(a);

        b.commit();

        checkSameView(a);

        b.execute("insert into names values (15, null)");
        b.execute("update names set id=16 where id=12");
        b.execute("delete from names where id=6");

        checkSameView(a);

        b.rollback();

        checkSameView(a);

        b.execute("delete from names");

        checkSameView(a);

        b.commit();

        checkSameView(a);
    }

    public void testSimpleDirtyRead() {
        TestSession a = createSession();  //  reading thread
        TestSession b = createSession();
        TestSession c = createSession();

        a.execute("set transaction isolation level read uncommitted");
            b.execute("insert into names values (6)");
            b.execute("update names set id=10 where id=2");
            b.execute("delete from names where id=1");
        a.execute("select sum(id) from names");
        a.checkSingleReturnValue(10 + 3 + 4 + 6);

                c.execute("insert into names values (100)");
                c.execute("update names set id=11 where id=3");
                c.execute("delete from names where id=4");

        a.execute("select sum(id) from names");
        a.checkSingleReturnValue(100 + 10 + 11 + 6);           
    }

    public void testComplexDirtyRead() {
        TestSession a = createSession();
        TestSession b = createSession();

        a.execute("set transaction isolation level read uncommitted");
            b.execute("delete from names where id=1");
        a.execute("select * from names");
        a.checkCount(3);
        a.execute("insert into names values (1)");
            b.commit();
        a.execute("select * from names");
        a.checkCount(4);
    }


    //  TODO:  test serialable failure in writes!!!

    /**
     * performs select sum(id) from names => 10
     * @param a
     */
    private void checkSameView(TestSession a) {
        a.execute("select sum(id) from names");
        a.checkValues(0, new int[] {10}, 0);
    }

    /**
     * TODO:  use this to test hard typing into various fields
     */
    public void testBasicPreparedStatement() {
        TestSession a = createSession();

        a.executePrepared("insert into names (ID, NAME) values (?,?)",
                new Comparable[] {new BigDecimal(5), "Jones"});
        a.execute("select count(*) from names");
        a.checkSingleReturnValue(5);
    }

    public void testBatch() {
        TestSession a = createSession();

          //  test basics
        for(int i = 0; i < 3; i++) {
            a.executeBatch(new String[] {"insert into names values (5, 'Bob')",
                                         "insert into names values (6, 'Bill')",
                                         "insert into names values (7, 'Willy')"});
            a.execute("select count(*) from names");
            a.checkSingleReturnValue(7);
            a.rollback();
        }

          //  test atomicity
        a.executeBatch(new String[] {"insert into names values (5, 'Bob')",
                                     "insert into names values (6, 'Bill')",
                                     "insert into GARBAGE"});
        a.checkException();
        a.execute("select count(*) from names");
        a.checkSingleReturnValue(4);
    }

    public void testPreparedBatch() {
        TestSession a = createSession();

        for(int i = 0; i < 3; i++) {
            a.executePreparedBatch("insert into names (ID, NAME) values (?,?)", new Comparable[][]
                            {{new BigDecimal(5), "Jones"},
                             {new BigDecimal(6), "BigMomma"}});
            a.execute("select sum(id) from names");
            a.checkSingleReturnValue(21);
            a.rollback();
        }

          //  test atomicity of prepared batch
        a.executePreparedBatch("insert into names (ID, NAME) values (?,?)", new Comparable[][]
                        {{new BigDecimal(5), "Jones"},
                         {new BigDecimal(1), "BigMomma"}});
        a.checkException();
        a.execute("select sum(id) from names");
        a.checkSingleReturnValue(10);
    }

    /**
     * This is a pretty subtle test.  It involves having an update determine that a row
     * that it intially thought was to be included in the update set later turn out to not
     * match the condition (any longer) thus, it should be excluded... and thus with its exclusion
     * involves forcing a rollback of the entire statement and a retry... a bit extreme but indeed required.
     */
    public void testUpdateFailedView() {
        TestSession a = createSession();
        TestSession b = createSession();

        for(int i = 0; i < 2; i++) {  //  do 2 times
            a.execute("create table view_test (label varchar(2), id int)");

            a.execute("insert into view_test values ('A', 1)");
            a.execute("insert into view_test values ('B', 2)");
            a.execute("insert into view_test values ('C', 3)");
            a.commit();

            if(i == 1) {  //  test serializable mode as well
                b.execute("set transaction isolation level serializable");  //  view locked in here
            }
            a.execute("update view_test set label='BB' where id=2");  //  prevent b from passing
                b.execute("update view_test set id=id-100 where id>1");  //  jam up b
            a.execute("update view_test set id=5 where id=1");  //  make a row past where b has been interesting to b's condition
            a.execute("update view_test set id=0 where id=3");  //  make row after b's current pos no longer interesting to b... will force inconsistent view
            a.commit();
                if(i == 1) {
                    b.checkException(ErrorCode.CANT_SERIALIZE);
                } else {
                    b.checkUpdateCount(2);
                    b.execute("select sum(id) from view_test");
                    b.checkSingleReturnValue(-95 + -98);
                }

                b.execute("drop table view_test");
        }
    }

    /**
     * Another subtle one but with delete instead of update
     */
    public void testDeleteFailedView() {
        TestSession a = createSession();
        TestSession b = createSession();

        for(int i = 0; i < 2; i++) {
            a.execute("create table view_test (label varchar(2), id int)");

            a.execute("insert into view_test values ('A', 1)");
            a.execute("insert into view_test values ('B', 2)");
            a.execute("insert into view_test values ('C', 3)");
            a.commit();
            if(i == 1) {  //  test serializable mode as well
                b.execute("set transaction isolation level serializable");  //  view locked in here
            }
            a.execute("update view_test set label='BB' where id=2");  //  prevent b from passing
                b.execute("update view_test set id=id-100 where id>1");  //  jam up b
            a.execute("update view_test set id=5 where id=1");  //  make a row past where b has been interesting to b's condition
            a.execute("delete from view_test where id=3");  //  make row after b's current pos no longer interesting to b... will force inconsistent view
            a.commit();
                if(i == 1) {
                    b.checkException(ErrorCode.CANT_SERIALIZE);
                } else {
                    b.checkUpdateCount(2);
                    b.execute("select sum(id) from view_test");
                    b.checkSingleReturnValue(-95 + -98);
                }

                b.execute("drop table view_test");
        }
    }

    /**
     * another edge case.  In the case of a rollback, It is assumed that it doesn't count against serialization.
     * In the case of a block based storage, is space for the old transaction always available?
     * TODO:  do same tests as these but with select for update.  Oracle behaves same way as we expect to
     */
    public void testDeleteSerializable() {
        TestSession a = createSession();
        TestSession b = createSession();
        TestSession c = createSession();

        a.execute("create table view_test (label varchar(2), id int)");

        a.execute("insert into view_test values ('A', 1)");
        a.execute("insert into view_test values ('B', 2)");
        a.execute("insert into view_test values ('C', 3)");
        a.commit();
            b.execute("set transaction isolation level serializable");  //  view locked in here (unfortunately)
        a.execute("delete from view_test where id=2");  //  prevent b from passing
            b.execute("update view_test set id=id-100 where id>1");  //  jam up b
        a.commit();
            b.checkException(ErrorCode.CANT_SERIALIZE);

            b.execute("drop table view_test");

        //  same test, but with rollback.... should work

        a.execute("create table view_test (label varchar(2), id int)");

        a.execute("insert into view_test values ('A', 1)");
        a.execute("insert into view_test values ('B', 2)");
        a.execute("insert into view_test values ('C', 3)");
        a.commit();
            b.execute("set transaction isolation level serializable");  //  view locked in here (unfortunately)
        a.execute("delete from view_test where id=2");  //  prevent b from passing
            b.execute("update view_test set id=id-100 where id>1");  //  jam up b
        a.rollback();
            b.checkNoException();

            b.execute("drop table view_test");
    }

    /*******************************************************************************************
     *
     */

    /**
     *
     * @param cleanupSession
     */
    public void afterEachTestCleanup(TestSession cleanupSession) {
          //  rebuild the tables
        cleanupSession.execute("DELETE FROM NAMES");
        cleanupSession.execute("INSERT INTO NAMES VALUES (1, 'JJ')");
        cleanupSession.execute("INSERT INTO NAMES VALUES (2, 'Pranav')");
        cleanupSession.execute("INSERT INTO NAMES VALUES (3, 'Tito')");
        cleanupSession.execute("INSERT INTO NAMES VALUES (4, 'Smoove B')");

        //  TODO:  cleanup nums too

        cleanupSession.commit();
    }

    /**
     * main entry point
     */
    public static void main(String[] args) throws Throwable {
        RegressionTest.runTestAndExit(UpdatesTest.class);
    }
}
