/*
 * Copyright JJ Furman
 * User: user
 * Date: Mar 19, 2002
 * Time: 6:10:16 AM
 */
package freebase.util;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Timestamp;


/**
 * This is an alternate to Java Serialization for the values currently used
 * internally.  Eventually this may support the tuple concept.<BR><BR>
 *
 * Currently Strings are written out as bytes
 *
 * Currently supports 255 digit numbers total, with 255 digits of precision.
 * To save making a new array, might be possible to subclass from BigInteger and BigDecimal
 */
public class ObjectIO implements ObjectIOTypes {


    /**
     * @return the size including the codes and everything
     * currently supports int size, probably good enough for a while
     * but come next century there is going to be a lot of data!<BR><BR>
     *
     * a null object[] is supported by saying that there are -1 columns
     *
     */
    public static int getSize(Object[] data) {
        if(data == null) {
            return 2;
        }

        int size = 2 + data.length;

        for(int i = 0; i < data.length; i++) {
            size += getObjectSize(data[i]);
        }

        return size;
    }

    /**
     *    Format is:  2 bytes for numEntries, then for each one:
     *        a byte for the type code and a variable size for each one
     */
    public static void writeObjects(DataOutput out, Object[] data) throws IOException {
        if(data == null) {
            out.writeShort((short)-1);
            return;
        }

        out.writeShort((short)data.length);

        for(int i = 0; i < data.length; i++) {
            writeObject(out, data[i]);
        }
    }

    /**
     * See writeObjects for format
     */
    public static Object[] readObjects(DataInput in) throws IOException {
        short unmaskedColumns = in.readShort();
        if(unmaskedColumns == -1) {
            return null;
        }

        int len = unmaskedColumns & 0xffff;
        Object[] data = new Comparable[len];

        for(int i = 0; i < data.length; i++) {
            data[i] = readObject(in);
        }

        return data;
    }

    /**
     * Return entire amount required to encode this object
     */
    public static int getSize(Object o) {
        return 1 + getObjectSize(o);
    }

    /**
     * @return the pure object size, not including the code
     */
    private static int getObjectSize(Object o) {
        if(o == null) {
            return 0;
        } else if(o instanceof Bool) {
            return 1;
        } else if(o instanceof Byte) {
            return 1;
        } else if(o instanceof Short) {
            return 2;
        } else if(o instanceof Integer) {
            return 4;
        } else if(o instanceof Long) {
            return 8;
        } else if(o instanceof Float) {
            return 4;
        } else if(o instanceof Double) {
            return 8;
        } else if(o instanceof String) {
              //  TODO:  handle Unicode, pref UTF-8
              //  1 for code, 4 for lenth and then the data
           return 4 + ((String)o).length();
        } else if(o instanceof java.sql.Date) {
            return 8;
        } else if(o instanceof java.sql.Time) {
            return 8;
        } else if(o instanceof java.sql.Timestamp) {
            return 12;
        } else if(o instanceof byte[]) {
            return 4 + ((byte[])o).length;
        } else if(o instanceof BigInteger) {
              //  one number for length, then the data
            return 1 + getBigIntLength((BigInteger)o);
        } else if(o instanceof BigDecimal) {
              //  one for length, another for scale then data
            return 2 + getBigIntLength(((BigDecimal)o).unscaledValue());
        } else {
            throw new RuntimeException("CRITICAL ERROR: don't support object : " + o.getClass());
        }
    }

    /**
     * Returns the number of bytes required to store this big integer.
     * NOTE:  this only returns the length of the byte[] that corresponds to this, not
     * including space for any other number!!!
     */
    public static byte getBigIntLength(BigInteger i) {
        if(i == null) {
            return 0;
        }
        return (byte)(i.bitLength() / 8 + 1);
    }

    /**
     * Maybe all of the instanceof's are not the fasted execution in the world, but
     * I can hardly imagine it being so aweful.  Eventually we will fill in these
     * performance cracks.
     */
    public static void writeObject(DataOutput out, Object o) throws IOException {
        if(o == null) {
            out.writeByte(NULL);
        } else if(o instanceof Bool) {
            out.writeByte(BOOLEAN);
            out.writeBoolean(((Bool)o).getValue());
        } else if(o instanceof Byte) {
            out.writeByte(BYTE);
            out.writeByte(((Byte)o).byteValue());
        } else if(o instanceof Short) {
            out.writeByte(SHORT);
            out.writeShort(((Short)o).shortValue());
        } else if(o instanceof Integer) {
            out.writeByte(INT);
            out.writeInt(((Integer)o).intValue());
        } else if(o instanceof Long) {
            out.writeByte(LONG);
            out.writeLong(((Long)o).longValue());
        } else if(o instanceof Float) {
            out.writeByte(FLOAT);
            out.writeFloat(((Float)o).floatValue());
        } else if(o instanceof Double) {
            out.writeByte(DOUBLE);
            out.writeDouble(((Double)o).doubleValue());
        } else if(o instanceof String) {
            out.writeByte(STRING);
            out.writeInt(((String)o).length());
            out.writeBytes((String)o);
        } else if(o instanceof java.sql.Date) {
            out.writeByte(SQLDATE);
            out.writeLong(((java.sql.Date)o).getTime());
        } else if(o instanceof java.sql.Time) {
            out.writeByte(SQLTIME);
            out.writeLong(((java.sql.Time)o).getTime());
        } else if(o instanceof java.sql.Timestamp) {
            out.writeByte(SQLTIMESTAMP);
            Timestamp ts = (Timestamp)o;
            out.writeLong(ts.getTime());
              //  always thinking about the future of Nanotech
            out.writeInt(ts.getNanos());
        } else if(o instanceof byte[]) {
            out.writeByte(BYTES);
            out.writeInt(((byte[])o).length);
            out.write((byte[])o);
        } else if(o instanceof BigInteger) {
            out.writeByte(BIG_INTEGER);
            writeBigInteger(out, (BigInteger)o);
        } else if(o instanceof BigDecimal) {
            out.writeByte(BIG_DECIMAL);
            BigDecimal d = (BigDecimal)o;
            out.writeByte((byte)d.scale());
            writeBigInteger(out, d.unscaledValue());
        } else {
            throw new RuntimeException("CRITICAL ERROR: don't support writing : " + o.getClass());
        }
    }

    public static Object readObject(DataInput in) throws IOException {
        byte type = in.readByte();

        switch(type) {
            case NULL:     return null;
            case BOOLEAN:  return Bool.getBool(in.readBoolean());
            case BYTE:     return new Byte(in.readByte());
            case SHORT:    return new Short(in.readShort());
            case INT:      return new Integer(in.readInt());
            case LONG:     return new Long(in.readLong());
            case FLOAT:    return new Float(in.readFloat());
            case DOUBLE:   return new Double(in.readDouble());
            case STRING:   int len = in.readInt();
                           byte[] data = new byte[len];
                           in.readFully(data);
                           return new String(data);
            case SQLDATE:  return new java.sql.Date(in.readLong());
            case SQLTIME:  return new java.sql.Time(in.readLong());
            case SQLTIMESTAMP:
                Timestamp ts = new Timestamp(in.readLong());
                ts.setNanos(in.readInt());
                return ts;

            case BYTES:
                len = in.readInt();
                data = new byte[len];
                in.readFully(data);
                return data;

            case BIG_INTEGER:
                return readBigInteger(in);

            case BIG_DECIMAL:
                int scale = in.readByte() & 0x000000ff;
                BigInteger unscaledValue = readBigInteger(in);
                return new BigDecimal(unscaledValue, scale);

            default:
                    throw new RuntimeException("CRITICAL ERROR: don't support reading type : " + type);
        }
    }

    /**
     *  For directly reading a BigInteger from DataInput
     */
    public static BigInteger readBigInteger(DataInput in) throws IOException {
        int length = in.readByte() & 0x000000ff;
        if(length == 0) {
            return null;
        }
        byte[] buf = new byte[length];
        in.readFully(buf);
        return new BigInteger(buf);
    }

    /**
     *  For directly writing a BigInteger from DataInput
     */
    public static void writeBigInteger(DataOutput out, BigInteger i) throws IOException {
        if(i == null) {
            out.writeByte((byte)0);
        } else {
            out.writeByte(getBigIntLength(i));
            out.write(i.toByteArray());
        }
    }
}
