package freebase.util;

import freebase.api.ColumnInfo;

import java.math.BigDecimal;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.text.ParseException;
import java.util.Calendar;
import java.util.GregorianCalendar;

/**
 *  TODO:  implement the method:
 *  boolean isCompatibleType(Object o, int type);
 *  boolean isCompatibleType(int source, int target);  //  ie, set target with value of source
 *             //  possibly return value of yes, no, yes but needs conversion or something (1.1 -> 1)
 *  Comparable convert(Comparable value, int source, int target)
 *
 * TODO:  support UTF Strings -or- a new UTF object, in ObjectIO as well
 */
public class SqlTypes {
    /**
     * Some initial high level categories.  All data fits into
     * one of these categories, really.
     */
    final public static int BOOLEAN = 0;  //  not a number to stay consistent with java :)
    final public static int NUMBER  = 1;
    final public static int STRING  = 2;
    final public static int TIME    = 3;
    final public static int BYTES   = 4;
//   NSTRING ?  international

    final public static String NULL_STRING = "null";
    
    private SqlTypes() {}
    
    /**
     * tries to parse the text into a SqlValue
     * TODO:  return an SqlValue
     * TO D:  put date parsers configuration in a file: make it an ordered list
     */
    public static SqlValue parseSqlValue(String text, int sqlType) throws ParseException, SQLException {
        Object value = parseObject(text, sqlType);

        return new SqlValue(sqlType, (Comparable)value);
    }

    /**
     * Parse the data into a basic object type
     */
    public static Comparable parseObject(String text, int sqlType) throws ParseException, SQLException {
        try{
        Comparable value = null;

          //  currently NULL is parsed in the actual parser
        if(text == null)
            return null;

        //  TODO:  this looks dubious and from another application.  look at it more closely
        if(text.equalsIgnoreCase(NULL_STRING))
            return null; //new SqlValue(sqlType, null);

        switch(sqlType) {
              //  Boolean is not acceptable as a val since it is not comparable.  DUH!
            case Types.BIT:           value = Bool.getBool(Boolean.getBoolean(text)); break;
            case Types.TINYINT:       value = Byte.valueOf(text);    break;
            case Types.SMALLINT:      value = Short.valueOf(text);   break;
            case Types.INTEGER:       value = Integer.valueOf(text); break;
            case Types.BIGINT:        value = Long.valueOf(text);    break;

            case Types.FLOAT:         value = Double.valueOf(text);  break;
            case Types.REAL:          value = Float.valueOf(text);   break;
            case Types.DOUBLE:        value = Double.valueOf(text);  break;

            case Types.NUMERIC:
            case Types.DECIMAL:       value = new BigDecimal(text);  break;

            case Types.CHAR:
            case Types.VARCHAR:
            case Types.LONGVARCHAR:   value = deSqlize(text);        break;

                //  TODO:  the data formats here need to be delegated out
            case Types.DATE:          value = java.sql.Date.valueOf(deSqlize(text));  break;
            case Types.TIME:          value = Time.valueOf(deSqlize(text));  break;
            case Types.TIMESTAMP:     value = Timestamp.valueOf(deSqlize(text));  break;

            case Types.BINARY:
            case Types.VARBINARY:
            case Types.LONGVARBINARY: throw new ParseException("Parsing not legal for BLOBS", 0);

            case Types.NULL:
            default:
                throw new ParseException("Fatalerror parsing '" + text + "' as typeid " + sqlType, 0);
        }

        return value;
        }catch(NumberFormatException nfe) {
            throw new SQLException(text + " not a number");
        }
    }

    /**
     * TODO:expand single quotes jj's  ->  jj''s
     */
    public static String sqlize(String st) {
        StringBuffer sb = new StringBuffer();
        sqlize(st, sb);
        return sb.toString();
    }

    /**
     * This is here just in case you are already writing to a string buffer
     * that way it saves extra objects
     */
    public static void sqlize(String st, StringBuffer sb) {
        if(st == null) {
            sb.append("null");
            return;
        }

        sb.append('\'');

        int curPos = 0;
        int nextQuote;
        while((nextQuote = st.indexOf('\'', curPos)) != -1) {
            sb.append(st.substring(curPos, nextQuote + 1));
            sb.append('\'');
            curPos = nextQuote + 1;
        }

        sb.append(st.substring(curPos));
        sb.append('\'');
    }

    /**
     * do the reverse of above
     */
    public static String deSqlize(String s) throws ParseException {
        if(s.charAt(0) != '\'') {
            throw new ParseException("val must begin with '", 0);
        }

        if(s.charAt(s.length() -1) != '\'') {
            throw new ParseException("val must end with '", s.length() -1);
        }

        StringBuffer sb = new StringBuffer();

        int end = s.length() - 1;
        int curPos = 1;
        int nextQuote;
        while((nextQuote = s.indexOf('\'', curPos)) != end) {
            if(nextQuote == (end - 1) || s.charAt(nextQuote + 1) != '\'')
                throw new ParseException("String not properly formatted", nextQuote);

            sb.append(s.substring(curPos, nextQuote));
            sb.append('\'');
            curPos = nextQuote + 2;
        }

        sb.append(s.substring(curPos, end));

        return sb.toString();
    }


    public static int compare(Comparable a, Comparable b) {
        if(a instanceof Number) {
            return NumberUtil.compareNumbers((Number)a, (Number)b);
        } else if(GlobalProps.isStringsCaseInsensitive() && a instanceof String) {
            String aString = (String)a;
            String bString = (String)b;

            return aString.compareToIgnoreCase(bString);
        } else {
            return a.compareTo(b);
        }
    }

    /**
     * Just like compare above, but null are equal to each other.  This is used in index
     * entry comparison, since null == null must be considered true.  Null is sorted after
     * the other value.
     * @param a
     * @param b
     * @return
     */
    public static int compareNullsEqual(Comparable a, Comparable b) {
          //  currently arbitrary ordering, sort it out later.
        if(a == null) {
            if(b != null) {
                return 1;
            }
            return 0;
        }
        if(b == null) {
            return -1;
        }

        return compare(a, b);
    }

    /**
     * a compare method useful for putting records into sorted record sets
     * 1  if o1 > o2
     * 0  if o1 = o2
     * -1 if o1 < o2
     *
    public static int compare(Object o1, Object o2) {
        return -2;
    }
     */

    /**
     * puts the Object into a string that can be directly used in
     * an sql query.
     *
     * TODO:  take into account the current time format
     */
    public static String sqlFormat(Object value) {
        if(value == null || value.equals(NULL_STRING))
            return "NULL";
        else if(value instanceof String)
            return '\'' + sqlize((String)value) + '\'';
        else if(value instanceof java.sql.Date ||
                value instanceof java.sql.Time ||
                value instanceof java.sql.Timestamp)
            return '\'' + value.toString() + '\'';
        else
            return value.toString();
    }

    /**
     * Will determine if data from the source can go into the target.  This is used to ensure that
     * statements are legal even if there is no actual data passing through.
     * TODO:  maybe this should directly throw an exception
     * @param source
     * @param target
     * @return eventually doesn't need to return anything.  TODO:  rename checkCompatible
     */
    public static boolean isCompatibleType(ColumnInfo source, ColumnInfo target) throws SQLException {
        if(source.getType() == Types.NULL) {
            if(! target.isNullable()) {
                ErrorCode.throwException(ErrorCode.COLUMN_NOT_NULLABLE, target.toString());
            } else {
                return true;
            }
        }

        //System.out.println("test compatibility " + source + " into " + target);

        if(isNumericType(target.getType())) {
            if(! isNumericType(source.getType())) {
                ErrorCode.throwException(ErrorCode.EXPRESSION_NOT_NUMBER, source.toString());
            }
        }

        if(isStringType(target.getType())) {
              //  TODO:  coerce numbers to strings automatically
            if(! isStringType(source.getType())) {
                return false;
            }
        }

        if(isTemporal(target.getType())) {
            if(! isStringType(source.getType()) && ! isTemporal(source.getType())) {
                ErrorCode.throwException(ErrorCode.EXPRESSION_NOT_NUMBER, source.toString());
            }
        }

        return true;
    }


    public static boolean isNumericType(int sqlType) {
        switch(sqlType) {
            case Types.BIT: case Types.TINYINT: case Types.SMALLINT:
            case Types.INTEGER: case Types.BIGINT:
            case Types.FLOAT: case Types.REAL:  case Types.DOUBLE:
            case Types.NUMERIC: case Types.DECIMAL:       return true;
            default:
                return false;
        }
    }

    public static boolean isStringType(int sqlType) {
        return sqlType == Types.CHAR || sqlType == Types.VARCHAR || sqlType == Types.LONGVARCHAR ||
               sqlType == Types.CLOB;
    }

    public static boolean isTemporal(int sqlType) {
        return sqlType == Types.DATE || sqlType == Types.TIME || sqlType == Types.TIMESTAMP;
    }

    /**
     * It is assumed that before this is called, the types have already been checked.  This is simply to
     * get the value into the exact correct type for this entity.
     *
     * TODO:  here we can put a nice switch to decide if everything is a BigDecimal or whatever
     */
    public static Comparable convertValue(Comparable value, ColumnInfo type) throws SQLException {
        if(value == null) {
            return null;
        }

        if(isStringType(type.getType())) {
            String s = (String)value;
            if(s.length() > type.getSize()) {
                ErrorCode.throwException(ErrorCode.STRING_TOO_LONG, "size " + s.length() +
                        " into column " + type.getFullName());
            }

            //  TODO:  pad char[] values (!!!! why, stupid field anyway)
        }

        if(type.getType() == Types.DATE) {
            if(value instanceof java.sql.Date) {
                return value;
            }

            if(value instanceof String) {
                try{
                    return java.sql.Date.valueOf((String)value);
                }catch(IllegalArgumentException pe) {
                    ErrorCode.throwException(ErrorCode.EXPRESSION_NOT_DATE, value.toString());
                }
            }

            if(value instanceof Timestamp) {
                throw new SQLException("Don't know what to do with timestamp into date:  truncate or error? (for now error)");
            }

            ErrorCode.throwException(ErrorCode.EXPRESSION_NOT_DATE);
        }

        if(type.getType() == Types.TIMESTAMP) {
            if(value instanceof Timestamp) {
                return value;
            }

            if(value instanceof java.util.Date) {
                return new Timestamp(((java.util.Date)value).getTime());
            }

            if(value instanceof String) {
                try{
                    String st = (String)value;
                    if(st.length() == 10) {  //  2000-12-12
                        return java.sql.Date.valueOf(st);
                    }

                    return Timestamp.valueOf(st);  //  any other
                }catch(IllegalArgumentException pe) {
                    ErrorCode.throwException(ErrorCode.EXPRESSION_NOT_TIMESTAMP, value.toString());
                }
            }

            ErrorCode.throwException(ErrorCode.EXPRESSION_NOT_TIMESTAMP, value.toString());
        }

          // all else?
        return value;
    }

    /**
     * returns a default val for any given sql type.
     */
    public static Object getDefault(int sqlType) {
        switch(sqlType) {
            case Types.BIT:           return Bool.FALSE;
            case Types.TINYINT:       return new Byte((byte)0);
            case Types.SMALLINT:      return new Short((short)0);
            case Types.INTEGER:       return new Integer(0);
            case Types.BIGINT:        return new Long(0);

            case Types.FLOAT:         return new Double(0);
            case Types.REAL:          return new Float(0);
            case Types.DOUBLE:        return new Double(0);

            case Types.NUMERIC:       
            case Types.DECIMAL:       return new BigDecimal(0);

            case Types.CHAR:          
            case Types.VARCHAR:       
            case Types.LONGVARCHAR:   return "";

            case Types.DATE:          return new java.sql.Date(System.currentTimeMillis());
            case Types.TIME:          return new java.sql.Time(System.currentTimeMillis());
            case Types.TIMESTAMP:     return new java.sql.Timestamp(System.currentTimeMillis());

            case Types.BINARY:        
            case Types.VARBINARY:     
            case Types.LONGVARBINARY: return new Bytes() ;
 
            case Types.NULL: 
            default:
                return null;
        }
    }

    /*********************************************************************
     * a couple of methods for getting names of things
     */

    /**
     * Returns the java class name for the given sql type.  This is currently used in ResultSetMetaData so
     * it returns Boolean for BOOLEAN and BIT, not Bool the internal type.
     * TODO:  add CLOB AND BLOB!
     * @param scale is ignored except for number types
     */
    public static String getJavaClassForType(int type, int scale) throws SQLException {
        switch(type) {
            case Types.BOOLEAN:
            case Types.BIT:           return "freebase.util.Bool";
            case Types.TINYINT:       return "java.lang.Byte";
            case Types.SMALLINT:      return "java.lang.Short";
            case Types.INTEGER:       return "java.lang.Integer";
            case Types.BIGINT:        return "java.lang.Long";

            case Types.FLOAT:         return "java.lang.Double";
            case Types.REAL:          return "java.lang.Float";
            case Types.DOUBLE:        return "java.lang.Double";

            case Types.NUMERIC:
            case Types.DECIMAL:
                if(scale == 0) return "java.math.BigInteger";
                return "java.math.BigDecimal";

            case Types.CHAR:
            case Types.VARCHAR:
            case Types.LONGVARCHAR:   return "java.lang.String";

            case Types.DATE:          return "java.sql.Date";
            case Types.TIME:          return "java.sql.Time";
            case Types.TIMESTAMP:     return "java.sql.Timestamp";

            case Types.BINARY:
            case Types.VARBINARY:
            case Types.LONGVARBINARY: return "freebase.util.Bytes";

            case Types.NULL:
                return "null";
            default:
                throw new SQLException("Currently do not support java type knowledge for jdbc type: " + type);
        }
    }

    /**
     * Returns the readable name for the given sql type.
     */
    public static String getNameForType(int type) throws SQLException {
        switch(type) {
            case Types.BOOLEAN:       return "BOOLEAN";
            case Types.BIT:           return "BIT";
            case Types.TINYINT:       return "TINYINT";
            case Types.SMALLINT:      return "SMALLINT";
            case Types.INTEGER:       return "INTEGER";
            case Types.BIGINT:        return "BIGINT";

            case Types.FLOAT:         return "FLOAT";
            case Types.REAL:          return "REAL";
            case Types.DOUBLE:        return "DOUBLE";

            case Types.NUMERIC:       return "NUMERIC";
            case Types.DECIMAL:       return "DECIMAL";

            case Types.CHAR:          return "CHAR";
            case Types.VARCHAR:       return "VARCHAR";
            case Types.LONGVARCHAR:   return "LONGVARCHAR";

            case Types.DATE:          return "DATE";
            case Types.TIME:          return "TIME";
            case Types.TIMESTAMP:     return "TIMESTAMP";

            case Types.BINARY:        return "BINARY";
            case Types.VARBINARY:     return "VARBINARY";
            case Types.LONGVARBINARY: return "LONGVARBINARY";

            case Types.NULL:          return "NULL";
            default:
                throw new SQLException("Currently do not support java type knowledge for jdbc type: " + type);
        }
    }

    /*********************************************************************
     * calendar method
     */

    //  done the same way that it is done in the java.util.Date
    static GregorianCalendar staticCal = new GregorianCalendar();
    public static String formatHumanReadable(Object o) {
        if(o instanceof java.sql.Date) {
            synchronized (staticCal) {
                staticCal.setTime((java.sql.Date)o);
                return (staticCal.get(Calendar.MONTH) + 1) + "/" + staticCal.get(Calendar.DAY_OF_MONTH) + "/" +staticCal.get(Calendar.YEAR);
            }
        } else 
            return o.toString();
        
    }

    /**
     *  @return the binary size of the object,
     *   -1 if it a true variable or -2 if it is constant but flexible storage size
     *  TODO:  rewrite this method to pass in the column descriptor
     */
    public static int getWidth(int type) {
        switch(type) {
            case Types.BIT:           return 1;  //  yep, we will store bools as bytes.  Too bad
            case Types.TINYINT:       return 1;
            case Types.SMALLINT:      return 2;
            case Types.INTEGER:       return 4;
            case Types.BIGINT:        return 8;

            case Types.FLOAT:         return 8;
            case Types.REAL:          return 4;
            case Types.DOUBLE:        return 8;

            case Types.NUMERIC:       // TODO:  work out the storage format of these guys
            case Types.DECIMAL:       return -2;

            case Types.CHAR:          return -2;
            case Types.VARCHAR:       
            case Types.LONGVARCHAR:   return -1;

                //  TODO:  the data formats here need to be delegated out
            case Types.DATE:          return 4;  //  y y m d  (I think this is enuf!
            case Types.TIME:          return 8;  //  millis
            case Types.TIMESTAMP:     return 10;  //  TODO:  these times all need work

            case Types.BINARY:        return -2;
            case Types.VARBINARY:     
            case Types.LONGVARBINARY: return -1;
         
            case Types.NULL:
            
            default:
                throw new OhShitException("WHAT?  THIS CAN'T BE");
        }   
    }
    
    public static byte getVariableColumnSizeFieldSize(int type) {
        switch(type) {
            case Types.VARCHAR:       return 1;  //  short varchar has max of 254 (255 means null)
            case Types.LONGVARCHAR:   return 2;  //  max varchar is 65534 (last means null)
            case Types.VARBINARY:     return 2;  //  TODO:  decide on these later
            case Types.LONGVARBINARY: return 4;
            default:
                throw new OhShitException(type + " Not a variable type");
        }
    }
    
    /**
     *  read the bytes and turn it into an object.  len might not be needed
     *
    public SqlValue parseBytes(ColumnFormatInfo col, byte[] bytes, int offset, int len) {
        switch(type) {
            case Types.BIT:           return 1;  //  yep, we will store bools as bytes.  Too bad
            case Types.TINYINT:       return 1;
            case Types.SMALLINT:      return 2;
            case Types.INTEGER:       return 4;
            case Types.BIGINT:        return 8;

            case Types.FLOAT:         return 8;
            case Types.REAL:          return 4;
            case Types.DOUBLE:        return 8;

            case Types.NUMERIC:       // TODO:  work out the storage format of these guys
            case Types.DECIMAL:       return -2;

            case Types.CHAR:          return -2;
            case Types.VARCHAR:       
            case Types.LONGVARCHAR:   return -1;

                //  TODO:  the data formats here need to be delegated out
            case Types.DATE:          return 4;  //  y y m d  (I think this is enuf!
            case Types.TIME:          return 8;  //  millis
            case Types.TIMESTAMP:     return 10;  //  TODO:  these times all need work

            case Types.BINARY:        return -2;
            case Types.VARBINARY:     
            case Types.LONGVARBINARY: return -1;
         
            case Types.NULL:
            
            default:
                throw new OhShitException("WHAT?  THIS CAN'T BE:  " + type);
        }
    }
    */
}